/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: SerializableCookie.java,v 1.7.12.1 2000/10/19 17:58:53 jasona Exp $
 */





package com.lutris.http;

import java.io.Serializable;
import javax.servlet.http.Cookie;


/**
 *   This class is a substitute for Cookies, but it is serializable.
 *   It is used to pass cookies back and forth when using RMI.
 *   Rather than pasing or returning a Cookie, which would cause a
 *   marshalling error, the Cookies are turned into SerializabeCookies,
 *   passed through RMI, then turned back into Cookies at the other end. <P>
 *
 *   There is no reason why Cookies could not be serializable. Sun may well
 *   fix this oversight in version 2 of Java, in which case this class will
 *   no longer be necissary to use RMI with the MultiServer.
 *
 * @see javax.servlet.http.Cookie
 * @see com.lutris.servlet.connectionMethod.RMIConnectionMethodImpl
 * @see com.lutris.servlet.rmi.http.RMIHttpServletRequestImpl
 * @see com.lutris.servlet.rmi.http.RMIHttpServletResponseImpl
 */
public class SerializableCookie implements Serializable {

    private String name;
    private String value;
    private String comment;
    private String domain; 
    private int maxAge; 
    private String path; 
    private boolean secure; 
    private int version; 

    /**
     * Constructor. Saves all the data about the Cookie, so it can be
     * reconstructed later.
     *
     * @param cookie The cookie to make a copy of. 
     */
    public SerializableCookie(Cookie cookie) {
        name = cookie.getName();
        value = cookie.getValue();
        comment = cookie.getComment();
        domain = cookie.getDomain();
        maxAge = cookie.getMaxAge();
        path = cookie.getPath();
        secure = cookie.getSecure();
        version = cookie.getVersion();
    }


    /**
     * Recreate the cookie that was passed into the constructor.
     *
     * @return A Cookie just like the one passed into the constructor.
     */
    public Cookie getCookie() {
        Cookie cookie = new Cookie(name, value);
        if (comment != null)
            cookie.setComment(comment);
        if (domain != null)
            cookie.setDomain(domain);
        cookie.setMaxAge(maxAge);
        if (path != null)
            cookie.setPath(path);
        cookie.setSecure(secure);
        cookie.setVersion(version);
        return cookie;
    }


    /**
     *  Returns a multi-line description of the data stored about
     *  the cookie. Usefull for debugging.
     *
     *  @return A string description of the cookie.
     */
    public String toString() {
        return "SerializableCookie:" +
               "\n    name = " + name +
               "\n    value = " + value +
               "\n    comment = " + comment +
               "\n    domain = " + domain +
               "\n    maxAge = " + maxAge +
               "\n    path = " + path +
               "\n    secure = " + secure +
               "\n    version = " + version;
    }

}


