/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: MimeHeaderTests.java,v 1.7.8.1 2000/10/19 17:58:53 jasona Exp $
 */





package com.lutris.mime.tests;
import	com.lutris.mime.*;
import	com.lutris.testFramework.*;

/**
 * Tests for KeywordValue Table.
 *
 * @version	$Revision: 1.7.8.1 $
 * @author	Mark Diekhans
 * @since	Jolt1.0
 */
public class MimeHeaderTests {
    private static boolean verbose=true;
    private static int failures = 0;
    private static int testNum = 0;
    private static String BOGUS = "BOGUS";
    private static String[][][] tests = {
	{   {"Correct header with no parameters"},
	    {"Content-Length: 1024", "content-length", "1024"}
	}, 
	{   {"Correct header with no parameters"},
	    {"Content-Type: application/octet-stream", "content-type",
		    "application/octet-stream"}
	}, 
	{   {"Correct header with single parameter"},
	    {"Content-Type: multipart/form-data; boundary=\"------FOO\"",
		    "content-type", "multipart/form-data"},
	    {"boundary", "------FOO"}
	},
	{   {"Correct header with two parameters"},
	    {"Content-Disposition: form-data; name=upload1; " +
				   "filename=\"/tmp/tmpf0933E4C8\"",
		    "content-disposition", "form-data"},
	    {"name", "upload1"},
	    {"filename", "/tmp/tmpf0933E4C8"}
	},
	{   {"Correct header with two parameters (no whitespace)"},
	    {"Content-Disposition:form-data;name=upload1;" +
				   "filename=\"/tmp/tmpf0933E4C8\"",
		    "content-disposition", "form-data"},
	    {"name", "upload1"},
	    {"filename", "/tmp/tmpf0933E4C8"}
	},
	{   {"Correct simple header with no parameters (comments)"},
	    {"Content-Length: 1024 (Test comments)", "content-length", "1024"}
	}, 
	{   {"Correct simple header with 1 parameter (comments)"},
	    {"Content-Length: 1024; foo=bar (Test comments)", "content-length", "1024"},
	    {"foo", "bar"}
	}, 
	{   {"Correct simple header with 2 parameters (comments)"},
	    {"Content-Length: 1024; foo=bar (Test comments) ; boo=bah", "content-length", "1024"},
	    {"foo", "bar"},
	    {"boo", "bah"}
	}, 
	{   {"Correct header with two parameters (comments)"},
	    {"Content-Disposition: form-data; name=upload1; " + "( comment )" +
				   " filename=\"/tmp/tmpf0933E4C8\"" +
				   " (another comment )",
		    "content-disposition", "form-data"},
	    {"name", "upload1"},
	    {"filename", "/tmp/tmpf0933E4C8"}
	},
	{   {"Correct header with two parameters (comments, no whitespace)"},
	    {"Content-Disposition: form-data; name=upload1; " + "( comment )" +
				   " filename=\"/tmp/tmpf0933E4C8\"" +
				   " (another comment )",
		    "content-disposition", "form-data"},
	    {"name", "upload1"},
	    {"filename", "/tmp/tmpf0933E4C8"}
	},
	{   {"Correct header with two parameters (comments 2)"},
	    {"Content-Disposition: form-data ;name=upload1; " + "( comment )" +
				   " filename=\"/tmp/tmpf0933E4C8\"" +
				   " (another comment )",
		    "content-disposition", "form-data"},
	    {"name", "upload1"},
	    {"filename", "/tmp/tmpf0933E4C8"}
	},
	{   {"Correct header with two parameters (comments, no whitespace 2)"},
	    {"Content-Disposition: form-data ;name=upload1; " + "( comment )" +
				   " filename=\"/tmp/tmpf0933E4C8\"" +
				   " (another comment )",
		    "content-disposition", "form-data"},
	    {"name", "upload1"},
	    {"filename", "/tmp/tmpf0933E4C8"}
	},
	{   {"Correct header with multivalued parameter"},
	    {"Weird-Header: foobar; name=upload1; name=\"upload2\"",
		    "weird-header", "foobar"},
	    {"name", "upload1", "upload2"}
	},
	{   {"Correct header, unknown parameter name"},
	    {"Content-Length: 1024", "content-length", "1024"},
	    {"missing", BOGUS},
	}, 
	{   {"The last test."},
	    {"Content-Length: 1024", "content-length", "1024"}
	} 
    };
    
    private static
    void printTest(String[][] test)
    {
	String[] s;

	System.out.print("    Test #" + (testNum + 1) + ": === ");
	s = test[0];
	if (s != null) {
	    if (s.length >= 1) {
		for (int i=0; i<s.length; i++) {
		    if (s[i] != null) System.out.print(" " + s[i]);
		    else System.out.print("NULL");
		}
	    } else System.out.print("EMPTY");
	} else System.out.print("NULL");
	System.out.println(" ===");
	
	s = test[1];
	if (s.length >= 1) {
	    System.out.print("\tInput=");
	    if (s[0] != null) System.out.print("'" + s[0] + "'; ");
	    else System.out.print("NULL; ");
	}
	System.out.println("");
	if (s.length >= 2) {
	    System.out.print("\tExpect type to be ");
	    if (s[1] != null) System.out.print("'" + s[1] + "'; ");
	    else System.out.print("NULL; ");
	}
	if (s.length >= 3) {
	    System.out.print(" and value to be ");
	    if (s[2] != null) System.out.print("'" + s[2] + "'.");
	    else System.out.print("NULL.");
	}
	System.out.println("");
	
	for (int i=2; i<test.length; i++) {
	    s = test[i];
	    if (s != null) {
		System.out.print("\tExpect parameter ");
		if (s[0] != null) System.out.print("'"+s[0]+"' to be ");
		else System.out.print("NULL");
		if ((s.length <= 1) || (s[1] == null))
		    System.out.print("to be null.");
		else if (s[1] == BOGUS)
		    System.out.print("to be nonexistent.");
		else
		    System.out.print("'" + s[1] + "'.");
		for (int j=2; j<s.length; j++) {
		    if (s[j] != null) System.out.print("??='"+s[j]+"'");
		    else System.out.print("??=NULL");
		}
		System.out.println("");
	    }
	}
    }
    
    private static
    String printArray(String[] s, int offset)
    {
	String line = "{ ";
	String comma = "";
	for (int j=offset; j<s.length; j++) {
	    line += comma;
	    comma = ", ";
	    if (s[j] == null) line += "NULL";
	    else line += "'" + s[j] + "'";
	}
	line += " }";
	return line;
    }
    
    private static
    void fail(String s)
    {
	System.out.println("    Result #" + (testNum + 1) + ": FAIL -- " + s);
	failures++;
    }
    
    public static
    void runTest(String[][] test)
    {
	String[] s;
	String	input;

	System.out.print("    Running #" + (testNum + 1) + ": === ");
	s = test[0];
	if (s != null) {
	    if (s.length >= 1) {
		for (int i=0; i<s.length; i++) {
		    if (s[i] != null) System.out.print(" " + s[i]);
		    else System.out.print("NULL");
		}
	    } else System.out.print("EMPTY");
	} else System.out.print("NULL");
	System.out.println(" ===");
	
	s = test[1];
	if (s.length >= 1) {
	    input = s[0];
	} else {
	    fail("No input specified.");
	    return;
	}

	ContentHeader header = new ContentHeader(input);
	
	String hl = header.getRawHeaderLine();
	if (!input.equals(hl)) {
	    if (hl != null)
		fail("Call to getRawHeaderLine() returned '" +
			hl + "'; expected '" + "null" + "'.");
	    else
		fail("Call to getRawHeaderLine() returned '" +
			hl + "'; expected '" + input + "'.");
	    return;
	}

	if (s.length >= 2) {
	    String type = header.getHeaderType();
	    if ((s[1] == BOGUS) || (s[1] == null)) {
		if (type != null) {
		    fail("Unexpected header type: '" + type + "'; expected none.");
		    return;
		}
	    } else {
		if (!type.equals(s[1])) {
		    fail("Got header type: '" + type + "'; expected '"+s[1]+"'.");
		}
	    }
	}

	if (s.length >= 3) {
	    String value = header.getValue();
	    if ((s[2] == BOGUS) || (s[2] == null)) {
		if (value != null) {
		    fail("Unexpected value: '" + value + "'; expected none.");
		    return;
		}
	    } else {
		if (!value.equals(s[2])) {
		    fail("Got value: '" + value + "'; expected '"+s[2]+"'.");
		    return;
		}
	    }
	}
	
	for (int i=2; i<test.length; i++) {
	    s = test[i];
	    if (s != null) {
		if (s[0] == null) {
		    fail("Test bug: null cannot be used as a parameter name.");
		    return;
		}
		if (s[0] == BOGUS) {
		    fail("Test bug: BOGUS cannot be used as a parameter name.");
		    return;
		}
		String param = header.getParameter(s[0]);
		String[] params = header.getParameters(s[0]);
		if ((s.length < 2) || (s[1] == null) || (s[1] == BOGUS)) {
		    if (param != null) {
			fail("Parameter '" + s[0] + "' expected no value; got '"
				+ s[1] + "'.");
			return;
		    }
		} else {
		    if (param == null) {
			String sx = s[s.length - 1];
			if (sx == null) sx = "NULL";
			else sx = "'" + sx + "'";
			fail("getParameter() returned null; expected " + sx);
			return;
		    }
		    if (params == null) {
			fail("getParameters() returned null; expected "
				+ printArray(s, 1));
			return;
		    }
		    if (s[s.length - 1] == null) {
			fail("getParameter() returned '" + param +
				"'; expected null.");
			return;
		    }
		    if (!param.equals(s[s.length - 1])) {
			fail("getParameter() returned '" + param +
				"'; expected '" + s[s.length - 1] + "'.");
			return;
		    }
		    if (params.length != (s.length - 1)) {
			fail("getParameters() returned wrong #parameters; " + 
				"got " + printArray(params, 0) + "; expected "
				+ printArray(s, 1));
			return;
		    }
		    for (int j=0; j<params.length; j++) {
			if (!((params[j] == null)&&(s[j+1] == null))) {
			    if (((params[j] == null) && (s[j+1] != null))
			         || ((params[j] != null) && (s[j+1] == null))
			         || (!params[j].equals(s[j+1]))) {
				fail("Parameter mismatch; " +
				"got " + printArray(params, 0) + "; expected "
				+ printArray(s, 1));
				return;
			    }
			}
		    }
		}
	    } else {
		fail("Test bug: Param: Element#" + i + " was null.");
		return;
	    }
	}
	System.out.println("    Result #" + (testNum + 1) + ": OK");
    }

    public static
    void runTests()
    throws Exception
    {
	for (testNum=0; testNum<tests.length; testNum++) {
	    if (verbose) printTest(tests[testNum]);
	    runTest(tests[testNum]);
	}
	if (failures > 0) { 
	    throw new Exception("FAIL: " + failures + " occured.");
	}
    }
}

