/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ResManager.java,v 1.1.2.3.2.1 2000/10/19 17:58:53 jasona Exp $
 */

package org.enhydra.i18n;


import java.util.*;
import java.text.MessageFormat;

import com.lutris.logging.*;


/**
 *
 * Some helper functions for handeling i18n issues. One instance of this class
 * should be created for each resource bundle.<P>
 *
 * The ResManager is created by a call to <CODE>getResourceManager()</CODE>
 * the parameter is the name of the package that contails the Res class.
 * e.g. ResManager rez = getResourceBundle("org.enhydra.mypackagename");<P>
 *
 * To use the ResManager make a call to any of the <CODE>format()</CODE>
 * methods. In the default resource bundle the key is the same as the value.
 * So to display "I am 2 years old" call rez.format("I am {0} years old",2);
 * If the string "I am {0} years old" is in the bundle the value is returned. If
 * string is not found in the bundle the key is returned and an error is logged
 * to I18N_DEBUG. To see these errors add DEBUG_I18N to Server.LogToFile[] and
 * Server.LogToStderr[] in multiserver.conf.
 *
 *
 * @author Peter Johnson
 */

public class ResManager {

    /**
     * The ResourceBundle for this locale.
     */

    private ResourceBundle bundle;
    private String bundleName;
    private LogChannel logChannel;
    private int logLevelDebug;
    private int logLevelInfo;

    static private Hashtable bundles = new Hashtable();


    private ResManager(String packageName) {
	bundleName = packageName + ".Res";
        // not ready yet.
	// bundle = ResourceBundle.getBundle(bundleName);

        // Add DEBUG_I18N to the multiserver.conf file to log ResManager issues
 	logChannel = Logger.getCentralLogger().getChannel("Multiserver");
        logLevelDebug = logChannel.getLevel("I18N_DEBUG");
        logLevelInfo = logChannel.getLevel("I18N_INFO");
    }


   /**
     * Returns a resource manager assocated with the package name.
     * An instance of the Res class is created the first time the method is
     * called.
     *
     * @param clazz     A class from the package that Rez is in.
     *
     */
    public static ResManager getResManager(Class clazz) {

        // build a package name from the
        String packageName = clazz.getName();
        int lastDot = packageName.lastIndexOf('.');
        if (lastDot != -1)
            packageName = packageName.substring(0,lastDot);

        return getResManager(packageName);
    }

   /**
     * Returns a resource manager assocated with the package name.
     * An instance of the Res class is created the first time the method is
     * called.
     *
     * @param packageName	The package name that holds the Res class
     *
     */
    public static ResManager getResManager(String packageName) {
        ResManager rez = (ResManager) bundles.get(packageName);
        if (rez==null) {
            rez = new ResManager(packageName);
            bundles.put(packageName,rez);
        }
        return rez;
    }


    private String getString(String key) {
        // just return the key for now.
        return key;
        /*
        String msg;
        try {
            msg = bundle.getString(key);
        }
        catch (MissingResourceException ex) {
            msg = key;

            // Couldn't find the key log to I18N_INFO
            //logChannel.write(logLevelInfo,key + " (String is missing from resource bundle \"" + bundleName+"\" )");
        }
        logChannel.write(logLevelDebug,key + " => " + msg);
        return msg;
        */
    }


   /**
     * Returns a string that has been obtained from the resource manager
     *
     * @param key           The string that is the key to the translated message
     *
     */
    public String format(String key) {
        return getString(key);
    }

   /**
     * Returns a string that has been obtained from the resource manager then
     * formatted using the passed parameters.
     *
     * @param key           The string that is the key to the translated message
     * @param o0            The param passed to format replaces {0}
     *
     */
    public String format(String pattern, Object o0) {
        return MessageFormat.format(getString(pattern), new Object[] {o0});
    }

   /**
     * Returns a string that has been obtained from the resource manager then
     * formatted using the passed parameters.
     *
     * @param key           The string that is the key to the translated message
     * @param o0            The param passed to format replaces {0}
     * @param o1            The param passed to format replaces {1}
     *
     */
    public String format(String pattern, Object o0, Object o1) {
        return MessageFormat.format(getString(pattern), new Object[] {o0,o1});
    }


   /**
     * Returns a string that has been obtained from the resource manager then
     * formatted using the passed parameters.
     *
     * @param key           The string that is the key to the translated message
     * @param o0            The param passed to format replaces {0}
     * @param o1            The param passed to format replaces {1}
     * @param o2            The param passed to format replaces {2}
     *
     */
    public String format(String pattern, Object o0, Object o1, Object o2) {
        return MessageFormat.format(getString(pattern), new Object[] {o0,o1,o2});
    }

   /**
     * Returns a string that has been obtained from the resource manager then
     * formatted using the passed parameters.
     *
     * @param key           The string that is the key to the translated message
     * @param o0            The param passed to format replaces {0}
     * @param o1            The param passed to format replaces {1}
     * @param o2            The param passed to format replaces {2}
     * @param o3            The param passed to format replaces {3}
     *
     */
    public String format(String pattern, Object o0, Object o1, Object o2, Object o3) {
        return MessageFormat.format(getString(pattern), new Object[] {o0,o1,o2,o3});
    }

    // add more if you need them...
}

