/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Builder.java,v 1.32.6.1 2000/10/05 00:56:03 jay Exp $
 */

/*
 *   Builder.java
 *
 * @author    Jay Gunter
 * @since     LBS1.8
 * @version   $Revision: 1.32.6.1 $
 *
 */

package com.lutris.dods.builder.engine;

import com.lutris.dods.store.*;

import com.lutris.dods.store.doa.*;
import com.lutris.dods.gui.tree.*;
import com.lutris.dods.builder.generator.base.SourceGenerator;
import com.lutris.dods.builder.engine.GeneratorInfo;

import  com.lutris.appserver.server.session.*;
import  com.lutris.util.Config;
import  com.lutris.util.ConfigFile;
import  com.lutris.util.ConfigException;
import  java.io.*;
import  java.util.*;
import  java.math.BigDecimal;
import  java.net.URLEncoder;

import  javax.swing.JTextArea;

/**
 *
 *   For each DataObject in the current DODSProject,
 *   the run() method invokes the source code generators
 *   described in the dods.conf file.
 *   After generating the tree of source code for for the DataObjects,
 *   a makefile is created in each directory in the tree.
 *   Finally, 'make' is run at the top directory.
 *
 */
public class Builder
{
    /**
     * DODS_HOME is the name of a property that is set with -D
     * when DODS is run.  The value for the property should be
     * the directory where DODS can find its config file, dods.conf.
     */
    private static final String DODS_HOME = "DODS_HOME";

    /**
     * dodsDir is the value of the DODS_HOME property.
     */
    private static String dodsDir = null;

    /**
     * ENHYDRA_DIR is the name of a property that is set with -D
     * when DODS is run.  The value for the property should be
     * the directory where DODS can find the Lutris stdlib.mk file.
     */
    private static final String ENHYDRA_DIR = "ENHYDRA_DIR";

    /**
     * lutrisDir is the value of the ENHYDRA_DIR property.
     */
    private static String lutrisDir = null;

    /**
     * CLASS_PATH is an optional key in dods.conf that specifies where
     * the SourceGenerator classes are located.
     */
    private static final String CLASS_PATH = "ClassPath";

    /**
     * SOURCE_GEN is a required key in dods.conf that informs the builder
     * which SourceGenerator classes are to be loaded and run.
     * @see RUN_SOURCE_GEN for example line in dods.conf file.
     */
    private static final String SOURCE_GEN = "SourceGenerator";

    /**
     * GEN_RUN is a required key in dods.conf that tells the builder
     * if a given SourceGenerator class is to be run.
     *   e.g.
     *      SourceGenerator.DO.RunMe= true
     */
    private static final String GEN_RUN = "RunMe";

    /**
     * GEN_RUN_FOR_ABSTRACT is an optional key in dods.conf
     * that tells the builder if a given SourceGenerator class is to be run
     * for Data Objects marked in DODS as "abstract".
     * If not defined, default value is "false".
     *   e.g.
     *      SourceGenerator.DO.RunForAbstract= true
     */
    private static final String GEN_RUN_FOR_ABSTRACT = "RunForAbstract";

    /**
     * GEN_HANDLES_ARRAYS is an optional key in dods.conf
     * that tells the builder if a given SourceGenerator class should
     * perform any actions for Attributes with isArray() == true.
     * If not defined, default value is "false".
     *   e.g.
     *      SourceGenerator.DO.HandleArrays= true
     */
    private static final String GEN_HANDLES_ARRAYS = "HandleArrays";

    /**
     * GEN_CLASS_NAME is a required key in dods.conf.
     * It tells the builder the name of a given SourceGenerator class.
     *   e.g.
     *      SourceGenerator.DO.GeneratorClass= SourceGenerator_DO
     */
    private static final String GEN_CLASS_NAME = "GeneratorClass";

    /**
     * GEN_OUTPUT_SUFFIX is a required key in dods.conf.
     * It tells the builder the suffix of the file
     * that a given SourceGenerator class will create.
     *   e.g.
     *      SourceGenerator.DO.OutputFileSuffix= java
     */
    private static final String GEN_OUTPUT_SUFFIX = "OutputFileSuffix";

    /**
     * GEN_OBJECT_SUFFIX is an optional key in dods.conf.
     * It tells the builder the SourceGenerator what string to append
     * to the object name.
     *   e.g.
     *      SourceGenerator.HomeInterface.ObjectSuffix= Home
     */
    private static final String GEN_OBJECT_SUFFIX = "ObjectSuffix";

    /**
     * GEN_OUTPUT_CLASS_PARENT is an optional key in dods.conf.
     * It tells the builder what (if any) class is extended by the
     * classes that will be generated.
     *   e.g.
     *      SourceGenerator.DO.OutputClassParent= com.lutris.blah.CoreDO
     */
    private static final String GEN_OUTPUT_CLASS_PARENT = "OutputClassParent";

    /**
     * GEN_EXTRA_CLASSPATH is an optional key in dods.conf.
     * It tells the builder where the GEN_EXTRA_CLASSPATH can be found.
     *   e.g.
     *      SourceGenerator.DO.ExtraClasspath= /users/fred/LBS/Harmony
     */
    private static final String GEN_EXTRA_CLASSPATH = "ExtraClasspath";

    /**
     * GEN_TEMPLATE_DIR is a required key in dods.conf that tells the builder
     * the location of the templates used by a given SourceGenerator class.
     *   e.g.
     *      SourceGenerator.DO.TemplateDir= .../templates
     */
    private static final String GEN_TEMPLATE_DIR = "TemplateDir";

    /**
     * sourceGenerators is a Vector of GeneratorInfo objects.
     * GeneratorInfo objects are used to initialized SourceGenerator objects.
     */
    private static Vector sourceGenerators = null;

    /**
     * We perform some special post-processing on the .sql files
     * produced by the SourceGenerator_SQL class,
     * but only if that class was run.
     */
    private static boolean generating_sql_files = false;

    /**
     * OUTPUT_DIR is the name of a property that is set with -P
     * when DODS is run.  The value for the property is the top of the
     * directory tree where DODS will write the source files it generates.
     */
    private static final String OUTPUT_DIR = "DODS_OUTPUT_DIR";

    /**
     * outputDir is the value of the OUTPUT_DIR property.
     */
    private static String outputDir = null;

    /**
     * ROOT_DIR is the name of a property that is set with -P
     * when DODS is run.  The value for the property is the value of the
     * ROOT_DIR macro that is set in a top-level Lutris makefile.
     */
    private static final String ROOT_DIR = "ROOT_DIR";

    /**
     * rootDir is the value of the ROOT_DIR property.
     */
    private static String rootDir = null;

    /**
     * MAKEFILE_SCRIPT is a required key in dods.conf that tells the builder
     * facts about the script that generates the Makefiles for the
     * generated source code.
     */
    private static final String MAKEFILE_SCRIPT = "MakefileGeneratorScript";

    /**
     * MAKEFILE_SCRIPT_PATH is a required key in dods.conf that tells
     * the builder the pathname of the script.
     */
    private static final String MAKEFILE_SCRIPT_PATH = "Path";

    /**
     * makefileScriptPath is the value of the MAKEFILE_SCRIPT_PATH property.
     */
    private static String makefileScriptPath = null;

    /**
     * MAKEFILE_SCRIPT_RUN is a required key in dods.conf that tells
     * the builder whether to run the script.
     */
    private static final String MAKEFILE_SCRIPT_RUN = "RunAutomatically";

    /**
     * makefileScriptRun is the value of the MAKEFILE_SCRIPT_RUN property.
     */
    private static String makefileScriptRun = null;

    /**
     * MAKEFILE_SCRIPT_CLEANUP is a required key in dods.conf that tells the
     * script whether to delete the generated source code after compilation.
     */
    private static final String MAKEFILE_SCRIPT_CLEANUP = "DeleteSourceCode";

    /**
     * makefileScriptCleanup is the value of the
     * MAKEFILE_SCRIPT_CLEANUP property.
     */
    private static String makefileScriptCleanup = null;

    /**
     * MAKEFILE_EXTRA_CLASSPATH is an optional key in dods.conf that tells the
     * script of any value for the EXTRA_CLASSPATH macro
     * in the generated Makefiles.
     */
    private static final String MAKEFILE_EXTRA_CLASSPATH = "ExtraClasspath";

    /**
     * makefileExtraClasspath is the value of the
     * MAKEFILE_EXTRA_CLASSPATH property.
     */
    private static String makefileExtraClasspath = null;

    /**
     * dirNames is the list of directories that were created by mkdir()
     * during this run of the Builder.
     */
    //private static Hashtable dirNames;
    private static Vector dirNames;

    /**
     * numDirs is the number of elements in dirNames.
     */
    private static int numDirs;

    /*
     * dodsProject is the contents of the DODS project file saved to disk
     */
    private static DODSProject dodsProject;

    /*
     * treeRoot is the top of the DataObject hierarchy for the DODS project
     */
    private static DOTreeNode treeRoot;

    /*
     * createdFiles is the list of files generated by Builder.
     */
    private static Vector createdFileNames = new Vector();

    /*
     * Write error and halt JVM.
     */
    static private void fatalErr(String msg) {
        System.err.println(msg);
        Runtime vm = Runtime.getRuntime();
	// FIX
//        vm.exit(1);
    }

    public static void main( String argv[] ) throws Exception {
	run( argv[1], argv[0] );
    }


   /**
     * Initializes the DODS Builder, which generates source code
     * for each DataObject defined in a DODS project file.
     *
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     * 1) Get the location of the top of the output tree
     * 2) Get the location of the dods.conf file
     * 3) Load the dods.conf file.
     * 4) Get the classpath from the config file. It is an optional field
     * 5) Load the SourceGenerator classes
     * 6) Get info for running the MakefileGeneratorScript.
     */
    private static void init() throws Exception {

	/*
	 * 1) Get the location of the top of the output tree.
	 */
	if ( null == outputDir ) {
	    outputDir = System.getProperty( OUTPUT_DIR );
	    if ( null == outputDir )
		fatalErr( "Cannot find property:  " + OUTPUT_DIR);
	}

	/*
	 * The property ROOT_DIR contains the value of the ROOT_DIR macro
	 * that is set at the top of Lutris makefiles.
	 */
	rootDir = System.getProperty( ROOT_DIR );
	if ( null == rootDir )
	    rootDir = outputDir;

	/*
	 * 2) Get the location of the dods.conf file.
	 */
	dodsDir = System.getProperty( DODS_HOME );
	if ( null == dodsDir )
	    fatalErr( "Cannot find property:  " + DODS_HOME);

        String confFilename = dodsDir + File.separator + "dods.conf";
        File confFile = new File(confFilename);
        if (!confFile.exists())
            fatalErr("No conf file " + confFilename);
        if (!confFile.isFile())
            fatalErr("Not a file: " + confFilename);

	/*
	 * 3) Load the dods.conf file.
	 */
        Config cfg = null;
        try {
            FileInputStream cfgFIS = new FileInputStream(confFilename);
            ConfigFile cfgF = new ConfigFile(cfgFIS);
            cfgFIS.close();
            cfg = cfgF.getConfig();
        } catch (IOException ioe) {
            fatalErr("IO error reading " + confFilename +
                     ":\n" + ioe.getMessage());
        } catch (ConfigException ce) {
            fatalErr("Bad file format reading " + confFilename +
                     ":\n" + ce.getMessage());
        }

        /*
         * 4) Get the classpath from the config file. It is an optional field.
         */
        String[] strArray = new String[1];
        String classPath = null;
        if (cfg.containsKey(CLASS_PATH)) {
            try {
                strArray = cfg.getStrings(CLASS_PATH);
                classPath = strArray[0];
            } catch (ConfigException ce) {
                fatalErr("Invalid config file " +
                      confFilename + ". No key " + CLASS_PATH);
            }
        }

	/*
	 * 5) Load the SourceGenerator classes.
	 */
	SourceGenerator.reset();
	Config srcGen = null;
        try {
            srcGen = (Config) cfg.getSection( SOURCE_GEN );
        } catch (ConfigException e) {
            // Do nothing.
        } catch (Exception e) {
            fatalErr("Internal error parsing config file (section " +
                     SOURCE_GEN + ").");
        }
        if ( null == srcGen )
            fatalErr("Illegal config file. Section " + SOURCE_GEN
			+ " not found.");
	sourceGenerators = new Vector();
        String[] genTypes = srcGen.keys();
	String runIt, templateDir, className, suffix, objSuffix;
	String parentClass, extraCP, runForAbstract;
	boolean genAbstract, handlesArrays;
	makefileExtraClasspath = "";
	Hashtable extraClasspaths = new Hashtable();
	for ( int k = 0; k < genTypes.length; k++ ) {
	    Config props = null;
	    try {
		props = (Config) srcGen.getSection( genTypes[k] );
	    } catch ( Exception e ) {
	    }
	    if ( null == props )
            	fatalErr("Illegal config file. Section " + SOURCE_GEN
			+ "." + genTypes[k] + " not found.");
	    runIt = props.getString( GEN_RUN );
	    if ( null != runIt && runIt.equals( "true" ) ) {
		// This SourceGenerator is enabled; prepare for using it.
		templateDir    = fixDodsPath( props,  GEN_TEMPLATE_DIR );
		className      = props.getString(     GEN_CLASS_NAME );
                try {
		    objSuffix  = props.getString(     GEN_OBJECT_SUFFIX );
                } catch ( Exception e ) {
		    objSuffix  = null;
                }
		suffix         = props.getString(     GEN_OUTPUT_SUFFIX );
		parentClass    = fixIfMissing( props, GEN_OUTPUT_CLASS_PARENT );
		extraCP        = fixDodsPath( fixIfMissing( props, GEN_EXTRA_CLASSPATH ) );
		genAbstract    = getBoolean(   props, GEN_RUN_FOR_ABSTRACT );
		handlesArrays  = getBoolean(   props, GEN_HANDLES_ARRAYS );
		extraClasspaths.put( extraCP, "x");
		Class generator = Class.forName( className );
		GeneratorInfo info = new GeneratorInfo(
			cfg, genAbstract, handlesArrays,
			outputDir, genTypes[k],
			suffix, parentClass, extraCP,
			generator, templateDir, objSuffix );
		sourceGenerators.addElement( info );
		if ( suffix.equals( "sql" ) )
		    generating_sql_files = true;
	    }
	}

	/*
	 * 6) Get info for running the MakefileGeneratorScript.
	 */
	lutrisDir = System.getProperty( ENHYDRA_DIR );
	if ( null == lutrisDir )
	    fatalErr( "Cannot find property:  " + ENHYDRA_DIR);

        File dir = new File( lutrisDir );
        if (!dir.exists())
            fatalErr("ENHYDRA_DIR " + lutrisDir + " does not exist.");
        if ( dir.isFile() )
            fatalErr("ENHYDRA_DIR " + lutrisDir + " is not a directory.");

	Config mkGen = null;
        try {
            mkGen = (Config) cfg.getSection( MAKEFILE_SCRIPT );
        } catch (ConfigException e) {
            // Do nothing.
        } catch (Exception e) {
            fatalErr("Internal error parsing config file (section " +
                     MAKEFILE_SCRIPT + ").");
        }
        if ( null == mkGen )
            fatalErr("Illegal config file. Section " + MAKEFILE_SCRIPT
			+ " not found.");
	makefileScriptPath    = fixDodsPath( mkGen, MAKEFILE_SCRIPT_PATH );
	makefileScriptRun     = mkGen.getString(    MAKEFILE_SCRIPT_RUN );
	makefileScriptCleanup = mkGen.getString(    MAKEFILE_SCRIPT_CLEANUP );
	String cp             = mkGen.getString(    MAKEFILE_EXTRA_CLASSPATH );
	if ( null != cp ) makefileExtraClasspath += cp + ":";
	//dirNames = new Hashtable();
	dirNames = new Vector();
	numDirs = 0;
    }

    private static String fixDodsPath( Config props, String key )
    throws Exception {
	String dir = props.getString( key );
	if ( null == dir || 0 == dir.length() )
	    throw new Exception( "Bad value for key '" + key + "'" );
	return fixDodsPath( dir );
	/*
	if ( dir.startsWith( DODS_HOME ) )
	    dir = dodsDir + dir.substring( DODS_HOME.length() );
	return dir;
	*/
    }

    private static String fixDodsPath( String dir )
    throws Exception {
	if ( null == dir || 0 == dir.length() )
	    return "";
	if ( dir.startsWith( DODS_HOME ) )
	    dir = dodsDir + dir.substring( DODS_HOME.length() );
	return dir;
    }

    private static String fixIfMissing( Config props, String key )
    throws Exception {
	String str = "";
	try {
	    str = props.getString( key );
	    if ( null == str ) str = "";
	} catch ( Exception e ) {
	    str = "";
	}
	return str;
    }

    private static boolean getBoolean( Config props, String key ) {
	boolean ret = false;
	String value = null;
	try {
	    value = props.getString( key );
	    if ( null != value && value.equals("true") )
		ret = true;
	} catch ( Exception e ) {
	}
	return ret;
    }


   /**
     * Runs the DODS Builder, creating source code for each DataObject.
     *
     * This version of the run() method can be used for invoking
     * the DODS Builder from the command-line.
     *
     * @param dir
     *   The directory containing the DODS project file.
     * @param dodsFilename
     *   The name of the DODS project file.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     * open DODS project file to get DataObject hierarchy;
     * invoke processDO on the top node in the hierarchy;
     * run the makefileGenerator script;
     * invoke make on the generated makefile at the top level directory;
     */
    public static void run( String dir, String dodsFilename)
    throws Exception {
	try {
	    //File file = new File( dir, dodsFilename );
	    File file = new File( dodsFilename );
	    /*
	    FileInputStream finput = new FileInputStream( file );
	    ObjectInputStream oInput = new ObjectInputStream( finput );
	    dodsProject = (DODSProject) oInput.readObject();
	    dodsProject.setFile( file );
	    //run( dodsProject );
	    */
	    Store store = new Store();
	    store.setInputFile( file );
	    store.read();
	    dodsProject = store.getProject();
/* 	    dodsProject.setFile( file ); */
	    run( dir, dodsProject );
	} catch ( Exception e ) {
	    System.err.println("ERROR run: " + e.getMessage() );
	    throw e;
	}
    }

   /**
     * Runs the DODS Builder, creating source code for each DataObject.
     *
     * This version of the run() method is used by the Build button on the GUI.
     *
     * @param targetDir
     *   The directory where the source code will be written.
     * @param project
     *   The project object containing the DataObject hierarchy.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     * Record the targetDir value.
     * Invoke the run(DODSProject) method.
     */
    public static void run( String targetDir, DODSProject project )
    throws Exception {
	outputDir = targetDir;
	run( project );
    }

   /**
     * Runs the DODS Builder, creating source code for each DataObject.
     *
     * This version of the run() method supports invoking
     * the DODS Builder both from the DODS GUI ('Build' button)
     * and from the the command-line (via the other run() method).
     *
     * @param dodsProject
     *   The project object containing the DataObject hierarchy.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     * open DODS project file to get DataObject hierarchy;
     * invoke processDO on the top node in the hierarchy;
     * run the makefileGenerator script;
     * invoke make on the generated makefile at the top level directory;
     */
    public static void run( DODSProject project ) throws Exception {
	DOgraphNode.reset();

	init();
        dodsProject = project;
	clearDOarray();
        DataObject[] a = getAllDataObjects( project );
	for ( int k = 0; k < a.length; k++ )
	    processDO( a[k] );
/*
	treeRoot = new DOTreeNode( dodsProject, true );
	treeRoot.updateChildrenOrder();
	Vector children = treeRoot.getChildren();
*/
//System.err.println("Node: " + treeRoot.getName() + ", children = "+children );
/* System.err.println("run: children.size()="+children.size()); */
/*
	for ( int k = 0; k < children.size(); k++ ) {
	    processNode( (DOTreeNode)children.elementAt(k) );
	}
*/

//System.err.println("DOgraphNode.getCreateOrder()="+DOgraphNode.getCreateOrder());
//System.err.println("DOgraphNode.getDropCommands()="+DOgraphNode.getDropCommands());
	if ( generating_sql_files )
	    generateSQLcommands();

	generateMakefiles();
    }

    /**
     * Ideally, the Builder class would not be concerned with the
     * generation of any particular type of file.
     * But the SourceGenerator classes are designed to generate
     * one file per DataObject; they are not well suited to handle
     * final roundup actions like we're doing here.
     * Uglier contortions would be necessary to place this method in the
     * SourceGenerator_SQL class.
     */
    private static void generateSQLcommands() {
	String objectid_tablename = "objectid";
	String create_sql = outputDir + "/create_tables.sql";
	try {
	    FileWriter create_tables = new FileWriter( create_sql );
	    char[] buf = new char[ 512 ];
	    int n;
	    String[] sql_files = DOgraphNode.getOrderedFilenames();
	    for ( int i = 0; i < sql_files.length; i++ ) {
		try {
		    FileReader in = new FileReader( sql_files[ i ] );
		    while ( 0 <= ( n = in.read( buf ) ) )
			create_tables.write( buf, 0, n );
		    in.close();
		} catch ( FileNotFoundException e ) {
		    // assume DO was abstract, thus having no sql file.
		}
	    }
	    String s = "create table " + objectid_tablename +
			"(\n next " + SourceGenerator.oidDbType +
			" NOT NULL \n);\n";
	    create_tables.write( s.toCharArray() );
	    create_tables.flush();
	    create_tables.close();
	} catch ( Exception e ) {
	    print( "ERROR IN GENERATING FILE " + create_sql + ": " + e.getMessage() );
	}

	String drop_sql = outputDir + "/drop_tables.sql";
	try {
	    FileWriter drop_tables = new FileWriter( drop_sql );
	    drop_tables.write( DOgraphNode.getDropCommands().toCharArray() );
	    String s = "drop table " + objectid_tablename + ";\n";
	    drop_tables.write( s.toCharArray() );
	    drop_tables.flush();
	    drop_tables.close();
	} catch ( Exception e ) {
	    print( "ERROR IN GENERATING FILE " + drop_sql );
	}
    }

   /** Walk the project tree to build an array of all DataObjects.
     *
     */
    private static DataObject[] DOarray;
    private static void clearDOarray() {
        DOarray = null;
    }
    public static DataObject[] getAllDataObjects( DODSProject project ) {
	if ( null != DOarray )
	    return DOarray;
	Vector all = new Vector();
	DOTreeNode root = new DOTreeNode( project, true );
        root.updateChildrenOrder();
/*         Vector children = root.getChildren(); */
//System.err.println("Node: " + root.getName() + ", children = "+children );
/* System.err.println("run: children.size()="+children.size()); */
	/*
        for ( int k = 0; k < children.size(); k++ ) {
            addChildrenOf( all, (DOTreeNode)children.elementAt(k) );
        }
	*/
	addChildrenOf( all, root );
	DOarray = new DataObject[ all.size() ];
	all.copyInto( DOarray );
//for ( int j = 0; j < DOarray.length; j++ )
//System.err.println("DOarray["+j+"].getName()="+DOarray[j].getName());
	return DOarray;
    }
    private static void addChildrenOf( Vector all, DOTreeNode node ) {
	DataObject DO = node.getDataObject();
	if ( ! ( DO instanceof DODSProject || DO instanceof DOPackage ) ) {
	    all.addElement( DO );
	}
	Vector children = node.getChildren();
	if ( null != children )
	    for ( int i = 0; i < children.size(); i++ ) {
		addChildrenOf( all, (DOTreeNode) children.elementAt( i ) );
	    }
    }

   /**
     * Recursively processes a DOTreeNode,
     * invoking SourceGenerators to write source code files.
     *
     * for the DOTreeNode and each of its children {
     *    if the node is a DataObject, process that DataObject
     * }
     *
     * @param node
     *   A node of the DataObject hierarchy.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     */
/* DELETE, run now using getAllDataObjects

    private static void processNode( DOTreeNode node ) throws Exception {
	DataObject DO = node.getDataObject();

//System.err.println("Name: "+DO.getName());
//System.err.println("DO.getName()="+DO.getName());

	if ( ! ( DO instanceof DODSProject || DO instanceof DOPackage ) ) {
	    processDO( DO );
	}
	Vector children = node.getChildren();
	if ( null != children ) {
//System.err.println("Children: "+children);
//System.err.println("children.size()="+children.size());
	    for ( int k = 0; k < children.size(); k++ ) {
		// FIX hopefully we don't need to do this instanceof.
		//if ( children.elementAt( k ) instanceof DataObject )
		processNode( (DOTreeNode) ( children.elementAt( k ) ) );
	    }
	}
    }
*/


   /**
     * Processes a DataObject.
     *
     *    create the directory for the source files to be generated;
     *    for each SourceGenerator {
     *       initialize the SourceGenerator;
     *       invoke the writeFileHeader method, passing the DataObject;
     *       for each Attribute belonging to the DataObject {
     *          invoke the writeMemberStuff method, passing the Attribute;
     *       }
     *       invoke the writeFileFooter method, passing the DataObject;
     *       invoke the adjustOutputFile method;
     *    }
     *
     * @param DO
     *   The DataObject to be processed
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     */
    private static void processDO( DataObject DO ) throws Exception {

	DOgraphNode.addNode( DO );

//System.err.println("DO.getName()="+DO.getName());

	    Vector members = DO.getAttributes();
//System.err.println("Members: "+members.size());

	    // Each array Attribute is made into a separate db table
	    // with separate classes to access it.
	    for ( int j = 0; j < members.size(); j++ ) {
//System.err.println("j ="+j );
		Attribute attr = (Attribute) members.elementAt( j );
		if ( attr.isArray() )
		    processDO( SourceGenerator.handleArray( DO, attr ) );
	    }

	    for ( int i = 0; i < sourceGenerators.size(); i++ ) {
		GeneratorInfo   info  =
		    (GeneratorInfo) sourceGenerators.elementAt( i );

		if ( DO.isAbstract() && ! info.runForAbstract )
		    continue;


		String outDir = fixOutputDir( DO );
		String subDir = mkdir( outDir, DO.getAdjustedPackageName() );
		info.outputDir = outDir + subDir;
//System.err.println("info.outputDir ="+info.outputDir );
//if(true)return;
		SourceGenerator gen =
		    (SourceGenerator) info.processor.newInstance();
		gen.objectPrep( info, dodsProject, DO );
		gen.writeFileHeader( DO );

		for ( int j = 0; j < members.size(); j++ ) {
//System.err.println("j ="+j );
		    Attribute attr = (Attribute) members.elementAt( j );
		    if ( ! ( attr.isArray() && ! info.handlesArrays ) ) {
			gen.memberPrep( attr );
			gen.writeMemberStuff( attr );
		    }
		}
		gen.writeFileFooter( DO );
		gen.adjustOutputFile();
		createdFileNames.addElement( gen.getOutputFilename() );
	    }
    }

    /**
     * Return the list of filenames created.
     * 
     * @return the list of filenames created by all generators.
     * @author Jay Gunter
     */
    public static String[] getcreatedFileNames() {
	String[] a = new String[ createdFileNames.size() ];
	createdFileNames.copyInto( a );
	return a;
    }


   /**
     * Creates a directory corresponding to the passed package path.
     *
     * @param pkgName
     *   The package for the DataObject currently being processed.
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     */
    private static String mkdir( String outputDir, String pkgName )
    throws Exception {
	// FIX replace the all this exec stuff with
	//     File dir = new File( outputDir + File.separator + dirName );
	//     dir.mkdir();
	// Will it create missing intermediate dirs like mkdir -p?
/* System.err.println("mkdir pkgName="+pkgName); */

	String dirName = pkgName;
	dirName = dirName.replace( '.', File.separatorChar );
	File dir = new File( outputDir + File.separator + dirName );
	dir.mkdirs();
	//dirNames.put( dirName, "" );
	boolean found = false;
	for ( int i = 0; ! found && i < dirNames.size(); i++ ) {
	    String s = (String) dirNames.elementAt( i );
	    if ( null != s && s.equals( dirName ) )
		found = true;
	}
	if ( ! found )
	    dirNames.addElement( dirName );
	numDirs++;
	return dirName;
    }

   /**
     * Invokes a shell script that creates a makefile in each output directory.
     * The script then runs 'make' in the top output directory.
     *
     * @exception Exception
     *   If an unrecoverable error occurs.
     *
     */
    private static void generateMakefiles() throws Exception {
	if ( numDirs == 0 ) return;

	//Enumeration dirs = dirNames.keys();
	//String[] argv = new String[ numDirs + 7 ];
	//String[] argv = new String[ dirNames.size() + 7 ];
	Vector args = new Vector();

	//argv[0] = makefileScriptPath;
//System.err.println( "makefileScriptPath="+ makefileScriptPath);
	String os = System.getProperty( "os.name" );
	if ( -1 != os.indexOf( "Windows" ) )
	    args.addElement( "bash" );
	/*
	else
	    args.addElement( "/bin/sh" );
	    */
	args.addElement( fixDodsPath( "DODS_HOME/makefileGenerator.sh" ) );

	//if ( null == argv[0] || 0 == argv[0].length() )
	if ( null == makefileScriptPath || 0 == makefileScriptPath.length() )
            throw new Exception ("Bad value for MAKEFILE_GENERATOR_SCRIPT." );

	args.addElement( fixScriptOption( makefileScriptRun ) );
	args.addElement( fixScriptOption( makefileScriptCleanup ) );
	args.addElement( fixPath( outputDir ) );
	args.addElement( fixPath( rootDir ) );
//System.err.println( "makefileExtraClasspath="+ makefileExtraClasspath );

//	args.addElement( fixDodsPath( "DODS_HOME/../../../../../output/lib/DODesignStudio.jar"  ) + makefileExtraClasspath );
	args.addElement( fixPath( makefileExtraClasspath ) );

	args.addElement( fixPath( lutrisDir ) );

	for ( int i = 0; i < dirNames.size(); i++ )
	    //argv[ i + 7 ] = (String) dirNames.elementAt( i );
	    args.addElement( dirNames.elementAt( i ) );
	String[] argv = new String[ args.size() ];
	args.copyInto( argv );

//for ( int i = 0; i < argv.length; i++ )
//System.err.println( "i="+i+", s="+argv[i]);

        Process p = Runtime.getRuntime().exec( argv );

	/* Giuseppe Persiani (gpersiani@insinet.it) contributed the following
	 * code to read the output of the makefile generation process.
	 * Jay made minor modifications.
	 * Thanks also to Ben Willis for determining that the bug is
	 * actually in the JDK (waitFor sometimes never returns.)
	 */
	/*
	BufferedInputStream bins = new BufferedInputStream(p.getInputStream());
	BufferedInputStream bers = new BufferedInputStream(p.getErrorStream());
	*/
	InputStream bins = p.getInputStream();
	InputStream bers = p.getErrorStream();

	int ex = 777;
	while( true ) {
	    try {
	       // Sleep for 1 second waiting the completing of the process
	       Thread.currentThread().sleep(1000);
		for ( int ch = bins.read(); 0 <= ch; ch = bins.read() )
		    print( (char) ch );
		for ( int ch = bers.read(); 0 <= ch; ch = bers.read() )
		    print( (char) ch );
		// Every call to Process.exitValue() will throw an
		// IllegalThreadStateException if the process is running.
		// We catch and ignore it to continuing waiting.
		// When the process finishes ex will contain the return value.
		ex = p.exitValue();
		print( "\nDODS BUILD COMPLETE (exit code " + ex + ")\n" );
		break;
	    }
	    catch( IllegalThreadStateException  e ) {
		    //System.err.println("ERROR run: " + e.getMessage() );
	    }
	}
	bins.close();
	bers.close();

/* 	int status = p.waitFor(); */
        if (ex != 0 && ex != 777) {
            throw new Exception ("Cannot run " + argv[0] );
        }
    }

    static private JTextArea textArea = null;
    static public void setTextArea( JTextArea t ) {
	textArea = t;
    }

    static private final int bufSize = 200;
    static private StringBuffer buf = new StringBuffer( bufSize );
    static private int size = 0;
    public static void print( char ch ) {
	buf.append( ch );
	size++;
	if ( '\n' == ch /* || size >= bufSize */ ) {
	    print( buf.toString() );
	    size = 0;
	    buf = new StringBuffer( bufSize );
	}
    }

    public static void print( String str ) {
	if ( ! str.endsWith( "\n" ) )
	    str += "\n";
	if ( null != textArea ) {
	    textArea.append( str );
	} else {
	    System.out.print( str );
	}
    }

    private static String fixScriptOption( String optionValue )
    throws Exception {
	if ( null == optionValue || 0 == optionValue.length() )
	    return "-unset";
	else
	    return optionValue;
    }

    private static String fixPath( String path ) {
	return path.replace( '\\', '/' );
    }

    /**
     * fixOutputDir
     *
     * The Enhydra utility 'newapp' creates a stub source tree
     * for a new Enhydra application.  Some directories in that tree are:
     *     app/app/presentation
     *     app/app/business
     *     app/app/data
     * When the DODS user clicks the Build button, DODS asks for
     * the directory into which the generated source code will be written.
     * It is often desirable to enter the path to the data/ directory
     * of an Enhydra application being developed:
     *     /home/jay/Projects/App/app/app/data/
     *
     * When the package hierarchy in the DODS project is
     *                                app.data[...]
     * then fixOutputDir silently truncates the output directory to
     *     /home/jay/Projects/App/app/
     * so that the source files are written to
     *     /home/jay/Projects/App/app/app/data/*.java
     *
     * But if the package hierarchy in the DODS project is
     *                                notapp.data[...]
     *                                       or
     *                                app.notdata[...]
     * then fixOutputDir does not alter the output directory,
     * and the source files are written to
     *     /home/jay/Projects/App/app/app/data/notapp/data/*.java
     *        or
     *     /home/jay/Projects/App/app/app/data/app/notdata/*.java
     */
    public static String fixOutputDir( String pkg ) {
	/*		outputDir	pkg	return	    pkg goes in
	 *		---------	---	------	    -----------
	 * case 1:	x/a/d		a.d	x/	    x/a/d
	 * case 2:	x/a		a.d	x/	    x/a/d
	 * case 3:	x		a.d	x/	    x/a/d
	 * case 4:	x/x/a		a.d	x/x/	    x/x/a/d
	 * case 5:	x/a/a		a.d	x/a/	    x/a/a/d
	 * case 6:	x/a/x/a		a.d	x/a/x/	    x/a/x/a/d
	 * case 7:	x/a/d/q		a.d	x/a/d/q/    x/a/d/q/a/d
	 * case 8:	x/a/d		a	x/a/d/	    x/a/d/a
	 * case 9:	x/a		b.d	x/a/	    x/a/b/d
	 * 
	 * Cases 1 thru 6 are "normal".
	 * 
	 * In other cases, the output directory hierarchy does not match
	 * the package hierarchy, and so we cannot assume the user
	 * wants the package hierarchy integrated into the output tree;
	 * so, we simply place the package hierarchy under the output directory.
	 */ 
	String	dir	= outputDir.replace( '\\', '/' ).trim();
	String	pl	= pkg.replace( '.', '/' ).trim();
	int	plen	= pl.length();
	int	x	= -1;
	int	sp	= -1;
	while ( -1 == ( x = dir.lastIndexOf( pl ) ) ) {
	    // pkg hierarchy does not match the ending of the outputDir
	    if ( -1 == ( sp = pl.lastIndexOf( '/' ) ) )
		break; // we've run out of pkg components
	    // truncate last pkg component
	    pl = pl.substring( 0, pl.lastIndexOf( '/' ) );
	}

	if ( -1 != x && x + plen >= dir.length() )
	    // all or leading part of pkg hierarchy matches ending of dir
	    dir = dir.substring( 0, x );  // truncate dir at point of match
	else
	    dir += '/';
	if ( -1 == x )
	    print(  "WARNING!!!  Package " + pkg + 
		    "\ndoes not fit into target directory " + outputDir + "." );
	return dir;
    }
    public static String fixOutputDir( DataObject DO ) {
	String pp = DO.getAdjustedPackageName();
	return fixOutputDir( pp );
    }
    public static void testFix( String[] argv ) {
	    outputDir = argv[0];
System.err.println("argv[0]="+argv[0]);
System.err.println("argv[1]="+argv[1]);
	    String s = fixOutputDir( argv[1] );
System.err.println("s ="+s+"\n" );
    }
}
