/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: DOTreeModel.java,v 1.9 2000/01/29 07:01:42 markd Exp $
 *-----------------------------------------------------------------------------
 */



package com.lutris.dods.gui.tree;

import com.lutris.dods.gui.*;
import com.lutris.dods.exceptions.*;
import com.lutris.dods.store.doa.*;

import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.event.TreeModelListener;
import java.util.Vector;
/**
 * This Tree Model organizes everything for the tree in how it is layed out
 * and what you see.
 *
 * @author  Philip Martin
 * @version $Revision: 1.9 $
 */
public class DOTreeModel implements TreeModel {
    /**
     * The root node of the tree.
     */
    protected DOTreeNode root;
    /**
     * The listeners for the model.
     */
    protected Vector listeners;
    /**
     * The DataObjectApplictation parent.
     */
    protected DataObjectApplication doa;
    /**
     * Initializes the appropriate variables.
     */
    public DOTreeModel(DOTreeNode r, DataObjectApplication da) {
        root = r;
        listeners = new Vector();
        doa = da;
    }
    /**
     * When the tree is edited, this updates the DataObjects.
     */
    public void valueForPathChanged(TreePath path, Object newValue) {
        try {
            ((DOTreeNode)path.getLastPathComponent()).getDataObject().setName(newValue.toString());
        }
        catch(DataObjectNameException e) {
           e.printStackTrace();
            doa.namingExceptionOccurred(e, doa);
            return;
        }
        ((DOPackage)root.getDataObject()).updatePackageName();
        doa.updateTree();
    }
    /**
     * Removes the given Listener from this model.
     */
    public void removeTreeModelListener(TreeModelListener l) {
        listeners.removeElement(l);
    }
    /**
     * Returns true if the given node is a leaf.
     */
    public boolean isLeaf(Object node) {
        if(node instanceof DOTreeNode) {
            return ((DOTreeNode)node).isLeaf();
        }
        return true;
    }
    /**
     * Returns the root node.
     */
    public Object getRoot() {
        return root;
    }
    /**
     * Returns the index of the given child or -1 if it is not in the parent.
     */
    public int getIndexOfChild(Object parent, Object child) {
        if(parent instanceof DOTreeNode) {
            try {
                return ((DOTreeNode)parent).getChildren().indexOf(child);
            }
            catch(Exception e) {
               e.printStackTrace();
            }
        }
        return -1;
    }
    /**
     * Returns the number of children in the parent node.
     */
    public int getChildCount(Object parent) {
        if(parent instanceof DOTreeNode) {
            try {
                return ((DOTreeNode)parent).getChildren().size();
            }
            catch(Exception e) {
              e.printStackTrace();
            }
        }
        return 0;
    }
    /**
     * returns the child at the given index.
     */
    public Object getChild(Object parent, int index) {
        if(parent instanceof DOTreeNode) {
            try {
                return ((DOTreeNode)parent).getChildren().elementAt(index);
            }
            catch(Exception e) {
              e.printStackTrace();
            }
        }
        return null;
    }
    /**
     * Adds the Listener to this model.
     */
    public void addTreeModelListener(TreeModelListener l) {
        if(l != null && !listeners.contains(l)) {
            listeners.addElement(l);
        }
    }
    /**
     * This method looks through all of the tree which descends from this
     * node, and returns the TreePath which contains the given DataObject.
     * in it's final node.
     */
    public TreePath findPathWithDO(DOTreeNode dotnode, DataObject data, Vector objects) {
        if(objects == null) {
            objects = new Vector();
        }
        objects.addElement(dotnode);
        Object[] objectArray = new Object[objects.size()];
        objects.copyInto(objectArray);
        TreePath returnValue = new TreePath(objectArray);
        Vector children = dotnode.getChildren();
        int childrenSize = 0;
        if(children == null) {
            return returnValue;
        }
        else {
            childrenSize = children.size();
        }
        for(int i = 0; i < childrenSize ; i++) {
            DOTreeNode dotn = (DOTreeNode)children.elementAt(i);
            if(dotn.contains(data)) {
                return findPathWithDO(dotn, data, objects);
            }
        }
        return returnValue;
    }

    public TreePath getPathContaining(DataObject dataObject) {
      TreePath   foundPath = null;
      DOTreeNode node      = null;
      Vector      objects  = new Vector();
      node = (DOTreeNode)getRoot();
      foundPath = findPathWithDO(node, dataObject, objects);
      return foundPath;
    }

    public DOTreeNode getNodeContaining(Object dataObject) {
      DOTreeNode node = null;
      node = (DOTreeNode)getRoot();
      node = node.getNodeContaining(dataObject);
      return node;
    }
}
