/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: DOMLUpgradeFactory.java,v 1.4.12.1 2000/10/12 00:18:15 markd Exp $
 *-----------------------------------------------------------------------------
 */

package dods.doml;

import dods.data.*;

import java.util.Vector;
import java.util.Enumeration;

import org.w3c.dom.*;

import org.apache.xerces.dom.DOMImplementationImpl;
import org.apache.xerces.parsers.DOMParser;

/**
 * This class is used to create DOML files out of DODS Project trees.
 *
 * @author Jason Abbott (jason@lutris.com)
 */
public class DOMLUpgradeFactory {
  private void appendTextElement(Element target,
                                 String  elementName,
                                 String  text) {
    Document doc = target.getOwnerDocument();
    Element  textElement = doc.createElement(elementName);
    Text     textNode = doc.createTextNode(text);

    target.appendChild(textElement);
    textElement.appendChild(textNode);
  }

  private Element generateAttributeElement(Document  doc,
                                           Attribute attr) {
    Element retVal = doc.createElement("column");

    retVal.setAttribute("id", attr.getName());

    if (attr.isIndex()) {
      retVal.setAttribute("isIndex", "true");
    }
    if (attr.isUsedForQuery()) {
      retVal.setAttribute("usedForQuery", "true");
    }
    if (attr.isConstant()) {
      retVal.setAttribute("isConstant", "true");
    }
    if (attr.hasError()) {
      appendTextElement(retVal, "error", attr.getErrorText());
    }
    String javadoc = attr.getJavadoc();
    if (javadoc != null) {
      String tmpJavaDoc = javadoc.replace('*', ' ').replace('/', ' ');
      if (tmpJavaDoc.trim().length() > 0) {
        appendTextElement(retVal, "javadoc", javadoc);
      }
    }
    DataObject objectRef = (DataObject) attr.getObjectRefDO();
    if (objectRef != null) {
      Element objectRefElement = doc.createElement("referenceObject");
      retVal.appendChild(objectRefElement);
      if (attr.isRefConstraint()) {
        objectRefElement.setAttribute("constraint", "true");
      }
      objectRefElement.setAttribute("reference",
                                    objectRef.getPackageName() +
                                    "." + objectRef.getName());
    }

    Element typeElement = null;
    String javatype = attr.getJavaType();
    if (javatype != null && javatype.trim().length() > 0) {
        if (typeElement == null) {
            typeElement = doc.createElement("type");
        }
        typeElement.setAttribute("javaType", javatype);
    }
    String dbtype = attr.getDBType();
    if (dbtype != null && dbtype.trim().length() > 0) {
        if (typeElement == null) {
            typeElement = doc.createElement("type");
        }
        typeElement.setAttribute("dbType", dbtype);
    }
    String size = attr.getSize();
    if (size != null && size.trim().length() > 0) {
        if (typeElement == null) {
        typeElement = doc.createElement("type");
        }
        typeElement.setAttribute("size", size);
    }

    if (attr.canBeNull()) {
      if (typeElement == null) {
        typeElement = doc.createElement("type");
      }
      typeElement.setAttribute("canBeNull", "true");
    }

    if (typeElement != null) {
      retVal.appendChild(typeElement);
    }


    String initialValue = attr.getInitialValue();
    if (initialValue != null && initialValue.length() > 0) {
      appendTextElement(retVal, "initialValue", initialValue);
    }
    return retVal;
  }

  private void placeDataObjectInDocument(DataObject dataObject,
                                         Element    parentElement) {
    Document doc = parentElement.getOwnerDocument();
    String   nodeType = "table";
    boolean  printTable = true;
    String   dataObjectName = dataObject.getName();

    if (dataObject instanceof DOPackage) {
      printTable = false;
      nodeType = "package";
    }

    Element element = doc.createElement(nodeType);

    parentElement.appendChild(element);
    if (dataObject.getPackageName() != null &&
          dataObject.getPackageName().trim().length() > 0) {
      element.setAttribute("id",
                           dataObject.getPackageName().trim() + "." +
                           dataObjectName);
    } else {
      element.setAttribute("id", dataObjectName);
    }

    if (printTable) {
        if (dataObjectName.equals(dataObject.getDBTableName()) == false) {
            element.setAttribute("dbTableName",
                                 dataObject.getDBTableName());
        }
    }

    if (dataObject.isFinal()) {
      element.setAttribute("isFinal", "true");
    }
    if (dataObject.isLazyLoading()) {
      element.setAttribute("isLazyLoading", "true");
    }
    if (dataObject.isFullCaching())
      element.setAttribute("caching", "full");
    else if (dataObject.isCaching())
      element.setAttribute("caching", "partial");
    if (dataObject.isIndex()) {
      element.setAttribute("isIndex", "true");
    }
    if (dataObject.isAbstract()) {
      element.setAttribute("isAbstract", "true");
    }
    if (false)    {  /* Will be implemented in the future */
      element.setAttribute("isEJB", "true");
    }
    if (false)    {  /* Will be implemented in the future */
      element.setAttribute("isView", "true");
    }
    DataObject extension = dataObject.getExtensionOf();
    if (extension != null) { 
      element.setAttribute("extensionOf",
                           extension.getPackageName() + "." +
                           extension.getName());
    }

    Vector attributes = dataObject.getAttributes();
    if (attributes != null) {
      Enumeration e = attributes.elements();
      while (e.hasMoreElements()) {
        Attribute attribute = (Attribute) e.nextElement();
        element.appendChild(generateAttributeElement(doc, attribute));
      }
    }

    if (dataObject instanceof DOPackage) {
      DOPackage dopackage = (DOPackage) dataObject;
      Vector    vector = dopackage.getDataObjects();
      if (vector != null) {
        Enumeration e = vector.elements();
        while (e.hasMoreElements()) {
          DataObject childDataObject = (DataObject) e.nextElement();
          placeDataObjectInDocument(childDataObject, element);
        }
      }
    }
  }

  /**
   * Constructor. Each factory object should be able to create many documents
   * out of many given projects without having to call for new factory
   * objects (although there is no real saving in doing so).
   */
  public DOMLUpgradeFactory() {
  }

  /**
   * Create a DOM Document out of an existing project data structure.
   *
   * @param project
   *  The DODSProject tree structure that will be read.
   * @return
   *  A Document which is a DOML style DOM object.
   */
  public Document createDocument(DODSProject project)
          throws UpException {

    DOMImplementation domImpl = DOMImplementationImpl.getDOMImplementation();
    DocumentType docType = domImpl.createDocumentType("doml", "doml",
                                                      "doml.dtd");
    Document doc = domImpl.createDocument("project.doml",
                                          "doml",
                                          docType);


    Element  dbElement = doc.createElement("database");
    String   database = project.getDatabase();

    doc.getDocumentElement().appendChild(dbElement);
    placeDataObjectInDocument(project, dbElement);

    if (database != null && database.trim().length() > 0) {
      dbElement.setAttribute("database", database);
    }

    return doc;
  }
}
