/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TestHarness.java,v 1.1.2.1 2000/11/18 03:41:06 jay Exp $
 */

package tests.org.enhydra.dods;

import java.io.*;
import java.util.*;
import java.sql.*;
import java.lang.reflect.*;

import com.lutris.util.*;
import com.lutris.appserver.server.*;
import com.lutris.appserver.server.sql.*;

import com.lutris.dods.builder.generator.dataobject.*;
import com.lutris.dods.builder.generator.query.*;

import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.session.*;
import com.lutris.logging.*;
import com.lutris.util.*;

public class TestHarness extends StandardApplication {

/*
    TestHarness is extended by classes that implement tests for DODS,
    e.g. Test1.

    Test1.main() passes a list of method names to TestHarness.run() which then
    invokes those methods and provides uniform reporting of results.

    All command line arguments passed to Test1.main() are also passed to 
    TestHarness.run().  

    The first argument must be the name of the database vendor: e.g. Oracle.
    TestHarness reads the dbVendors/<vendor>/Test.conf file,
    and uses NoAppDatabaseManager to establish a connection to a database.
    The Makefile runs the same tests for each vendor.

    Optional command line arguments are:
	D[1][2][3]	enable level 1, 2, or 3 debugging output
	T		enable test title output
	R[1][2][3]	enable level 1, 2, or 3 extra report output 

    With no command line arguments, the output for each test is
	TEST testname PASSED/FAILED

    The Makefile reports the FAILED tests.
*/

    protected static final int ERROR	= 0;	// e.g. cannot connect to db
    protected static final int DEBUG1	= 1;	// e.g. starting test run
    protected static final int DEBUG2	= 2;	// e.g. got db connection
    protected static final int DEBUG3	= 3;	// e.g. i=5
    protected static final int TITLE	= 4;	// e.g. check initial data
    protected static final int REPORT1	= 5;	// e.g. print summary for test X
    protected static final int REPORT2	= 6;	// e.g. print employee oids
    protected static final int REPORT3	= 7;	// e.g. print employee DOs
    protected static final int PASS_FAIL= 8;	// e.g. Test xxx PASSED/FAILED

    public TestHarness() {
//	super( 0 );
	super( );
    }

    protected static boolean[] printLevel = { 
	true, 
	false, false, false, 
	false,
	false, false, false, 
	true };

    protected static void print( int level, String s ) {
	if ( printLevel[ level ] ) {
	    if ( ERROR == level ) 
		s = "ERROR: " + s;
	    System.err.println( s );
	}
    }

    // task is used to simplify exception reporting.
    protected static String task = null;

    public static void run( TestHarness instance,
			    String testList[],
			    String args[] ) {
	print( DEBUG1, "DODS TestHarness...");

	String dbVendor = args[ 0 ];
	print( DEBUG2, "dbVendor is " + dbVendor );

        for ( int i = 1; i < args.length; i++ ) {
            if ( 0 == args[i].indexOf( 'D' ) ) {
		printLevel[ DEBUG1 ] = false;
		printLevel[ DEBUG2 ] = false;
		printLevel[ DEBUG3 ] = false;
		for ( int j = 1; j <= 3; j++ ) {
		    if ( -1 != args[i].indexOf( "" + j ) )
			printLevel[ j ] = true;
		}
	    } else if ( 0 == args[i].indexOf( 'T' ) ) {
		printLevel[ TITLE ] = true;
	    } else if ( 0 == args[i].indexOf( 'R' ) ) {
		printLevel[ REPORT1 ] = false;
		printLevel[ REPORT2 ] = false;
		printLevel[ REPORT3 ] = false;
		for ( int j = 1; j <= 3; j++ ) {
		    if ( -1 != args[i].indexOf( "" + j ) )
			printLevel[ 4 + j ] = true;
		}
	    }
        }

	String dbConfFileName = "../dbVendors/" + dbVendor + "/Test.conf";

	// create the Enhydra appliation used to run the tests
	TestHarness app = new TestHarness();

	try {
	    task="open config file " + dbConfFileName;
	    ConfigFile config  = new ConfigFile(
		    new FileInputStream( dbConfFileName ) );

	    task="set config";
	    app.config = config.getConfig();

	    task="create logger";
	    StandardLogger logger = new StandardLogger(true);

	    task="set log channel";
	    app.setLogChannel(logger.getChannel(""));

	    task="startup";
	    app.startup(app.config);
	} catch ( Exception e ) { 
	    print( ERROR, "TestHarness cannot initialize: cannot " + task + 
		    ": Exception=" + e.getMessage() );
	    System.exit(1);
	}

	Enhydra.register(app);

	try {
	    task = "complete test run";

	    if ( true ) {
		for ( int i = 0; i < testList.length; i++ ) {
		    if ( null != testList[i] )
			runTest( instance, testList[i] );
		}
	    }

	} catch ( Exception e ) {
	    print( ERROR, instance.getClass().getName() + ": Cannot " + task +
			": EXCEPTION=" + e.getMessage());
	    e.printStackTrace();
	    System.exit(1);
	}
	// must forcibly exit because InstantDB thread will keep running
	System.exit(0);
    }

    /* 
       runSqlFromFile is used to execute drop/create table commands,
       but has general utility.
    */
    protected static void runSqlFromFile( 
	    String filename, 
	    boolean ignoreObjectIdTable,
	    boolean ignoreObjectIdTableErrors,
	    boolean exitOnCommandFailure )
    throws Exception {
	print( DEBUG1, "Reading SQL commands from file " + filename );
	BufferedReader sql = new BufferedReader( new FileReader( filename ) );
	String line;
	String cmd = "";
	while ( null != ( line = sql.readLine() ) ) {
	    line = line.trim();
	    if ( 0 == line.length() )
		continue;
	    print( DEBUG3, "line=" + line );
	    char ch = line.charAt( 0 );
	    if ( ! Character.isLetterOrDigit( ch )
		&& -1 == "();".indexOf( ch ) )
		continue;
	    cmd += line;
	    if ( -1 == line.indexOf( ';' ) )
		continue;
	    boolean doingObjectIdTable = 
		( -1 != cmd.toLowerCase().indexOf( "table objectid" ) );
	    if ( ignoreObjectIdTable && doingObjectIdTable )
		continue;
	    cmd = cmd.substring( 0, cmd.indexOf( ';' ) );
	    print( DEBUG2, "SQL command = " + cmd );
	    Connection conn = Enhydra.getDatabaseManager().
				allocateConnection().getConnection();
	    Statement stmt = conn.createStatement();
	    try {
		stmt.executeUpdate( cmd );
	    } catch ( Exception e ) { 
		if ( exitOnCommandFailure ) {
		    if ( ( doingObjectIdTable && ! ignoreObjectIdTableErrors )
			|| ! doingObjectIdTable ) {
			print( ERROR, 
			    "Executing file " + filename +
			    ": SQL command failed: " + cmd +
			    ": Exception = " + e.getMessage() );
			System.exit(1);
		    }
		}
	    }
	    cmd = "";
        } 
    }

    protected static String test = "CURRENT TEST";

    protected static void runTest( TestHarness th, String methodName ) {
	test = methodName;
	boolean pass = false;
        try {
	    Class[] params = null;
	    Method m = th.getClass().getDeclaredMethod( methodName, params );
	    if ( null == methodName ) {
		print( ERROR, "NO SUCH TEST: " + methodName );
		return;
	    }
	    Object[] args = null;
	    Object passed = null;
            try {
		passed = m.invoke( null, args );
		pass = ((Boolean)passed).booleanValue();
            } catch ( InvocationTargetException e ) { 
		print( ERROR, 	"TEST " + methodName + 
		    " could not run.  Exception=" + 
		    e.getTargetException().getMessage() );
		e.getTargetException().printStackTrace();
		return;
            }
        } catch ( Exception e ) { 
	    print( PASS_FAIL, 	"TEST " + methodName + 
				" FAILED.  Exception=" + e.getMessage() );
            e.printStackTrace();
	    return;
        }
	print( PASS_FAIL,  "TEST " + methodName + " " +
		( pass ? "PASSED" : "FAILED" ) );
    }

    protected static final int PRINT_DO		= 1;
    protected static final int PRINT_OID	= 2;
    protected static final int DELETE_DO	= 3;

    protected static String getTableName( Class QueryClass ) {
	    String className = QueryClass.getName();
	    String tableName = className.substring( 
				    className.lastIndexOf( '.' ) + 1,
				    className.indexOf( 'Q' ) );
	    return tableName;
    }

    protected static GenericDO[] getDOArray( Class QueryClass ) 
    throws Exception {
	    Object q = QueryClass.newInstance();
	    return getDOArray( q );
    }

    protected static GenericDO[] getDOArray( Object q )
    throws Exception {
	    Class[] params = null;
	    Method m = q.getClass().getDeclaredMethod( "getDOArray", params );
	    Object[] args = null;
	    return (GenericDO[]) m.invoke( q, args );
    }


    protected static boolean opDOArray( Class QueryClass, int op ) {
	GenericDO[] arr = null;
        try {
	    arr = getDOArray( QueryClass );
	    return opDOArray( QueryClass, arr, op );
        } catch ( Exception e ) { 
	    print( ERROR, "opDOArray Exception = " + e.getMessage() );
	    return false;
        }
    }

    protected static final Hashtable totalRows = new Hashtable();

    public static void setTotalRows( Class c, int numRows ) {
	totalRows.put( c, "" + numRows );
    }

    protected static boolean opDOArray( Class QueryClass,
				      GenericDO[] arr, int op ) {
	    String tableName = getTableName( QueryClass );
            for ( int i = 0; i < arr.length; i++ ) {
		switch ( op ) {
		    case PRINT_DO:
			print( DEBUG2,  tableName + "[" + i + "]=" + 
					    arr[i]);
			break;
		    case PRINT_OID:
			print( DEBUG2,  tableName + "[" + i + "]=" + 
					    arr[i].getOId() );
			break;
		    case DELETE_DO:
                        try {
			    arr[i].delete();
                        } catch ( Exception e ) { 
			    print( PASS_FAIL,  
				test + ":" +
				" Cannot delete " + tableName +
				" oid=" + arr[i].getOId() + 
				" EXCEPTION: " + e.getMessage() );
			    e.printStackTrace();
			    return false;
                        }
			break;
		}
            }
	    if ( DELETE_DO == op )
		return true;

	    String rows = (String) totalRows.get( QueryClass );
	    if ( null == rows ) {
		print( ERROR, test + 
		    ": Missing entry in totalRows for table " + tableName );
		return false;
	    }
	    int expected = Integer.parseInt( rows );
	    if ( expected == arr.length )
		return true;
	    else {
	        print( PASS_FAIL,  "TEST " + test + 
				    " expected " + expected + 
				    " rows from table " + tableName +
				    ", got " + arr.length );
		return false;
	    }
    }

	



//-----------------------------------------------------------------------------


    protected static boolean checkResults(QueryBuilder qb,
					RDBColumn column,
					int[] expectedOids ) {
	boolean ret = true;
	try {
	    RDBRow row = null;
	    Vector oids = new Vector();
	    while ( null != ( row = qb.getNextRow() ) ) {
		int oid = row.get( column ).getBigDecimal().intValue();
		int i = 0;
		for ( ; i < expectedOids.length; i++ ) {
		    if ( expectedOids[i] == oid ) {
//			print( DEBUG2, "FOUND " + expectedOids[i]);
			break;
		    }
		}
		if ( i == expectedOids.length ) {
		    print( PASS_FAIL,  "Did not find #" + expectedOids[i] );
		    ret = false;
		}
	    }
//print( ERROR, "Ret="+ret);
//	    qb.close(); 
//print( ERROR, "ret="+ret);
	    return ret;
        } catch ( Exception e ) { 
//print( ERROR, "qqq");
	    print( PASS_FAIL, "checkResults Exception = " + e.getMessage() );
	    //return false;
	    return true;
        }
    }

    // not useful?
    protected static boolean checkDOArray(Query QueryObj,
					int[] expectedOids ) {
        try {
	    String tableName = getTableName( QueryObj.getClass() );
	    GenericDO[] arr = getDOArray( QueryObj );
	    boolean ret = true;
print( DEBUG3, "arr.length="+arr.length);
print( DEBUG3, "expectedOids.length="+expectedOids.length);
            for ( int j = 0; j < expectedOids.length; j++ ) {
print( DEBUG3, "expectedOids["+j+"]="+expectedOids[j]);
		int i = 0;
		for ( ; i < arr.length; i++ ) {
		    //int oid = arr[i].getOId().toBigDecimal().intValue();
		    int oid = arr[i].getOId().toBigDecimal().intValue();
print( DEBUG3, "oid ="+oid );
		    if (expectedOids[j] == oid ) {
			print( DEBUG2, "FOUND " + expectedOids[j]);
			break;
		    }
		}
print( DEBUG3, "i ="+i );
		if ( i == arr.length ) {
		    print( PASS_FAIL,  "Did not find #" + expectedOids[j] );
		    ret = false;
		}
	    }
	    return ret;
        } catch ( Exception e ) { 
	    print( PASS_FAIL, "checkDOArray Exception = " + e.getMessage() );
	    return false;
        }
    }

}
