/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: InformixDBQuery.java,v 1.9.12.1 2000/10/19 17:59:05 jasona Exp $
 */





package	com.lutris.appserver.server.sql.informix;

import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.*;
import com.lutris.logging.*;
import java.sql.*;
import java.util.*;

/**
 * An Informix database query object.
 *
 * @author	Kyle Clark
 * @since	LBS1.7
 * @version	$Revision: 1.9.12.1 $
 */
public class InformixDBQuery extends StandardDBQuery {

    /**
     * Amount of time (seconds) to wait on locks before
     * throwing an exception.
     * @see setQueryTimeOut
     */
    private int queryTimeout = 0;

    /**
     * The informix database connection.
     */
    protected InformixDBConnection conn;

    /**
     * Private constructor.
     *
     * @param dbConnection
     *   The database connection to use.
     * @exception SQLException
     *   If a database access error occurs. 
     */
    protected InformixDBQuery(InformixDBConnection dbConnection)
        throws SQLException
    {
        super(dbConnection);
        this.conn = dbConnection;
        try {
            setQueryTimeout(dbConnection.logicalDatabase.getDefaultQueryTimeout());
        } catch (SQLException sqlExcept) {
            handleException(sqlExcept);
            dbConnection.release();
            throw sqlExcept;
        }
    }


    /**
     * Public constructor.  Establishes a connection with the default
     * logical database.  The connection manager must have been
     * initialized with a default logical database name.
     *
     *
     * @exception SQLException
     *   If a database access error occurs. 
     * @exception DatabaseManagerException
     *   If the default logical database wasn't initialized in the
     *   Database Manager.
     * @see
     *   com.lutris.appserver.server.sql.DatabaseManager#setDefaultDatabase
     */
    public InformixDBQuery()
        throws SQLException, DatabaseManagerException
    {
        this((InformixDBConnection)
             LBS.getApplication().getDatabaseManager().allocateConnection());
    }


    /**
     * Public constructor.  Establishes a connection to the specified
     * logical database name.
     *
     * @param dbName
     *   Logical database name to access.
     * @exception SQLException
     *   If a database access error occurs. 
     * @exception DatabaseManagerException
     *   If a nonexistant logical database name is supplied.
     */
    public InformixDBQuery(String dbName)
        throws SQLException, DatabaseManagerException
    {
        this((InformixDBConnection)LBS.getApplication().getDatabaseManager().
             allocateConnection(dbName));
    }

    /**
     * Set query time out.  If a query blocks, then it will only
     * block as long as specified here.
     *
     * @param seconds
     *   The amount of time to block.
     * @exception SQLException if a database error occurs.
     */
    public void setQueryTimeout(int seconds) throws SQLException
    {
        try {
            logDebug("set query time out to " + seconds);
            conn.setLockModeToWait(queryTimeout);
            queryTimeout = seconds;
        }
        catch (SQLException e) {
            handleException(e);
            throw e;
        }
    }

}
