/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: InformixDBTransaction.java,v 1.9.12.1 2000/10/19 17:59:05 jasona Exp $
 */





package	com.lutris.appserver.server.sql.informix;

import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.*;
import com.lutris.logging.*;
import java.sql.*;
import java.util.*;

/**
 * Informix implementation of SQL database transaction.
 *
 * @author	Kyle Clark
 * @since	LBS1.8
 * @version	$Revision: 1.9.12.1 $
 */
public class InformixDBTransaction extends StandardDBTransaction {

    /**
     * The connection to the informix database.
     */
    protected InformixDBConnection conn;

    /**
     * Private constructor. Used if connection is already known.
     *
     * @param conn
     *   The database connection to use.
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     */
    protected InformixDBTransaction(InformixDBConnection conn)
        throws SQLException
    {
        super(conn);
        this.conn = conn;
        try {
            setTransactionTimeout(conn.logicalDatabase.
                                  getDefaultTransactionTimeout());
        } catch (SQLException sqlExcept) {
            conn.handleException(sqlExcept);
            conn.release();
            throw sqlExcept;
        }
    }


    /**
     * Public constructor.  Establishes a connection with the default
     * logical database.  
     * The connection manager must have been initialized with
     * a default logical database name.
     *
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     * @exception DatabaseManagerException
     *   If the default logical database wasn't initialized in the
     *   Database Manager.
     * @exception ClassCastException
     *   If the default database manager doesn't allocate connections
     *   of type InformixDBConnection.
     * @see DatabaseManager#setDefaultDatabase
     */
    public InformixDBTransaction()
	throws SQLException, DatabaseManagerException {

        this((InformixDBConnection)
             LBS.getApplication().getDatabaseManager().allocateConnection());

    }

    /**
     * Public constructor.  Establishes a connection with the specified
     * logical database name.
     *
     * @param dbName
     *   Logical database name to access.
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     */
    public InformixDBTransaction(String dbName)
	throws SQLException, DatabaseManagerException {

        this((InformixDBConnection)LBS.getApplication().getDatabaseManager().
             allocateConnection(dbName));
    }

    /**
     * Set transaction time out.  If a transaction blocks, then it will only
     * block as long as specified here.
     *
     * @param seconds
     *   The amount of time to block in seconds.  If less than zero then blocks
     *   indefinitly.
     * @exception SQLException if a database error occurs.
     */
    public void setTransactionTimeout(int seconds) throws SQLException {
        try {
            logDebug("set transaction time out to " + seconds);
            conn.setLockModeToWait(seconds);
        } catch (SQLException e) {
            handleException(e);
            throw e;
        }
    }
}
