/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: InformixLogicalDatabase.java,v 1.8.4.1 2000/10/19 17:59:05 jasona Exp $
 */





package	com.lutris.appserver.server.sql.informix;

import java.util.Date;	// Must be before java.sql.Date
import java.sql.*;
import com.lutris.util.*;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.*;

/**
 * Represents a logical database with Informix capabilities.
 *
 * @author	Kyle Clark
 * @since	LBS1.8
 * @version	$Revision: 1.8.4.1 $
 */
public class InformixLogicalDatabase extends StandardLogicalDatabase {

    /**
     * The amount of time (in seconds) that
     * a query will block before throwing an exception. If <= 0 then the
     * query will not block. Optional, if not specified, then the value
     * defaults to 0.
     */
    protected int defaultQueryTimeout = 0;

    /**
     * The amount of time (in seconds)
     * that a transaction will block before throwing an exception. If
     * <= 0 then the transaction will not block. Optional, if not specified,
     * then the value defaults to 0.
     */
    protected int defaultTransactionTimeout = 0;

    /**
     * Default constructor to configure a single logical database.
     * Note that the <CODE>init()</CODE> method must be called to
     * configure the database.
     */
    public InformixLogicalDatabase () {
        super();
    }

    /**
     * Creates and configures a single logical database.
     *
     * @param dbName
     *   The logical name of the database.  Used to
     * @param dbConfig
     *   The configuration object for this logical Informix database.
     * <p>
     * The configuration data is specified in the section:
     * <CODE>DatabaseManager.DB.<I>dbName</I></CODE>
     * <p>
     * The Informix specific configuration parameters are:
     * <UL>
     * <LI> <CODE>QueryTimeout</CODE> - The amount of time (in seconds) that
     *	a query will block before throwing an exception. If < 0 then the
     *	query will block indefinitly. Optional, if not specified, then the value
     *	defaults to 0.
     * <LI> <CODE>TransactionTimeout</CODE> - The amount of time (in seconds)
     *	that a transaction will block before throwing an exception. If
     *	< 0 then the transaction will block indefinitly. Optional, if not specified,
     *	then the value defaults to 0.
     * </UL>
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @exception SQLException
     *   If a SQL error occurs.
     */
    public InformixLogicalDatabase (String dbName, Config dbConfig)
        throws ConfigException, SQLException {

        super(dbName, dbConfig);
        this.defaultQueryTimeout = dbConfig.getInt("QueryTimeout", 0);
        this.defaultTransactionTimeout =
            dbConfig.getInt("TransactionTimeout", 0);
    }
    

    /**
     * Return the connection allocator.
     *
     * @param connectionConfig
     *   The configuration object for the connection allocator.
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @return
     *   The connection allocator.
     */
    public ConnectionAllocator loadConnectionAllocator(Config connectionConfig)
        throws ConfigException {

	return new InformixConnectionAllocator (this, connectionConfig);

    }


    /**
     * Return the object id allocator.
     *
     * @param objIdConfig
     *   The configuration object for the object id allocator.
     * @exception ConfigException
     *   If there is an error in the configuration file.
     * @return
     *   The object id allocator.
     */
    public ObjectIdAllocator loadObjectIdAllocator (Config objIdConfig)
        throws ConfigException {

	return new InformixObjectIdAllocator (this, objIdConfig);

    }


    /**
     * Return a description of the logical database type.
     *
     * @return The type.
     */
    public String getType () {
	return "Informix JDBC";
    }

    /**
     * Returns the configured default amount of time (in seconds) that
     * a query should block before throwing an exception. If < 0 then the
     * query should block indefinitly.
     *
     * @return the default query timeout.
     */
    protected int getDefaultQueryTimeout() {
        return defaultQueryTimeout;
    }

    /**
     * Returns the configured default amount of time (in seconds)
     * that a transaction should block before throwing an exception. If
     * less than zero then the transaction should block indefinitly.
     *
     * @return the default transaction timeout.
     */
    protected int getDefaultTransactionTimeout() {
        return defaultTransactionTimeout;
    }

}
