/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardDBTransaction.java,v 1.7.12.3 2000/11/30 00:00:22 jay Exp $
 */





package	com.lutris.appserver.server.sql.standard;

import com.lutris.appserver.server.Enhydra;
import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.logging.*;
import java.sql.*;
import java.util.*;

/**
 * Standard implementation of SQL database transaction.
 *
 * @see		DBTransaction
 * @author	Kyle Clark
 * @since	LBS1.8
 * @version	$Revision: 1.7.12.3 $
 */
public class StandardDBTransaction implements DBTransaction {

    // Identifier for this transaction object.
    private int id;

    // Next available identifier.
    private static int nextId;

    // Connection used by this transaction object.
    private DBConnection conn;

    // Is this object still active, or has it been released.
    private boolean released = false;

    // Objects that make up this transaction
    private Transaction[] trans = {};

    // Action (INSERT, UPDATE, DELETE) to apply to each object
    // in the objects array.
    private int[] transAction = {};

    // Index of next action to apply
    private int transIdx = 0;

    // Amount by which to increment the array size if it is full.
    private final int SIZE_DELTA = 10;

    /**
     * The log channel.
     */
    private LogChannel log;

    private static final int INSERT = 1;
    private static final int UPDATE = 2;
    private static final int DELETE = 3;


    /**
     * Construct a transaction object for use on the supplied dB connection.
     *
     * @param conn
     *   The database connection to use.
     * @exception java.sql.SQLException
     *   If a database access error occurs.
     */
    protected StandardDBTransaction(DBConnection conn)
            throws SQLException {

        try {
            id = nextId++;
            logDebug("new instance");
            this.conn = conn;
            this.conn.setAutoCommit(false);
        } catch (SQLException sqlExcept) {
            this.conn.handleException(sqlExcept);
            this.conn.release();
            throw sqlExcept;
        }
    }


//    /**
//     * Public constructor.  Establishes a connection with the default
//     * logical database.
//     * The connection manager must have been initialized with
//     * a default logical database name.
//     *
//     * @exception java.sql.SQLException
//     *   If a database access error occurs.
//     * @exception DatabaseManagerException
//     *   If the default logical database wasn't initialized in the
//     *   Database Manager.
//     * @see DatabaseManager#setDefaultDatabase
//     */
//    public StandardDBTransaction()
//	throws SQLException, DatabaseManagerException {
//
//        this(LBS.getApplication().getDatabaseManager().allocateConnection());
//    }
//
//
//    /**
//     * Public constructor.  Establishes a connection with the specified
//     * logical database name.
//     *
//     * @param dbName
//     *   Logical database name to access.
//     * @exception java.sql.SQLException
//     *   If a database access error occurs.
//     * @exception DatabaseManagerException
//     *   If a nonexistent logical database name is supplied.
//     */
//    public StandardDBTransaction(String dbName)
//	throws SQLException, DatabaseManagerException {
//
//        this(LBS.getApplication().getDatabaseManager().allocateConnection(dbName));
//    }


    /**
     * Method to add an object to the transaction list.
     *
     * @param transaction
     *   Object that implements transaction interface.
     */
    private synchronized void add(Transaction transaction, int action) {
        while (trans.length <= transIdx) {
           Transaction [] newTrans =
             new Transaction[trans.length + SIZE_DELTA];
           int [] newTransAction = new int[transAction.length + SIZE_DELTA];
           for (int idx=0; idx<trans.length; idx++) {
             newTrans[idx] = trans[idx];
             newTransAction[idx] = transAction[idx];
           }
           trans = newTrans;
           transAction = newTransAction;
         }
         trans[transIdx] = transaction;
         transAction[transIdx] = action;
         transIdx++;
    }


    /**
     * Method to update an object in the database.
     *
     * @param transaction
     *   Object that implements transaction interface.
     */
    public void update(Transaction transaction) {
        add(transaction, UPDATE);
    }


    /**
     * Method to delete an object in the database.
     *
     * @param transaction
     *   Object that implements transaction interface.
     */
    public void delete(Transaction transaction) {
	// If the transaction is an instance of CoreDO and we're
	// doing a DELETE, then check if the oid of this CoreDO
	// is the same as the oid of an existing CoreDO in our
	// transaction. If this is true, do not add this
	// transaction because we are already deleting the DO.
	// This will happen occassionally during complex CASCADE DELETE
	// scenarios when the programmer passes their own
	// DBTransaction to the ***DO.delete() methods.
	if(transaction instanceof CoreDO) {
	    ObjectId oid = ((CoreDO)transaction).getOId();
	    for(int idx=0; idx < transIdx; idx++) {
	    	if(trans[idx] instanceof CoreDO)
		    if(((CoreDO)trans[idx]).getOId().equals(oid))
			return;
	    }
	}
        add(transaction, DELETE);
    }


    /**
     * Method to insert an object in the database.
     *
     * @param transaction
     *   Object that implements transaction interface.
     */
    public void insert(Transaction transaction) {
        add(transaction, INSERT);
    }


    /**
     * Method to commit upates.
     *
     * @exception java.sql.SQLException If a database access error occurs.
     * @exception DBRowUpdateException If a version error occurs.

     */
    public void commit() throws SQLException, DBRowUpdateException {
        try {
            logDebug("commit");
            validate();
	    conn.incrRequestCount(); // Record request
            for (int idx=0; idx<transIdx; idx++) {
              switch (transAction[idx]) {
              case INSERT:
                trans[idx].executeInsert(conn);
                break;
              case UPDATE:
                trans[idx].executeUpdate(conn);
                break;
              case DELETE:
                trans[idx].executeDelete(conn);
                break;
              }
            }
            conn.commit();
            // Notify all objects that the transaction succeeded.
            transactionNotify(true);
        } catch (SQLException sqlExcept) {
            handleException(sqlExcept);
            rollback();
            throw sqlExcept;
        } finally {
            transIdx = 0;
        }
    }


    /**
     * Method to rollback changes.
     *
     * @exception java.sql.SQLException
     *   If a database access error occurs.
     */
    public void rollback() throws SQLException {
        try {
            logDebug("rollback");
            validate();
            conn.rollback();
        } catch (SQLException sqlExcept) {
            handleException(sqlExcept);
            throw sqlExcept;
        } finally {
            // Notify all objects that the transaction failed.
            transactionNotify(false);
        }
    }


    /**
     * Notify all objects in this transaction if the
     * transaction was successfully commited to the database.
     * An object will only be notified once of success or failure
     * even if this method is called multiple times.
     *
     * @param succeeded true if the transaction commited successfully.
     */
    private synchronized void transactionNotify(boolean succeeded) {
        for (int idx=0; idx<transIdx; idx++) {
          switch(transAction[idx]) {
          case INSERT:
            trans[idx].finalizeInsert(succeeded);
            break;
          case UPDATE:
            trans[idx].finalizeUpdate(succeeded);
            break;
          case DELETE:
            trans[idx].finalizeDelete(succeeded);
            break;
          }
        }
        transIdx = 0;
    }


    /**
     * Frees all resources consumed by this transaction
     * Connections are returned to the connection pool.
     * Subsequent transactions via this object,
     * will allocate a new set of resources (i.e. connection).
     */
    public synchronized void release() {
        try {
            logDebug("release");
            if (!released) {
                conn.reset();
            }
        } catch (SQLException sqlExcept) {
            handleException(sqlExcept);
        } finally {
            conn.release();
            conn = null;
            released = true;
            transIdx = 0;
        }
    }


    /**
     * Exception handeler.  This object is should not be
     * used for subsequent queries if this method returns
     * false.
     *
     * @return boolean True if the exception can be handeled
     *   and the object is still valid, false otherwise.
     */
    public synchronized boolean handleException(SQLException e) {
        logDebug("handle exception");
        return conn.handleException(e);
    }


    /**
     * Method to ensure this object is still valid.
     * Once this object has been released it cannot be
     * used any more.
     *
     * @exception java.sql.SQLException
     *   If a database access error occurs.
     */
    private void validate() throws SQLException {
        if (released)
            throw new SQLException("Cannot access DBTransaction object " +
                                   "once it has been released.");
    }


    /**
     * If this object has not been <A HREF=#release>released</A>,
     * this method ensures that garbage collection does so.
     */
    protected void finalize() {
        if (!released) {
            release();
        }
    }


    /**
     * Logging. For debuging only, since it effects all Query objects.
     *
     * @param str
     *   The data to log.
     */
    protected void logDebug(String str) {
        if (Enhydra.getDatabaseManager().debug) {
	    if (log == null) log = Enhydra.getLogChannel();
            log.write(Logger.DEBUG, "DBTransaction[" + id + "]: " + str);
        }
    }
}
