/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardAppUtil.java,v 1.24.2.4.4.2 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server;

import java.net.*;
import java.util.StringTokenizer;
import com.lutris.util.*;
import com.lutris.http.*;
import com.lutris.appserver.server.session.*;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.xml.xmlc.html.*;
import org.w3c.dom.html.*;
import org.w3c.dom.*;
import javax.servlet.http.Cookie;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;


/**
 * Static methods used by <CODE>StandardApplication</CODE>.  These mostly
 * are used to manage sessions.
 *
 * @version	$Revision: 1.24.2.4.4.2 $
 * @author	Mark Diekhans
 * @author	Richard Kunze <kunze@opus5.de>
 */
public class StandardAppUtil {

    /*
     * Constant used for encoding the session in a url rewrite.
     */
    public static final String ENHYDRA_SESSION_ID_URL = "jsessionid";
    public static final String ENHYDRA_SESSION_ID_COOKIE = "JSESSIONID";

    /**
     * Prevent instantiation.
     */
    private StandardAppUtil() {
    }

    /**
     * Get internet peer address.
     *
     * @param comms The presentations manager comms object with
     *  all request/response information in it.
     * @return The peer address.
     * @exception ApplicationException If address can't be obtained.
     */
    //FIXME: no longer needed, but might make a nice general util function.
    public static InetAddress[] getPeerAddress(HttpPresentationComms comms)
            throws ApplicationException {

        InetAddress[] peer = new InetAddress[1];
        String remoteHost;
        try {
            remoteHost = comms.request.getRemoteHost();
        } catch (HttpPresentationException except) {
            throw new ApplicationException(except);
        }
        if (remoteHost == null) {
            throw new ApplicationException("Can't find remote host for request.");
        }
        if (remoteHost != null) {
            try {
                peer[0] = InetAddress.getByName(remoteHost);
            } catch (UnknownHostException except) {
                throw new ApplicationException("Can't find remote address for \"" + remoteHost + "\"", except);
            }
        }
        return peer;
    }

    /**
     * Search for a session cookie with the specified application name.
     *
     * @param applicationName Application name that is saved in the cookie.
     * @param cookies Array of cookies for the current request.
     * @return The session key decoded from the cookie, or null if
     *  the session was not found.
     * @deprecated
     */
    private static String findSessionKey(String applicationName,
                                         Cookie[] cookies) {
	return findSessionKey(cookies);
    }

    /**
     * Search for a session cookie with the name jsessionid.
     *
     * @param applicationName Application name that is saved in the cookie.
     * @param cookies Array of cookies for the current request.
     * @return The session key decoded from the cookie, or null if
     *  the session was not found.
     * @deprecated
     */
    private static String findSessionKey(Cookie[] cookies) {
        if (cookies == null) {
            return null;
        }
	for (int i=0; i<cookies.length; i++) {
	    if (cookies[i].getName().equalsIgnoreCase(ENHYDRA_SESSION_ID_COOKIE)) {
                return cookies[i].getValue();
	    }
	}
        return null;
    }

    /**
     * Encode a URL with a session identifier.
     *
     * @param url The url to encode.
     * @param id The session identifier to encode with the url.
     */
    public static String encodeUrl(String url, String id) {
        String pre = url;
        String post = "";
        int pos;
        // Need to put the parameters before a local anchor
        if ((pos = url.indexOf('?')) != -1 ||
            (pos = url.indexOf('#')) != -1) {
            pre  = url.substring(0, pos);
            post = url.substring(pos);
        }
	return pre + ';' + ENHYDRA_SESSION_ID_URL + '=' + id + post;
    }

    /**
     * Determine if a URL references a presentation object. Used for
     * URL rewriting to test if a session ID should be appended to a
     * given URL. Takes the cheap way out at the moment in that it
     * simply returns true iff (if and only if, for the
     * non-mathematicians :-) the URL contains the character sequence
     * <code>.po</code>.
     * <br>
     * TODO: Implement a better check
     *
     * @param url The url to check
     * @return True if the URL points to a PO, false otherwise
     */
    public static boolean pointsToPO(String url) {
        int idx = url.indexOf(".po");
        if (idx == -1) {
            return false;
        }
        if (idx == url.length()-3) {
            return true;
        }
        switch (url.charAt(idx+3)) {
            case '?':
            case '#':
                return true;
            case '/':
                // A bit more difficult: Need to decide if it's in the
                // the server or URI part of the URL.
                try {
                    URL u = new URL(url);
                    int i = u.getFile().indexOf(".po");
                    return (i != -1);
                } catch (MalformedURLException e) {
                    // url doesn't contain a schema and/or server
                    // part, so it must be relative. OK, found a PO.
                    return true;
                }
            default:
                return false;
        }
    }


    /**
     * Determine if a session exists for this request.  Does not
     * validate that a session is logged in.
     *
     * @param comms The presentations manager comms object with
     *  all request/response information in it.
     * @return The session object or null if a session cookie or
     *  matching session is not found.
     * @exception ApplicationException If address can't be obtained.
     */
    public static Session getRequestSession(HttpPresentationComms comms)
            throws ApplicationException {
        Session s = null;
        String sessionKey = null;
        Cookie[] cookies;
	StandardSessionManager sessionManager =
	    (StandardSessionManager)comms.application.getSessionManager();

        try {
	    if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_ALWAYS)) {
		cookies = comms.request.getCookies();
		sessionKey = findSessionKey(ENHYDRA_SESSION_ID_COOKIE,
					    cookies);
		if (sessionKey != null) {
		    // These are set so that application programmers
		    // can determine the state of session cookies if not
                    // using automatic url encoding
		    comms.request.setRequestedSessionIdFromCookie(true);
		    comms.request.setRequestedSessionIdFromUrl(false);
		    comms.response.setSessionIdCookieRequired(true);
		    comms.response.setSessionIdEncodeUrlRequired(false);
		}
            }

	    if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_NEVER) &&
		sessionKey == null) {

                //
                // This needs to be redone.
                //
//                if (comms.request.getHttpServletRequest() instanceof EnhydraRequest ) {
//                    EnhydraRequest er = (EnhydraRequest) comms.request.getHttpServletRequest();
//                }

		// Check if sessionKey is encoded as a parameter in the path
		// info (the form is http://localhost/foo.po;jsessionid=XYZ
                String pathInfo = comms.request.getPathInfo();
		StringTokenizer tokens =
		    new StringTokenizer(pathInfo, "/\\;=", false);
		while (tokens.hasMoreTokens()) {
		    String param = tokens.nextToken();
		    if (param.equalsIgnoreCase(ENHYDRA_SESSION_ID_URL)) {
			sessionKey = tokens.nextToken();
			break;
		    }
		}

        	if (sessionKey != null) {
                    // These are set so that application programmers
                    // can determine the state of session cookies if not
                    // using automatic url encoding
        	    comms.request.setRequestedSessionIdFromUrl(true);
        	    comms.request.setRequestedSessionIdFromCookie(false);
		    comms.response.setSessionIdEncodeUrlRequired(true);
		    comms.response.setSessionIdCookieRequired(false);
        	}
            }
        } catch (HttpPresentationException except2) {
            throw new ApplicationException(except2);
        }
	if (sessionKey == null) {
	    return null;
	}

	if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_NEVER)) {
	    comms.response.setSessionKey(sessionKey);
	    comms.response.setSessionManager(comms.application.getSessionManager());
	}

        try {
            s = comms.application.getSessionManager().getSession(
		    Thread.currentThread(), sessionKey);
        } catch (SessionException e) {
            throw new ApplicationException(e);
        }
        return s;
    }

    /**
     *  This is a rarely used utility function. Give a raw ServletRequest,
     *  the session cookie used to look up the session object that would be
     *  used with the request if it were passed into the Enhydra framework.
     *  Returns null if no cookie or session is found; a new session will
     *  not be created.
     *  Currently this is only used to support the debugger.
     *
     *  @param request
     *    The raw request.
     *  @param application
     *    The application that will be servicing the request.
     *  @return
     *    The session object that would be used if the request were
     *    passed into the Enhydra framework, or null if not found (no new
     *    session will be created).
     */
    public static Session getRequestSession(ServletRequest request,
                                            Application application)
            throws ApplicationException {
        if (application == null) {
            return null;
        }
        Cookie[] cookies;
        Session s;
        HttpServletRequest req;
        try {
            req = (HttpServletRequest)request;
        } catch (ClassCastException e) {
            return null;
        }
        cookies = req.getCookies();
        String sessionKey = findSessionKey(cookies);
        if (sessionKey == null) {
            return null;
        }
        SessionManager sm = application.getSessionManager();
        if (sm == null) {
            return null;
        }
        try {
            s = sm.getSession(Thread.currentThread(), sessionKey);
        } catch (SessionException e) {
            throw new ApplicationException(e);
        }
        return s;
    }


    /**
     * Associate a new <CODE>StandardSession<CODE> to a client cookie.
     *
     * @param comms The presentations manager comms object with
     *  all request/response information in it.
     * @exception ApplicationException If address can't be obtained.
     */
    public static void bindSessionToClient(HttpPresentationComms comms)
            throws ApplicationException {

        StandardSession session = (StandardSession)comms.session;
	StandardSessionManager sessionManager =
	    (StandardSessionManager)session.getSessionManager();
	if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_ALWAYS)) {
	    try {
		Cookie cookie = new Cookie(ENHYDRA_SESSION_ID_COOKIE,
					   session.getSessionKey());
		cookie.setPath(comms.request.getApplicationPath());
		String serverName = comms.request.getServerName();
  // Do not set the domain (PJ).
/*
		int first = serverName.indexOf('.');
		int last = serverName.lastIndexOf('.');
		if ((first > -1) && (last > -1) && (first != last)) {
		    cookie.setDomain(serverName.substring(first));
		}
*/
		comms.response.addCookie(cookie);
	    } catch (HttpPresentationException except) {
		throw new ApplicationException(except);
	    }
	}
	if (!sessionManager.getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_NEVER)) {
	    comms.response.setSessionKey(comms.session.getSessionKey());
	    comms.response.setSessionManager(comms.application.getSessionManager());
	}
    }
}
