/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: CopyFilePresentation.java,v 1.11.14.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation;

import java.io.*;
import com.lutris.util.*;
import com.lutris.classloader.*;

/**
 * Presentation for file data.  This should not be cached, as it copies
 * the input stream and closes it.
 */
class CopyFilePresentation implements HttpPresentation {
    private InputStream input;
    private int contentLength;
    private String mimeType;

    private static final int FILE_BUFFER_SIZE = 4096;

    /*
     * Copy the input stream to the HTTP output stream.
     */
    private void copyFile(HttpPresentationComms comms)
        throws FilePresentationException {

        try {
            byte [] buffer = new byte[FILE_BUFFER_SIZE];
            HttpPresentationOutputStream output = comms.response.getOutputStream();
            while (true) {
                int readLen = input.read(buffer, 0, FILE_BUFFER_SIZE);
                if (readLen < 0) {
                    break;
                }
                output.write(buffer, 0, readLen);
            }
        } catch (Exception except) {
            throw new FilePresentationException(except);
        }
    }

    /**
     * Construct a new copy file presentation, given an input stream.
     */
    protected CopyFilePresentation(ClassLoader classLoader,
                                   String urlPath,
                                   String fileMimeType) 
            throws FilePresentationException {
        mimeType = fileMimeType;

        /*
         * Get the stream from the classloader.  If we are using the
         * Lutris classloader, we maybe able to get the file size from
         * the classloader.
         */
        try {
            if (classLoader instanceof MultiClassLoader) {
                Resource resource = ((MultiClassLoader)classLoader).getResourceAsIs(urlPath);
                if (resource != null) {
                    input = resource.getInputStream();
                    contentLength = (int)resource.getSize();
                }
            } else {
                input = classLoader.getResourceAsStream(urlPath);
                contentLength = -1;
            }
        } catch (IOException except) {
            throw new FilePresentationException("Error accessing  \"" + urlPath + "\"", except);
        }
        if (input == null) {
            throw new FilePresentationException
                (new FileNotFoundException ("File \"" + urlPath
                                            + "\" not found on application class path"));
        }
    }
    
    /**
     * Finalizer to close input file if we haven't done it already.
     */
    protected void finalize() throws Throwable {
        if (input != null) {
            input.close();
        }
        super.finalize();
    }

    /**
     * Entry point for copy file presenation.
     */
    public void run(HttpPresentationComms comms)
            throws FilePresentationException {
        try {
            comms.response.setContentType(mimeType);
            comms.response.setContentLength(contentLength);
            try {
                if (!comms.request.getMethod().equals("HEAD")) {
                    copyFile(comms);
                }
            } finally {
                input.close();
                input = null;
            }
        } catch (FilePresentationException except) {
            throw except;
        } catch (Exception except) {
            throw new FilePresentationException(except);
        }
    }
}
