/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ExceptionHandler.java,v 1.43.14.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation;

import com.lutris.appserver.server.*;
import com.lutris.util.*;
import com.lutris.logging.*;
import org.enhydra.servlet.ServletIOException;
import java.io.*;

/**
 * Various routines used in handling exceptions.
 */
class ExceptionHandler {
    /*
     * Disallow instances.
     */
    private ExceptionHandler() {
    }

    /**
     * Generate an HTML page describing an exception and log the exception.
     * If it looks like a client disconnect, log at the DEBUG level.
     */
    private static void sendExceptionPage(HttpPresentationComms comms,
                                          Application application,
                                          Throwable except)
            throws HttpPresentationException, IOException {

        String title = "Error in " + application.getName();

        comms.response.setStatus(HttpPresentationResponse.SC_INTERNAL_SERVER_ERROR);
        comms.response.setContentType("text/html");

        HttpPresentationOutputStream out = comms.response.getOutputStream();

        out.println("<HTML><TITLE>");
        out.println(title);
        out.println("</TITLE>");

        out.println("<BODY BGCOLOR=white>");
        out.println("<H2><P><CENTER><FONT COLOR=red> - " + title + " - </FONT></CENTER></H2>");

        // Reason...
        if (except instanceof java.io.FileNotFoundException) {
            out.println("<B>File not found</B>: " +
                        comms.request.getRequestURI());
        } else if (except instanceof DirectoryException) {
            out.println("<B>Directory access error</B>: " +
                        comms.request.getRequestURI());
        } else if (except instanceof java.lang.IllegalAccessException) {
            out.println("<B>Illegal access</B>: " +
                        comms.request.getRequestURI());
        } else if ((except instanceof java.lang.ClassNotFoundException)
                   || (except instanceof java.lang.NoClassDefFoundError)) {
            out.println("<B>Class not found: " + except.getMessage() + "</B>");
        } else {
            // catch all
            out.println("<B>Reason: </B>" + except.getClass() +
		": " + except.getMessage());
        }

        // Details...
        out.println("<P>");
        out.println("<B>Java Call Stack:</B><P><FONT size=-1><PRE>");
        ByteArrayOutputStream b = new ByteArrayOutputStream(); //FIX: This should be a utility.
        PrintWriter p = new PrintWriter(b);
        except.printStackTrace(p);
        p.flush();
        out.println(b.toString());
        out.println("</PRE></FONT>\n");

        // Back button
        out.println("<FORM>");
        out.println("<CENTER><INPUT TYPE=BUTTON VALUE=\"back\"");
        out.println("onClick=\"history.go(-1)\"></CENTER>\n");
        out.println("</FORM>");

        // EOF
        out.println("</BODY>");
        out.println("</HTML>");
        comms.response.flush();
    }


    /**
     * Output an HTML page describing an exception and log the exception.
     * If it looks like a client disconnect, log at the DEBUG level.
     */
    protected static void logDisplayException(HttpPresentationComms comms,
                                              LogChannel logChannel,
                                              long requestId,
                                              Application application,
                                              Throwable except,
                                              String presObjPath) {

        /*
         * Log at DEBUG if client I/O error; but generate page anyway.  It
         * will most likely be discarded, but it will prevent just a blank
         * page if the client is still connected.
         */
        try {
            if (HttpPresentationIOException.isClientIOException(except)) {
                logChannel.write(Logger.DEBUG, "RID:" + requestId
                                 + ": Client appeared to drop connection: "
                                 + presObjPath);
                sendExceptionPage(comms, application, except);
            } else {
                logChannel.write(Logger.ERROR, "RID:" + requestId
                                 + ": Unhandled Application error in "
                                 + application.getName(), except);
                sendExceptionPage(comms, application, except);
            }
        } catch (IllegalStateException except2) {
            // May not be able set set headers if output has begun.
        } catch (IOException except2) {
            // Ignore, client is gone
        } catch (HttpPresentationException except2) {
            // Ignore, client maybe gone
        }
    }

    /**
     * Get the path of the error handler for an exception that occured during
     * the loading and execution of a presentation or the handling of an error
     * The handler is a presentation named `ErrorHandler' in the current or
     * a higher level directory.
     *
     * This method just returns the path for the error handler for the
     * specified presentation path.  This attempts to load the error handler
     * just to see if it exists. This also handles errors that are throw while
     * handling an error (which is often just an ErrorHandler passing an error up
     * to a higher level).
     *
     * @param presObjPath
     *   The path of the presentation object that threw the error.  This
     *   may not be the same as the original request URL.  This is relative
     *   to <CODE>presentationRoot</CODE>.
     * @param requestedPresObjPath
     *   Original presentation object that the error hander is being called
     *   for.
     * @param whileHandlingError
     *   True an error was already being handled.
     * @param logChannel
     *   Log channel to use.
     * @return The path to the presentation object to invoke or null if no
     *  is available (reached top of tree).
     */
    protected static String getErrorHandler(String presObjPath,
                                            String requestedPresObjPath,
                                            boolean whileHandlingError,
                                            PresentationLoader loader,
                                            LogChannel logChannel,
                                            long requestId) {
        String handlerDir;

        logChannel.write(Logger.DEBUG, "RID:" +requestId
                         + ": Searching for error handler for "
                         + requestedPresObjPath);

        // Find first place to look.
        if (!whileHandlingError) {
            // Look in current package; its ok for dir to be null here.
            handlerDir = (new File(presObjPath)).getParent();
        } else {
            // Look in parent package.
            String currentDir = (new File(presObjPath)).getParent();
            if (currentDir == null) {
                handlerDir = null;
            } else {
                handlerDir = (new File(currentDir)).getParent();
                // already at top level, don't try that handler again
                if (handlerDir == null) {
                    logChannel.write(Logger.DEBUG, "RID:" + requestId
                                     + ": No error handler found for: "
                                     + presObjPath);
                    return null;
                }
            }
        }

        // Try to load, searching up if a handler was not found.
        while (true) {
            // handlerDir can be null here (trying the top).
            String handlerPath = (new File(handlerDir, "ErrorHandler.po")).getPath();
            logChannel.write(Logger.DEBUG, "RID:" + requestId
                             + ": Checking for error handler: " + handlerPath);
            try {
                loader.loadPresentation(handlerPath);
                return handlerPath;  // Found and loaded a handler.
            } catch (ClassNotFoundException noClass) {
                if (handlerDir == null) {
                    break;  // Not found
                }
                String currentDir = (new File(handlerPath)).getParent();
                if (currentDir == null) {
                    handlerDir = null;
                } else {
                    handlerDir = (new File(currentDir)).getParent();
                }
            } catch (Throwable except) {
                logChannel.write(Logger.DEBUG, "RID:" + requestId
                                 + ": Error while loading error handler: "
                                 + handlerPath, except);
            }
        }

        logChannel.write(Logger.DEBUG, "RID:" + requestId
                         + ": No error handler found for: " + presObjPath);
        return null;
    }


    /**
     * Send a simple HTML page as part of a response. This method
     * composes a complete page with header and body sections.
     */
    private static void sendHtml(HttpPresentationComms comms,
                                 String title,
                                 String htmlText)
            throws HttpPresentationException, IOException {

        comms.response.setContentType("text/html");
        HttpPresentationOutputStream out = comms.response.getOutputStream();

        out.println("<HTML>");
        out.println("<TITLE>" + title + "</TITLE>");
        out.println("<BODY>");
        out.println(htmlText);
        out.println("</BODY>");
        out.println("</HTML>");
        comms.response.flush();
    }


    /**
     * Process an unauthorized page exception, sending the approriate
     * HTTP response. <P>
     *
     * Two actions are taken: a HTTP header is set telling the browser that it
     * needs to use basic auth authentication and which relm to ask for, and
     * the response code is set to 401, which means unauthorized. The response
     * is now ready to return to the user, no further proccessing needs to be
     * done on it.  Most browsers will see the 401 response, and will create a
     * dialog box asking the user to enter a username and password. The relm
     * appears in this dialog, so the user knows what he or she is logging in
     * to. After gathering the data, most browsers will re-issue the response,
     * with the username and password in an HTTP header.
     */
    protected static void sendUnauthorizedPage(HttpPresentationComms comms,
                                               LogChannel logChannel,
                                               long requestId,
                                               Application application,
                                               String presObjPath,
                                               PageUnauthorizedException unauthExcept)
            throws HttpPresentationException, IOException {

        try {
            logChannel.write(Logger.DEBUG, "RID:" + requestId
                             + ": Authentication required: " + presObjPath);

            comms.response.setHeader("WWW-authenticate",
                                     "basic realm=\"" +
				     unauthExcept.getRelm() +"\"");
            comms.response.setStatus(HttpPresentationResponse.SC_UNAUTHORIZED);
            sendHtml(comms,
                     unauthExcept.getHtmlTitle(),
                     unauthExcept.getHtmlText());
            comms.response.flush();
        } catch (Exception except) {
            // Make chained exception
            HttpPresentationException presExcept =
                new HttpPresentationException("Exception while sending page unauthorized response", except);
            logDisplayException(comms, logChannel, requestId, application, presExcept, presObjPath);
        }
    }

    /**
     * Generate an error for the maximum number of redirect's being
     * reached.
     */
    protected static void maxRedirectsReached(HttpPresentationComms comms,
                                              LogChannel logChannel,
                                              long requestId,
                                              int maxRedirectErrorLoops,
                                              Application application,
                                              String presObjPath)
            throws HttpPresentationException, IOException {

        String msg = "Maximum number of server redirects or ErrorHandler "
            + "calls per request reached (" + maxRedirectErrorLoops
            + "); probable endless loop: " + presObjPath;
        HttpPresentationException except =
            new HttpPresentationException(msg, comms.exception);
        logDisplayException(comms, logChannel, requestId,  application, except, presObjPath);
    }

    /**
     * Process an exception that has not been handled by an error handler or
     * so other mechanism.
     */
    protected static void processUnhandledException(HttpPresentationComms comms,
                                                    LogChannel logChannel,
                                                    long requestId,
                                                    Application application,
                                                    String presObjPath,
                                                    Throwable except) {

        try {
            if ((except instanceof ClassNotFoundException)
                || (except instanceof NoClassDefFoundError)
                || (except instanceof FilePresentationException)) {

                logChannel.write(Logger.WARNING, "RID:" + requestId
                                 + ": page not found: " + presObjPath);
                comms.response.setStatus(HttpPresentationResponse.SC_NOT_FOUND);
                sendHtml(comms, "Page Not Found",
                         "<B>Page was not found on this server</B>");
            } else {
                logDisplayException(comms, logChannel, requestId, application, except, presObjPath);
            }
        } catch (Throwable except2) {
            // Error handling the error; hang it up.
        }
    }
}
