/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpPresentationResponse.java,v 1.21.2.2.4.4 2000/11/24 21:03:23 markd Exp $
 */





package com.lutris.appserver.server.httpPresentation;

import com.lutris.appserver.server.session.SessionManager;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletResponse;
import org.w3c.dom.html.HTMLDocument;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.io.OutputOptions;

/**
 * Object passed to presentation objects that is used to generate HTTP
 * responses.
 *
 * @author	Mark Diekhans
 * @version	$Revision: 1.21.2.2.4.4 $
 */
public interface HttpPresentationResponse {
    /**
     * Returns the original HttpServletResponse.
     */
    public HttpServletResponse getHttpServletResponse();

    /**
     * Sets the content length for this response.
     *
     * @param len the content length
     */
    public void setContentLength(int len)
	throws HttpPresentationException;

    /**
     * Sets the content type for this response.
     *
     * @param type the content's MIME type
     */
    public void setContentType(String type)
	throws HttpPresentationException;


    /**
     * Create an OutputOptions object for a document.  Options are default for
     * the specified document.  The object maybe then modified as needed to
     * override the default values.
     * <P>
     * The following attributes are set in the object:
     * <UL>
     * <LI> encoding
     * <LI> MIME type - Defaults for document.
     * <LI> URLRewriter - Set if URL encoding of sessions is enabled.
     * </UL>
     */
    public OutputOptions createOutputOptions(XMLObject document)
        throws HttpPresentationException;

    /**
     * Returns an output stream for writing response data.
     */
    public HttpPresentationOutputStream getOutputStream()
	throws HttpPresentationException;

    /**
     * Adds the specified cookie to the response. It can be called multiple
     * times to set more than one cookie.
     *
     * @param cookie - The Cookie to return to the client.
     */
    public void addCookie(Cookie cookie)
	throws HttpPresentationException;

    /**
     * Returns true if the response message header has a field with
     * the specified name.
     *
     * @param name the header field name
     */
    public boolean containsHeader(String name)
	throws HttpPresentationException;

    /**
     * Sets the status code and message for this response.
     *
     * @param sc the status code
     * @param sm the status message
     */
    public void setStatus(int sc, String sm)
	throws HttpPresentationException;

    /**
     * Sets the status code and a default message for this response.
     *
     * @param sc the status code
     */
    public void setStatus(int sc)
	throws HttpPresentationException;

    /**
     * Adds a field to the response header with a given name and
     * value.  If the field had already been set, the new value
     * overwrites the previous one.  The containsHeader method can be
     * used to test for the presence of a header before setting its
     * value.
     *
     * @param name the header field name
     * @param value the header field value
     */
    public void setHeader(String name, String value)
	throws HttpPresentationException;

    /**
     * Adds a field to the response header with a given name and
     * integer value.  If the field had already been set, the new
     * value overwrites the previous one.  The containsHeader method
     * can be used to test for the presence of a header before setting
     * its value.
     *
     * @param name the header field name
     * @param value the header field integer value
     */
    public void setIntHeader(String name, int value)
	throws HttpPresentationException;

    /**
     *
     * Adds a field to the response header with a given name and
     * date-valued field.  The date is specified in terms of
     * milliseconds since the epoch.  If the date field had already
     * been set, the new value overwrites the previous one.  The
     * containsHeader method can be used to test for the presence of a
     * header before setting its value.
     *
     * @param name the header field name
     * @param value the header field date value
     */
    public void setDateHeader(String name, long date)
	throws HttpPresentationException;

    /**
     * Sends an error response to the client using the specified status
     * code and descriptive message.
     *
     * @param sc the status code
     * @param msg the detail message
     */
    public void sendError(int sc, String msg)
	throws HttpPresentationException;

    /**
     * Sends an error response to the client using the specified
     * status code and a default message.
     *
     * @param sc the status code
     */
    public void sendError(int sc)
	throws HttpPresentationException;

    //Omitted: sendRedirect(String location)

    //Omitted: public abstract String encodeUrl(String url)

    //Omitted: public abstract String encodeRedirectUrl(String url)

    /**
     * Called at the end of processing a response to force any cached headers to
     * be written and buffers flushed.  This maybe a no-op if no buffering is
     * implemented.  This is normally not called by a client.
     */
    public void flush ()
	throws HttpPresentationException;

    /**
     * Output an an XMLC document object (DOM). The document is formatted
     * according to it's type.  The MIME type of the response is automatically
     * set.
     *
     * @param outputFormat Object use to specify options controling formatting
     *  of the document.
     * @param doc The DOM object to be returned as response.
     */
    public void writeDOM(OutputOptions outputOptions,
                         XMLObject document) throws HttpPresentationException;

    /**
     * Output an an XMLC document object (DOM). The document is formatted
     * according to it's type.  The MIME type of the response is automatically
     * set.
     *
     * @param doc The DOM object to be returned as response.
     */
    public void writeDOM(XMLObject document) throws HttpPresentationException;

    /**
     * Utility method to output an HTML page. The appropriate headers are
     * set for MIME type and to disable caching of the HTML by the broswer.
     */
    public void writeHTML(String html) throws HttpPresentationException;

    /**
     * Utility method to output an HTML page from a DOM object. The
     * appropriate headers are set for MIME type and to disable caching
     * of the HTML by the broswer.
     *
     * @param doc The DOM object to be returned as response
     * @deprecated use writeDOM
     * @see #writeDOM
     */
    public void writeHTML(HTMLDocument doc) throws HttpPresentationException;

    /**
     * Set the output character encoding.
     *
     * @param enc character encoding
     */
    public void setEncoding(String enc);

    /**
     * Get the output character encoding.
     *
     * @return the current encoding
     */
    public String getEncoding();

    /**
     * Sets the current session key for this response
     *
     * @param sessionKey
     *        The current sessionKey
     **/
    public void setSessionKey(String sessionKey);

    /**
     * Sets the current application name for this response
     *
     * @param appName
     *        The current application name
     **/
    public void setSessionManager(SessionManager sessionManager);

    /**
     * Indicates whether client response requires a sessionId cookie
     * @return true if client response requires a sessionId cookie
     *         false otherwise
     */
    public boolean isSessionIdCookieRequired()
	throws HttpPresentationException;

    /*
     * set boolean flag for sessionId response cookie.  Indicates whether
     * client response requires a sessionId cookie
     * @param isFromCookie boolean flag
     */
    public void setSessionIdCookieRequired(boolean sessionIdCookie)
	throws HttpPresentationException;

    /**
     * Indicates whether client response requires url encoding for
     * sessionId
     * @return true if client response requires url encoding for
     *         sessionId; false otherwise
     */
    public boolean isSessionIdEncodeUrlRequired()
	throws HttpPresentationException;

    /*
     * set boolean flag url encoding for sessionId.  Indicates whether
     * client response requires url encoding for sessionId
     * @param sessionIdUrl boolean flag
     */
    public void setSessionIdEncodeUrlRequired(boolean sessionIdUrl)
	throws HttpPresentationException;

    /*
     * SERVER status codes; see RFC 1945.
     */

    /**
     * Status code (200) indicating the request succeeded normally.
     */
    public static final int SC_OK = 200;

    /**
     * Status code (201) indicating the request succeeded and created
     * a new resource on the server.
     */
    public static final int SC_CREATED = 201;

    /**
     * Status code (202) indicating that a request was accepted for
     * processing, but was not completed.
     */
    public static final int SC_ACCEPTED = 202;

    /**
     * Status code (204) indicating that the request succeeded but that
     * there was no new information to return.
     */
    public static final int SC_NO_CONTENT = 204;

    /**
     * Status code (301) indicating that the resource has permanently
     * moved to a new location, and that future references should use a
     * new URI with their requests.
     */
    public static final int SC_MOVED_PERMANENTLY = 301;

    /**
     * Status code (302) indicating that the resource has temporarily
     * moved to another location, but that future references should
     * still use the original URI to access the resource.
     */
    public static final int SC_MOVED_TEMPORARILY = 302;

    /**
     * Status code (304) indicating that a conditional GET operation
     * found that the resource was available and not modified.
     */
    public static final int SC_NOT_MODIFIED = 304;

    /**
     * Status code (400) indicating the request sent by the client was
     * syntactically incorrect.
     */
    public static final int SC_BAD_REQUEST = 400;

    /**
     * Status code (401) indicating that the request requires HTTP
     * authentication.
     */
    public static final int SC_UNAUTHORIZED = 401;

    /**
     * Status code (403) indicating the server understood the request
     * but refused to fulfill it.
     */
    public static final int SC_FORBIDDEN = 403;

    /**
     * Status code (404) indicating that the requested resource is not
     * available.
     */
    public static final int SC_NOT_FOUND = 404;

    /**
     * Status code (500) indicating an error inside the HTTP service
     * which prevented it from fulfilling the request.
     */
    public static final int SC_INTERNAL_SERVER_ERROR = 500;

    /**
     * Status code (501) indicating the HTTP service does not support
     * the functionality needed to fulfill the request.
     */
    public static final int SC_NOT_IMPLEMENTED = 501;

    /**
     * Status code (502) indicating that the HTTP server received an
     * invalid response from a server it consulted when acting as a
     * proxy or gateway.
     */
    public static final int SC_BAD_GATEWAY = 502;

    /**
     * Status code (503) indicating that the HTTP service is
     * temporarily overloaded, and unable to handle the request.
     */
    public static final int SC_SERVICE_UNAVAILABLE = 503;
}
