/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpSerialized.java,v 1.7.14.1 2000/10/19 17:59:06 jasona Exp $
 */




package com.lutris.appserver.server.httpPresentation;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *  This class provides two utility methods for presentation objects
 *  that want to read and/or write Java serialized objects. <P>
 *
 *  Reading is accomplished by having the client issue a POST request,
 *  with the serialized data as the body of the request. <P>
 *
 *  Writing is accomplished by simply sending out the serialized bytes.
 *
 * @see java.io.Serializable
 */
public class HttpSerialized {

    /**
     *  The mime type used to send and recieve Java serialized objects.
     */
    public static final String serializedMimeType = 
                                             "application/java-serialized";

    /**
     *  Read in a Java object from a POST request. The object is returned.
     *  
     *  @param request
     *   HTTP communication request.
     *  @return 
     *   The java object sent in the POST request.
     *  @exception HttpPresentationException 
     *   If the request is not of type POST, or the content type is not
     *   correct, or there is an IO or serialization error.
     */
    public static Object readSerializedObject(
                             HttpPresentationRequest request)
                             throws HttpPresentationException {
        /*
         *  Is this a POST request?
         */
        String method = request.getMethod();
        if ((method == null) || 
            !method.equalsIgnoreCase("POST")) 
            throw new HttpPresentationException("Request method is not POST.");
        /*
         *  Is the content type correct?
         */
        String type = request.getContentType();
        if ((type == null) ||
           !type.equalsIgnoreCase(serializedMimeType))
            throw new HttpPresentationException("POSTed data is not of type " +
                                                serializedMimeType + ".");
        /*
         *  Try to read in the object.
         */
        Object result = null;
        try {
            HttpPresentationInputStream hpis = request.getInputStream();
            ObjectInputStream ois = new ObjectInputStream(hpis);
            result = ois.readObject();
        } catch (InvalidClassException e) {
            throw new HttpPresentationException(
                          "Invalid class sent: " + e.getMessage(), e);
        } catch (StreamCorruptedException e) {
            throw new HttpPresentationException(
                          "Invalid data sent: " + e.getMessage(), e);
        } catch (OptionalDataException e) {
            throw new HttpPresentationException(
                          "Invalid data sent: " + e.getMessage(), e);
        } catch (ClassNotFoundException e) {
            throw new HttpPresentationException(
                          "Invalid class sent: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new HttpPresentationException(
                          "IO error reading object: " + e.getMessage(), e);
        }
        /*
         *  Success!
         */
        return result;
    }

    /**
     *  Read in a Java object from a POST request. The object is returned.
     *  
     *  @param comms
     *   HTTP communications object. Contains objects and interfaces to read
     *   the request and send a response.
     *  @return 
     *   The java object sent in the POST request.
     *  @exception HttpPresentationException 
     *   If the request is not of type POST, or the content type is not
     *   correct, or there is an IO or serialization error.
     */
    public static Object readSerializedObject(
                             HttpPresentationComms comms)
                             throws HttpPresentationException {
        return readSerializedObject(comms.request);
    }

    /**
     *  Return a serialized Java object to the client. 
     *
     *  @param comms
     *   HTTP Response object. interface to send a response.
     * @param object 
     *  The object to return to the client.
     * @exception HttpPresentationException 
     *  If there is an error serializing the object, or an error in the
     *  underlying calls to response.
     */
    public static void writeSerializedObject(
                           HttpPresentationResponse response,
                           Serializable object)
                           throws HttpPresentationException {
        /*
         * First try to serialize the object, so we can catch any exceptions
         * up front, and so we know the length.
         */
        byte[] objectData = null;
        try {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(object); 
            objectData = baos.toByteArray();
        } catch (InvalidClassException e) {
            throw new HttpPresentationException(
                          "Invalid class: "  + e.getMessage(), e);
        } catch (NotSerializableException e) {
            throw new HttpPresentationException(
                          "Invalid class: "  + e.getMessage(), e);
        } catch (IOException e) {
            throw new HttpPresentationException(
                          "Error serializing object: " + e.getMessage(), e);
        }
        if (objectData == null)
            throw new HttpPresentationException(
                         "Error serializing object: no data generated.");
        /*
         *  Set up the response.
         */
        response.setContentType(serializedMimeType);
        response.setContentLength(objectData.length);
        /*
         *  Transmit the object data.
         */
        HttpPresentationOutputStream hpos = response.getOutputStream();
        try {
            hpos.write(objectData);
        } catch (IOException e) {
            throw new HttpPresentationException(
                         "Error sending object: " + e, e);
        }
        /*
         * Success!
         */
    }

    /**
     *  Return a serialized Java object to the client. 
     *
     *  @param comms
     *   HTTP communications object. Contains objects and interfaces to read
     *   the request and send a response.
     * @param object 
     *  The object to return to the client.
     * @exception HttpPresentationException 
     *  If there is an error serializing the object, or an error in the
     *  underlying calls to response.
     */
    public static void writeSerializedObject(
                           HttpPresentationComms comms,
                           Serializable object)
                           throws HttpPresentationException {
        writeSerializedObject(comms.response,object);
    }
    
}

