/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PresentationLoader.java,v 1.18.2.2.4.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation;

import com.lutris.http.*;
import com.lutris.logging.*;
import com.lutris.appserver.server.*;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.StringTokenizer;
import java.net.URL;
import java.net.URLEncoder;
import java.io.*;

/**
 * Loader for presentation objects and resources associated with an
 * application.  It also manages a cache of resources.
 *
 * @version	$Revision: 1.18.2.2.4.1 $
 * @author	Mark Diekhans
 * @since	1.8
 * @see HttpPresentationManager
 */
class PresentationLoader {
    /**
     * Prefix added to the file name portion of the URL.  The
     * resulting names is used to search for classes and
     * files on the class path.
     */
    private final String presentationPrefix;

    /**
     * Mime type converter.
     */
    private Mime mime;

    /**
     * Class loader for the application's classes.
     */
    private ClassLoader appClassLoader;

    /**
     * Cache for loaded objects.  If null, presentation classes are not cached.
     */
    private Hashtable presObjCache;

    /**
     * Cache for resource presentation objects.  If null, resource presentation
     * classes are not cached.
     */
    private Hashtable resourceObjCache;

    /**
     * Applications log channel.  No logging is done if null.
     */
    private LogChannel logChannel;

    /**
     * Is DEBUG1 logging enabled?
     */
    private boolean debug1LoggingEnabled;


    /**
     * Constructor.
     *
     * @param appPresentationPrefix Prefix added to the file name portion
     * of the URL.  The resulting names is used to search for classes and
     * files on the class path.
     * @param applicationClassLoader The class loader to use for the application.
     * @param cacheClasses Enables or disables caching of presentation object
     *        classes in memory.
     * @param cacheFiles Enables or disables caching of files (html, gif, etc)
     * that are servered as part of the application.
     * @param presLogChannel Log channel for the presentation.  Maybe null.
     */
    protected PresentationLoader(String appPresentationPrefix,
                                 ClassLoader applicationClassLoader,
                                 boolean cacheClasses,
                                 boolean cacheFiles,
                                 LogChannel presLogChannel)
            throws HttpPresentationException {
        logChannel = presLogChannel;
        if (logChannel != null) {
            debug1LoggingEnabled = logChannel.isEnabled(Logger.DEBUG1);
        }
        appClassLoader = applicationClassLoader;
        presentationPrefix = cleanUrlPath(appPresentationPrefix);
        mime = new Mime();
        if (cacheClasses) {
            presObjCache = new Hashtable();
        }
        if (cacheFiles) {
            resourceObjCache = new Hashtable();
        }
    }

    /**
     * Clean up a path name taken from a URL, removing such uglyness
     * as `./', `//', `//' and `\'.  Strips off leading and trailing
     * `/' and `\'.  Uses `/', since this is really a URL, not a file path.
     */
    private String cleanUrlPath(String urlPath)
            throws HttpPresentationException {
        // Tokenize on `/' and `\', but drop leading and training spaces,
        // since they sometimes get dragged in from broken config files.

        // Do a quick scan through the string to look to look for anything
        // that is amiss. The idea is to avoid calling the string tokenizer
        // unless we have to. If there is nothing to do then this quick test
        // make this routine about 4 times faster.
        int i=0;
        int len = urlPath.length();
        if (len != 0) {
            if (urlPath.charAt(0)=='/') {
                // This is pretty common and quick to handle here.
                urlPath = urlPath.substring(1);
                len--;
            }

            // if we start or end with a '/' then we are ugly.
            // otherwsie scan through the string looking for '\' or "./" or "//"
            boolean uglyness = (urlPath.charAt(0) == '/' || urlPath.charAt(len-1) == '/');
            while (!uglyness && i < len) {
                char c = urlPath.charAt(i);
                if (c =='\\') {
                    uglyness = true;
                }
                else if (c=='/' && i > 0 ) {
                    char c0 = urlPath.charAt(i-1);
                    uglyness = ( c0 == '.' || c0 == '/');
                }
                i++;
            }

            if (!uglyness)  {
                return urlPath;
            }
        }
        // end quick test.

        // The quick test found a problem...

        StringTokenizer tokens =
            new StringTokenizer(urlPath.trim(), "/\\", false);
        StringBuffer newPath = new StringBuffer();


        // Loop through path elements, dropping uninteresting parts
        // of the name.
        while (tokens.hasMoreTokens()) {
            String name = tokens.nextToken();
            if (!(name.equals("/") || name.equals("\\")
                  || name.equals(".") || name.equals(""))) {
                if (newPath.length() > 0) {
                    newPath.append('/');
                }
                newPath.append(name);
            }
        }
        return newPath.toString();
    }


    /**
     * Concert a URL path name to class names. If the path appears invalid, a
     * bogus class (but reasonable looking) name will be generated which will
     * lead to an error when loading.
     */
    private String convertToClassName(String urlPath) {
        String className = urlPath;
        int idx, start, end;

        // Drop MIME extension
        idx = className.lastIndexOf('.');
        if (idx > 0) {
            className = className.substring(0, idx);
        }

        // Convert the URL separator to `.'.  Note `\' is not a legal
        // URL separator
        className = className.replace('/','.');

        // Drop leading or training `.'
        start = 0;
        end = className.length()-1;
        if (className.charAt(start) == '.') {
            start++;
        }
        if (className.charAt(end) == '.') {
            end--;
        }
        if (start >= end) {
            // Bad name; put it back so it generates a better error.
            start = 0;
            end = className.length()-1;
        }

        return className.substring(start, end+1);
    }

    /**
     * Load a presentation object by searching the applications class path.
     *
     * @param urlPath The path to the presentation object extracted from
     *        the URL.  It will converted to a class name.
     * @exception ClassNotFoundException if this loader or the system loader
     *            can not find or successfully load the class.
     */
    private synchronized HttpPresentation loadPresentationObject(String urlPath)
            throws ClassNotFoundException, IllegalAccessException,
            InstantiationException {
        String className = convertToClassName(urlPath);

        Class classObj = null;
	if (presObjCache != null) {
            classObj = (Class)presObjCache.get(className);
	}

        if (classObj == null) {
            // Load class, instantiate object and add to cache.
            if (appClassLoader != null) {
                classObj = appClassLoader.loadClass(className);
            } else {
                //FIXME: not needed when our own class loader is working.
                classObj = Class.forName(className);
            }
	    if (presObjCache != null) {
                presObjCache.put(className, classObj);
	    }
        }
        return (HttpPresentation)classObj.newInstance();
    }

    /**
     * Load create a static file presentation for a standard file.
     * The class loader is used as find and open the file.
     */
    private HttpPresentation loadResourcePresentation(String urlPath,String mimeType)
        throws ClassNotFoundException, IllegalAccessException,
               InstantiationException, HttpPresentationException,
               FilePresentationException, IOException {
        HttpPresentation resourceObj;

        if (resourceObjCache != null) {
            resourceObj = (HttpPresentation)resourceObjCache.get(urlPath);
            if (resourceObj == null) {
                resourceObj = new CachedFilePresentation(appClassLoader,
                                                         urlPath, mimeType);
                resourceObjCache.put(urlPath, resourceObj);
            }
        } else {
            resourceObj = new CopyFilePresentation(appClassLoader,
                                                   urlPath, mimeType);
        }
        return resourceObj;
    }

    /**
     * Determine if a request URL references a presentation object.
     *
     * @param request The request for the presentation.
     * @return true if the URL has a presentation object MIME type.
     */
    protected boolean isPresentationRequest(HttpPresentationRequest request)
            throws HttpPresentationException {
        return mime.getType(request.getPresentationObjectPath()).equals("object/presentation");
    }

    /**
     * Load a presentation object or create a presentation object to
     * deliver a file.
     */
    protected HttpPresentation loadPresentation(String urlPath)
            throws ClassNotFoundException, IllegalAccessException,
            InstantiationException, HttpPresentationException,
            FilePresentationException, IOException {
        HttpPresentation presObj;
        String presPath;
        String mimeType = mime.getType(urlPath);

        try {
            if (mimeType.equals("object/presentation")) {
                presPath = presentationPrefix + "/" + cleanUrlPath(urlPath);
                if (debug1LoggingEnabled) {
                    logChannel.write(Logger.DEBUG1, "loadPresentationObject: " + urlPath);
                }
                presObj = loadPresentationObject(presPath);
            } else {
                if (debug1LoggingEnabled) {
                    logChannel.write(Logger.DEBUG1, "loadResourcePresentation: " + urlPath
                                     + " " + mimeType);
                }
                /*
                 * If a class is getting loaded by the client; don't use the
                 * presentation prefix.  Class must be requested by full
                 * class name.
                 */
                if (mimeType.equals("application/java-vm")
                    || mimeType.equals("application/java-archive")) {
                    presPath = cleanUrlPath(urlPath);
                } else {
                    presPath = presentationPrefix + "/" + cleanUrlPath(urlPath);
                }
                presObj = loadResourcePresentation(presPath,mimeType);
            }
        } catch (ClassNotFoundException except) {
            if (debug1LoggingEnabled) {
                logChannel.write(Logger.DEBUG1, "    not found: " + urlPath);
            }
            throw except;
        }
        return presObj;
    }

    /**
     * Determine if the PO cache is enabled.
     *
     * @return <code>true</code> is the cache is enabled.
     */
    protected boolean isPOCacheEnabled() {
        return (presObjCache != null);
    }

    /**
     * Return the number of entries in the PO cache.
     *
     * @return the number of entries in the cache or 0 is disabled.
     */
    protected int sizeofPOCache() {
        if (presObjCache != null) {
            return presObjCache.size();
        } else {
            return 0;
        }
    }

    /**
     * Determine if the file resource class cache is enabled.
     *
     * @return <code>true</code> is the cache is enabled.
     */
    protected boolean isResourceCacheEnabled() {
        return (resourceObjCache != null);
    }

    /**
     * Return the number of entries in the resource cache.
     *
     * @return the number of entries in the cache or 0 is disabled.
     */
    protected int sizeofResourceCache() {
        if (resourceObjCache != null) return resourceObjCache.size();
	else return 0;
    }

    /**
     * Get a file associated with the application.  The file is located
     * in the same manner as presentation objects are located.  That is,
     * prepending the <CODE>presentationPrefix</CODE> to the specified
     * path and searching the class path for a directory or JAR containing
     * the file.
     *
     * @param appfileName The file name relative to the
     *  to the application's <CODE>presentationPrefix</CODE>.
     * @return An input stream associated with the file.
     * @exception IOException If the file can not be found or opened.
     */
    protected InputStream getAppFileAsStream(String appFileName)
            throws IOException, HttpPresentationException {
        String path = presentationPrefix + "/" + cleanUrlPath(appFileName);
        InputStream input = appClassLoader.getResourceAsStream(path);
        if (input == null) {
            throw new HttpPresentationException("File \"" + appFileName + "\" not found on application class path");
        }
        return input;
    }

    /**
     * Flush the presentation object and resource caches.
     */
    protected void flushCache() {
        if (presObjCache != null) {
            presObjCache = new Hashtable();
        }
        if (resourceObjCache != null) {
            resourceObjCache = new Hashtable();
        }
    }

    /**
     * Returns the mime type for the URL path.
     *
     * @param urlPath the URL path.
     */
    protected String getMimeType(String urlPath) {
        return mime.getType(urlPath);
    }

    /** Add a new mime type to extension mapping. */
    public void addMimeType(String mimeType, String extension) {
        mime.addType(mimeType, extension);
    }


    /**
     * Get the application class loader.
     *
     * @return The application class loader.
     */
    protected ClassLoader getAppClassLoader() {
        return appClassLoader;
    }
}
