/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ServletHttpPresentationRequest.java,v 1.32.14.2 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.httpPresentation.servlet;

import java.io.IOException;
import java.io.File;
import java.util.Enumeration;
import java.util.StringTokenizer;
import javax.servlet.http.*;
import com.lutris.http.*;
import com.lutris.http.debug.*;
import com.lutris.appserver.server.httpPresentation.*;
import javax.servlet.http.*;
import java.util.Stack;

/**
 * Servlet implementation of the object passed to Presentation objects that
 * is used to access HTTP request data.
 */
public class ServletHttpPresentationRequest
	implements HttpPresentationRequest, DebugRequest {

    /*
     * Servlet objects.
     */
    private HttpServletRequest request;
    private HttpPresentationInputStream inputStream = null;  // Lazy allocation.
    private String presentationPath;
    private String presentationPathInfo;

    /*
     * Indicates if sessionId is found in a request cookie
     */
    private boolean isRequestedSessionIdFromCookie = false;

    /*
     * Indicates if sessionId is found in a request url
     */
    private boolean isRequestedSessionIdFromUrl = false;

    /**
     * Construct an object associated with a servlet request.
     * @param request Servlet request object that this object will front-end.
     */
    protected ServletHttpPresentationRequest(HttpServletRequest request) {
        this.request = request;
        StringBuffer presPath = new StringBuffer();
        StringBuffer presPathInfo = new StringBuffer();

        /*
         * Split the request URI into the presentation path and path info.
         */
        StringTokenizer tokens =
            new StringTokenizer(request.getRequestURI().trim(), "/\\", false);

        while (tokens.hasMoreTokens()) {
            String name = tokens.nextToken();
	    // A presentation object is found if a token ends with *.po or
	    // *.po; where the session id would follow the semi-colon
            if (name.endsWith(".po")) {
		presPath.append("/");
		presPath.append(name);
		break;
	    } else if (name.indexOf(".po;") != -1) {
		presPath.append("/");
		presPath.append(name.substring(0, name.indexOf(";")));
		presPathInfo.append(name.substring(name.indexOf(";")));
                break;
	    } else {
		presPath.append("/");
		presPath.append(name);
	    }
        }
        if (presPath.length() == 0) {
            presPath.append("/");
        }

        if (tokens.hasMoreTokens()) {
            //presPathInfo = new StringBuffer();
            while (tokens.hasMoreTokens()) {
                presPathInfo.append("/");
                presPathInfo.append(tokens.nextToken());
            }
        }
        presentationPath = presPath.toString();
        presentationPathInfo = presPathInfo.toString();

    }

    /**
     * Returns the original HttpServletRequest.
     */
    public HttpServletRequest getHttpServletRequest() {
	return this.request;
    }

    /**
     * Returns the size of the request entity data, or -1 if not known. Same
     * as the CGI variable CONTENT_LENGTH.
     */
    public int getContentLength() throws HttpPresentationException {
        return request.getContentLength();
    }

    /*
     * Returns the Internet Media Type of the request entity data, or null if
     * not known. Same as the CGI variable CONTENT_TYPE.
     */
    public String getContentType() throws HttpPresentationException {
        return request.getContentType();
    }

    /**
     * Returns the protocol and version of the request as a string of
     * the form <code>&lt;protocol&gt;/&lt;major version&gt;.&lt;minor
     * version&gt</code>.  Same as the CGI variable SERVER_PROTOCOL.
     */
    public String getProtocol() throws HttpPresentationException {
        return request.getProtocol();
    }

    /**
     * Returns the scheme of the URL used in this request, for example
     * "http", "https", or "ftp".  Different schemes have different
     * rules for constructing URLs, as noted in RFC 1738.  The URL used
     * to create a request may be reconstructed using this scheme, the
     * server name and port, and additional information such as URIs.
     */
    public String getScheme() {
        return request.getScheme();
    }

    /**
     * Returns the host name of the server that received the request.
     * Same as the CGI variable SERVER_NAME.
     */
    public String getServerName() {
        return request.getServerName();
    }

    /**
     * Returns the port number on which this request was received.
     * Same as the CGI variable SERVER_PORT.
     */
    public int getServerPort() {
        return request.getServerPort();
    }

    /**
     * Returns the IP address of the agent that sent the request. Same as the
     * CGI variable REMOTE_ADDR.
     */
    public String getRemoteAddr() throws HttpPresentationException {
        return request.getRemoteAddr();
    }

    /**
     * Returns the fully qualified host name of the agent that sent the
     * request. Same as the CGI variable REMOTE_HOST.
     */
    public String getRemoteHost() throws HttpPresentationException {
        return request.getRemoteHost();
    }

    /**
     * Applies alias rules to the specified virtual path and returns the
     * corresponding real path, or null if the translation can not be
     * performed for any reason.  This resolves the path using the virtual
     * docroot, if virtual hosting is enabled, and with the default docroot
     * otherwise.  Calling this method with the string "/" as an argument
     * returns the document root.
     *
     * @param path the virtual path to be translated to a real path
     */
/*
    public String getRealPath(String path) throws HttpPresentationException {
        return request.getRealPath(path);
    }
*/

    /**
     * Returns an input stream for reading the request body.
     */
    public HttpPresentationInputStream getInputStream() throws HttpPresentationException {
        if (inputStream == null) {
            try {
                inputStream = new ServletHttpPresentationInputStream(request.getInputStream());
            } catch (IOException except) {
                // Want to throw a HttpPresentationIOException to record the fact that
                // the client disconnected, but want to maintain the HttpPresentationException
                // API.
                throw new HttpPresentationException(new HttpPresentationIOException(except));
            }
        }
        return inputStream;
    }

    /**
     * Returns a string containing the lone value of the specified query
     * parameter, or null if the parameter does not exist. Presentation
     * writers should use this method only when they are sure that there is
     * only one value for the parameter.  If the parameter has (or could have)
     * multiple values, then use getParameterValues. If a multiple valued
     * parameter name is passed as an argument, the return value is
     * implementation dependent.
     *
     * @param name the name of the parameter whose value is required.
     * @see HttpPresentationRequest#getParameterValues
     */
    public String getParameter(String name) throws HttpPresentationException {
        return request.getParameter(name);
    }

    /**
     * Returns the values of the specified query parameter for the request as
     * an array of strings, or a 0 length array if the named parameter does
     * not exist.
     *
     * @param name the name of the parameter whose value is required.
     */
    public String[] getParameterValues(String name)
	throws HttpPresentationException
    {
	Object values = request.getParameterValues(name);
	if (values == null) {
	    return new String[0];
	}
        if (values instanceof String) {
            String[] valArray = new String[1];
            valArray[0] = (String)values;
            return valArray;
        }
        return (String[])values;
    }

    /**
     * Returns the parameter names for this request as an enumeration
     * of strings, or an empty enumeration if there are no parameters.
     */
    public Enumeration getParameterNames() throws HttpPresentationException {
        return request.getParameterNames();
    }

    /**
     * Returns the method with which the request was made. The returned
     * value can be "GET", "HEAD", "POST", or an extension method. Same
     * as the CGI variable REQUEST_METHOD.
     */
    public String getMethod() throws HttpPresentationException {
        return request.getMethod();
    }

    /**
     * Returns the request URI.
     */
    public String getRequestURI() throws HttpPresentationException {
        return request.getRequestURI();
    }

    /**
     * Returns the presentation URI.
     */
    public String getPresentationURI() throws HttpPresentationException {
        return request.getRequestURI();
    }

    /**
     * Returns the part of the request URI that refers to the application
     * object being invoked. Analogous to the CGI variable SCRIPT_NAME.
     *
     * @deprecated This method was named in a confusing manner; it
     *  returns the application, not presentation object path. Use
     *  <A HREF="getApplicationPath.html">getApplicationPath()</A>.
     */
    public String getPresentationPath() throws HttpPresentationException {
        //return request.getServletPath();
        return request.getContextPath();
    }

    /**
     * Returns the part of the request URI that refers to the presentation
     * object being invoked.
     */
    public String getPresentationObjectPath() throws HttpPresentationException {
        return presentationPath;
    }

    /**
     * Returns the part of the request URI after the presentation
     * manager servlet, upto and including the presentation object .po,
     * but not any path info.
     */
    public String getPresentationObjectRelativePath() throws HttpPresentationException {
        String reqPathInfo = request.getPathInfo();
        String poPathInfo = getPathInfo();

//System.err.println("@1@" +reqPathInfo+"@"+ poPathInfo+"@");
        // Return the relative path less path information.
        if (poPathInfo == null) {
            return "/";
        }
        if (poPathInfo.endsWith("/")) {
            poPathInfo = poPathInfo.substring(0, poPathInfo.length()-1);
        }
        if (reqPathInfo == null) {
            return poPathInfo;
        }
        if (!reqPathInfo.startsWith("/")) {
            reqPathInfo = "/" + reqPathInfo;
        }
        if (reqPathInfo.endsWith("/")) {
            reqPathInfo = reqPathInfo.substring(0, reqPathInfo.length()-1);
        }
//System.err.println("@2@" +reqPathInfo+"@"+ poPathInfo+"@" + reqPathInfo.substring(0, reqPathInfo.length()-poPathInfo.length()) + "@");

	int pos;
	if ((pos = reqPathInfo.indexOf(';')) < 0) {
	    return reqPathInfo;
	} else {
	    return reqPathInfo.substring(0, pos);
	}
    }

    /**
     * Returns the part of the request URI that refers to the application.
     * Analogous to the CGI variable SCRIPT_NAME.
     */
    public String getApplicationPath() throws HttpPresentationException {
        String contextPath = request.getContextPath();
        if (contextPath == null) {
            contextPath = "/";
        }
        else if (!contextPath.endsWith("/")) {
            contextPath += "/";
        }
        return contextPath + request.getServletPath();

    }

    /**
     * Returns optional extra path information following the presentation
     * path, but immediately preceding the query string. Returns null if
     * not specified. Same as the CGI variable PATH_INFO.
     */
    public String getPathInfo() throws HttpPresentationException {
        if (presentationPathInfo == null) {
            return null;
        } else {
            return presentationPathInfo;
        }
    }

    /**
     * Returns extra path information translated to a real path. Returns
     * null if no extra path information specified. Same as the CGI variable
     * PATH_TRANSLATED.
     */
    public String getPathTranslated() throws HttpPresentationException {
        return request.getPathTranslated();
    }

    /**
     * Returns the query string part of the presentation URI, or null if none.
     * Same as the CGI variable QUERY_STRING.
     */
    public String getQueryString() throws HttpPresentationException {
        return request.getQueryString();
    }

    /**
     * Returns the name of the user making this request, or null if not
     * known.  The user name is set with HTTP authentication.  Whether
     * the user name will continue to be sent with each subsequent
     * communication is browser-dependent.  Same as the CGI variable
     * REMOTE_USER.
     */
    public String getRemoteUser() throws HttpPresentationException {
        return request.getRemoteUser();
    }

    /**
     * Returns the authentication scheme of the request, or null if none.
     * Same as the CGI variable AUTH_TYPE.
     */
    public String getAuthType() throws HttpPresentationException {
        return request.getAuthType();
    }

    /**
     * Gets the array of cookies found in this request.
     *
     * @return The array of cookies found in this request.
     */
    public Cookie[] getCookies() throws HttpPresentationException {
        return request.getCookies();
    }


    /*
     * set the flag indicating whether the sessionId came from a cookie
     *@param isFromCookie boolean indicating whether sessionId came from cookie
     */
    public void setRequestedSessionIdFromCookie(boolean isFromCookie)
      throws HttpPresentationException {
      isRequestedSessionIdFromCookie = isFromCookie;
    }


    /**
     * Indicates whether client submitted their session id through a cookie
     * @return true if client submitted their sessionId via a cookie,
     *         false otherwise
     */
    public boolean isRequestedSessionIdFromCookie()
	throws HttpPresentationException {
       return isRequestedSessionIdFromCookie;
    }


    /*
     * set the flag indicating whether the sessionId came from a url
     * @param isFromUrl boolean indicating whether sessionId came from url
     */
    public void setRequestedSessionIdFromUrl(boolean isFromUrl)
      throws HttpPresentationException {
      isRequestedSessionIdFromUrl = isFromUrl;
    }



    /**
     * Indicates whether client submitted their sessionId through a
     * rewritten url
     * @return true if client submitted their sessionId via a rewritten url
     */
    public boolean isRequestedSessionIdFromUrl()
	throws HttpPresentationException {
	return isRequestedSessionIdFromUrl;
    }

    /**
     * Returns the value of a header field, or null if not known.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public String getHeader(String name) throws HttpPresentationException {
        return request.getHeader(name);
    }

    /**
     * Returns the value of an integer header field, or -1 if not found.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public int getIntHeader(String name) throws HttpPresentationException {
        return request.getIntHeader(name);
    }

    /**
     * Returns the value of a date header field, or -1 if not found.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public long getDateHeader(String name) throws HttpPresentationException {
        return request.getDateHeader(name);
    }

    /**
     * Returns an enumeration of strings representing the header names
     * for this request. Some server implementations do not allow headers
     * to be accessed in this way, in which case this method will return null.
     */
    public Enumeration getHeaderNames() throws HttpPresentationException {
        return request.getHeaderNames();
    }

    /**
     * Get the URI path for a file in the application.  This converts a path
     * to the file part of the URL.  It adds in the reference to application
     * servlet.
     *
     * @param file File with in the application.  Currently this must
     *  be a path relative to the presentation prefix.
     * @return The file path portion of the URL, starting with
     * a <CODE>/</CODE>.
     */
    public String getAppFileURIPath(String file)
	throws HttpPresentationException {

        StringBuffer uriPath;
        String presPath = getApplicationPath();

        if (presPath.charAt(presPath.length()-1) == '/') {
            uriPath = new StringBuffer(presPath.substring(0, presPath.length()-1));
        } else {
            uriPath = new StringBuffer(presPath);
        }

        // Convert system dependent file path into a URL file path.
        Stack names = new Stack();
        File fparse = new File(file);
        String parent;
        while ((parent = fparse.getParent()) != null) {
            names.push(fparse.getName());
            fparse = new File(parent);
        }
        names.push(fparse.getName());

        while (!names.empty()) {
            uriPath.append('/');
            uriPath.append((String)names.pop());
        }
        return uriPath.toString();
    }

    /**
     * Returns the total number of bytes processed as part of this request.
     * This is not supported for Jolt Presentations and is necessary
     * to support the DebugRequest interface.
     *
     * @see com.lutris.http.debug.DebugRequest
     */
    public int getTotalBytes() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
        return -1;
    }

    /**
     * Returns the part of the request URI that refers to the presentation
     * object being invoked. This is the same as getPresentationPath(),
     * but is necessary to support the DebugRequest interface.
     *
     * @see com.lutris.http.debug.DebugRequest
     */
    public String getPath() throws HttpPresentationException {
        return getPresentationPath();
    }
}
