/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ServletHttpPresentationResponse.java,v 1.31.2.8.4.8 2000/12/01 02:11:51 markd Exp $
 */

package com.lutris.appserver.server.httpPresentation.servlet;

import java.util.Enumeration;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import javax.servlet.http.*;
import com.lutris.http.debug.*;
import com.lutris.util.StringEnum;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.StandardAppUtil;
import com.lutris.appserver.server.session.SessionManager;
import com.lutris.appserver.server.session.StandardSessionManager;
import com.lutris.appserver.server.session.SessionException;
import org.w3c.dom.Document;
import org.w3c.dom.html.HTMLDocument;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.io.Encoding;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.URLRewriter;
import org.enhydra.xml.dom.DOMStats;

/**
 * Object passed to presentation objects that is used to generate HTTP
 * responses.
 */
public class ServletHttpPresentationResponse
    implements HttpPresentationResponse, DebugResponse {

    /*
     * Servlet objects.
     */
    private HttpServletResponse response;
    private HttpPresentationOutputStream outputStream = null; // Lazy allocation

    /*
     * Default encoding of the VM
     */
    private static final String defaultEncoding = System.getProperty("file.encoding");

    /*
     * Encoding for output
     */
    private String encoding = defaultEncoding;

    /*
     * The current session key
     */
    private String sessionKey = null;

    /*
     * The current session manager
     */
    private SessionManager sessionManager = null;

    /*
     * Boolean flag indicating if response requires a sessionId cookie
     */
    private boolean sessionIdCookie = true;

    /*
     * Boolean flag indicating if response requires url encoding for sessionId
     */
    private boolean sessionIdUrl = true;

    /**
     * If not null, log DOM statistics here after each DOM write.
     */
    private PrintWriter domStatsLogWriter;

    /**
     * Construct an object associated with a servlet response.
     *
     * @param response Servlet response object that this object will front-end.
     */
    protected ServletHttpPresentationResponse(HttpServletResponse response,
                                              PrintWriter domStatsLogWriter) {
        this.response = response;
        this.domStatsLogWriter = domStatsLogWriter;
    }

    /**
     * Returns the original HttpServletResponse.
     */
    public HttpServletResponse getHttpServletResponse() {
	return this.response;
    }

    /**
     * Sets the content length for this response.
     *
     * @param len the content length
     */
    public void setContentLength(int len)
	throws HttpPresentationException {

        response.setContentLength(len);
    }


    /**
     * Sets the content type for this response.
     *
     * @param type the content's MIME type
     */
    public void setContentType(String type)
	throws HttpPresentationException {

        response.setContentType(type);
    }


    /**
     * Returns an output stream for writing response data.
     */
    public HttpPresentationOutputStream getOutputStream()
	throws HttpPresentationException {

        if (outputStream == null) {
            try {
                outputStream = new ServletHttpPresentationOutputStream
			(this, response.getOutputStream());
            } catch (IOException except) {
                // Want to throw a HttpPresentationIOException to record the fact that
                // the client disconnected, but want to maintain the HttpPresentationException
                // API.
                throw new HttpPresentationException(new HttpPresentationIOException(except));
            }
        }
        return outputStream;
    }


    /**
     * Adds the specified cookie to the response. It can be called multiple
     * times to set more than one cookie.
     *
     * @param cookie - The Cookie to return to the client.
     */
    public void addCookie(Cookie cookie)
	throws HttpPresentationException {
        response.addCookie(cookie);
    }


    /**
     * Returns true if the response message header has a field with
     * the specified name.
     * @param name the header field name
     */
    public boolean containsHeader(String name)
	throws HttpPresentationException {

        return response.containsHeader(name);
    }


    /**
     * Sets the status code and message for this response.
     *
     * @param sc the status code
     * @param sm the status message
     */
    public void setStatus(int sc, String sm)
	throws HttpPresentationException {

        response.setStatus(sc, sm);
    }


    /**
     * Sets the status code and a default message for this response.
     *
     * @param sc the status code
     */
    public void setStatus(int sc)
	throws HttpPresentationException {

        response.setStatus(sc);
    }


    /**
     * Adds a field to the response header with a given name and
     * value.  If the field had already been set, the new value
     * overwrites the previous one.  The containsHeader method can be
     * used to test for the presence of a header before setting its
     * value.
     *
     * @param name the header field name
     * @param value the header field value
     */
    public void setHeader(String name, String value)
	throws HttpPresentationException {

        response.setHeader(name, value);
    }


    /**
     * Adds a field to the response header with a given name and
     * integer value.  If the field had already been set, the new
     * value overwrites the previous one.  The containsHeader method
     * can be used to test for the presence of a header before setting
     * its value.
     *
     * @param name the header field name
     * @param value the header field integer value
     */
    public void setIntHeader(String name, int value)
	throws HttpPresentationException {

        response.setIntHeader(name, value);
    }


    /**
     * Adds a field to the response header with a given name and
     * date-valued field.  The date is specified in terms of
     * milliseconds since the epoch.  If the date field had already
     * been set, the new value overwrites the previous one.  The
     * containsHeader method can be used to test for the presence of a
     * header before setting its value.
     *
     * @param name the header field name
     * @param value the header field date value
     */
    public void setDateHeader(String name, long date)
	throws HttpPresentationException {

        response.setDateHeader(name, date);
    }


    /**
     * Sends an error response to the client using the specified status
     * code and descriptive message.
     *
     * @param sc the status code
     * @param msg the detail message
     */
    public void sendError(int sc, String msg)
	throws HttpPresentationException {

	try {
	    response.sendError(sc, msg);
	} catch (IOException except) {
            // Want to throw a HttpPresentationIOException to record the fact that
            // the client disconnected, but want to maintain the HttpPresentationException
            // API.
            throw new HttpPresentationException(new HttpPresentationIOException(except));
	}
    }

    /**
     * Sends an error response to the client using the specified
     * status code and a default message.
     *
     * @param sc the status code
     */
    public void sendError(int sc)
	throws HttpPresentationException {

	try {
            response.sendError(sc);
	} catch (IOException except) {
            // Want to throw a HttpPresentationIOException to record the fact that
            // the client disconnected, but want to maintain the HttpPresentationException
            // API.
            throw new HttpPresentationException(new HttpPresentationIOException(except));
	}
    }

    /**
     * Called at the end of processing a response to force any cached
     * headers to be written and buffers flushed.
     */
    public void flush() throws HttpPresentationException {

	try {
            // Before version 3.0 of enhydra this was getOutputStream().flush()
            // the getOutStream() can throw an exception if the writer was used
            // instead of the outputstream.
            response.flushBuffer();
        } catch (IOException except) {
            // Want to throw a HttpPresentationIOException to record the fact that
            // the client disconnected, but want to maintain the HttpPresentationException
            // API.
            throw new HttpPresentationException(new HttpPresentationIOException(except));
        }
    }

    /**
     * Encodes a url with session id for use with cookieless browsers.
     *
     * @param url the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     */
    public String encodeUrl(String url) {
	return response.encodeUrl(url);
    }

    /**
     * Encodes a url with session id for sendRedirect for use with
     * cookieless browsers.
     *
     * @param url the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     *
     * @see #sendRedirect
     * @see #encodeUrl
     */
    public String encodeRedirectUrl(String url) {
	return response.encodeUrl(url);
    }




    /*
     * The following methods are for the debugging interface...
     */

    /**
     * Returns the size of the request entity data, or -1 if not available.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the content length.
     */
    public int getContentLength() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
	return -1;
    }


    /**
     * Returns the Internet Media Type of the response, or null if not known
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the content type.
     */
    public String getContentType() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
	return null;
    }


    /**
     * Returns the status code for the response, or -1 if not available.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the status code.
     */
    public int getStatusCode() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
	return -1;
    }


    /**
     * Returns the status message for the response, or null if not available.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the status message.
     */
    public String getStatusMessage() {
	// FIX: When we have our own servlet implementation, we can
	// FIX: support this.
	return null;
    }


    /**
     * Returns an array of cookies to be set by this response, or
     * a 0 length array if the information is not available.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return The array of cookies found in this response.
     */
    public Cookie[] getCookies() {
	// FIX: Not supported - our servlet response implementation
        // FIX: should implement this debug interface directly to save
	// FIX: storing this stuff here.
	return new Cookie[0];
    }


    /**
     * Returns the values of the specified header for the response as
     * an array of strings, or a 0 length array if the named header does
     * not exist.
     * This is necessary to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @param name the case-insensitive header field name
     * @return the value of the supplied header.
     */
    public String getHeader(String name) {
	// FIX: Not supported - our servlet response implementation
        // FIX: should implement this debug interface directly to save
	// FIX: storing this stuff here.
	return "";
    }


    /**
     * Returns an array of strings representing the header names
     * for this response. This does not have return cookies - use
     * getCookies().
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return name the case-insensitive header field name
     */
    public Enumeration getHeaderNames() {
	// FIX: Not supported - our servlet response implementation
        // FIX: should implement this debug interface directly to save
	// FIX: storing this stuff here.
	return new StringEnum(new String[0]);
    }


    /**
     * Returns the total number of bytes that make up this response.
     * This is not supported for Jolt Presentations and is necessary
     * to support the DebugResponse interface.
     *
     * @see com.lutris.http.debug.DebugResponse
     * @return the total number of bytes in the response.
     */
    public int getTotalBytes() {
	// FIX: Not supported - our servlet response implementation
        // FIX: should implement this debug interface directly to save
	// FIX: storing this stuff here.
        return -1;
    }

    /**
     * Returns null.
     * Normally this function returns the data written to the output
     * stream associated with this request.
     */
    public String getResponseData() {
        return null;
    }

    /**
     * Set the output character encoding.
     *
     * @param enc character encoding
     */
    public void setEncoding(String enc) {
        encoding = enc;
    }

    /**
     * Get the output character encoding.
     */
    public String getEncoding() {
        return encoding;
    }

    /**
     * Sets the current session key for this response
     *
     * @param sessionKey
     *        The current sessionKey
     **/
    public void setSessionKey(String sessionKey) {
	this.sessionKey = sessionKey;
    }

    /**
     * Sets the session manager.
     **/
    public void setSessionManager(SessionManager sessionManager) {
	this.sessionManager = sessionManager;
    }

    /**
     * Indicates whether client response requires a sessionId cookie
     * @return true if client response requires a sessionId cookie
     *         false otherwise
     */
    public boolean isSessionIdCookieRequired()
	throws HttpPresentationException {
	return sessionIdCookie;
    }

    /*
     * set boolean flag for sessionId response cookie.  Indicates whether
     * client response requires a sessionId cookie
     * @param isFromCookie boolean flag
     */
    public void setSessionIdCookieRequired(boolean sessionIdCookie)
	throws HttpPresentationException {
	this.sessionIdCookie = sessionIdCookie;
    }

    /**
     * Indicates whether client response requires url encoding for
     * sessionId
     * @return true if client response requires url encoding for
     *         sessionId; false otherwise
     */
    public boolean isSessionIdEncodeUrlRequired()
	throws HttpPresentationException {
	return sessionIdUrl;
    }
    /*
     * Set boolean flag url encoding for sessionId.  Indicates whether
     * client response requires url encoding for sessionId
     * @param sessionIdUrl boolean flag
     */
    public void setSessionIdEncodeUrlRequired(boolean sessionIdUrl)
	throws HttpPresentationException {
	this.sessionIdUrl = sessionIdUrl;
    }

    /**
     * Determine if a session key should be encoded in URLs, based on
     * what is requested and session being valid.
     */
    private boolean shouldEncodeSessionKey() throws HttpPresentationException {
        try {
            return (sessionIdUrl && (sessionManager != null) && (sessionKey != null)
                    && sessionManager.sessionExists(sessionKey)
                    && (sessionManager instanceof StandardSessionManager)
                    && !((StandardSessionManager)sessionManager).getEncodeUrlState().equalsIgnoreCase(StandardSessionManager.ENCODE_URL_NEVER));
        } catch (SessionException except) {
            throw new HttpPresentationException(except);
        }
    }

    /**
     * Create an OutputOptions object for a document.  Options are default for
     * the specified document.  The object maybe then modified as needed to
     * override the default values.
     * <P>
     * The following attributes are set in the object:
     * <UL>
     * <LI> encoding
     * <LI> MIME type - Defaults for document.
     * <LI> URLRewriter - Set if URL encoding of sessions is enabled.
     * </UL>
     */
    public OutputOptions createOutputOptions(XMLObject document)
        throws HttpPresentationException {

        OutputOptions options = DOMFormatter.getDefaultOutputOptions(document);

        // Encoding encoding
        String mimeEncoding = document.getEncoding();
        if (encoding != null) {
            // Will return null if its invalid
            mimeEncoding = Encoding.java2XmlEncoding(encoding);
        }
        options.setXmlEncoding(mimeEncoding);

        // Get the default MIME type from object
        options.setMIMEType(document.getMIMEType());

        // Enable encoding of session id in URL, if needed.
        if (shouldEncodeSessionKey()) {
            options.setURLRewriter(new URLRewriter() {
                    public String rewriteURL(String urlAttrValue) {
                        if (StandardAppUtil.pointsToPO(urlAttrValue)) {
                            return StandardAppUtil.encodeUrl(urlAttrValue, sessionKey);
                        } else {
                            return urlAttrValue;
                        }
                    }
                });
        }
        return options;
    }

    /**
     * Write a string document represented as a byte array.
     */
    private void outputDocumentBytes(byte[] docBytes,
                                     String mimeEncoding,
                                     String mimeType)
        throws HttpPresentationException, IOException {

        // Set headers
        setContentLength(docBytes.length);
        if (mimeType != null) {
            if (mimeEncoding != null) {
                setContentType(mimeType + "; charset=" + mimeEncoding);
            } else {
                setContentType(mimeType);
            }
        }
        
        // Disable caching of page unless otherwise specified.
        if (getHeader("Cache-Control") == null) {
            setHeader("Cache-Control", "no-cache");
        }
        if (getHeader("Expires") == null) {
            setHeader("Expires", "0");
        }

        // Output
        getOutputStream().write(docBytes);
        getOutputStream().flush();
    }

    /**
     * Do common operations for writing DOM document.  Sets approriate HTTP
     * header, including setting of cache-control headers, if not already set.
     * Handles encoding conversion and output.
     */
    private void outputDocument(XMLObject document,
                                OutputOptions outputOptions)
        throws HttpPresentationException, IOException {

        // Convert to bytes and output
        DOMFormatter formatter = new DOMFormatter(outputOptions);
        outputDocumentBytes(formatter.toBytes(document), 
                            outputOptions.getXmlEncoding(),
                            outputOptions.getMIMEType());
        // Log if requested
        if (domStatsLogWriter != null) {
            // Buffer results so it gets logged as a single entry
            StringWriter buf = new StringWriter(2048);
            PrintWriter writer = new PrintWriter(buf);
            DOMStats.printStats("Write " + document.getClass().getName(),
                                document, 0, writer);
            writer.flush();
            domStatsLogWriter.println(buf.toString());
        }
    }

    /**
     * Output an an XMLC document object (DOM). The document is formatted
     * according to it's type.  The MIME type of the response is automatically
     * set.
     *
     * @param outputFormat Object use to specify options controling formatting
     *  of the document.
     * @param doc The DOM object to be returned as response.
     */
    public void writeDOM(OutputOptions outputOptions,
                         XMLObject document) throws HttpPresentationException {
        try {
            outputDocument(document, outputOptions);
        } catch (IOException except) {
            throw new HttpPresentationException(except);
        }
    }

    /**
     * Output an an XMLC document object (DOM). The document is formatted
     * according to it's type.  The MIME type of the response is automatically
     * set.
     *
     * @param doc The DOM object to be returned as response.
     */
    public void writeDOM(XMLObject document) throws HttpPresentationException {
        
        OutputOptions outputOptions = createOutputOptions(document);
        if (document instanceof HTMLDocument) {
            // special, ugly hack for IE
            outputOptions.setDropHtmlSpanIds(true);
        }
        try {
            outputDocument(document, outputOptions);
        } catch (IOException except) {
            throw new HttpPresentationException(except);
        }
    }

    /**
     * Utility method to output an HTML page. The appropriate headers are
     * set for MIME type and to disable caching of the HTML by the broswer.
     */
    public void writeHTML(String html) throws HttpPresentationException {
        byte bytes[];

        // Determine encoding to use.
        String mimeEncoding = null;
        if (encoding != null) {
            // Will return null if its invalid
            mimeEncoding = Encoding.java2XmlEncoding(encoding);
        }

        try {
            if (encoding != null) {
                bytes = html.getBytes(encoding);
            } else {
                bytes = html.getBytes();
            }
            outputDocumentBytes(bytes, mimeEncoding, "text/html");
        } catch (IOException except) {
            throw new HttpPresentationException(except);
        }
    }

    /**
     * Utility method to output an HTML page from a DOM object. The
     * appropriate headers are set for MIME type and to disable caching
     * of the HTML by the broswer.
     *
     * @param doc The DOM object to be returned as response
     */
    public void writeHTML(HTMLDocument doc) throws HttpPresentationException {
        writeDOM((XMLObject)doc);
    }
}
