/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BasicSessionHome.java,v 1.5.2.1.4.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session;

import com.lutris.appserver.server.Enhydra;
import com.lutris.util.*;
import com.lutris.logging.*;
import java.util.*;

/**
 * BasicSessionHome creates instances of BasicSession and
 * manages the 'active'/'passive' state of those sessions.
 * All sessions created by BasicSessionHome are held in
 * memory.<p>
 *
 * The following parameters can be used to configure
 * the BasicSessionHome.  They should be grouped together in a section,
 * normally <code>SessionManager.SessionHome</code>, which is specified to
 * the constructor.<p>
 *
 * <ul>
 * <li><code>MaxSessions: {int}</code><p>
 * Specifies the maximum number of in use concurrent sessions.  If this value
 * is exceeded then CreateSession() will throw a CreateSessionException.
 * -1 indicates unlimited session.  If MaxSessions is not set
 * in the application's configuration then MaxSessions defaults to
 * unlimited sessions.<p>
 * </ul>
 *
 * @see         BasicSession
 * @see         StandardSession
 * @see         StandardSessionManager
 * @version	$Revision: 1.5.2.1.4.1 $
 * @author	Kyle Clark
 */
public class BasicSessionHome implements StandardSessionHome {

    /**
     * The active cache holds sessions that are in the
     * 'active' state.
     */
    private Hashtable activeCache = new Hashtable();

    /**
     * The active thread cache holds all the association
     * between active sessions and the threads that have
     * checked them out.
     */
    private Hashtable activeThreadCache = new Hashtable();

    /**
     * The active cache holds sessions that are in the
     * 'active' state.  There is a single mapping
     * of keys to session objects: sessionKey -> session
     */
    private Hashtable passiveCache = new Hashtable();

    /**
     * Maximum number of allowed sessions.
     */
    private long maxSessions = -1;

    /**
     * Configuration keys.
     */
    private final String MAX_SESSIONS_KEY = "MaxSessions";
    private final long UNDEFINED_MAX_SESSIONS = -1;

    /**
     * The session manager associated with the session home.
     */
    private StandardSessionManager sessionMgr;

    /**
     * @param sessionMgr
     *   The session manager associated with this session home.
     * @param config
     *   Object parsed from configuration file.  This should be
     *   for the section constaining the standard session home configuration.
     * @exception ConfigException
     *   signifies a problem in the configuration file.
     */
    public BasicSessionHome(StandardSessionManager sessionMgr,
                            Config config)
        throws ConfigException {
        this.sessionMgr = sessionMgr;
        if (config.containsKey(MAX_SESSIONS_KEY)) {
            maxSessions = config.getLong(MAX_SESSIONS_KEY);
            debug(MAX_SESSIONS_KEY + " = " + maxSessions);
        }
        if (maxSessions <= 0) {
            maxSessions = UNDEFINED_MAX_SESSIONS;
        }
    }

    /**
     * @param sessionMgr
     *   The session manager associated with this session home.
     * @param config
     *   Object parsed from configuration file.  This should be
     *   for the section constaining the standard session home configuration.
     * @param loader
     *   The class load to use when load objects from persistent store.
     * @exception ConfigException
     *   signifies a problem in the configuration file.
     */
    public BasicSessionHome(StandardSessionManager sessionMgr,
                            Config config, ClassLoader loader)
        throws ConfigException {
        this(sessionMgr, config);
    }

    /**
     * Creates and returns a new session instance.  The
     * session is bound to the specified session key.  The
     * session is also associated with the current thread
     * and is considered in the 'active' state.  The
     * session remains in the 'active' state until the
     * thread puts the session into the 'passive' state..
     * Only this thread will be able to put the
     * session into the 'passive' state.
     *
     * @param sessionKey the key to associate with the session.
     * @return the newly created session.
     * @exception CreateSessionException if the session cannot be
     *   created.
     * @exception DuplicateKeyException if the session cannot
     *   be created because the key is already in use.
     * @exception SessionException if an error occurs.
     * @see #passivateSession
     */
    public synchronized StandardSession createSession(String sessionKey)
        throws CreateSessionException, DuplicateKeyException, SessionException {
        if (containsKey(sessionKey)) {
            throw new DuplicateKeyException("Session key " +
                                            sessionKey + " is already in use.");
        }
        if ((maxSessions != UNDEFINED_MAX_SESSIONS) &&
            (maxSessions <= size())) {
            cleanupNewSession();
            if (maxSessions <= size()) {
                throw new CreateSessionException("Maximum session limit ("
                                                 + maxSessions +
                                                 ") has been reached.");
            }
        }
        BasicSession session = new BasicSession(sessionMgr, sessionKey);
        SessionThread activeKey = new SessionThread(Thread.currentThread(),
                                                    sessionKey);
        // keep track of # of threads that reference session
        session.incrementRefCount();
        activeCache.put(sessionKey, session);
        activeThreadCache.put(activeKey, session);
        debug(3, "create session: key = " + sessionKey);
        return (StandardSession)session;
    }

    /**
     * Returns the session bound to the session key.
     * The session must already be in the 'active' state
     * and associated with the current thread,
     * otherwise null is returned.
     *
     * @param sessionKey
     *   the session key for the session.  If the session
     *   doesn't exist or is not is not bound to the current
     *   thread then null is returned.
     * @return
     *   the session.
     * @exception SessionException
     *   if the session cannot be retrieved.
     * @see #getSession(Thread, String)
     */
    public synchronized StandardSession getSession(String sessionKey)
        throws SessionException {
        debug(3, "get session: key = " + sessionKey);
        SessionThread activeKey = new SessionThread(Thread.currentThread(),
                                                    sessionKey);
        return (BasicSession)activeThreadCache.get(activeKey);
    }

    /**
     * Returns the session bound to the specified
     * session key.  The session is put into the 'active' state.
     * The session is also associated with the specified thread.
     * Only this thread will be able to put the session
     * back into the 'passive' state once it is done with the
     * session.
     *
     * @param thread the thread that should be associated with
     *   the session while it is in the active state.  Only this
     *   thread can put the session back into the passive state.
     * @param sessionKey
     *   the session key for the session that will be made
     *   'active' and returned.  If the session doesn't exist
     *   then null is returned.
     * @return
     *   the session.
     * @exception SessionException
     *   if the session cannot be retrieved.
     * @see #passivateSession
     */
    public synchronized StandardSession getSession(Thread thread, String sessionKey)
        throws SessionException {
        debug(3, "get session: key = " + sessionKey);
        SessionThread activeKey = new SessionThread(thread, sessionKey);
        BasicSession s = null;
        // First check if the session is in the passive state.
        if (passiveCache.containsKey(sessionKey)) {
            s = (BasicSession)passiveCache.remove(sessionKey);
            activeCache.put(sessionKey, s);
            activeThreadCache.put(activeKey, s);
            s.incrementRefCount();
        } else {
            // The session must already be active.
            if (activeThreadCache.containsKey(activeKey)) {
                // It's already checked out by the thread
                s = (BasicSession)activeThreadCache.get(activeKey);
            } else {
                // It's being checked out by a new thread
                s = (BasicSession)activeCache.get(sessionKey);
                if (s != null) {
                    activeThreadCache.put(activeKey, s);
                    // Only increment the ref count when a new thread
                    // checks out the session.
                    s.incrementRefCount();
                }
            }
        }
        return (StandardSession)s;
    }


    /**
     * Removes a session from the cache.  If the session
     * doesn't exist the opration is ignored.
     *
     * @param sessionKey
     *   the session key associated with the session.
     * @exception SessionException
     *   if the session cannot be retrieved.
     */
    public synchronized void removeSession(String sessionKey)
        throws SessionException {
        debug(3, "remove session: key = " + sessionKey);
        Enumeration e = activeThreadCache.keys();
        while (e.hasMoreElements()) {
            SessionThread key = (SessionThread)e.nextElement();
            if (key.sessionKey.equals(sessionKey)) {
                activeThreadCache.remove(key);
            }
        }
        activeCache.remove(sessionKey);
        passiveCache.remove(sessionKey);
    }

    /**
     * Puts a session into the 'passive' state.  A 'passive'
     * session may be made persistent.  Only the thread that
     * put the session into the 'active' state may put
     * the session into the 'passive' state.
     *
     * @param thread the thread that is currently associated
     *   with the session.
     * @param sessionKey
     *   the session key for the session that will be made passive.
     * @exception SessionException
     *   if the session cannot be retrieved.
     */
    public synchronized void passivateSession(Thread thread, String sessionKey)
        throws SessionException {
        // This is a bad place for this debug() to be.
        // We should avoid any delay inside the synchronized block
        //debug(3, "passivate session: key = " + sessionKey);
        SessionThread activeKey = new SessionThread(thread, sessionKey);
        if (activeThreadCache.containsKey(activeKey)) {
            BasicSession s = (BasicSession)activeThreadCache.remove(activeKey);
            if (s.decrementRefCount() == 0) {
                activeCache.remove(sessionKey);
                passiveCache.put(sessionKey, s);
            }
        }
    }

    /**
     * Specifies if a key is currently bound to a session.
     *
     * @param sessionKey
     *   the session key to be tested.
     * @return
     *   true if the session key is in use.
     * @exception SessionException
     *   if the existence of the key cannot be determined.
     */
    public boolean containsKey(String sessionKey)
        throws SessionException {
        return (activeCache.containsKey(sessionKey)
                || passiveCache.containsKey(sessionKey));
    }

    /**
     * Returns the current number of sessions.
     *
     * @return
     *   the 'active' session count.
     * @exception SessionException
     *   if the size cannot be determined
     */
    public int size()
        throws SessionException {
        return (activeCache.size() + passiveCache.size());
    }

    /**
     * Returns the current number of sessions that are paged to
     * persistent store.
     *
     * @return
     *   the 'paged' session count.
     * @exception SessionException
     *   if the size cannot be determined
     */
    public int pagedSize() throws SessionException {
        return 0;
    }

    /**
     * Returns an enumeration of the keys for all the sessions.
     *
     * @return
     *   the enumeration of session keys.
     * @exception SessionException
     *   if the session enumeration cannot be retrieved.
     */
    public synchronized Enumeration keys() throws SessionException {
        Vector v = new Vector();
        Enumeration e;
        e = activeCache.keys();
        while (e.hasMoreElements()) {
            v.addElement(e.nextElement());
        }
        e = passiveCache.keys();
        while (e.hasMoreElements()) {
            v.addElement(e.nextElement());
        }
        return v.elements();
    }

    /**
     * Cleans up (removes) the oldest unused session from the passive cache.
     *
     * @exception SessionException if an error occurs.
     */
    private void cleanupNewSession() throws SessionException {
        Enumeration e = passiveCache.keys();
        StandardSession oldest = null;
        while (e.hasMoreElements()) {
            StandardSession s = (StandardSession)passiveCache.get(e.nextElement());
            if (s.isNew()) {
                if ((oldest == null)
                    || (s.getTimeCreated() < oldest.getTimeCreated())) {
                    oldest = s;
                }
            }
        }
        if (oldest != null) {
            removeSession(oldest.getSessionKey());
        }
    }

    /**
     * Shuts dows the session home.
     */
    public void shutdown() {
        // noop
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param msg the message to print.
     */
    private void debug(String msg) {
        debug(0, msg);
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param level the debug level.
     * @param msg the message to print.
     */
    private void debug(int level, String msg) {
        int dbg = Logger.DEBUG;
        switch (level) {
        case 1:
            dbg = Logger.DEBUG1;
            break;
        case 2:
            dbg = Logger.DEBUG2;
            break;
        case 3:
            dbg = Logger.DEBUG3;
            break;
        case 4:
            dbg = Logger.DEBUG4;
            break;
        case 5:
            dbg = Logger.DEBUG5;
            break;
        case 6:
            dbg = Logger.DEBUG6;
            break;
        case 7:
            dbg = Logger.DEBUG7;
            break;
        case 8:
            dbg = Logger.DEBUG8;
            break;
        case 9:
            dbg = Logger.DEBUG9;
            break;
        default:
            dbg = Logger.DEBUG;
            break;
        }

        Enhydra.getLogChannel().write(dbg, "BasicSessionHome("
                                      + Thread.currentThread().getName()
                                      + "): " + msg);
    }

}
