/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BasicSessionUserTable.java,v 1.3.2.1.4.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session;
import com.lutris.appserver.server.Enhydra;
import java.util.Hashtable;
import java.util.Enumeration;
import com.lutris.logging.*;
import com.lutris.util.*;
import com.lutris.appserver.server.user.User;

/**
 * Table used by StandardSessionManager to cross reference <CODE>User</CODE>
 * objects and sessions.  This table does not hold references to the
 * user and session objects.  Instead it only cross-references the
 * session keys and user names.  An underlying assumption is that the
 * user name can be used as a unique reference to the user.
 *
 * N.B.  It is assumed that this class is only used by StandardSessionManager
 * and that it is responsible for providing high level locks instead of
 * synchronizing
 *
 * @version	$Revision: 1.3.2.1.4.1 $
 * @author      Mark Diekhans
 * @author      Kyle Clark
 */
class BasicSessionUserTable implements StandardSessionUserTable {

    /**
     * Table indexed by a <CODE>User</CODE> object. Each entry contains
     * a Hashtable of sessions keys associated with the user name.
     * The hash table is indexed by the <CODE>Session</CODE> object.
     */
    private Hashtable userSessionTable = new Hashtable();

    /**
     * A empty hash table used to return empty enumerations.
     */
    static final private Hashtable emptyTable = new Hashtable();

    /**
     * Creates an instance of this table.  The config options
     * are ignored.
     *
     * @param config
     *   configuration options for this table.
     */
    public BasicSessionUserTable(Config config) {
    }

    /**
     * Add a session key to the user name of the session xref table.
     *
     * @param session
     *   The session key.  
     * @param user
     *   the user to associated with the session key.
     */
    public void add(String sessionKey, User user) {
	Hashtable sessions = (Hashtable)userSessionTable.get(user.getName());
        if (sessions == null) {
            sessions = new Hashtable();
            userSessionTable.put(user.getName(), sessions);
        }
        debug(2, "add session: user=" + user.getName()
              + ": sessionKey=" + sessionKey);
        sessions.put(sessionKey, sessionKey); // Have to put something!
    }

    /**
     * Remove a session from the user to session mapping table.
     * If the session is not it the table, it is ignored.
     *
     * @param sessionKey The session key, with the user already
     * filled in.
     * @param user the user associated with the session.
     */
    public void remove(String sessionKey, User user) {
	Hashtable sessions = (Hashtable)userSessionTable.get(user.getName());
        if (sessions != null) {
            sessions.remove(sessionKey);
            if (sessions.size() == 0) {
                // Remove the user.
                userSessionTable.remove(user.getName());
            }
        }
        debug(2, "remove session: user=" + user.getName()
              + ": sessionKey=" + sessionKey);
    }

    /**
     * Removes all references to a session from the user session table.
     *
     * @param sessionKey The session key, with the user already
     * filled in.
     */
    public void remove(String sessionKey) {
        Enumeration users = userSessionTable.keys();
        while (users.hasMoreElements()) {
            String userName = (String)users.nextElement();
            Hashtable sessions = (Hashtable)userSessionTable.get(userName);
            if (sessions != null) {
                sessions.remove(sessionKey);
                if (sessions.size() == 0) {
                    // Remove the user.
                    userSessionTable.remove(userName);
                }
            }
        }
        debug(2, "remove sessionKey from user table:" + sessionKey);
    }

    /**
     * Get the number of sessions for a user.
     *
     * @param user The user object to check for.
     * @return The count of the number of sessions associated with this
     *  user.
     */
    public int numSessions(User user) {
	Hashtable sessions = (Hashtable)userSessionTable.get(user.getName());
        if (sessions == null) {
            return 0;
        } else { 
            return sessions.size();
        }
    }

    /**
     * Returns the session keys associated with a particular user.
     *
     * @param user The user object to check for.
     * @return An enumeration of the session keys associated
     *   with the user.
     */
    public Enumeration getSessionKeys(User user) {
	Hashtable sessions = (Hashtable)userSessionTable.get(user.getName());
        if (sessions == null) {
            debug(2, "getSessionKeys: " + user.getName()
                  + ": totalKeys=0");
            return emptyTable.elements();
        } else { 
            debug(2, "getSessionKeys: " + user.getName()
                  + ": totalKeys=" + sessions.size());
            return sessions.elements();
        }
    }

    /**
     * Shutdown this session user table as required.  The session
     * user table should not be used after this method has been called
     */
    public void shutdown() {
        // TODO
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param msg the message to print.
     */
    private void debug(String msg) {
        debug(0, msg);
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param level the debug level.
     * @param msg the message to print.
     */
    private void debug(int level, String msg) {
        int dbg = Logger.DEBUG;
        switch (level) {
        case 1:
            dbg = Logger.DEBUG1;
            break;
        case 2:
            dbg = Logger.DEBUG2;
            break;
        case 3:
            dbg = Logger.DEBUG3;
            break;
        case 4:
            dbg = Logger.DEBUG4;
            break;
        case 5:
            dbg = Logger.DEBUG5;
            break;
        case 6:
            dbg = Logger.DEBUG6;
            break;
        case 7:
            dbg = Logger.DEBUG7;
            break;
        case 8:
            dbg = Logger.DEBUG8;
            break;
        case 9:
            dbg = Logger.DEBUG9;
            break;
        default:
            dbg = Logger.DEBUG;
            break;
        }
        
        Enhydra.getLogChannel().write(dbg, "StandardSessionUserTable("
                                      + Thread.currentThread().getName()
                                      + "): " + msg);
    }
}
