/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DiskPagedSessionHome.java,v 1.2.14.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session;

import com.lutris.appserver.server.Enhydra;
import com.lutris.util.*;
import com.lutris.logging.*;
import java.util.*;
import java.io.IOException;

/**
 * The StandardSessionManager uses PagedSessionHome to manage
 * a collection of sessions that can be paged to disk.<p>
 *
 * PagedSessionHome will page sessions to disk
 * as soon as a pre-configured threshold of sessions has been
 * reached.  Only sessions in the 'passive' state will be
 * paged to disk.  If all sessions are in the 'active' state
 * and the threshold has been reached, then a request to create
 * a new session will block until one of the 'active' sessions
 * goes into the 'passive' state.  At this point the session
 * that just became 'passive' is paged to disk and a new
 * session is created.<p>
 *
 * Sessions are paged to disk by serializing
 * the all data (excluding the session manager) that is
 * associated with a session.  This requires that
 * the session data and user associated with a session
 * are serializable.<p>
 *
 * The following parameters can be used to configure
 * the PagedSessionHome.  They should be grouped together in a section,
 * normally <code>SessionManager.SessionHome</code>, which is specified to
 * the constructor.<p>
 *
 * <ul>
 *
 * <li><code>PageThreshold: {int}</code><p>
 * Sessions are paged if the total number of sessions exceeds PageThreshold.
 * If not set, defaults to 100.<p>
 *
 * <li><code>PageWait: {int}</code><p>
 * Specifies the maximum time (in milliseconds) that a thread will
 * wait for a session to be created or retrieved.  If the
 * page threshold has been reached, then a thread has to wait for
 * an existing session to be paged before a new session can be
 * created or a pre-existing session (that has been paged to
 * disk) can be activated.  If not set, defaults to 1 minute.
 * If this time is exceeded then an exception is thrown - this
 * prevents a possible deadlock situation.<p>
 *
 * <li><code>PageDir: {String}</code><p>
 * The directory where sessions will be paged.  This setting is
 * required.<p>
 *
 * <li><code>MaxSessions: {int}</code><p>
 * Specifies the maximum number of in use concurrent sessions.  If this value
 * is exceeded then CreateSession() will throw a CreateSessionException.
 * -1 indicates unlimited session.  If MaxSessions is not set
 * in the application's configuration then MaxSessions defaults to
 * unlimited sessions.<p>
 *
 * </ul>
 *
 * @see         StandardSession
 * @see         StandardSessionManager
 * @version	$Revision: 1.2.14.1 $
 * @author	Kyle Clark
 */
public class DiskPagedSessionHome extends PagedSessionHome {

    /**
     * The paged cache contains handles to passive
     * sessions that have been paged to disk.
     */
    private Hashtable pagedCache = new Hashtable();

    /**
     * The persistent storage interface.
     */
    private PersistentStore store;
        
    /**
     * Configuration keys.
     */
    private final String PAGE_DIR_KEY = "PageDir";
    
    /**
     * @param sessionMgr
     *   The session manager associated with this session home.
     * @param config
     *   Object parsed from configuration file.  This should be
     *   for the section containing the standard session home configuration.
     * @param loader
     *   The class load to use when load objects from persistent store.
     * @exception ConfigException
     *   signifies a problem in the configuration file.
     * @exception SessionException
     *   if the initialization fails.
     */
    public DiskPagedSessionHome(StandardSessionManager sessionMgr,
                                Config config, ClassLoader loader)
        throws SessionException, ConfigException {
        super(sessionMgr, config, loader);
        String pageDir = config.getString(PAGE_DIR_KEY);
        debug(PAGE_DIR_KEY + " = " + pageDir);
        try {
            store = new FilePersistentStore(pageDir, loader);
        } catch (Exception e) {
            throw new SessionException(e);
        }
    }

    /**
     * Creates a new session object.  This method is intended to be
     * overriden by classes that extend PagedSessionHome.
     *
     * @return a new session.
     */
    protected PagedSession newSession(StandardSessionManager mgr,
                                      String sessionKey)
        throws SessionException {
        return new PagedSession(mgr, sessionKey);
    }

    /**
     * Deletes a paged session.  If the session doesn't exist then this
     * is a noop.
     *
     * @param sessionKey the key identifying the session
     *   that should be deleted.  
     */
    protected void deleteSession(String sessionKey)
        throws SessionException {
        PagedSessionHandle pagedSession =
            (PagedSessionHandle)pagedCache.remove(sessionKey);
        if (pagedSession != null) {
            pagedSession.delete();
        }
    }

    /**
     * Pages a session to disk.
     *
     * @param session the session to page.
     * @exception SessionException if the paged session could not be
     *   paged out.
     */
    protected synchronized void pageOut(PagedSession s) throws SessionException {
        debug(3, "page: write session to disk: " + s.getSessionKey());
        PagedSessionHandle page = new PagedSessionHandle(s, store);
        page.write();
        pagedCache.put(s.getSessionKey(), page);
    }

    /**
     * Reads a paged session from disk.
     *
     * @param sessionKey the key identifying the session that should
     *   be paged in.
     * @return the paged session that was read in.
     * @exception SessionException if the paged session could not be
     *   read in or does not exist.
     */
    protected synchronized PagedSession pageIn(String sessionKey) throws SessionException {
        PagedSessionHandle sessHandle =
            (PagedSessionHandle)pagedCache.get(sessionKey);
        PagedSession session = null;
        try {
            if (sessHandle != null) {
                debug(3, "page: read session from disk: " + sessHandle.getSessionKey());
                session = sessHandle.read();
                debug(3, "page: read data: " + session);
                // delete paged version from disk...
                sessHandle.delete();
            }
        } catch (SessionException e) {
            Enhydra.getLogChannel().write(Logger.ALERT,
                                          "Session not found on disk.", e);
        }
        pagedCache.remove(sessionKey);
        return session;
    }

    /**
     * Returns the number of paged sessions.
     * @exception SessionException if the paged session count
     *   cannot be retrieved.
     */
    protected synchronized int getPagedSessionCount() throws SessionException {
        return pagedCache.size();
    }

    /**
     * Returns true if the specified session key is in use
     * by a session that has been paged out.
     *
     * @param sessionKey the session key to test.
     * @return true if the session key is in use by a paged session.
     */
    protected boolean pagedSessionKeyExists(String sessionKey)
        throws SessionException {
        return pagedCache.containsKey(sessionKey);
    }

    /**
     * Returns an enumeration of the keys of all the sessions that have
     * been paged out to persistent storage.
     *
     * @return the session key enumeration.
     * @exception SessionException if an error occurs.
     */
    protected Enumeration getPagedSessionKeys() throws SessionException {
        return pagedCache.keys();
    }

    /**
     * Removes a session that is new and paged.
     *
     * @exception SessionException if an error occurs.
     */
    protected boolean cleanupNewPagedSession() throws SessionException {
        long oldestTime = -1;
        String key = null;
        Enumeration e = pagedCache.keys();
        while (e.hasMoreElements()) {
            PagedSessionHandle s = (PagedSessionHandle)pagedCache.get(e.nextElement());
            if (s.isNew()) {
                if ((oldestTime < 0)
                    || (s.getTimeCreated() < oldestTime)) {
                    oldestTime = s.getTimeCreated();
                    key = s.getSessionKey();
                }
            }
        }
        if (key != null) {
            removeSession(key);
            return true;
        }
        return false;
    }

    /**
     * Shuts dows the session home.  Removes paged sessions
     * from disk.
     */
    public void shutdown() {
        Enumeration enum = pagedCache.elements();
        for (int i=0; i<pagedCache.size(); i++) {
            try {
                ((PagedSessionHandle)enum.nextElement()).delete();
            } catch (Exception e) {
                // ignore
            }
        }
    }

}




