/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PagedSessionHome.java,v 1.5.14.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session;

import com.lutris.appserver.server.Enhydra;
import com.lutris.util.*;
import com.lutris.logging.*;
import java.util.*;
import java.io.IOException;

/**
 * The StandardSessionManager uses PagedSessionHome to manage
 * a collection of sessions that can be paged to disk.<p>
 *
 * PagedSessionHome will page sessions to disk
 * as soon as a pre-configured threshold of sessions has been
 * reached.  Only sessions in the 'passive' state will be
 * paged to disk.  If all sessions are in the 'active' state
 * and the threshold has been reached, then a request to create
 * a new session will block until one of the 'active' sessions
 * goes into the 'passive' state.  At this point the session
 * that just became 'passive' is paged to disk and a new
 * session is created.<p>
 *
 * Sessions are paged to disk by serializing
 * the all data (excluding the session manager) that is
 * associated with a session.  This requires that
 * the session data and user associated with a session
 * are serializable.<p>
 *
 * The following parameters can be used to configure
 * the PagedSessionHome.  They should be grouped together in a section,
 * normally <code>SessionManager.SessionHome</code>, which is specified to
 * the constructor.<p>
 *
 * <ul>
 *
 * <li><code>PageThreshold: {int}</code><p>
 * Specifies the maximum number of session that may reside in memory
 * simultaneously.  If set to zero the the paged session home acts
 * as a write through cache.  If set to -1, then the page threshold
 * is assumed to be infinite and paging will never take place.
 * If not set, defaults to -1.<p>
 *
 * <li><code>PageTimeThreshold: {long}</code><p>
 * Specifies the amount of time a session may remain in memory
 * before it is paged out to disk.
 * If this parameter is not set, or, it is set to -1, then the session
 * will not be paged out unless the <i>PageThreshold</i> is reached.<p>
 *
 * <li><code>PageWait: {int}</code><p>
 * Specifies the maximum time (in milliseconds) that a thread will
 * wait for a session to be created or retrieved.  If the
 * page threshold has been reached, then a thread has to wait for
 * an existing session to be paged before a new session can be
 * created or a pre-existing session (that has been paged to
 * disk) can be activated.  If not set, defaults to 1 minute.
 * If this time is exceeded then an exception is thrown - this
 * prevents a possible deadlock situation.<p>
 *
 * <li><code>PageDir: {String}</code><p>
 * The directory where sessions will be paged.  This setting is
 * required.<p>
 *
 * <li><code>MaxSessions: {int}</code><p>
 * Specifies the maximum number of in use concurrent sessions.  If this value
 * is exceeded then CreateSession() will throw a CreateSessionException.
 * -1 indicates unlimited session.  If MaxSessions is not set
 * in the application's configuration then MaxSessions defaults to
 * unlimited sessions.<p>
 *
 * </ul>
 *
 * @see         StandardSession
 * @see         StandardSessionManager
 * @version	$Revision: 1.5.14.1 $
 * @author	Kyle Clark
 */
public abstract class PagedSessionHome implements StandardSessionHome {

    /**
     * The active cache contains sessions that are
     * actively in use (associated with) by a request.
     */
    private Hashtable activeCache = new Hashtable();

    /**
     * The active thread cache holds all the association
     * between active sessions and the threads that have
     * checked them out.
     */
    private Hashtable activeThreadCache = new Hashtable();

    /**
     * The passive cache contains sessions that
     * aren't currently associated with a request.
     */
    private Hashtable passiveCache = new Hashtable();

    /**
     * Config options for paged session home.
     */
    private Config config;

    /**
     * Maximum number of allowed sessions.
     */
    private long maxSessions = -1;

    /**
     * The threshold where paging beings.
     */
    private long pageThreshold = -1;

    /**
     * The amount of time a session may remain in memory
     * unused before it is paged out.  In seconds.
     */
    private long pageTimeThreshold = -1;

    /**
     * The pageTimeThreshold timer.
     */
    private PagedSessionThresholdTimer thresholdTimer;

    /**
     * If a thread has to wait for a session to be paged
     * before it can retrieve (get) or create a session then
     * pageWait specifies that maximum amount of time
     * that the thread will wait before an exception is
     * thrown.
     */
    private long pageWait = 60000; // milliseconds
        
    /**
     * Configuration keys.
     */
    private final String MAX_SESSIONS_KEY = "MaxSessions";
    private final String PAGE_THRESHOLD_KEY = "PageThreshold";
    private final String PAGE_TIME_THRESHOLD_KEY = "PageTimeThreshold";
    private final String PAGE_WAIT_KEY = "PageWait";
    private final long UNDEFINED_MAX_SESSIONS = -1;

    /**
     * The class loader to use when reading in paged data.
     */
    protected ClassLoader loader;

    /**
     * The session manager associated with the session home.
     */
    protected StandardSessionManager sessionMgr;
    
    /**
     * @param sessionMgr
     *   The session manager associated with this session home.
     * @param config
     *   Object parsed from configuration file.  This should be
     *   for the section containing the standard session home configuration.
     * @param loader
     *   The class load to use when load objects from persistent store.
     * @exception ConfigException
     *   signifies a problem in the configuration file.
     * @exception SessionException
     *   if the initialization fails.
     */
    public PagedSessionHome(StandardSessionManager sessionMgr,
                            Config config, ClassLoader loader)
        throws SessionException, ConfigException {
        this.config = config;
        this.loader = loader;
        this.sessionMgr = sessionMgr;
        if (config.containsKey(MAX_SESSIONS_KEY)) {
            maxSessions = config.getLong(MAX_SESSIONS_KEY);
        }
        if (maxSessions <= 0) {
            maxSessions = UNDEFINED_MAX_SESSIONS;
        }
        debug(MAX_SESSIONS_KEY + " = " + maxSessions);
        pageThreshold = config.getLong(PAGE_THRESHOLD_KEY);
        debug(PAGE_THRESHOLD_KEY + " = " + pageThreshold);
        if (config.containsKey(PAGE_WAIT_KEY)) {
            pageWait = config.getLong(PAGE_WAIT_KEY);
            if (pageWait <= 0) {
                throw new ConfigException("PagedSessionHome: "
                                          + "PageWait cannot be <= 0.");
            }
        }
        if (config.containsKey(PAGE_TIME_THRESHOLD_KEY)) {
            pageTimeThreshold = config.getLong(PAGE_TIME_THRESHOLD_KEY) * 1000;
            if (!isWriteThroughCache() && (pageTimeThreshold > 0)) {
                thresholdTimer = new PagedSessionThresholdTimer(this, sessionMgr.app);
                thresholdTimer.start();
            }
        }
        debug(PAGE_TIME_THRESHOLD_KEY + " = " + pageTimeThreshold);
    }

    /**
     * Creates a new session instance in the cache.  The
     * session is associated with the session key and the
     * current thread.
     *
     * @param sessionKey the key to associate with the session.
     * @return the newly created session.
     * @exception DuplicateKeyException
     *   if the session cannot
     *   be created because the key is already in use.
     * @exception CreateSessionException
     *   if the session could not be created.
     * @exception SessionException
     *   any other internal error.
     */
    public synchronized StandardSession createSession(String sessionKey)
        throws CreateSessionException, DuplicateKeyException, SessionException {
        if (containsKey(sessionKey)) {
            throw new DuplicateKeyException("Session key "
                                            + sessionKey + " is already in use.");
        }
        if ((maxSessions != UNDEFINED_MAX_SESSIONS) &&
            (maxSessions <= size())) {
            cleanupNewSession();
            if (maxSessions <= size()) {
                throw new CreateSessionException("Maximum session limit of "
                                                 + maxSessions
                                                 + " sessions has been reached.");
            }
        }
        ensureNewSession();
        PagedSession session = newSession(sessionMgr, sessionKey);
        SessionThread activeKey = new SessionThread(Thread.currentThread(),
                                                    sessionKey);
        session.incrementRefCount(); // keeps track of # of threads that reference session
        activeThreadCache.put(activeKey, session);
        activeCache.put(sessionKey, session);
        debug(3, "createSession: " + activeKey);
        return (StandardSession)session;
    }

    /**
     * Creates a new session object.  This method is intended to be
     * overriden by classes that extend PagedSessionHome.
     *
     * @return a new session.
     * @exception SessionException if an error occurs.
     */
    protected abstract PagedSession newSession(StandardSessionManager mgr,
                                               String sessionKey)
        throws SessionException;


    /**
     * Deletes a paged session.  If the session doesn't exist then this
     * is a noop.
     *
     * @param sessionKey the key identifying the session
     *   that should be deleted.  
     */
    protected abstract void deleteSession(String sessionKey)
        throws SessionException;

    /**
     * Returns the session associated with the session key.
     * The session must already be associated with the current thread
     * otherwize null is returned.
     *
     * @param sessionKey the session key
     * @return the session.
     * @see #getSession(Thread, String)
     * @exception SessionException
     *   if the session could not be retrieved.
     */
    public synchronized StandardSession getSession(String sessionKey)
        throws SessionException {
        debug(3, "get session: key = " + sessionKey);
        SessionThread activeKey = new SessionThread(Thread.currentThread(),
                                                    sessionKey);
        PagedSession s = (PagedSession)activeThreadCache.get(activeKey);
        return s;
    }

    /**
     * Returns the session associated with the session key.
     * The session is associated with the specified thread.
     * The session is put into the 'active' state.
     *
     * @param thread the thread to associate with the session.
     * @param sessionKey the session key for the session
     *   that will be made 'active' and returned.  If the
     *   session doesn't exist or it hasn't been
     *   associated with this thread then null is returned.
     * @return the session.
     * @exception SessionException
     *   if the session could not be retrieved.
     */
    public synchronized StandardSession getSession(Thread thread, String sessionKey)
        throws SessionException {
        debug(3, "getSession: " + sessionKey);
        SessionThread activeKey = new SessionThread(thread, sessionKey);
        PagedSession s = (PagedSession)activeThreadCache.get(activeKey);
        if (s == null) {
            s = (PagedSession)activeCache.get(sessionKey);
            if (s == null) {
                s = activateSession(sessionKey);
            }
            if (s != null) {
                activeThreadCache.put(activeKey, s);
                // keeps track of # of threads that reference session
                s.incrementRefCount(); 
            }
        }
        return s;
    }

    /**
     * Removes the session associated with the session key.
     *
     * @param sessionKey the session key for the session
     *   that will be made 'active' and returned.
     * @return the session.
     * @exception SessionException
     *   if the session couldn't be removed.
     */
    public synchronized void removeSession(String sessionKey) throws SessionException {
        debug(3, "removeSession: " + sessionKey);
        Enumeration e = activeThreadCache.keys();
        while (e.hasMoreElements()) {
            SessionThread key = (SessionThread)e.nextElement();
            if (key.sessionKey.equals(sessionKey)) {
                activeThreadCache.remove(key);
            }
        }
        activeCache.remove(sessionKey);
        passiveCache.remove(sessionKey);
        deleteSession(sessionKey);
        notify();
    }

    /**
     * Puts an 'active' session into the 'passive' state.  
     *
     * @param thread the thread associated with the session.
     * @param sessionKey the session key for the session
     *   that will be made persistent.
     * @exception SessionException
     *   if the session coulnd not be put into the passive state.
     */
    public synchronized void passivateSession(Thread thread, String sessionKey)
        throws SessionException {
        SessionThread activeKey = new SessionThread(thread, sessionKey);
        try {
            debug(3, "passivateSession: " + activeKey);
            if (activeThreadCache.containsKey(activeKey)) {
                PagedSession s = (PagedSession)activeThreadCache.remove(activeKey);
                s.decrementRefCount();
                // If there aren't any other threads associated with the session
                if (s.getRefCount() == 0) {
                    if (isWriteThroughCache()) {
                        pageOut(s);
                    } else if ((pageThreshold < 0)
                        || ((activeCache.size() + passiveCache.size()) < pageThreshold)) {
                        debug(3, "active -> passive: " + sessionKey);
                        passiveCache.put(sessionKey, s);
                    } else {
                        pageOut(s);
                    }
                    activeCache.remove(sessionKey);
                }
            } else {
                debug(3, "passivateSession called but active session "
                      + " doesn't exist: " + sessionKey);
            }
        } catch (Exception e) {
            throw new SessionException(e);
        } finally {
            notify();
        }
    }

    /**
     * Pages a session to disk.
     *
     * @param session the session to page.
     * @exception SessionException if the paged session could not be
     *   paged out.
     */
    protected abstract void pageOut(PagedSession s) throws SessionException;

    /**
     * Reads a paged session from disk.
     *
     * @param sessionKey the key identifying the session that should
     *   be paged in.
     * @return the paged session that was read in.  May be null if
     *   the session referenced by sessionKey doesn't exist.
     * @exception SessionException if the paged session could not be
     *   read in.
     */
    protected abstract PagedSession pageIn(String sessionKey) throws SessionException;

    /**
     * Returns the number of paged sessions.
     */
    protected abstract int getPagedSessionCount() throws SessionException;

    /**
     * Puts a session into the 'active' state.  Reads paged version
     * if necessary. 
     *
     * @param activeKey the session-thread key for the session
     *   that should be activated.
     * @exception SessionException
     *   if the session cannot be activated.
     */
    private synchronized PagedSession activateSession(String sessionKey)
        throws SessionException {
        PagedSession session = null;
        boolean isPaged = !activeCache.containsKey(sessionKey)
            && !passiveCache.containsKey(sessionKey);
        debug(3, "activateSession: " + sessionKey);
        if (isPaged) {
            ensureNewSession();
            session = pageIn(sessionKey);
        } else if (passiveCache.containsKey(sessionKey)) {
            debug(3, "passive -> active : " + sessionKey);
            session = (PagedSession)passiveCache.remove(sessionKey);
        } else if (activeCache.containsKey(sessionKey)) {
            debug(3, "session is active : " + sessionKey);
            session = (PagedSession)activeCache.get(sessionKey);
        }
        if (session != null) {
            activeCache.put(sessionKey, session);
        }
        return session;
    }

    /**
     * Specifies if a key is currently in use.
     *
     * @param sessionKey the session key to be tested.
     * @return true if the session key is in use.
     * @exception SessionException if an error occurs.
     */
    public boolean containsKey(String sessionKey) throws SessionException {
        if (activeCache.containsKey(sessionKey)
            || passiveCache.containsKey(sessionKey)
            || pagedSessionKeyExists(sessionKey)) {
            return true;
        }
        return false;
    }

    /**
     * Returns true if the specified session key is in use
     * by a session that has been paged out.
     *
     * @param sessionKey the session key to test.
     * @return true if the session key is in use by a paged session.
     * @exception SessionException if an error occurs.
     */
    protected abstract boolean pagedSessionKeyExists(String sessionKey)
        throws SessionException;

    /**
     * Returns the current number of sessions.
     *
     * @return the session count.
     * @exception SessionException if an error occurs.
     */
    public int size() throws SessionException {
        return activeCache.size() + passiveCache.size()
            + getPagedSessionCount();
    }

    /**
     * Returns the current number of sessions that are paged to
     * persistent store.
     *
     * @return
     *   the 'paged' session count.
     * @exception SessionException
     *   if the size cannot be determined
     */
    public int pagedSize() throws SessionException {
        return getPagedSessionCount();
    }

    /**
     * Returns an enumeration of keys for the cached sessions.
     *
     * @return the session keys enumeration.
     * @exception SessionException if an error occurs.
     */
    public synchronized Enumeration keys() throws SessionException {
        try {
            Vector v = new Vector();
            String sessionKey;
            Enumeration e;
            e = activeCache.keys();
            while (e.hasMoreElements()) {
                sessionKey = (String)e.nextElement();
                if (!v.contains(sessionKey)) {
                    v.addElement(sessionKey);
                }
            }
            e = passiveCache.keys();
            while (e.hasMoreElements()) {
                sessionKey = (String)e.nextElement();
                if (!v.contains(sessionKey)) {
                    v.addElement(sessionKey);
                }
            }
            e = getPagedSessionKeys();
            while (e.hasMoreElements()) {
                sessionKey = (String)e.nextElement();
                if (!v.contains(sessionKey)) {
                    v.addElement(sessionKey);
                }
            }
            return v.elements();
        } catch (Exception e) {
            throw new SessionException(e);
        }
    }

    /**
     * Returns an enumeration of the keys of all the sessions that have
     * been paged out to persistent storage.
     *
     * @return the session key enumeration.
     * @exception SessionException if an error occurs.
     */
    protected abstract Enumeration getPagedSessionKeys() throws SessionException;

    /**
     * Cleans up (removes) the oldest unused (new) session from the
     * passive cache.
     *
     * @return true if an unused session was cleaned up.
     */
    private void cleanupNewSession() throws SessionException {
        if (!cleanupNewPagedSession()) {
            Enumeration e;
            String key = null;
            long oldestTime = -1;
            e = passiveCache.keys();
            while (e.hasMoreElements()) {
                PagedSession s = (PagedSession)passiveCache.get(e.nextElement());
                if (s.isNew()) {
                    if ((oldestTime < 0)
                        || (s.getTimeCreated() < oldestTime)) {
                        oldestTime = s.getTimeCreated();
                        key = s.getSessionKey();
                    }
                }
            }
            if (key != null) {
                removeSession(key);
            }
        }
    }

    /**
     * Pages out any sessions that are passive and have exceeded
     * the page time threshold.
     *
     * @return returns the amount of time in milliseconds when
     *   the next check should occur.
     */
    long checkPassiveSessions() throws SessionException {
        return checkPassiveSessions(pageTimeThreshold);
    }

    /**
     * Pages out any sessions that are passive and have exceeded
     * the page time threshold.
     *
     * @param threshold the threshold when a passive session
     *   should be paged.
     * @return returns the amount of time in milliseconds when
     *   the next check should occur.
     */
    private synchronized long checkPassiveSessions(long threshold)
        throws SessionException {
        long nextCheck = threshold;
        Enumeration e = passiveCache.elements();
        long now = System.currentTimeMillis();
        while (e.hasMoreElements()) {
            PagedSession s = (PagedSession) e.nextElement();
            long passiveTime = now - s.getTimeLastUsed();
            if (passiveTime >= threshold) {
                pageOut(s);
                passiveCache.remove(s.getSessionKey());
            } else if (nextCheck > (threshold - passiveTime)) {
                nextCheck = threshold - passiveTime;
            }
        }
        return nextCheck;
    }

    /**
     * Removes a session that is new and paged.
     *
     * @return true is a new session was found and deleted.
     * @exception SessionException if an error occurs.
     */
    protected abstract boolean cleanupNewPagedSession() throws SessionException;


    /**
     * Ensures that a new session can be put into the list
     * in-core sessions that are being managed.
     * If necessary, this method will page a passive session disk in
     * in order to guarantee that another session can be
     * read into memory (or created).  This method may block
     * if a session first needs to be paged to disk.
     */
    private void ensureNewSession() throws SessionException {
        if (pageThreshold > 0) { // If not a write-through cache....
            while ((activeCache.size() + passiveCache.size()) > pageThreshold) {
                if (passiveCache.size() == 0) {
                    try {
                        debug(3, "waiting for a passive session that can be paged...");
                        wait(pageWait);
                        if (passiveCache.size() <= 0) {
                            throw new SessionException("Unable to activate session: "
                                                       + "no sessions are ready to be paged.");
                        }
                    } catch (InterruptedException e) {
                        // ignore
                    }
                } else {
                    debug(3, "a passive session ready to be paged");
                    Enumeration keys = passiveCache.keys();
                    PagedSession s = (PagedSession)passiveCache.get((String)keys.nextElement());
                    pageOut(s);
                    passiveCache.remove(s.getSessionKey());
                }
            }
        }
    }

    /**
     * Shuts down the session home.  Pages out all active sessions.
     */
    public synchronized void shutdown() {
        try {
            if (thresholdTimer != null) {
                thresholdTimer.shutdown();
                thresholdTimer = null;
            }
            checkPassiveSessions(0);
        } catch (Exception e) {
            Enhydra.getLogChannel().write(Logger.ERROR,
                                          "PagedSessionHome: "
                                          + "failed to page out active and passive sessions.",
                                          e);
        }
    }


    /**
     * Returns true if the current configuration is for
     * a write trhough cache.
     *
     * @return true if write through cache is configured.
     */
    private boolean isWriteThroughCache() {
        return (pageThreshold == 0);
    }        

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param msg the message to print.
     */
    protected void debug(String msg) {
        debug(0, msg);
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param level the debug level.
     * @param msg the message to print.
     */
    protected void debug(int level, String msg) {
        int dbg = Logger.DEBUG;
        switch (level) {
        case 1:
            dbg = Logger.DEBUG1;
            break;
        case 2:
            dbg = Logger.DEBUG2;
            break;
        case 3:
            dbg = Logger.DEBUG3;
            break;
        case 4:
            dbg = Logger.DEBUG4;
            break;
        case 5:
            dbg = Logger.DEBUG5;
            break;
        case 6:
            dbg = Logger.DEBUG6;
            break;
        case 7:
            dbg = Logger.DEBUG7;
            break;
        case 8:
            dbg = Logger.DEBUG8;
            break;
        case 9:
            dbg = Logger.DEBUG9;
            break;
        default:
            dbg = Logger.DEBUG;
            break;
        }
        
        Enhydra.getLogChannel().write(dbg, "PagedSessionHome("
                                      + Thread.currentThread().getName()
                                      + "): " + msg);
    }

}


