/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: SessionManager.java,v 1.19.14.1 2000/10/19 17:59:06 jasona Exp $
 */




package com.lutris.appserver.server.session;

import java.util.Date;
import java.util.Enumeration;
import com.lutris.appserver.server.user.User;

/**
 * The session management object interface. This class implementing this
 * interface manage the session state for currently active users
 * of one or more applications.  It allocates sessions and maintains the
 * mapping between session keys and Session objects.
 *
 * @version	$Revision: 1.19.14.1 $
 * @author	Shawn McMurdo
 * @author	Mark Diekhans
 */
public interface SessionManager {
    /**
     * Create a new <CODE>Session</CODE> object and an associated unique
     * random key.  No <CODE>User</CODE> object is initially associated with
     * the session (<CODE>getUser()</CODE> returns <CODE>null</CODE>).
     * 
     * @return	session The new <code>Session</code> object.
     * @exception SessionException if the session cannot be created.
     * @see	Session
     */
    public Session createSession() throws SessionException;

    /**
     * Create a new <CODE>Session</CODE> object and an associated unique
     * random key.  No <CODE>User</CODE> object is initially associated with
     * the session (<CODE>getUser()</CODE> returns <CODE>null</CODE>).
     * 
     * @param ipPortToken
     *        The base64 encoded IP and Port number to include in session key
     * @return	session The new <code>Session</code> object.
     * @exception SessionException if the session cannot be created.
     * @see	Session
     */
    public Session createSession(String ipPortToken) throws SessionException;

    /**
     * Removes a session from the session manager.
     * 
     * @param session The session object to delete.
     * @exception SessionException
     *   if the session cannot be deleted.
     */
    public void deleteSession(Session session) throws SessionException;
    
    /**
     * Removes a session from the manager.
     * 
     * @param sessionKey
     *   the session key.
     * @exception SessionException
     *   if the session cannot be deleted.
     */
    public void deleteSession(String sessionKey) throws SessionException;
    
    /**
     * Returns whether the <code>Session</code> object associated with
     * the specified session key exists.
     * 
     * @param	sessionKey
     *   The String used to reference a	<code>Session</code> object.
     * @return
     *   If the key is associated with an active session, then
     * 	 return true, otherwise return false.
     * @exception SessionException
     *   if the existence of the  session cannot be determined.
     */
    public boolean sessionExists(String sessionKey) throws SessionException;

    /**
     * Returns the <code>Session</code> object associated with the
     * specified session key.  The session is put in the 'active'
     * state.  If no <code>Session</code> object is
     * associated with the key then this method returns <code>null</code>.
     * 
     * @param	sessionKey The String used to reference a
     *		<code>Session</code> object.
     * @return	If the key is associated with an active session, then
     * 		the corresponding <code>Session</code> object is
     * 		returned.  Otherwise <code>null</code> is returned.
     * @see	Session
     * @exception SessionException
     *   if the session cannot be retrieved.
     */
    public Session getSession(String sessionKey) throws SessionException;

    /**
     * Returns the <code>Session</code> object associated with the
     * specified session key.  The session is put in the 'active'
     * state.  If no <code>Session</code> object is
     * associated with the key then this method returns <code>null</code>.
     * 
     * @param   thread the thread that should be associated with
     *          the session.  Only this thread can subsequently
     *          put the session into the passive state.
     * @param	sessionKey The String used to reference a
     *		<code>Session</code> object.
     * @return	If the key is associated with an active session, then
     * 		the corresponding <code>Session</code> object is
     * 		returned.  Otherwise <code>null</code> is returned.
     * @see	Session
     * @exception SessionException
     *   if the session cannot be retrieved.
     */
    public Session getSession(Thread thread, String sessionKey)
        throws SessionException;

    /**
     * Lookup active sessions keys for a user.  A given user may have
     * multiple sessions associated with it.
     *
     * @param user The user to search for.
     * @return An enumeration of the active sessions keys for the user.
     * @exception SessionException
     *   if the sessions cannot be retrieved.
     */
    public Enumeration getSessionKeys(User user) throws SessionException;

    /**
     * Puts a session into the 'passive' state.  A 'passive'
     * session may be made persistent.
     *
     * @param thread the thread currently associate with the
     *   the session.
     * @param sessionKey the session key for the session
     *   that will be made persistent.
     * @exception SessionException
     *   if the session cannot be put into the passive state..
     */
    public void passivateSession(Thread thread, String sessionKey)
        throws SessionException;

    /**
     * Gets the number of currently active sessions.
     * 
     * @return	The number of currently active sessions.
     * @exception SessionException
     *   if the active session count cannot be determined.
     */
    public int activeSessionCount() throws SessionException;
   
    /**
     * Gets the maximum number of concurent sessions that existed
     * at any time since this object was created, or
     * <CODE>resetMaxSessionCount()</CODE> was called.
     * This is a historical highwater mark. 
     * If you do not implement this feature, return -1.
     *
     * @return The highwater mark for number of sessions, or -1.
     */
    public int maxSessionCount();

    /**
     * Gets the time when the maximum refered to by 
     * <CODE>maxSessionCount()</CODE> occured.
     *
     * @return The Date of when the maximum number of sessions occured.
     */
    public Date maxSessionCountDate();

    /**
     * Reset the maximum session count. See <CODE>maxSessionCount()</CODE>.
     * The highwater mark should be reset to the current number of sessions.
     *
     * @exception SessionException
     *   if the max session count cannot be reset.
     */
    public void resetMaxSessionCount() throws SessionException; 

    /**
     * Get an enumeration of all session keys. 
     *
     * @return An enumeration of the session keys.
     * @exception SessionException
     *   if the session keys cannot be retrieved.
     */
    public Enumeration getSessionKeys() throws SessionException;

    /**
     * Shutdown this session manager as required.  The session manager
     * should not be used after this method has been called
     */
    public void shutdown();

}
