/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: SessionUtil.java,v 1.2.14.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session;
import com.lutris.appserver.server.user.User;
import com.lutris.util.*;
import java.util.*;

/**
 * Static methods that can be used by applications to manage sessions.
 *
 * @version	$Revision: 1.2.14.1 $
 * @author	Kyle Clark
 */
public class SessionUtil {

    /**
     * Prevent instantiation.
     */
    private SessionUtil() {
    }

    /**
     * Logs a user in.
     *
     * @param user the user to log in.
     * @param session the session to associated with the user.
     * @param multiple if true the user can have more than one session
     *   associated with him/her.  If false, all previous sessions
     *   associated with the user will be removed.
     * @exception SessionException if an error occurs.
     */
    public static void logIn(User user, Session session, boolean multiple)
        throws SessionException {
        SessionManager mgr = session.getSessionManager();
        synchronized(mgr) {
            if (!multiple) {
                Enumeration e = mgr.getSessionKeys(user);
                while (e.hasMoreElements()) {
                    String key = (String)e.nextElement();
                    mgr.deleteSession(key);
                }
            }
            session.setUser(user);
        }
    }

    /**
     * Dissasociates the user from the session and all
     * references to the session are deleted from the
     * session manager.
     *
     * @param session the session associated with the user.
     * @exception SessionException if an error occurs.
     */
    public static void logOut(Session session)
        throws SessionException {
	if (session != null) {
            SessionManager mgr = session.getSessionManager();
            mgr.deleteSession(session);
        }
    }

    /**
     * If a user attempts to login but has not logged
     * out from a previous session then this method can be used
     * to copy data from the previous session into this session
     * and to log out the previous session.
     *
     * @param user the user
     * @param session the session
     * @return true if the user was already logged in and the session
     *   has been resumed.
     * @exception SessionException
     *   if an error occurs.
     */
    public static boolean resumeLogIn(User user, Session session)
        throws SessionException {
        SessionManager mgr = session.getSessionManager();
        try {
            synchronized (mgr) {
                Enumeration e = mgr.getSessionKeys(user);
                if (e.hasMoreElements()) {
                    String key = (String)e.nextElement();
                    if (e.hasMoreElements()) {
                        // More than one login - we don't deal with it.
                        return false;
                    }
                    Session oldSession = mgr.getSession(Thread.currentThread(), key);
                    SessionData dataSrc = oldSession.getSessionData();
                    SessionData dataDst = session.getSessionData();
                    // Copy src to dst
                    String[] k = dataSrc.keys();
                    for (int i=0; i<k.length; i++) {
                        dataDst.set(k[i], dataSrc.get(k[i]));
                    }
                    SessionUtil.logIn(user, session, false);
                    return true;
                }
            }
        } catch (SessionException e) {
            throw e;
        } catch (Exception e) {
            throw new SessionException(e);
        }
        return false;
    }
}
