/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardSessionKeyGen.java,v 1.14.6.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.session;

import java.security.*;
import java.io.*;
import java.util.*;
import com.lutris.util.*;

/**
 * The session random key generator.  This class implements a background thread
 * that wakes up and counts the number of Standard Session Manager requests
 * completed at one or more different interval periods, and
 * supplements the seed of the Manager's random number generator
 * in order to make the value of the cookies extremely unpredictable.
 * This is an absolute requirement if random cookie values are to
 * be used for any type of security purpose.
 *
 * This random number generator uses the JDK 1.1 <code>SecureRandom</code>
 * object, which implements a cryptographic grade random number
 * generator based on the RSA MD5 one-way hash.  In combination with
 * external user-generated time delay information, the numbers
 * generated by this object are highly unpredictable, and therefore
 * suitably secure for their use as session keys. 
 *
 * @version	$Revision: 1.14.6.1 $
 * @author	John Marco
 * @author	Shawn McMurdo
 */
public class StandardSessionKeyGen extends Thread {

    /**
     * Random number generator used to generate session keys.
     * 
     * @see com.lutris.StandardSessionManager#randomThread
     * @see java.security.SecureRandom
     */
    private SecureRandom randomizer = new SecureRandom();

    // If you want to improve startup performance significantly
    // use the following constructor.
    // new SecureRandom(new Long(System.currentTimeMillis()).toString().getBytes());

    /**
     * Counter that is incremented by the user of this class to provide a
     * random external value.  Usually done on requests.
     */
    private int randomCounter = 0;

    /*
     * Timer intervals.
     */
    private long[] alarmVector;
    private long[] intervalVector;
    private int numIntervals;
    
    /**
     * Constructor a new key generator random number entropy
     * generator.  Initializes timers and counters and start a
     * thread.
     * 
     * @param manager The Standard session manager to be randomized.
     * @param intervals An array of one or more intervals, in seconds
     * 		in which to periodically supplement the random number
     *          generator with external user-generated entropy.
     */ 
    public StandardSessionKeyGen(long intervals[]) {
	int i;
	long now = System.currentTimeMillis();
	numIntervals = intervals.length;
	intervalVector = new long[numIntervals];

	//Convert each interval value from seconds to milliseconds.
	for (i=0; i<numIntervals; i++) {
	    intervalVector[i] = intervals[i] * 1000;  // Milliseconds.
	}
	alarmVector = new long[numIntervals];
	for (i=0; i<numIntervals; i++) {
	    alarmVector[i] = now + intervalVector[i];
	}
	setDaemon(true); // Don't require this thread to exit.
    }
    
    /**
     * The main code body of the Idle Timer Thread.  Enters an endless
     * loop that sleeps for a configurable period, periodically waking
     * up to modify the session manager's random seed. An externally
     * incremented count is used as a source of user-generated randomness.
     */
    public void run() {
	ByteArrayOutputStream bos = new ByteArrayOutputStream();
	DataOutputStream dos = new DataOutputStream(bos);
	MessageDigest md5 = null;
	int i;
	long now, least, interval;

	try {
	    md5 = MessageDigest.getInstance("MD5");
	} catch (NoSuchAlgorithmException e) { 
            // JDK1.1 always has MD5.
            throw new FatalExceptionError(e);
        }
	while (true) {
	    now = System.currentTimeMillis();
	    least = Long.MAX_VALUE;

	    // Check each alarm for expiry, and processes if needed.
	    for (i=0; i<numIntervals; i++) {
		if (now >= alarmVector[i]) {
		    try {
			dos.writeLong(randomCounter);
			dos.writeLong(now);
			dos.flush();
		    } catch (IOException e) {
                        throw new FatalExceptionError(e); //- won't happen.
                    }
		    randomizer.setSeed(md5.digest(bos.toByteArray()));

		    // Set the alarm to its next wakeup time.
		    alarmVector[i] = now + intervalVector[i];
		}

		// Find the next scheduled alarm.
		if (alarmVector[i] < least) {
		    least = alarmVector[i];
                }
	    }

	    // Wait for the next scheduled alarm.
	    interval = least - now;
	    if (interval <= 0) {
		interval = 1;
            }
	    try {
		sleep(interval);
	    } catch (InterruptedException e) {
                // Don't throw interrupts.
            }
	}
    }

    /**
     * Increment the random counter.  Used for randomization, so doesn't
     * have to be completely accurate and is not sychronized.  The normal
     * way to use this is to increment it on requests generated by external
     * sources.
     */
    public void incrementRandomCounter() {
        randomCounter++;
    }

    /**
     * Generates a new random key to identify a session.
     * This key represents a random integer that is large and sparse
     * enough to make it highly unlikely that a valid session key can
     * be guessed by an intruder.  The <code>randomizer</code> object is
     * used to generate this key.<P>
     *
     * This function is reentrant and does not need synchronization.
     * 
     * @return	A string representing a random key.  The characters
     * 		in this key are constrained to [A-Za-z0-9_-]. The
     *          encoding is more or less Base 64, but instead of '+'
     *          and '/' as defined in RFC1521, the characters '_' and
     *          '-' are used because they are safe in URLs and file names.
     */
    public String newSessionKey() {
	try {
	    byte[] rand32 = new byte[32];
	    byte[] rand2 = new byte[2];
	    byte[] md5result;
	    long now = System.currentTimeMillis();
	    ByteArrayOutputStream bos = new ByteArrayOutputStream();
	    DataOutputStream dos = new DataOutputStream(bos);
	    MessageDigest md5 = MessageDigest.getInstance("MD5");
	    randomizer.nextBytes(rand32);
	    randomizer.nextBytes(rand2);
	    dos.write(rand32, 0, rand32.length);
	    dos.writeLong(now);
	    dos.flush();
	    md5result = md5.digest(bos.toByteArray());
	    bos.reset();
	    dos = new DataOutputStream(bos);
	    dos.write(md5result, 0, md5result.length);
	    dos.write(rand2, 0, rand2.length);
	    return Base64Encoder.toBase64SessionKeyString(bos.toByteArray());
	} catch (IOException e) {
            // Byte stream; should never happen.
            throw new FatalExceptionError(e);
        } catch (NoSuchAlgorithmException e) {
            // JDK 1.1 has MD5m, should never happen.
            throw new FatalExceptionError(e);
        }
    }

    /**
     * Shutdown the thread associated with this object.
     */
    public void shutdown() {
        stop();
    }
}

