/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardSessionManager.java,v 1.42.2.4.4.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session;

import java.security.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.Constructor;
import com.lutris.util.*;
import com.lutris.logging.*;
import com.lutris.appserver.server.*;
import com.lutris.appserver.server.user.User;
import com.lutris.appserver.server.session.persistent.*;

/**
 * This session manager maintains the mapping between session keys
 * and sessions.  It generates secure session keys that are safe to use
 * as cookie values in web applications.  It also manages the life
 * of a session.  If a session is not used (touched) for a preconfigured
 * amount of time, then it is expired and removed from the session
 * manager.  The following parameters can be used to configure
 * the session manager.  They should be grouped together in a section,
 * normally <code>SessionManager</code>, which is specified to
 * the constructor.<p>
 * <ul>
 * <li><code>SessionLifetimeMax: {int}</code><p>
 *
 *     The maximum number of minutes a session is valid.  If this
 *     value us zero, then there is no lifetime limit.
 *     This parameter must be set in the configuration file.<p>
 *
 * <li><code>SessionIdleTimeMax: {int}</code><p>
 *
 *     Maximum number of minutes a session may be idle before
 *     the session is expired (deleted).  If this value
 *     is less than or equal to zero, then the session may
 *     be idle indefinitly.<p>
 *
 * <li><code>SessionNoUserIdleTimeMax: {int}</code><p>
 *
 *     Maximum number of minutes a session that does not have a
 *     <CODE>User</CODE> object assocaited with it may be idle before
 *     the session is expired (deleted).  If this value
 *     is less than or equal to zero, then the session may
 *     be idle indefinitly.  Default is the same as <code>MaxIdleTime</code><p>
 *
 * <li><code>IdleScanInterval: {int}</code><p>
 *
 *     How often, in seconds, the session manager tests if
 *     any sessions should be expired.
 *     This parameter must be greater than zero.<p>
 *
 * <li><code>RandomizerIntervals: {int [,int]...}</code><p>
 *
 *     The set of varying intervals to use for adding user-generated entropy
 *     to the random number generator.  It is a good idea to use several
 *     different prime numbers.  These intervals are in seconds.<p>
 *
 * <li><code>SessionHome.Mode: {BASIC | PAGE_TO_DISK | PAGE_TO_DB | CUSTOM}</code><p>
 *
 *     Indicates the mode (StandardSessionHome interface) that the standard
 *     session manager will use.
 *     In 'BASIC' mode the standard session manager uses the <code>BasicSessionHome</code>
 *     to manage sessions.  In 'PAGE_TO_DISK' the standard session manager uses
 *     the <code>DiskPagedSessionHome</code> to manage sessions.  In 'PAGE_TO_DB'
 *     the standard session manager used the <code>PersistentSessionHome</code>
 *     to manage sessions.  In 'CUSTOM' the application specifies the classes
 *     that should be loaded and provide the <code>StandardSessionHome</code>
 *     implementation.  If set to 'CUSTOM' then the settings
 *     <code>SessionHome.Class</code> and <code>SessionUserTable.Class</code>
 *     should be set.
 *
 * <li><code>SessionHome.Class: {String}</code><p>
 *
 *     The class that implements the StandardSessionHome interface.  This
 *     class is loaded by StandardSessionManager and used to managed
 *     the collection of sessions.  If this option is not specified then
 *     BasicSessionHome is used.  The SessionHome that is
 *     loaded by the session manager is expected to contain a consturctor
 *     that takes the following arguments:
 *     <code>SessionHome(StandardSessionManager mgr,
 *                       com.lutris.util.Config config,
 *                       ClassLoader loader)</code><p>
 *
 * <li><code>SessionUserTable.Class: {String}</code><p>
 *
 *     The class that implements the StandardSessiondUserTable interface.  This
 *     class is loaded by StandardSessionManager and used to managed
 *     the session to user relationships.  If this option is not specified then
 *     BasicSessionUserTable is used.  The SessionUserTable that is
 *     loaded by the session manager is expected to contain a consturctor
 *     that takes the following arguments:
 *     <code>SessionUserTable(com.lutris.util.Config config)</code><p>
 *
 * </ul>
 *
 * @version	$Revision: 1.42.2.4.4.1 $
 * @author	John Marco
 * @author	Shawn McMurdo
 * @author      Kyle Clark
 * @author      Mark Diekhans
 * @see         com.lutris.util.Config
 * @see         StandardSessionHome
 * @see         StandardSessionUserTable
 * @see         BasicSessionHome
 * @see         DiskPagedSessionHome
 * @see         com.lutris.appserver.server.session.persistent.PersistentSessionHome
 */
public class StandardSessionManager implements SessionManager,
                                               StandardSessionIdleHandler {

    /**
     * Represents the current session management mode.
     */
    protected int mode;

    /**
     * Indicates that the session manager is using
     * the basic session home interface to manage sessions.
     * @see #getMode
     * @see com.lutris.appserver.server.session.BasicSessionHome
     */
    public static final int MODE_BASIC = 1;

    /**
     * Indicates that the session manager is using
     * the page to disk home interface to manage sessions.
     * @see #getMode
     * @see com.lutris.appserver.server.session.DiskPagedSessionHome
     */
    public static final int MODE_PAGE_TO_DISK = 2;

    /**
     * Indicates that the session manager is using
     * the page to database home interface to manage sessions.
     * @see #getMode
     * @see com.lutris.appserver.server.session.persistent.PersistentSessionHome
     */
    public static final int MODE_PAGE_TO_DB = 3;

    /**
     * Indicates that the session manager is using
     * the custom home interface to manage sessions.
     * @see #getMode
     */
    public static final int MODE_CUSTOM = 4;

    /**
     * Indicates that url encoding of session ids is never preformed.
     * @see #getEncodeUrlState
     **/
    public static final String ENCODE_URL_NEVER = "Never";

    /**
     * Indicates that url encoding of session ids is always preformed.
     * @see #getEncodeUrlState
     **/
    public static final String ENCODE_URL_ALWAYS = "Always";

    /**
     * Indicates that url encoding of session ids is preformed only when
     * cookies are disabled on the client browser.  This is automatically
     * detected.
     * @see #getEncodeUrlState
     **/
    public static final String ENCODE_URL_AUTO = "Auto";

    /**
     * Default maximum session life time, in seconds.  Zero if there is no
     * limit.  A derived <CODE>SessionManager</CODE> may override this value
     * to use the default expiration logic or override
     * <CODE>isSessionExpired</CODE> to define custom expiration logic.
     *
     * @see lutris.session.StandardSessionManager#getMaxSessionLifeTime
     */
    protected static long defaultMaxSessionLifeTime = 0;

    /**
     * Default maximum session idle time, in seconds.  A derived
     * <CODE>SessionManager</CODE> may override this value to use the default
     * expiration logic or override <CODE>isSessionExpired</CODE>
     * to define custom expiration logic.
     * A value less-than or equal to zero disables idle checking.
     * Default value is 30 minutes.
     *
     * @see #getMaxSessionIdleTime
     */
    protected static long defaultMaxSessionIdleTime = 30*60;

    /**
     * Indicates url encoding status.  Assumes that
     * comms.response.writeHTML(HTMLDocument doc) is used to commit the
     * response to the client.  The default is to Auto.<br>
     * <code>Never</code> indicates urls are never encoded with session keys.
     * This indicates that session cookies have to be used or no session state
     * can be maintained.<br>
     * <code>Always</code> indicates that urls are always encoded with session
     * keys.  Session cookies are never used.<br>
     * <code>Auto</code> indicates that session cookies will be if available.
     * If not, urls will automatically be encoded.<br>
     * @see #getEncodeUrlState
     */
    protected static String defaultEncodeUrlState = "Auto";

    /**
     * Default interval, in seconds, to scan for sessions to expire.
     * A derived <CODE>SessionManager</CODE> may override this value.
     * Default value is 30 seconds.
     */
    protected static long defaultIdleScanInterval = 30;

    /**
     * Default list of randomize key generator time intervals, in seconds.
     * A list of prime numbers is recommended.
     * A derived <CODE>SessionManager</CODE> may override this value.
     */
    protected static long[] defaultRandomizerIntervals = {
        301, 1001, 5003
    };

    /**
     * The name of the config variable for the max session lifetime.
     */
    public static final String CFG_LIFE = "SessionLifetimeMax";

    /**
     * The name of the config variable for the max session idle time.
     */
    public static final String CFG_IDLE = "SessionIdleTimeMax";

    /**
     * The name of the config variable for the url encoding state
     */
    public static final String CFG_ENCODE_URL_STATE = "SessionEncodeUrlState";

    /**
     * The name of the config variable for the max idle time for sessions
     * with no user.
     */
    public static final String CFG_NOUSER_IDLE = "SessionNoUserIdleTimeMax";

    /**
     * The name of the config variable for the interval between scans for
     * idle sessions.
     */
    public static final String CFG_SCAN = "IdleScanInterval";

    /**
     * The name of the config variable for the interval between introduction
     * of randomness to the session key generator.
     */
    public static final String CFG_RANDOM = "RandomizerIntervals";

    /**
     * The name of the config variable for the session home settings.
     */
    public static final String CFG_SESSION_HOME = "SessionHome";

    /**
     * The name of the config variable for the session home type.
     */
    public static final String CFG_SESSION_HOME_TYPE = "SessionHome.Mode";

    /**
     * Indicates that a sesson is still active.
     *
     * @see #sessionDeleted
     */
    public static final int SESSION_ACTIVE = 0;

    /**
     * Indicates that a sesson was deleted due to max time
     * being exceeded.
     *
     * @see #sessionDeleted
     */
    public static final int SESSION_MAX_TIME = 1;

    /**
     * Indicates that a sesson was deleted due to idle time
     * being exceeded.
     *
     * @see #sessionDeleted
     */
    public static final int SESSION_IDLE_EXPIRE = 2;

    /**
     * Indicates that a sesson was explictly deleted.
     *
     * @see #sessionDeleted
     */
    public static final int SESSION_EXPLICT_DELETE = 3;

    /**
     * Table of currently active user sessions is managed
     * by the session home.  Different implementations of
     * StandardSessionHome can be loaded by this manager.  A specific
     * implementation can manage failover persistence, paging, etc.
     */
    private StandardSessionHome sessionHome;

    /**
     * Table of users objects.  This table tracks the number of simultaneous
     * sessions a user currently has active. Table entries are of type
     * <CODE>Vector</CODE>, constaining Session objects, keyed by User
     * object reference.
     */
    private StandardSessionUserTable sessionUserTable;

    // The maximum number of sessions (highwater mark).
    private int maxSessions;
    private Date maxSessionsDate;

    /**
     * Object that generates random session keys.  Manages a thread to
     * for increase randomization.
     *
     * @see StandardSessionKeyGen
     */
    private StandardSessionKeyGen keyGenerator;

    /**
     * Maximum session life time, in seconds.  Zero if there is no
     * limit.
     *
     * @see lutris.session.StandardSessionManager#getMaxSessionLifeTime
     */
    private long maxSessionLifeTime;

    /**
     * Maximum session idle time, in seconds.
     * A value less-than or equal to zero disables idle checking.
     *
     * @see #getMaxSessionIdleTime
     */
    protected long maxSessionIdleTime;

    /**
     * The url encoding state.  Either <code>Never</code>
     * <code>Always</code> <code>Auto</code>.
     */
    protected String encodeUrlState;

    /**
     * Background thread that keeps track of session idle time.
     * Sessions that have been idle for more than a configurable time
     * limit are terminated by this thread.
     *
     * @see StandardSessionIdleTimer
     */
    private StandardSessionIdleTimer idleTimer;

    /**
     * Maximum session idle time, in seconds for sessions
     * that don't have a <CODE>User</CODE> object associated with them.
     * A value less-than or equal to zero disables idle checking.
     *
     * @see #getMaxNoUserSessionIdleTime
     */
    protected long maxNoUserSessionIdleTime;

    /**
     * The time in seconds between scans for idle sessions to expire.
     */
    protected long scanInterval;

    /**
     * Log channel for debugging messages.
     */
    LogChannel logChannel = null;

    /**
     * The application using this session manager.
     */
    Application app = null;

    /**
     * The class loader used by this app/context.
     */

     ClassLoader classLoader;

    /**
     * The supported session home types.
     */
    private static final String BASIC = "BASIC";
    private static final String PAGE_TO_DISK = "PAGE_TO_DISK";
    private static final String PAGE_TO_DB= "PAGE_TO_DB";
    private static final String CUSTOM = "CUSTOM";

    /**
     * The session home interface type being used.
     */
    String sessionHomeType = BASIC;

    /**
     * Log configuration information.
     */
    private void logConfig() {
        if (logChannel != null) {
            logChannel.write(Logger.DEBUG,
		    "SessionManager." + CFG_LIFE + " = "
		    + maxSessionLifeTime + " sec");
            logChannel.write(Logger.DEBUG,
		    "SessionManager." + CFG_IDLE + " = "
		    + maxSessionIdleTime + " sec");
            logChannel.write(Logger.DEBUG,
		    "SessionManager." + CFG_NOUSER_IDLE + " = "
		    + maxNoUserSessionIdleTime + " sec");
            logChannel.write(Logger.DEBUG,
		    "SessionManager." + CFG_SCAN + " = "
		    + scanInterval + " sec");
            logChannel.write(Logger.DEBUG,
		    "SessionManager." + CFG_ENCODE_URL_STATE + " = "
		    + encodeUrlState);
        }
    }

    /**
     * Creates a new <code>SessionManager</code> object.
     * This constructor will first looks for the session manager
     * configuration parameters that have the specified configuration
     * prefix prepended to the standard session manager configuration
     * option.<p>
     *
     * @param app the application associate with this session
     *   manager.
     * @param config Object parsed from configuration file.  This should be
     *   for the section constaining the session manager configuration.
     * @param sessionMgrLogChannel If not <CODE>null</CODE>, channel to
     *   log debugging information to.
     * @exception ConfigException signifies a problem in the
     *	 configuration file.
     * @exception SessionException
     *   if all classes (Home and UserTable) couldn't be loaded
     *   by the session manager.
     */
    public StandardSessionManager(Application application,
                                  Config config,
                                  LogChannel sessionMgrLogChannel)
            throws ConfigException, SessionException {


       this(application.getClass().getClassLoader(),config,sessionMgrLogChannel);
       app = application;
    }

    public StandardSessionManager(ClassLoader classLoader,
                                  Config config,
                                  LogChannel sessionMgrLogChannel)
            throws ConfigException, SessionException {

        this.classLoader = classLoader;
        logChannel = sessionMgrLogChannel;

        // High-water mark
        maxSessions = 0;
        maxSessionsDate = new Date();

        /*
         * Config file specifies minutes, but the values are kept internally
         * as seconds.
         */

        // Session lifetime
        if (config.containsKey(CFG_LIFE)) {
            maxSessionLifeTime = config.getLong(CFG_LIFE) * 60;
        } else if (config.containsKey("Lifetime")) {
	    // backwards compatability
            maxSessionLifeTime = config.getLong("Lifetime") * 60;
        } else {
            maxSessionLifeTime = defaultMaxSessionLifeTime;
        }

        // Idle time.
        if (config.containsKey(CFG_IDLE)) {
            maxSessionIdleTime = config.getLong(CFG_IDLE) * 60;
        } else if (config.containsKey("MaxIdleTime")) {
	    // backwards compatability
            maxSessionIdleTime = config.getLong("MaxIdleTime") * 60;
        } else {
            maxSessionIdleTime = defaultMaxSessionIdleTime;
        }

        // Url encoding state
        if (config.containsKey(CFG_ENCODE_URL_STATE)) {
            encodeUrlState = config.getString(CFG_ENCODE_URL_STATE);
        } else if (config.containsKey("EncodeUrlState")) {
	    // backwards compatability
            encodeUrlState = config.getString("EncodeUrlState");
        } else {
            encodeUrlState = defaultEncodeUrlState;
        }
	if (!encodeUrlState.equalsIgnoreCase(ENCODE_URL_NEVER) &&
	    !encodeUrlState.equalsIgnoreCase(ENCODE_URL_ALWAYS) &&
	    !encodeUrlState.equalsIgnoreCase(ENCODE_URL_AUTO)) {
	    throw new ConfigException("EncodeUrlState must be one of the following: " + ENCODE_URL_NEVER + ", " + ENCODE_URL_ALWAYS + ", or " + ENCODE_URL_AUTO + ".");
	}

        if (config.containsKey(CFG_NOUSER_IDLE)) {
            maxNoUserSessionIdleTime = config.getLong(CFG_NOUSER_IDLE) * 60;
        } else if (config.containsKey("MaxNoUserIdleTime")) {
	    // backwards compatability
            maxNoUserSessionIdleTime = config.getLong("MaxNoUserIdleTime") * 60;
        } else {
            maxNoUserSessionIdleTime = maxSessionIdleTime;
        }

        scanInterval = defaultIdleScanInterval;
        if (config.containsKey(CFG_SCAN)){
            scanInterval = config.getLong(CFG_SCAN);
        } else if (config.containsKey("IdleScanInterval")){
	    // backwards compatability
            scanInterval = config.getLong("IdleScanInterval");
        }
        if (scanInterval <= 0) {
           throw new ConfigException("IdleScanInterval must be greater than zero.");
        }
	idleTimer = new StandardSessionIdleTimer(this, app, scanInterval);

        if (config.containsKey(CFG_SESSION_HOME_TYPE)) {
            sessionHomeType = config.getString(CFG_SESSION_HOME_TYPE);
        }
        if (!sessionHomeType.equalsIgnoreCase(BASIC)
            && !sessionHomeType.equalsIgnoreCase(PAGE_TO_DISK)
            && !sessionHomeType.equalsIgnoreCase(PAGE_TO_DB)
            && !sessionHomeType.equalsIgnoreCase(CUSTOM)) {
            throw new ConfigException("Invalid " + CFG_SESSION_HOME_TYPE
                                      + ": '"  + sessionHomeType + "'");
        }

        // Session Home
        sessionHome = loadSessionHome(config);

        // Session User Table
        sessionUserTable = loadSessionUserTable(config);

        // Random key generator.
        long[] intervals = config.getLongs(CFG_RANDOM,
                                           defaultRandomizerIntervals);
        if (intervals.length == 0) {
           throw new ConfigException (CFG_RANDOM + " must contain some values.");
        }
        for (int i = 0; i < intervals.length; i++) {
            if (intervals[i] <= 0) {
                throw new ConfigException (CFG_RANDOM + " must contain positive integers.");
            }
        }
	keyGenerator = new StandardSessionKeyGen(intervals);

        // Start threads.
	keyGenerator.start();
	idleTimer.start();

        logConfig();
    }

    /**
     * Loads the StandardSessionHome used by this session manager.
     *
     * @param config
     *   This session manager's config section.
     * @exception
     *   ConfigException if the configuration is invalid.
     * @exception
     *   SessionException if the StandardSessionHome could not be loaded.
     */
    protected StandardSessionHome loadSessionHome(Config config)
        throws ConfigException, SessionException {
        try {
            StandardSessionHome sessionHome = null;
            Config homeConfig = null;
            if (config.containsKey(CFG_SESSION_HOME)) {
                homeConfig = (Config)config.getSection(CFG_SESSION_HOME);
            } else {
                homeConfig = new Config();
            }
            if (sessionHomeType.equalsIgnoreCase(CUSTOM)) {
                if (homeConfig.containsKey("Class")) {
                    String homeClassName = homeConfig.getString("Class");
                    try {
                        Class [] paramTypes = new Class[3];
                        Object [] args = new Object[3];
                        paramTypes[0] = Class.forName("com.lutris.appserver.server.session.StandardSessionManager");
                        paramTypes[1] = Class.forName("com.lutris.util.Config");
                        paramTypes[2] = Class.forName("java.lang.ClassLoader");
                        args[0] = this;
                        args[1] = homeConfig;
                        args[2] = classLoader;
                        Constructor c = Class.forName(homeClassName).getConstructor(paramTypes);
                        sessionHome = (StandardSessionHome)c.newInstance(args);
                        if (logChannel != null) {
                            logChannel.write(Logger.DEBUG, "SessionMgr: "
                                             + "StandardSessionHome: " + homeClassName);
                        }
                    } catch (Exception e) {
                        throw new SessionException("Unable to load " + homeClassName, e);
                    }
                }
                mode = MODE_CUSTOM;
            } else if (sessionHomeType.equalsIgnoreCase(PAGE_TO_DISK)) {
                sessionHome = new DiskPagedSessionHome(this, homeConfig,classLoader);
                mode = MODE_PAGE_TO_DISK;
            } else if (sessionHomeType.equalsIgnoreCase(PAGE_TO_DB)) {
                sessionHome = new PersistentSessionHome(this, homeConfig,
                                                        classLoader);
                mode = MODE_PAGE_TO_DB;
            } else {
                sessionHome = new BasicSessionHome(this, homeConfig);
                mode = MODE_BASIC;
            }
            if (logChannel != null) {
                logChannel.write(Logger.DEBUG, "SessionMgr: "
                                 + "StandardSessionHome: " + sessionHomeType + "\n");
            }
            return sessionHome;
        } catch (KeywordValueException e) {
            e.printStackTrace();
            throw new ConfigException("SessionMgr: unable to load StandardSessionHome: "
                                      + e);
        }
    }

    /**
     * Loads the StandardSessiondUserTable used by this session manager.
     *
     * @param config
     *   This session manager's config section.
     * @exception
     *   ConfigException if the StandardSessionHome could not be loaded.
     */
    protected StandardSessionUserTable loadSessionUserTable(Config config)
        throws ConfigException, SessionException {
        try {
            StandardSessionUserTable sessionUserTable = null;
            Config userTableConfig = null;
            if (config.containsKey("SessionUserTable")) {
                userTableConfig = (Config)config.getSection("SessionUserTable");
            } else {
                userTableConfig = new Config();
            }
            if (sessionHomeType.equalsIgnoreCase(CUSTOM)) {
                String tableClassName = userTableConfig.getString("Class");
                try {
                    Class [] paramTypes = new Class[1];
                    Object [] args = new Object[1];
                    paramTypes[0] = Class.forName("com.lutris.util.Config");
                    args[0] = userTableConfig;
                    Constructor c = Class.forName(tableClassName).getConstructor(paramTypes);
                    sessionUserTable = (StandardSessionUserTable)c.newInstance(args);
                    if (logChannel != null) {
                        logChannel.write(Logger.DEBUG, "SessionMgr: "
                                         + "StandardSessionUserTable: " + tableClassName);
                    }
                } catch (NoSuchMethodException e) {
                    throw new SessionException("Unable to load " + tableClassName + ": "
                                               + "Constructor not found.", e);
                } catch (Exception e) {
                    throw new SessionException("Unable to create instance of " + tableClassName, e);
                }
            } else if (sessionHomeType.equalsIgnoreCase(PAGE_TO_DISK)) {
                sessionUserTable = new PagedSessionUserTable(userTableConfig);
            } else if (sessionHomeType.equalsIgnoreCase(PAGE_TO_DB)) {
                sessionUserTable = new PersistentSessionUserTable(userTableConfig);
            } else {
                sessionUserTable = new BasicSessionUserTable(userTableConfig);
            }
            return sessionUserTable;
        } catch (KeywordValueException e) {
            throw new ConfigException("SessionMgr: unable to load StandardSessionUserTable: "
                                      + e);
        }
    }

    /**
     * Shutdown this session manager, stopping threads that are
     * associated with it.
     */
    public synchronized void shutdown() {
        if (keyGenerator != null) {
            keyGenerator.shutdown();
            keyGenerator = null;
        }
        if (idleTimer != null) {
            idleTimer.shutdown();
            idleTimer = null;
        }
        if (sessionHome != null) {
            sessionHome.shutdown();
            sessionHome = null;
        }
        if (sessionUserTable != null) {
            sessionUserTable.shutdown();
            sessionUserTable = null;
        }
    }

    /**
     * Allocate a new <CODE>StandardSession</CODE> object.  This maybe
     * overridden by derived session managers to allocated a
     * <CODE>Session</CODE> object derived from <CODE>StandardSession</CODE>.
     * This is called by <CODE>createSession</CODE>.
     *
     * @param   sessionKey The session key to associate with the
     *          session.
     * @return	session The new <code>StandardSession</code> object.
     * @see	StandardSession
     * @deprecated The instance of StandardSessionHome should be replaced
     *   instead of extending this method.
     */
    protected StandardSession newSession(String sessionKey)
        throws CreateSessionException, SessionException {
        return (StandardSession) sessionHome.createSession(sessionKey);
    }

    /**
     * Create a new <CODE>Session</CODE> object and associate a unique random
     * key.  No <CODE>User</CODE> is initially associated with
     * (<CODE>Session.getUser()</CODE> returns <CODE>null</CODE>/
     *
     * @return	session The new <code>Session</code> object.
     * @exception SessionException
     *   if the session cannot be created.
     * @see	Session
     */
    public Session createSession() throws SessionException {
	return createSession(null);
    }

    /**
     * Create a new <CODE>Session</CODE> object and associate a unique random
     * key.  No <CODE>User</CODE> is initially associated with
     * (<CODE>Session.getUser()</CODE> returns <CODE>null</CODE>/
     *
     * @param ipPortToken
     *        The base64 encoded IP and Port number to include in session key
     * @return	session The new <code>Session</code> object.
     * @exception SessionException
     *   if the session cannot be created.
     * @see	Session
     */
    public Session createSession(String ipPortToken)
	throws SessionException {
        StandardSession session = null;
	String sessionKey = null;

        /*
         * Create a new key that does not already exist in
         * the session table.  The odds of this loop making more than
         * one pass are astronomically small, but not zero.
         */
        do {
            try {
                sessionKey = keyGenerator.newSessionKey();
		if (ipPortToken != null) {
		    sessionKey += ("(" + ipPortToken + ")");
		}
                session = sessionHome.createSession(sessionKey);
            } catch (DuplicateKeyException e) {
                // TODO - log message?
                session = null;
            }
        } while (session == null);

        int currentSize = sessionHome.size();
        if (currentSize > maxSessions) {
            maxSessions = currentSize;
            maxSessionsDate = new Date();
        }
        keyGenerator.incrementRandomCounter();

        if (logChannel != null) {
            logChannel.write(Logger.DEBUG, "SessionMgr: createSession: " + sessionKey);
        }

        return session;
    }

    /**
     * Log a user being registered or unregistered.
     */
    private void logRegisterUser(Session session, String which) {
        if (logChannel != null) {
            String userName = null;
            User user = session.getUser();
            if (user != null) {
                userName = user.getName();
            }
            logChannel.write(Logger.DEBUG, "SessionMgr: "
                             + which + ": " + session.getSessionKey() +
                             " user = \"" + userName + "\"");
        }
    }

    /**
     * Adds a session's user to the session-to-user table.
     * This is called by the <CODE>Session</CODE> object to associate
     * a session with a user.
     *
     * @param	session The now logged-in <code>Session</code> instance to
     * 			store in the session table.
     * @see	Session
     * @see	User
     * @see	#unregisterUser
     * @exception SessionException
     *   if an error occurs.
     */
    protected synchronized void registerUser(Session session)
        throws SessionException {
        sessionUserTable.add(session.getSessionKey(), session.getUser());
        logRegisterUser(session, "registerUser");
    }

    /**
     * Removes a session's user from the session-to-user table.
     * This is called by the <CODE>Session</CODE> object to disassoicate
     * a user from a session.
     *
     * @param	session The now <code>Session</code> instance
     * 			the user is being logged out from.
     * @see	Session
     * @exception SessionException
     *   if an error occurs.
     */
    protected synchronized void unregisterUser(Session session)
        throws SessionException {
        logRegisterUser(session, "unregisterUser");
        sessionUserTable.remove(session.getSessionKey(), session.getUser());
    }

    /**
     * Determine if a session should be expired based on the idle
     * time.  This method is designed to be overriden by derived classes that
     * want to use different idle time checks.
     *
     * @param session Session to check.
     * @return A reason code for the experation.
     *  One of <CODE>SESSION_MAX_TIME</CODE>,
     *  <CODE>SESSION_IDLE_EXPIRE</CODE>, or
     *  <CODE>SESSION_ACTIVE</CODE>, to indicate that the session
     *  is not expired.
     *
     * @see #sessionDeleted
     * @see #maxSessionIdleTime
     * @see #maxNoUserSessionIdleTime
     */
    protected int isSessionExpired(StandardSession session) {
        long now = System.currentTimeMillis();
        /*
         * Has the session reached it's maximum age, regardless of user
         * interaction? This is an optional feature.
         */
        long maxAge = session.getTimeExpires();
        if ((maxAge > 0) && (now > maxAge)) {
            return SESSION_MAX_TIME;
        }

        /*
         * Has the session been idle too long? There are two thresholds,
         * for whether or not a user is logged in.
         * Both are optional features.
         */
        long idle = now - session.getTimeLastUsed();
        long maxIdle;
        if (session.getUser() != null) {
            maxIdle = session.getMaxIdleTime();
        } else {
            maxIdle = session.getMaxNoUserIdleTime();
        }
        if ((maxIdle > 0) && (idle > maxIdle)) {
            return SESSION_IDLE_EXPIRE;
        }

        /*
         * The session is ok.
         */
        return SESSION_ACTIVE;
    }

    /**
     * Method called when a session is deleted.
     * This method is designed to be overriden by derived classes that
     * want need to take specific action, such as saving the sessions
     * state.
     * The default implementation of this method does nothing.
     *
     * @param session Session that is being delete.
     * @param reason The reason the session was being deleted.
     *  One of <CODE>SESSION_MAX_TIME</CODE>,
     *  <CODE>SESSION_IDLE_EXPIRE</CODE>,
     *  <CODE>SESSION_EXPLICT_DELETE</CODE>.
     * @see #isSessionExpired
     */
    protected void sessionDeleted(Session session, int reason) {
        // Default implementation does nothing.
    }

    /**
     * Does the actual work of deleting an entry from the session table.
     *
     * @param session The session to delete.
     * @param reason The reason the session was being deleted.
     * @exception SessionException
     *   If the session cannot be deleted.
     */
    private void doDeleteSession(StandardSession session, int reason)
        throws SessionException {
        keyGenerator.incrementRandomCounter();

        if (logChannel != null) {
            String userName = null;
            User user = session.getUser();
            if (user != null) {
                userName = user.getName();
            }
            String reasonMsg;
            switch (reason) {
            case SESSION_ACTIVE:
                reasonMsg = "SESSION_ACTIVE";
                break;
            case SESSION_MAX_TIME:
                reasonMsg = "SESSION_MAX_TIME";
                break;
            case SESSION_IDLE_EXPIRE:
                reasonMsg = "SESSION_IDLE_EXPIRE";
                break;
            case SESSION_EXPLICT_DELETE:
                reasonMsg = "SESSION_EXPLICT_DELETE";
                break;
            default:
                reasonMsg = "bad reason code " + reason;
                break;
            }

            logChannel.write(Logger.DEBUG, "SessionMgr deleteSession: "
                             + session.getSessionKey() + " user = \"" + userName
                             + "\" reason = " + reasonMsg);
        }
        if (session.getUser() != null) {
            sessionUserTable.remove(session.getSessionKey(), session.getUser());
        }
	sessionHome.removeSession(session.getSessionKey());
    }

    /**
     * Removes a session from the Session Table.  This method is used
     * to log a user off the application. Note that
     * <CODE>sessionDeleted()</CODE> is called.
     *
     * @param session Session object to remove from table removed.
     * @exception SessionException
     *   If the session cannot be deleted.
     */
    public synchronized void deleteSession(Session session) throws SessionException {
        sessionDeleted((StandardSession)session, SESSION_EXPLICT_DELETE);
	doDeleteSession((StandardSession)session, SESSION_EXPLICT_DELETE);
    }

    /**
     * Removes a session from the Session Table.  This method is used
     * to log a user off the application. Note that
     * <CODE>sessionDeleted()</CODE> is called.
     *
     * @param sessionKey A key corresponding to a user session to be
     * 			removed.
     * @exception SessionException
     *   If the session cannot be deleted.
     */
    public synchronized void deleteSession(String sessionKey) throws SessionException {
	Session session = (Session)sessionHome.getSession(sessionKey);
	if (session != null) {
            deleteSession(session);
        }
    }

    /**
     * Puts a session into the 'passive' state.  A 'passive'
     * session may be made persistent.
     *
     * @param thread the thread currently associated with the session.
     * @param sessionKey the session key for the session
     *   that will be made persistent.
     * @exception SessionException
     *   If the session cannot be put into the passive state.
     */
    public void passivateSession(Thread thread, String sessionKey) throws SessionException {
        sessionHome.passivateSession(thread, sessionKey);
    }


    /**
     * Scans session table to determine if any sessions need to be
     * expired.
     *
     * @exception SessionException
     *   if an error occurs.
     */
    public void cleanUpIdleSessions() throws SessionException {
	Enumeration e = sessionHome.keys();
	StandardSession session;
        String sessionKey;
        debug(3, "checking for idle sessions... ");
	while (e.hasMoreElements()) {
            sessionKey = (String)e.nextElement();
	    session = (StandardSession)sessionHome.getSession(Thread.currentThread(), sessionKey);
            if (session != null) {
                int stat = isSessionExpired(session);
                if (stat != SESSION_ACTIVE) {
                    debug(3, "cleaning up idle session: " + sessionKey);
                    sessionDeleted(session, stat);
                    doDeleteSession(session, stat);
                } else {
                    sessionHome.passivateSession(Thread.currentThread(), sessionKey);
                }
            }
	}
    }

    /**
     * Returns whether the <code>Session</code> object associated with
     * the specified session key exists.
     *
     * @param	sessionKey The String used to reference a
     *		<code>Session</code> object.
     * @return	If the key is associated with an active session, then
     * 		return true, otherwise return false.
     * @exception SessionException
     *   If existence of the session cannot be tested.
     */
    public synchronized boolean sessionExists(String sessionKey)
        throws SessionException {
	if (sessionKey == null) {
	    return false;
        }
	return sessionHome.containsKey(sessionKey);
    }

    /**
     * Lookup the <code>Session</code> object associated with the
     * specified session key. <P>
     *
     * Each time a session is returned via this method, the session's
     * last used (referenced) time is updated to the current time.
     *
     * @param	sessionKey The String used to reference a
     *		<code>Session</code> object.
     * @return	If the key is associated with an active session, then
     * 		the corresponding <code>Session</code> object is
     * 		returned.  Otherwise <code>null</code> is returned.
     * @see	Session
     * @see	StandardSession#touch
     * @exception SessionException
     *   If the session cannot be retrieved.
     */
    public synchronized Session getSession(String sessionKey) throws SessionException {
        keyGenerator.incrementRandomCounter();
	StandardSession session = (StandardSession)sessionHome.getSession(sessionKey);
        if (session != null) {
            session.touch();
        } else {
            sessionUserTable.remove(sessionKey);
        }
	return session;
    }

    /**
     * Lookup the <code>Session</code> object associated with the
     * specified session key. <P>
     *
     * Each time a session is returned via this method, the session's
     * last used (referenced) time is updated to the current time.
     *
     * @param   thread the thread to associate with the session.
     * @param	sessionKey The String used to reference a
     *		<code>Session</code> object.
     * @return	If the key is associated with an active session, then
     * 		the corresponding <code>Session</code> object is
     * 		returned.  Otherwise <code>null</code> is returned.
     * @see	Session
     * @see	StandardSession#touch
     * @exception SessionException
     *   If the session cannot be retrieved.
     */
    public synchronized Session getSession(Thread thread, String sessionKey)
        throws SessionException {
        keyGenerator.incrementRandomCounter();
	StandardSession session =
            (StandardSession)sessionHome.getSession(thread, sessionKey);
        if (session != null) {
            session.touch();
        } else {
            sessionUserTable.remove(sessionKey);
        }
	return session;
    }

    /**
     * Get all of the active sessions.
     *
     * @return An enumeration of the active sessions.
     * @exception SessionException
     *   If the sessions cannot be retrieved.
     */
    public synchronized Enumeration getSessionKeys() throws SessionException {
        return sessionHome.keys();
    }

    /**
     * Lookup the keys for the active sessions for a user.
     * A given user may have multiple sessions associated
     * with it (using FTP as an example, the "anonymous" account).
     *
     * @param user The user to search for.
     * @return An enumeration containing the session keys of the
     *   active sessions for the user.
     * @exception SessionException
     *   If the sessions cannot be retrieved.
     */
    public synchronized Enumeration getSessionKeys(User user) throws SessionException {
        return sessionUserTable.getSessionKeys(user);
    }

    /**
     * Gets the number of currently active sessions.
     *
     * @return	The number of currently active sessions.
     * @exception SessionException
     *   If the session cannot be calculated.
     */
    public int activeSessionCount() throws SessionException {
	return sessionHome.size();
    }

    /**
     * Gets the number of session that are paged to persistent store.
     *
     * @return	The number of currently paged sessions.
     * @exception SessionException
     *   If the session cannot be calculated.
     */
    public int pagedSessionCount() throws SessionException {
	return sessionHome.pagedSize();
    }

    /**
     * Returns the current mode for the session manager.
     *
     * @return	the mode.
     * @exception SessionException
     *   If the mode cannot be determined.
     * @see #MODE_BASIC
     * @see #MODE_PAGE_TO_DISK
     * @see #MODE_PAGE_TO_DB
     * @see #MODE_CUSTOM
     */
    public int getMode() throws SessionException {
	return mode;
    }

   /**
     * Gets the maximum number of concurent sessions that existed
     * at any time since this object was created, or
     * <CODE>resetMaxSessionCount()</CODE> was called.
     * This is a historical highwater mark.
     *
     * @return The highwater mark for number of sessions, or -1.
     */
    public int maxSessionCount() {
        return maxSessions;
    }

    /**
     * Gets the time when the maximum refered to by
     * <CODE>maxSessionCount()</CODE> occured.
     *
     * @return The Date of when the maximum number of sessions occured.
     */
    public Date maxSessionCountDate() {
        return maxSessionsDate;
    }

    /**
     * Reset the maximum session count. See <CODE>maxSessionCount()</CODE>.
     * The highwater mark should be reset to the current number of sessions.
     *
     * @exception SessionException
     *   if the max session count cannot be reset.
     */
    public void resetMaxSessionCount() throws SessionException {
        maxSessions = sessionHome.size();
        maxSessionsDate = new Date();
    }

    /**
     * Returns the maximum session life time, in seconds.
     */
    public long getMaxSessionLifeTime() {
        return maxSessionLifeTime;
    }

    /**
     * Sets the maximum session idle time, in seconds.
     */
    public void setMaxSessionIdleTime(long maxSessionIdleTime) {
        this.maxSessionIdleTime = maxSessionIdleTime;
    }

    /**
     * Returns the maximum session idle time, in seconds.
     */
    public long getMaxSessionIdleTime() {
        return maxSessionIdleTime;
    }

    /**
     * Returns the url encoding state.  Either <code>Never</code>
     * <code>Always</code> <code>Auto</code>
     */
    public String getEncodeUrlState() {
        return encodeUrlState;
    }

    /**
     * Returns the maximum no-user session idle time, in seconds.
     */
    public long getMaxNoUserSessionIdleTime() {
        return maxNoUserSessionIdleTime;
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param msg the message to print.
     */
    private void debug(String msg) {
        debug(0, msg);
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param level the debug level.
     * @param msg the message to print.
     */
    private void debug(int level, String msg) {
        int dbg = Logger.DEBUG;
        switch (level) {
        case 1:
            dbg = Logger.DEBUG1;
            break;
        case 2:
            dbg = Logger.DEBUG2;
            break;
        case 3:
            dbg = Logger.DEBUG3;
            break;
        case 4:
            dbg = Logger.DEBUG4;
            break;
        case 5:
            dbg = Logger.DEBUG5;
            break;
        case 6:
            dbg = Logger.DEBUG6;
            break;
        case 7:
            dbg = Logger.DEBUG7;
            break;
        case 8:
            dbg = Logger.DEBUG8;
            break;
        case 9:
            dbg = Logger.DEBUG9;
            break;
        default:
            dbg = Logger.DEBUG;
            break;
        }

        Enhydra.getLogChannel().write(dbg, "StandardSessionManager("
                                      + Thread.currentThread().getName()
                                      + "): " + msg);
    }

}
