/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PersistentSessionHome.java,v 1.4.14.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session.persistent;

import com.lutris.appserver.server.Enhydra;
import com.lutris.appserver.server.session.*;
import com.lutris.appserver.server.sql.*;
import com.lutris.logging.*;
import com.lutris.util.*;
import java.util.*;
import java.sql.*;

/**
 * PersistentSessionHome writes all passive sessions
 * to a database.  The sessions are written by serializing
 * all the data (excluding the session manager) that is
 * associated with a session.  This requires that
 * the session data and user associated with a session
 * are serializable.  PersistentSessionHome
 * should be used by applications that want failover support
 * or that want to run in a clustered environment.<p>
 *
 * The session data is written to a table in the database
 * that is defined as (Informix):<p>
 *
 * <pre>
 *     CREATE TABLE PersistentSession
 *     (
 *          sessionKey VARCHAR(64) NOT NULL,
 *          isNew CHAR(1) DEFAULT "1" NOT NULL,
 *          timeCreated DECIMAL(19,0),
 *          timeLastUsed DECIMAL(19,0),
 *          timeExpires DECIMAL(19,0),
 *          maxIdleTime DECIMAL(19,0),
 *          maxNoUserIdleTime INTEGER,
 *          userName VARCHAR(255),
 *          data BYTE,
 *          PRIMARY KEY(sessionKey)
 *     )
 * </pre>
 *
 * The configuration settings for PersistentSessionHome:<p>
 *
 * <ul>
 *
 * <li><code>DatabaseName: {String}</code><p>
 * The database that will be accessed.  This session is optional.  If
 * not set then the default database is accessed.
 *
 * <li><code>DBTableName: {String}</code><p>
 * The name of the table in the database where session data
 * will be written.  The default table name is "PersistentSession".<p>
 *
 * </ul>
 *
 * @see         StandardSessionManager
 * @version	$Revision: 1.4.14.1 $
 * @author	Kyle Clark
 */
public class PersistentSessionHome extends PagedSessionHome {

    /**
     * The name of the database that will be accessed.
     */
    static String dbName = null;

    /**
     * The name of the database table where session information
     * will be stored.
     */
    static String dbTableName = "PersistentSession"; // Default

    /**
     * Configuration keys.
     */
    private static final String DB_NAME_KEY = "DatabaseName";
    private static final String DB_TABLE_NAME_KEY = "DBTableName";

    /**
     * @param sessionMgr
     *   The session manager associated with this session home.
     * @param config
     *   Object parsed from configuration file.  This should be
     *   for the section containing the standard session home configuration.
     * @param loader
     *   The class load to use when load objects from persistent store.
     * @exception ConfigException
     *   signifies a problem in the configuration file.
     * @exception SessionException
     *   if the initialization fails.
     */
    public PersistentSessionHome(StandardSessionManager sessionMgr,
                                Config config, ClassLoader loader)
        throws SessionException, ConfigException {
        super(sessionMgr, config, loader);
        if (config.containsKey(DB_TABLE_NAME_KEY)) {
            dbTableName = config.getString(DB_TABLE_NAME_KEY);
        }
        debug(DB_TABLE_NAME_KEY + " = " + dbTableName);
        dbName = getDatabaseName(config);
        if (dbName != null) {
            debug(DB_NAME_KEY + " = " + dbName);
        } else {
            debug(DB_NAME_KEY + " = DEFAULT");
        }
    }

    /**
     * Returns the name of the database being accessed by this manager.
     *
     * @return the database name.  May be null if the default database is
     *   being accessed.
     */
    public String getDatabaseName() {
        return dbName;
    }

    /**
     * Returns the name of the database that should be accessed. 
     *
     * @param config the config in whihc to look up the database
     *   setting.
     * @return the database name.  May be null if the default database is
     *   being accessed.
     * @exception ConfigException if an error occurs.
     */
    public static String getDatabaseName(Config config) throws ConfigException {
        String s = null;
        if (config.containsKey(DB_NAME_KEY)) {
            s = config.getString(DB_NAME_KEY);
            if (s.length() == 0) {
                s = null;
            }
        }
        return s;
    }

    /**
     * Creates a new session object.  This method is intended to be
     * overriden by classes that extend PagedSessionHome.
     *
     * @return a new session.
     */
    protected PagedSession newSession(StandardSessionManager mgr,
                                      String sessionKey)
        throws SessionException {
        PersistentSession session = new PersistentSession(mgr, sessionKey, this);
        DBUtil.dbInsert(session, dbName);
        return session;
    }

    /**
     * Deletes a paged session.  If the session doesn't exist then this
     * is a noop.
     *
     * @param sessionKey the key identifying the session
     *   that should be deleted.  
     */
    protected void deleteSession(String sessionKey)
        throws SessionException {
        // Delete from database
        DBUtil.dbDelete(sessionKey, dbName);
    }

    /**
     * Pages a session to disk.
     *
     * @param session the session to page.
     * @exception SessionException if the paged session could not be
     *   paged out.
     */
    protected synchronized void pageOut(PagedSession s) throws SessionException {
        // N.B. it is possible for the session to be removed
        // from the database if running in a clustered environment.
        // The idle timer in another application instance may run
        // while this session is in the active state.
        // If it was removed then the update would fail and we
        // need to re-insert.
        if (DBUtil.dbUpdate((PersistentSession)s, dbName) <= 0) {
            DBUtil.dbInsert((PersistentSession)s, dbName);
        }
    }

    /**
     * Reads a paged session from disk.
     *
     * @param sessionKey the key identifying the session that should
     *   be paged in.
     * @return the paged session that was read in.
     * @exception SessionException if the paged session could not be
     *   read in or does not exist.
     */
    protected synchronized PagedSession pageIn(String sessionKey) throws SessionException {
        SessionQuery sessionQuery = new SessionQuery(sessionKey, loader);
        PersistentSession s = (PersistentSession)DBUtil.dbQuery(sessionQuery, dbName);
        if (s != null) {
            s.restoreSessionHome(this);
            s.restoreSessionManager(sessionMgr);
        }
        return s;
    }

    /**
     * Returns the number of paged sessions.
     */
    protected synchronized int getPagedSessionCount() throws SessionException {
        SizeQuery sizeQuery = new SizeQuery();
        Integer size = (Integer)DBUtil.dbQuery(sizeQuery, dbName);
        return size.intValue();
    }

    /**
     * Returns true if the specified session key is in use
     * by a session that has been paged out.
     *
     * @param sessionKey the session key to test.
     * @return true if the session key is in use by a paged session.
     */
    protected boolean pagedSessionKeyExists(String sessionKey)
        throws SessionException {
        KeyExistsQuery keyExistsQuery = new KeyExistsQuery(sessionKey);
        Boolean exists = (Boolean)DBUtil.dbQuery(keyExistsQuery, dbName);
        return exists.booleanValue();
    }

    /**
     * Returns an enumeration of the keys of all the sessions that have
     * been paged out to persistent storage.
     *
     * @return the session key enumeration.
     * @exception SessionException if an error occurs.
     */
    protected Enumeration getPagedSessionKeys() throws SessionException {
        return (Enumeration)DBUtil.dbQuery(new KeysQuery(), dbName);
    }

    /**
     * Removes a session that is new and paged.
     *
     * @exception SessionException if an error occurs.
     */
    protected boolean cleanupNewPagedSession() throws SessionException {
        String key = null;
        UnusedQuery unusedQuery = new UnusedQuery();
        key = (String)DBUtil.dbQuery(unusedQuery, dbName);
        if (key != null) {
            removeSession(key);
            return true;
        }
        return false;
    }

    /**
     * Shuts dows the session home.
     */
    public void shutdown() {
        // TODO
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param msg the message to print.
     */
    protected void debug(String msg) {
        debug(0, msg);
    }

    /**
     * Prints debug information under Logger.DEBUG.
     *
     * @param level the debug level.
     * @param msg the message to print.
     */
    protected void debug(int level, String msg) {
        int dbg = Logger.DEBUG;
        switch (level) {
        case 1:
            dbg = Logger.DEBUG1;
            break;
        case 2:
            dbg = Logger.DEBUG2;
            break;
        case 3:
            dbg = Logger.DEBUG3;
            break;
        case 4:
            dbg = Logger.DEBUG4;
            break;
        case 5:
            dbg = Logger.DEBUG5;
            break;
        case 6:
            dbg = Logger.DEBUG6;
            break;
        case 7:
            dbg = Logger.DEBUG7;
            break;
        case 8:
            dbg = Logger.DEBUG8;
            break;
        case 9:
            dbg = Logger.DEBUG9;
            break;
        default:
            dbg = Logger.DEBUG;
            break;
        }
        
        Enhydra.getLogChannel().write(dbg, "PersistentSessionHome("
                                      + Thread.currentThread().getName()
                                      + "): " + msg);
    }

}




