/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpDebug.java,v 1.27.14.1 2000/10/19 17:59:07 jasona Exp $
 */





package com.lutris.http.debug;

import java.util.Enumeration;
import javax.servlet.http.Cookie;
import com.lutris.html.*;
import com.lutris.http.HttpStatus;
import com.lutris.util.HtmlEncoder;

/**
 * This class is used for debugging. It generates html formatted output.
 * It can be used to display http request, http response information.
 * The interfaces DebugRequest and DebugResponse are used to define
 * how this debugging class interacts with the various type of
 * request and response classes including: HttpServletRequest,
 * HttpServletResponse, HttpPresentationRequest, HttpPresentationResponse,
 * HttpRequest, HttpResponse, etc.
 *
 * @author	Paul Morgan
 * @since	LBS1.7
 * @version	$Revision: 1.27.14.1 $
 */
public class HttpDebug {
    
    //
    // Define table colors.
    //
    private static final String SPACE_BG_COLOR = "#000000";
    private static final String HEADER_BG_COLOR = "#888888";
    private static final String BODY_BG_COLOR = "#BBBBBB";
    private static final String HIGHLIGHT_BG_COLOR = "#999999";
    private static final String HEADER_FG_COLOR = "#FFFFFF";
    private static final String FRAME_COLOR = "#EEEEEE";


    /**
     * Returns HTML containing all the useful information
     * in a HTTP Request.
     *
     * @param request The request object to be dumped.
     * @return The composed HTML table.
     * @exception Exception from underlying implementation of request object.
     */
    public static String
	getRequestDebugInfo(DebugRequest request)
	throws Exception
    {
	HtmlTable tbl = new HtmlTable();
	tbl.setCellPadding(1);
	tbl.setCellSpacing(2);
	tbl.setBackgroundColor(BODY_BG_COLOR);
	tbl.setBorderWidth(0);
	tbl.setPercentWidth(100);

	addRequestInformation(tbl, request);
	addHeaderInformation(tbl, request);
	addCookieInformation(tbl, request);
	addParameterInformation(tbl, request);

	/*
	 * Add frame background...
	 */
	HtmlTable frameTbl = new HtmlTable();
	frameTbl.setBackgroundColor(FRAME_COLOR);
	frameTbl.setPercentWidth(100);
	HtmlTableRow tr = new HtmlTableRow();
	HtmlTableDataCell td = new HtmlTableDataCell(tbl.toString());
	tr.addCell(td);
	frameTbl.addRow(tr);

	return frameTbl.toString();
    }

    /**
     * Returns HTML containing all the useful information
     * in a HTTP Response. 
     *
     * @param response The response object to be dumped.
     * @return The composed HTML table.
     * @exception Exception from underlying implementation of response object.
     */
    public static String
	getResponseDebugInfo(DebugResponse response)
	throws Exception
    {
	HtmlTable tbl = new HtmlTable();
	tbl.setCellPadding(1);
	tbl.setCellSpacing(2);
	tbl.setBackgroundColor(BODY_BG_COLOR);
	tbl.setBorderWidth(0);
	tbl.setPercentWidth(100);

	addResponseInformation(tbl, response);
	addHeaderInformation(tbl, response);
	addCookieInformation(tbl, response);
        addResponseDataInformation(tbl, response);

	/*
	 * Add frame white background...
	 */
	HtmlTable frameTbl = new HtmlTable();
	frameTbl.setBackgroundColor(FRAME_COLOR);
	frameTbl.setPercentWidth(100);
	HtmlTableRow tr = new HtmlTableRow();
	HtmlTableDataCell td = new HtmlTableDataCell(tbl.toString());
	tr.addCell(td);
	frameTbl.addRow(tr);

	return frameTbl.toString();
    }

    /**
     * Given a HTML table object, this method extracts all useful information
     * from a httpRequest object and composes a new table row for each
     * datum.
     *
     * @param tbl The HTML table object to attach information.
     * @param request The request object to be dumped.
     * @exception Any exception from underlying inplementation of
     *		  request object.
     */
    private static void
	addRequestInformation(HtmlTable tbl, DebugRequest request)
	throws Exception
    {
        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
				     HEADER_FG_COLOR +
				     ">Request Information</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Information</B>");
	td.setPercentWidth(10);		// Keeps first column narrow
        tr.addCell(td);
        td = new HtmlTableDataCell("<B>Value</B>");
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getPath
	 */
        /* FIXME: Not impelmented; paul, what's this suppose to be.  Mark ???
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Path");
        tr.addCell(td);
        String value = request.getPath();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);
        */

	/*
	 * getServerName
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Server Name");
        tr.addCell(td);
        String value = request.getServerName();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getServerPort
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Server Port");
        tr.addCell(td);
        td = new HtmlTableDataCell(Integer.toString(request.getServerPort()));
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getScheme
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("URI");
        tr.addCell(td);
        value = request.getRequestURI();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getRequestURI
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Scheme");
        tr.addCell(td);
        value = request.getScheme();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getTotalBytes
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Total Bytes");
        tr.addCell(td);
	int totalBytes = request.getTotalBytes();
	if (totalBytes == -1) {
	    td = new HtmlTableDataCell("<I>N/A</I>");
	} else {
	    td = new HtmlTableDataCell(Integer.toString(totalBytes));
	}
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getQueryString
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Query String");
        tr.addCell(td);
        value = request.getQueryString();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getRemoteAddr
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Remote IP Address");
        tr.addCell(td);
        value = request.getRemoteAddr();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getRemoteHost
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Remote Host");
        tr.addCell(td);
        value = request.getRemoteHost();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getRemoteUser
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Remote User");
        tr.addCell(td);
        value = request.getRemoteUser();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getAuthType
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Auth Type");
        tr.addCell(td);
        value = request.getAuthType();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getContentLength
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Content Length");
        tr.addCell(td);
	int contentLength = request.getContentLength();
	if (contentLength == -1) {
	    td = new HtmlTableDataCell("<I>N/A</I>");
	} else {
	    td = new HtmlTableDataCell(Integer.toString(contentLength));
	}
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getContentType
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Content Type");
        tr.addCell(td);
        value = request.getContentType();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getMethod
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Method");
        tr.addCell(td);
        value = request.getMethod();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getPathInfo
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Path Info");
        tr.addCell(td);
        value = request.getPathInfo();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getPathTranslated
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Path Translated");
        tr.addCell(td);
        value = request.getPathTranslated();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getPathProtocol
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Protocol");
        tr.addCell(td);
        value = request.getProtocol();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);
    }


    /**
     * Adds request headers data rows to a table.
     *
     * @param tbl HTML table to which to add the data.
     * @param request Object from which http request header data is compiled.
     * @exception Any exception from underlying inplementation of
     *		  request object.
     */
    private static void
	addHeaderInformation(HtmlTable tbl, DebugRequest request)
	throws Exception
    {
        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
				     HEADER_FG_COLOR +
				     ">Request Headers</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Header Name</B>");
	td.setPercentWidth(10);		// Keeps first column narrow
        tr.addCell(td);
        td = new HtmlTableDataCell("<B>Value(s)</B>");
        tr.addCell(td);
        tbl.addRow(tr);

	Enumeration headerNames = request.getHeaderNames();
	String headerName, headerValue;
	boolean first = true;


	if (!headerNames.hasMoreElements()) {
	    tr = new HtmlTableRow();
	    td = new HtmlTableDataCell("<I>None found</I>");
	    td.setColSpan(2);
	    td.setHorizontalAlignment(td.ALIGN_CENTER);
	    tr.addCell(td);
	    tbl.addRow(tr);
	} else {
	    while (headerNames.hasMoreElements()) {
		headerName = (String)headerNames.nextElement();
		headerValue = request.getHeader(headerName);
		tr = new HtmlTableRow();
		tr.addCell(new HtmlTableDataCell(capitalize(headerName)));
		tr.addCell(new HtmlTableDataCell(headerValue));
		tbl.addRow(tr);
	    }
        }
    }


    /**
     * Adds request cookie data rows to a table.
     *
     * @param tbl HTML table to which to add the data.
     * @param request Object from which http request cookie data is compiled.
     * @exception Any exception from underlying inplementation of
     *		  request object.
     */
    private static void
	addCookieInformation(HtmlTable tbl, DebugRequest request)
	throws Exception
    {
	Cookie[] cookies = request.getCookies();
	formatCookiesAsHtml(tbl, cookies, "Request Cookies", false);
    }


    /**
     * Format the cookies as HTML.
     *
     * @param tbl The HTML table.
     * @param cookies Cookies to be formatted.
     * @param header for table section.
     * @param showFields Should the fields (path, expires etc...) be shown?
     * @return The formatted HTML.
     * @exception Any exception from underlying inplementation of
     *		  request object.
     */
    private static void
	formatCookiesAsHtml(HtmlTable tbl, Cookie[] cookies, 
			    String header, boolean showFields)
	throws Exception
    {
        HtmlTable innerTbl;
        HtmlTableRow tr, innerTr;
        HtmlTableDataCell td, innerTd;
        HtmlTableHeaderCell th;
	String domain, path;
	int maxAge;
	
        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
				     HEADER_FG_COLOR +
				     ">" + header + "</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Cookie Name</B>");
	td.setPercentWidth(10);		// Keeps first column narrow
        tr.addCell(td);
        if (showFields)
            td = new HtmlTableDataCell("<B>Attributes</B>");
        else
            td = new HtmlTableDataCell("<B>Value</B>");
        tr.addCell(td);
        tbl.addRow(tr);

	if (cookies.length == 0) {
            tr = new HtmlTableRow();
            td = new HtmlTableDataCell("<I>None found</I>");
            td.setColSpan(2);
            td.setHorizontalAlignment(td.ALIGN_CENTER);
            tr.addCell(td);
            tbl.addRow(tr);
	} else {
	    for (int idx = 0; idx < cookies.length; idx++) {
		tr = new HtmlTableRow();
		td = new HtmlTableDataCell(cookies[idx].getName());
		tr.addCell(td);

                if (showFields) {
           	    /*
                     * Since this is an outgoing cookie, we have access to
                     * the other non-value fields, like maxAge and path.
                     */
                    innerTbl = new HtmlTable();
       		    innerTbl.setCellPadding(1);
       		    innerTbl.setCellSpacing(1);
       		    innerTbl.setBackgroundColor(HIGHLIGHT_BG_COLOR);
       		    innerTbl.setBorderWidth(0);
       		    innerTbl.setPercentWidth(100);

		    innerTr = new HtmlTableRow();
		    innerTd = new HtmlTableDataCell("Value:");
		    innerTd.setPercentWidth(10);  // Keeps first column narrow
		    innerTd.setHorizontalAlignment(innerTd.ALIGN_RIGHT);
		    innerTr.addCell(innerTd);
		    innerTd = new HtmlTableDataCell(cookies[idx].getValue());
		    innerTr.addCell(innerTd);
		    innerTbl.addRow(innerTr);

		    innerTr = new HtmlTableRow();
		    innerTd = new HtmlTableDataCell("Domain:");
		    innerTd.setHorizontalAlignment(innerTd.ALIGN_RIGHT);
		    innerTr.addCell(innerTd);
		    domain = cookies[idx].getDomain();
		    if (domain == null)
		        innerTd = new HtmlTableDataCell("Not specified");
		    else 
		        innerTd = new HtmlTableDataCell(domain);
		    innerTr.addCell(innerTd);
		    innerTbl.addRow(innerTr);

		    innerTr = new HtmlTableRow();
		    innerTd = new HtmlTableDataCell("Path:");
		    innerTd.setHorizontalAlignment(innerTd.ALIGN_RIGHT);
		    innerTr.addCell(innerTd);
		    path = cookies[idx].getPath();
		    if (path == null)
		        innerTd = new HtmlTableDataCell("Not specified");
		    else
		        innerTd = new HtmlTableDataCell(path);
		    innerTr.addCell(innerTd);
		    innerTbl.addRow(innerTr);

		    innerTr = new HtmlTableRow();
		    innerTd = new HtmlTableDataCell("Max Age:");
		    innerTd.setHorizontalAlignment(innerTd.ALIGN_RIGHT);
		    innerTr.addCell(innerTd);
		    maxAge = cookies[idx].getMaxAge();
		    if (maxAge == -1)
		        innerTd = new HtmlTableDataCell("Not specified");
		    else
		        innerTd = new HtmlTableDataCell(maxAge + " seconds");
		    innerTr.addCell(innerTd);
		    innerTbl.addRow(innerTr);

		    innerTr = new HtmlTableRow();
		    innerTd = new HtmlTableDataCell("Secure:");
		    innerTd.setHorizontalAlignment(innerTd.ALIGN_RIGHT);
		    innerTr.addCell(innerTd);
		    if (cookies[idx].getSecure())
		        innerTd = new HtmlTableDataCell("Yes");
		    else
		        innerTd = new HtmlTableDataCell("No");
		    innerTr.addCell(innerTd);
		    innerTbl.addRow(innerTr);

		    td = new HtmlTableDataCell(innerTbl.toString());
                } else {
                    /*
                     * Since this is an incomming cookie, there is only
                     * the value to show. The maxAge field, for example,
                     * is not sent back to the server.
                     */
		    td = new HtmlTableDataCell(cookies[idx].getValue());
                }
                tr.addCell(td);
                tbl.addRow(tr);
	    }
        }
    }


    /**
     * Adds request parameters data rows to a table.
     *
     * @param tbl HTML table to which to add the data.
     * @param request Object from which http request parameter data is compiled.
     * @exception Any exception from underlying inplementation of
     *		  request object.
     */
    private static void
	addParameterInformation(HtmlTable tbl, DebugRequest request)
	throws Exception
    {
        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
				     HEADER_FG_COLOR +
				     ">Request Parameters</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Parameter Name</B>");
        tr.addCell(td);
        td = new HtmlTableDataCell("<B>Value(s)</B>");
        tr.addCell(td);
        tbl.addRow(tr);

	Enumeration parameterNames = request.getParameterNames();
	String[] parameterValues;
	String parameterName;

	if (!parameterNames.hasMoreElements()) {
            tr = new HtmlTableRow();
            td = new HtmlTableDataCell("<I>None found</I>");
            td.setColSpan(2);
            td.setHorizontalAlignment(td.ALIGN_CENTER);
            tr.addCell(td);
            tbl.addRow(tr);
	} else {
	    while (parameterNames.hasMoreElements()) {
		parameterName = (String)parameterNames.nextElement();
		parameterValues = request.getParameterValues(parameterName);
		tr = new HtmlTableRow();
		td = new HtmlTableDataCell(parameterName);
		tr.addCell(td);
		for (int idx = 0; idx < parameterValues.length; idx++) {
		    if (idx == 0) {
			td = new HtmlTableDataCell(parameterValues[0].trim(),
                                                   "&nbsp");
			tr.addCell(td);
			tbl.addRow(tr);
		    } else {
			tr = new HtmlTableRow();
			td = new HtmlTableDataCell("");
			tr.addCell(td);
			td = new HtmlTableDataCell(parameterValues[idx].trim(),
                                                   "&nbsp");
			tr.addCell(td);
			tbl.addRow(tr);
		    }
		}
	    }
        }
    }


    /**
     * Given a HTML table object, this method extracts all useful information
     * from a httpResponse object and composes a new table row for each
     * datum.
     *
     * @param tbl The HTML table object to attach information.
     * @param response The response object to be dumped.
     * @exception Any exception from underlying inplementation of
     *		  response object.
     */
    private static void
	addResponseInformation(HtmlTable tbl, DebugResponse response)
	throws Exception
    {
        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
				     HEADER_FG_COLOR +
				     ">Response Information</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

        if (response == null || response.equals("")) {
            tr = new HtmlTableRow();
            td = new HtmlTableDataCell("<I>Response Not Available</I>");
            td.setColSpan(2);
            td.setHorizontalAlignment(td.ALIGN_CENTER);
            tr.addCell(td);
            tbl.addRow(tr);
        } else {
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Information</B>");
	td.setPercentWidth(10);		// Keeps first column narrow
        tr.addCell(td);
        td = new HtmlTableDataCell("<B>Value</B>");
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getContentLength
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Content Length");
        tr.addCell(td);
	int contentLength = response.getContentLength();
	if (contentLength == -1) {
	    td = new HtmlTableDataCell("<I>Not set</I>");
	} else {
	    td = new HtmlTableDataCell(Integer.toString(contentLength));
	}
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getContentType
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Content Type");
        tr.addCell(td);
        String value = response.getContentType();
        if (value == null)
            value = "<I>Not set</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getStatusCode
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Status Code");
        tr.addCell(td);
        int sc = response.getStatusCode();
        String code;
        if (sc == -1)
            code = "<i>Not set (defaults to 200 OK)</i>";
        else
            code = Integer.toString(sc) + " (" + 
                      HttpStatus.getStatusMsg(response.getStatusCode()) +
                      ")";
        td = new HtmlTableDataCell(code);
        tr.addCell(td);
        tbl.addRow(tr);

	/*
	 * getStatusMessage
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Status Message");
        tr.addCell(td);
        value = response.getStatusMessage();
        if (value == null)
            value = "<I>N/A</I>";
        td = new HtmlTableDataCell(value);
        tr.addCell(td);

	/*
	 * getTotalBytes
	 */
        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("Total Bytes");
        tr.addCell(td);
	int totalBytes = response.getTotalBytes();
	if (totalBytes == -1) {
	    td = new HtmlTableDataCell("<I>N/A</I>");
	} else {
	    td = new HtmlTableDataCell(Integer.toString(totalBytes));
	}
        tr.addCell(td);
        tbl.addRow(tr);
        }
    }


    /**
     * Given a HTML table object, this method adds a section for the
     * response data (what was writtent to the output stream).
     * If that data was not saved (most likeley scenerio), the table is
     * left untouched.
     *
     * @param tbl The HTML table object to attach information.
     * @param response The response object to be dumped.
     * @exception Any exception from underlying inplementation of
     *		  response object.
     */
    private static void
	addResponseDataInformation(HtmlTable tbl, DebugResponse response)
	throws Exception 
    {
        /*
         * Optionally, the data written to the output stream. This is optional,
         * and is more likely than not to be null.
         */
        String responseData = response.getResponseData();
        if (responseData == null) {
            return;
	}

        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
				     HEADER_FG_COLOR +
                                     ">Response Data</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("<FORM><TEXTAREA ROWS=20 COLS=80>" +
                                   HtmlEncoder.encode(responseData) +
                                   "</TEXTAREA></FORM>");
        td.setColSpan(2);
        tr.addCell(td);
        tbl.addRow(tr);
    }


    /**
     * Adds response headers data rows to a table.
     *
     * @param tbl HTML table to which to add the data.
     * @param response Object from which http response header data is compiled.
     * @exception Any exception from underlying inplementation of
     *		  response object.
     */
    private static void
	addHeaderInformation(HtmlTable tbl, DebugResponse response)
	throws Exception
    {
        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
				     HEADER_FG_COLOR +
				     ">Response Headers</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

        tr = new HtmlTableRow();
        td = new HtmlTableDataCell("<B>Header Name</B>");
	td.setPercentWidth(10);		// Keeps first column narrow
        tr.addCell(td);
        td = new HtmlTableDataCell("<B>Value(s)</B>");
        tr.addCell(td);
        tbl.addRow(tr);

	Enumeration headerNames = response.getHeaderNames();
	String headerName, headerValue;
	boolean first = true;


	if (!headerNames.hasMoreElements()) {
	    tr = new HtmlTableRow();
	    td = new HtmlTableDataCell("<I>None found</I>");
	    td.setColSpan(2);
	    td.setHorizontalAlignment(td.ALIGN_CENTER);
	    tr.addCell(td);
	    tbl.addRow(tr);
	} else {
	    while (headerNames.hasMoreElements()) {
		headerName = (String)headerNames.nextElement();
		headerValue = response.getHeader(headerName);
                tr = new HtmlTableRow();
                tr.addCell(new HtmlTableDataCell(capitalize(headerName)));
                tr.addCell(new HtmlTableDataCell(headerValue));
                tbl.addRow(tr);
	    }
        }
    }


    /**
     * Adds response cookie data rows to a table.
     *
     * @param tbl HTML table to which to add the data.
     * @param response Object from which http response cookie data is compiled.
     * @exception Any exception from underlying inplementation of
     *		  response object.
     */
    private static void
	addCookieInformation(HtmlTable tbl, DebugResponse response)
	throws Exception
    {
	Cookie[] cookies = response.getCookies();
	formatCookiesAsHtml(tbl, cookies, "Response Cookies", true);
    }


    /**
     * Adds a dump of the request object trace.
     *
     * @param tbl HTML table to which to add the data.
     * @param request Object from which http request trace is obtained.
     */
    private static void
	addTrace(HtmlTable tbl, DebugRequest request)
    {
        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
				     HEADER_FG_COLOR +
				     ">Request Trace</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);
    }


    /**
     * Adds a dump of the response object trace.
     *
     * @param trace The string to be formatted.
     * @return The formatted HTML.
     */
    public static String
	formatTraceDebugInfo(String trace)
    {
        HtmlTable tbl = new HtmlTable();
        tbl.setCellPadding(1);
        tbl.setCellSpacing(2);
        tbl.setBackgroundColor(BODY_BG_COLOR);
        tbl.setBorderWidth(0);
        tbl.setPercentWidth(100);

        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
                                     HEADER_FG_COLOR +
                                     ">Trace Information</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

	if (trace == null || trace.equals("")) {
            tr = new HtmlTableRow();
            td = new HtmlTableDataCell("<I>Trace Not Available</I>");
            td.setColSpan(2);
            td.setHorizontalAlignment(td.ALIGN_CENTER);
            tr.addCell(td);
            tbl.addRow(tr);
	} else {
	    String line;
	    int stop, idx = 0;
	    while ((stop = trace.indexOf("\n", idx)) > 0) {
		line = trace.substring(idx, stop);
		tr = new HtmlTableRow();
		td = new HtmlTableDataCell("<CODE>" + line + "</CODE>");
		td.setColSpan(2);
		tr.addCell(td);
                tbl.addRow(tr);
		idx = stop + 1;
	    }
	    line = trace.substring(idx);
	    tr = new HtmlTableRow();
	    td = new HtmlTableDataCell("<CODE>" + line + "</CODE>");
	    td.setColSpan(2);
	    tr.addCell(td);
            tbl.addRow(tr);
	}
        /*
         * Add frame background...
         */
        HtmlTable frameTbl = new HtmlTable();
        frameTbl.setBackgroundColor(FRAME_COLOR);
        frameTbl.setPercentWidth(100);
        HtmlTableRow tfr = new HtmlTableRow();
        HtmlTableDataCell tfd = new HtmlTableDataCell(tbl.toString());
        tfr.addCell(tfd);
        frameTbl.addRow(tfr);
 
        return frameTbl.toString();
}


    public static String
	formatSessionInfo(String info, String title)
    {
        HtmlTable tbl = new HtmlTable();
        tbl.setCellPadding(1);
        tbl.setCellSpacing(2);
        tbl.setBackgroundColor(BODY_BG_COLOR);
        tbl.setBorderWidth(0);
        tbl.setPercentWidth(100);

        HtmlTableRow tr;
        HtmlTableDataCell td;
        HtmlTableHeaderCell th;

        tr = new HtmlTableRow();
        th = new HtmlTableHeaderCell("<FONT SIZE=+1 COLOR=" +
                                     HEADER_FG_COLOR +
                                     ">" + title + "</FONT>");
        th.setColSpan(2);
        th.setHorizontalAlignment(th.ALIGN_CENTER);
        tr.addCell(th);
        tr.setBackgroundColor(HEADER_BG_COLOR);
        tbl.addRow(tr);

	if (info == null || info.equals("")) {
            tr = new HtmlTableRow();
            td = new HtmlTableDataCell("<I>N/A</I>");
            td.setColSpan(2);
            td.setHorizontalAlignment(td.ALIGN_CENTER);
            tr.addCell(td);
            tbl.addRow(tr);
	} else {
	    tr = new HtmlTableRow();
	    td = new HtmlTableDataCell("<CODE>" + info + "</CODE>");
	    td.setColSpan(2);
	    tr.addCell(td);
            tbl.addRow(tr);
	}
        /*
         * Add frame background...
         */
        HtmlTable frameTbl = new HtmlTable();
        frameTbl.setBackgroundColor(FRAME_COLOR);
        frameTbl.setPercentWidth(100);
        HtmlTableRow tfr = new HtmlTableRow();
        HtmlTableDataCell tfd = new HtmlTableDataCell(tbl.toString());
        tfr.addCell(tfd);
        frameTbl.addRow(tfr);
 
        return frameTbl.toString();
    }


    /**
     * Capitialize the first letter of a string and any character
     * following a "-" character. This is the convention for Http
     * headers.
     *
     * @param trace The string to be formatted.
     * @return The formatted HTML.
     */
    private static String capitalize (String s) {
	boolean convert = true;
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (convert) {
		sb.append(Character.toUpperCase(c));
		convert = false;
	    } else {
		sb.append(c);
	    }
	    if (c == '-') {
		convert = true;
	    }
	}
	return sb.toString();
    }
}
