/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BaseTestThread.java,v 1.6.2.2.2.1 2000/10/19 17:59:07 jasona Exp $
 */

package org.enhydra.httpServerTest.client;

import org.enhydra.httpServerTest.common.HttpResponseObject;
import org.enhydra.httpServerTest.common.HttpServerTestUtils;

import java.util.Enumeration;

/**
 * This class provides the base set of functionality that all test threads
 * share.  This functionality includes maintaining references to the time
 * the thread was created, the id of the thread, the thread configuration 
 * parameters, and the thread results collector.  All these references are
 * initialized in the constructor and are accessable through get methods.
 *
 * This class also provides initialization functions to construct and 
 * initialize the expected golden response based on the given thread's
 * configuration parameters.
 *
 * @author Mike Ward
 **/
public abstract class BaseTestThread 
    extends Thread {

    /**
     * Holds the time the thread was created.
     **/
    protected long startTime = (long)-1;

    /**
     * Holds the id value of this thread.
     **/
    protected String id = null;

    /**
     * Holds the thread's configuration parameters.
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     **/
    protected HttpTestClientConfig params = null;

    /**
     * Holds a reference to the results collector.
     * @see org.enhydra.httpServerTest.client.ThreadResultsCollector
     **/
    protected ThreadResultsCollector threadResultsCollector = null;

    /**
     * The expected response.  This object is initialized once, before the
     * initial request is sent.  The response is an XML structure that is
     * parsed into an object of the same type.  The two objects are compared
     * and any differences are reported.
     * @see org.enhydra.httpServerTest.common.HttpResponseObject
     * @see org.enhydra.httpServerTest.common.HttpResponseObjectParser
     **/
    protected HttpResponseObject goldenResponse = null;

    /**
     * Constructor initializes all member variables for this thread.
     *
     * @param id
     *        A String identification for this thread.
     * @param params
     *        The thread's configuration parameters; can not be null.
     * @param threadResultsCollector
     *        The result collector for this thread; can not be null.
     * @exception IllegalArgumentException
     *            Thrown if a parameter does not meet the above constraints.
     **/
    public BaseTestThread(String id, 
			  HttpTestClientConfig params,
			  ThreadResultsCollector threadResultsCollector)
	throws IllegalArgumentException {

	startTime = System.currentTimeMillis();

	if (id == null) { 
	    throw new IllegalArgumentException("ID argument can not be null.");
	} else {
	    this.id = id;
	}

	if (params == null) {
	    throw new IllegalArgumentException("Params argument can not be null.");
	} else {
	    this.params = params;
	}
	
	if (threadResultsCollector == null) {
	    throw new IllegalArgumentException("ThreadResultsCollector argument can not be null.");
	} else {
	    this.threadResultsCollector = threadResultsCollector;
	}
    }

    /**
     * Initializes the expected response object.  These attributes should 
     * be derived from the configuration parameters assigned to this thread.
     * This function runs through all mathods in this class matching the
     * signiture setGoldenResponse<type>().  If any of these methods
     * returns false, this function will return false indicating that the
     * initialization was not completed successfully.  Typically client
     * threads will override some of the setGoldenResponse<type>() methods
     * to customize the golden response attributes.
     *
     * @return boolean
     *         The success status of the initialization.
     **/
    protected boolean initGoldenResponse() {
	boolean init1 = setGoldenResponseObject();
	boolean init2 =	setGoldenResponseAuthType();
	boolean init3 =	setGoldenResponseMethod();
	boolean init4 =	setGoldenResponsePathInfo();
	boolean init5 =	setGoldenResponsePathTranslated();
	boolean init6 =	setGoldenResponseRemoteUser();
	boolean init7 =	setGoldenResponseRequestURI();
	boolean init8 =	setGoldenResponseServletPath();
	boolean init9 =	setGoldenResponseSessionInfo();

	boolean init10 = setGoldenResponseMD5HashKey();

	boolean init11 = setGoldenResponseCgiArgs();
	boolean init12 = setGoldenResponseQueryString();
	boolean init13 = setGoldenResponseCgiParams();
	boolean init14 = setGoldenResponseHeaders();
	boolean init15 = setGoldenResponseCookies();

	return (init1 && init2 && init3 && init4 && init5 && init6 &&
		init7 && init8 && init9 && init10 && init11 && init12 &&
		init13 && init14 && init15);
    }

    /**
     * This function instantiates the golden response object.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseObject() {
	goldenResponse = new HttpResponseObject();
	return true;
    }

    /**
     * This function sets the golden response authType attribute.  The
     * initialization here set this to 'null', authentication is not
     * used.  Client thread extending this class and using authenticated
     * connections should override this method.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseAuthType() {
	goldenResponse.authType = null;
	return true;
    }

    /**
     * This function sets the golden response method attribute.  This is
     * initialized to null here because this class does not preform any
     * actual requests.  Client thread extending this class should override
     * this function.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseMethod() {
	goldenResponse.method = null;
	return true;
    }

    /**
     * This function sets the golden response pathInfo attribute.  This
     * consists of a subset of the URI string.  This method is typically
     * fine for most request types and not usually extended.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponsePathInfo() {
	String uri = HttpTestClientUtils.getURI(params);
	if (uri.length() > 1) {
	    if (params.redirectedToFile == null) {
		goldenResponse.pathInfo = uri.substring(params.file.length());
	    } else {
		goldenResponse.pathInfo = 
		    uri.substring(params.redirectedToFile.length());
	    }
	} else {
	    goldenResponse.pathInfo = uri;
	}
	return true;
    }

    /**
     * This function sets the golden response pathTranslated attribute.  This
     * attribute is unknown until the first request is made so the value
     * set here is 'null'.  This attribute should be set after the first
     * response has been received with a call to 
     * HttpTestClientUtils.populatePathTranslated().
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientUtils
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponsePathTranslated() {
	goldenResponse.pathTranslated = null;
	return true;
    }

    /**
     * This function sets the golden response remoteUser attribute.  This
     * attribute is initialized to 'null' indicating no user information
     * is sent with the request.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseRemoteUser() {
	goldenResponse.remoteUser = null;
	return true;
    }

    /**
     * Thi function sets the golden response requestURI attribute.  This
     * attribute using a call to HttpTestClientUtils.getURI().
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientUtils
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseRequestURI() {
	goldenResponse.requestURI = HttpTestClientUtils.getURI(params);
	return true;
    }

    /**
     * This function sets the goldenResponse servletPath attribute.  This 
     * method is typically fine for most request types and not usually
     * extended.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseServletPath() {
	goldenResponse.servletPath = "";
	if (params.file.charAt(0) != '/') {
	    goldenResponse.servletPath = "/";
	}
	int servletPathEnd = params.file.lastIndexOf('/');
	if (servletPathEnd >= 0) {
	    goldenResponse.servletPath += 
		params.file.substring(0, (servletPathEnd + 1));
	}
	return true;
    }

    /**
     * This function sets the golden response session information.  Although
     * a session key may be specified specifically in the configuration file,
     * it is not recomended unless attempting to test persistent sessions.
     * Normal behavior for this function will be to initialize all session
     * information to 'null' and rely on 
     * HttpTestClientUtils.populateSessionInfo() to populate session 
     * information based on the initial response.  If a session key is found
     * in the configuration parameters (as a session cookie or a session CGI
     * argument, this key will be used to initialize session information and 
     * all responses will be validated against this key.  A session key is 
     * specified by including a CGI parameter of 'jsessionid' or a cookie
     * of 'JSESSIONID'.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientUtils
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseSessionInfo() {
	Enumeration cookieKeys = params.cookies.keys();
	while(cookieKeys.hasMoreElements()) {
	    String currentKey = (String)cookieKeys.nextElement();
	    if (currentKey.equals(HttpServerTestUtils.HTTP_COOKIE_SESSION_KEY)) {
		goldenResponse.session = 
		    (String)params.cookies.get(currentKey);
		goldenResponse.sessionIdFromCookie = true;
		goldenResponse.sessionIdFromURL = false;
		goldenResponse.sessionIdInvalid = false;
		return true;
	    }
	}

	Enumeration cgiParamKeys = params.cgiParams.keys();
	while(cgiParamKeys.hasMoreElements()) {
	    String currentKey = (String)cgiParamKeys.nextElement();
	    if (currentKey.equals(HttpServerTestUtils.CGI_PARAM_SESSION_KEY)) {
		goldenResponse.session = 
		    (String)params.cgiParams.get(currentKey);
		goldenResponse.sessionIdFromCookie = false;
		goldenResponse.sessionIdFromURL = true;
		goldenResponse.sessionIdInvalid = false;
		return true;
	    }
	}

	goldenResponse.session = HttpResponseObject.DEFAULT_SESSION;
	goldenResponse.sessionIdFromCookie = false;
	goldenResponse.sessionIdFromURL = false;
	goldenResponse.sessionIdInvalid = true;
	return true;
    }
    
    /**
     * This function sets the golden response md5HashKey attribute.  This
     * attribute is a vector of hash keys for all upload files specified
     * in the configuration file.  This function initially sets this vector
     * to 'null' and is typically extended by a client thread that wants to
     * preform file uploads.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseMD5HashKey() {
	goldenResponse.md5HashKey = null;
	return true;
    }

    /**
     * This function set the golden response CGI argument attribute.  This 
     * method is typically fine for most request types and not usually 
     * extended.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseCgiArgs() {
	goldenResponse.cgiArgs = params.cgiArgs;
	return true;
    }

    /**
     * This function sets the golden repsonse query string attribute.  Since,
     * query strings are only valid for GET requests, this function initializes
     * the attribute to 'null' and client threads issuing GET request should
     * override this method.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseQueryString() {
	goldenResponse.queryString = null;
	return true;
    }
    
    /**
     * This function set the golden response CGI parameter attribute.  This 
     * method is typically fine for most request types and not usually 
     * extended.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseCgiParams() {
	goldenResponse.cgiParams = params.cgiParams;
	return true;
    }

    /**
     * This function set the golden response header attribute.  This 
     * method is typically fine for most request types and not usually 
     * extended.
     *
     * @return boolean
     *         The success state of set function.
     **/    
    protected boolean setGoldenResponseHeaders() {
	goldenResponse.headers = params.headers;
	return true;
    }
    
    /**
     * This function set the golden response cookie attribute.  This 
     * method is typically fine for most request types and not usually 
     * extended.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseCookies() {
	goldenResponse.cookies = params.cookies;
	return true;
    }

    /**
     * Default run method for this thread, just calls initGoldenResponse
     * and exits.  This function will typically be extended in client threads.
     **/
    public void run() {
	boolean initSuccess = initGoldenResponse();
	if (!initSuccess) {
	    return;
	}
    }

    /**
     * Getter method for the start time of this thread; the time the thread
     * was created.
     *
     * @return long
     *         The start time in milliseconds.
     **/
    public long getStartTime() {
	return startTime;
    }

    /**
     * Getter method for thread's id.
     *
     * @return String
     *         The id.
     **/
    public String getId() {
	return id;
    }

    /**
     * Getter method for thread's confuration parameters.
     *
     * @return HttpTestClientConfig
     *         Configuration parameters.
     **/
    public HttpTestClientConfig getParams() {
	return params;
    }
}
