/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpTestClientConfig.java,v 1.1.2.3.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.client;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

/**
 * This class holds all configuration parameters for a test case thread.  This
 * class acts as a simple data structure that gets populated from a XML
 * configuration file with the HttpTestClientConfigParser.
 * @see org.enhydra.httpServerTest.client.HttpTestClientConfigParser
 *
 * Configuration parameters held within this class are all public and acessed
 * directly by the class objects when needed.
 * @see org.enhydra.httpServerTest.client.HttpTestClient
 *
 * @author Mike Ward
 **/
public class HttpTestClientConfig {

    /**
     * Static Test element tag and its attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final String TEST_ELEMENT = "Test";
    public static final String TEST_NAME_ATTR = "testName";
    public static final String TEST_ID_ATTR = "id";

    /**
     * Static TestParams element tag and its attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final String TEST_PARAMS_ELEMENT = "TestParams";
    public static final String TEST_PARAMS_NUM_THREADS_ATTR = "numThreads";
    public static final String TEST_PARAMS_REQUESTS_PER_THREAD_ATTR = 
	"requestsPerThread";
    public static final String TEST_PARAMS_DELAY_BETWEEN_REQUESTS_ATTR = 
	"delayBetweenRequests";
    public static final String TEST_PARAMS_SESSION_TRACKING = 
	"sessionTracking";

    /**
     * Static Url element tag and its attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final String URL_ELEMENT = "Url";
    public static final String URL_PROTOCOL_ATTR = "protocol";
    public static final String URL_KEY_STORE_PASSWORD_ATTR = 
	"keyStorePassword";
    public static final String URL_KEY_STORE_LOCATION_ATTR = 
	"keyStoreLocation";
    public static final String URL_HOST_ATTR = "host";
    public static final String URL_PORT_ATTR = "port";
    public static final String URL_FILE_ATTR = "file";
    public static final String URL_REDIRECTED_TO_FILE_ATTR = 
	"redirectedToFile";

    /**
     * Static CgiParam element tag and its attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final String CGI_PARAM_ELEMENT = "CgiParam";
    public static final String CGI_PARAM_NAME_ATTR = "name";
    public static final String CGI_PARAM_VALUE_ATTR = "value";

    /**
     * Static CgiArg element tag and its attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final String CGI_ARG_ELEMENT = "CgiArg";
    public static final String CGI_ARG_NAME_ATTR = "name";
    public static final String CGI_ARG_VALUE_ATTR = "value";

    /**
     * Static Header element tag and its attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final String HEADER_ELEMENT = "Header";
    public static final String HEADER_NAME_ATTR = "name";
    public static final String HEADER_VALUE_ATTR = "value";

    /**
     * Static Cookie element tag and its attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final String COOKIE_ELEMENT = "Cookie";
    public static final String COOKIE_NAME_ATTR = "name";
    public static final String COOKIE_VALUE_ATTR = "value";

    /**
     * Static FileUpload element tag and its attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final String FILE_UPLOAD_ELEMENT = "FileUpload";
    public static final String FILE_UPLOAD_FILE_DIR_ATTR = "fileDir";
    public static final String FILE_UPLOAD_FILE_NAME_ATTR = "fileName";

    /**
     * Static default or implied values for element attributes.
     * @see HttpTestClientConfig.dtd for details
     **/
    public static final int DEFAULT_NUM_THREADS = 1;
    public static final int DEFAULT_REQUEST_PER_THREAD = 1;
    public static final long DEFAULT_DELAY_BETWEEN_REQUESTS = (long)0;
    public static final String DEFAULT_SESSION_TRACKING = "none";
    public static final String SESSION_TRACKING_BY_COOKIE = "cookie";
    public static final String SESSION_TRACKING_BY_CGI_PARAM = "cgiParam";
    
    public static final String DEFAULT_PROTOCOL = "http";
    public static final String DEFAULT_HOST = "localhost";
    public static final int DEFAULT_HTTP_PORT = 80;
    public static final int DEFAULT_HTTPS_PORT = 443;

    /**
     * Test parameter varibles.
     **/
    public String id = null;
    public int numThreads = DEFAULT_NUM_THREADS;
    public int requestsPerThread = DEFAULT_REQUEST_PER_THREAD;
    public long delayBetweenRequests = DEFAULT_DELAY_BETWEEN_REQUESTS;
    public String sessionTracking = DEFAULT_SESSION_TRACKING;
    public String protocol = DEFAULT_PROTOCOL;
    public String keyStorePassword = "";
    public String keyStoreLocation = "";
    public String host = DEFAULT_HOST;
    public int port = DEFAULT_HTTP_PORT;
    public String file = null;
    public String redirectedToFile = null;
    public Hashtable cgiParams = new Hashtable();
    public Hashtable cgiArgs = new Hashtable();
    public Hashtable headers = new Hashtable();
    public Hashtable cookies = new Hashtable();
    public Vector fileUploads = new Vector();

    /**
     * This function provides a true clone of this object.  All objects
     * held in this class are realocated and a new instance containing
     * the exact values held in this class is returned.  This function is
     * used to hand each client thread its own, unique instance of the
     * configuration parameters.  A single instance of this object can not 
     * be shared by all client threads because some values such as session
     * cookies are unique to each thread.
     *
     * @return HttpTestClientConfig
     *         The newly cloned instance of this object.
     **/
    public HttpTestClientConfig cloneInstance() {
	HttpTestClientConfig newInstance = new HttpTestClientConfig();
	newInstance.id = new String(this.id);
	newInstance.numThreads = this.numThreads;
	newInstance.requestsPerThread = this.requestsPerThread;
	newInstance.delayBetweenRequests = this.delayBetweenRequests;
 	newInstance.sessionTracking = new String(this.sessionTracking);
 	newInstance.protocol = new String(this.protocol);
 	newInstance.keyStorePassword = new String(this.keyStorePassword);
 	newInstance.keyStoreLocation = new String(this.keyStoreLocation);
 	newInstance.host = new String(this.host);
 	newInstance.port = this.port;
	if (this.file != null) {
	    newInstance.file = new String(this.file);
	}
	if (this.redirectedToFile != null) {
	    newInstance.redirectedToFile = new String(this.redirectedToFile);
	}
 	newInstance.cgiParams = new Hashtable();
	Enumeration keys = this.cgiParams.keys();
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    newInstance.cgiParams.put(new String(currentKey), 
			  new String((String)this.cgiParams.get(currentKey)));
	}
 	newInstance.cgiArgs = new Hashtable();
	keys = this.cgiArgs.keys();
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    newInstance.cgiArgs.put(new String(currentKey), 
			  new String((String)this.cgiArgs.get(currentKey)));
	}
 	newInstance.headers = new Hashtable();
	keys = this.headers.keys();
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    newInstance.headers.put(new String(currentKey), 
			  new String((String)this.headers.get(currentKey)));
	}
 	newInstance.cookies = new Hashtable();
	keys = this.cookies.keys();
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    newInstance.cookies.put(new String(currentKey), 
			  new String((String)this.cookies.get(currentKey)));
	}
	
	newInstance.fileUploads = new Vector();
	Enumeration files = this.fileUploads.elements();
	while (files.hasMoreElements()) {
	    String currentFile = (String)files.nextElement();
	    newInstance.fileUploads.addElement(new String(currentFile));
	}

	return newInstance;
    }

    /**
     * Overloaded toString method, prints out all configuration parameter
     * variable held within this class.
     *
     * @return String
     *         A formatted string displaying the value of all configuration
     *         parameters.
     **/
    public String toString() {
	String cgiParamString = new String("'");
	String cgiArgString = new String("'");
	String headerString = new String("'");
	String cookieString = new String("'");
	String fileString = new String("'");

	Enumeration keys = cgiParams.keys();
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    cgiParamString += currentKey + "=" + 
		cgiParams.get(currentKey) + " ";
	}
	cgiParamString += "'";

	keys = cgiArgs.keys();
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    cgiArgString += currentKey + "=" + 
		cgiArgs.get(currentKey) + " ";
	}
	cgiArgString += "'";


	keys = headers.keys();
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    headerString += currentKey + "=" + 
		headers.get(currentKey) + " ";
	}
	headerString += "'";
	    
	keys = cookies.keys();
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    cookieString += currentKey + "=" + 
		cookies.get(currentKey) + " ";
	}
	cookieString += "'";

	Enumeration files = fileUploads.elements();
	while (files.hasMoreElements()) {
	    String currentFile = (String)keys.nextElement();
	    fileString += currentFile + ", ";
	}
	fileString += "'";
	    
	return "HttpTestClientConfig: id=" + id +
	    " numThreads=" + numThreads + 
	    " requestsPerThread=" + requestsPerThread +
	    " delayBetweenRequests=" + delayBetweenRequests +
	    " sessionTracking=" + sessionTracking +
	    " protocol=" + protocol +
	    " keyStorePassword=" + keyStorePassword +
	    " keyStoreLocation=" + keyStoreLocation +
	    " host=" + host +
	    " port=" + port + 
	    " file=" + file +
	    " redirectedToFile=" + redirectedToFile +
	    " cgiParams:" + cgiParamString +
	    " cgiArgs:" + cgiArgString + 
	    " headers:" + headerString + 
	    " cookies:" + cookieString + 
	    " fileUploads=" + fileString;
    }
}
