/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpTestClientConfigParser.java,v 1.1.2.3.2.2 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.client;
                    
import java.util.Hashtable;
import java.io.File;
import java.io.FileReader;

import org.xml.sax.InputSource;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This class uses the Xerces SAX parser to parse the test parameters 
 * configuration file and populate an HttpTestClientConfig object.  This
 * class implements a factory mechanism for retrieving a hashtable of
 * all test parameter sections.  Any errors encountered while parsing the 
 * XML file will cause an exception to be thrown.
 * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
 * @see org.apache.xerces.parsers.SAXParser
 *
 * This parser implementation will assume validation against 
 * HttpTestClientConfig.dtd.
 *
 * @author Mike Ward
 **/
public class HttpTestClientConfigParser
    extends DefaultHandler {

    /** 
     * Default parser name.  This value can be supplied if a different SAX
     * parser is desired.
     * @see org.apache.xerces.parsers.SAXParser
     **/
    public static final String DEFAULT_PARSER_NAME = 
	"org.apache.xerces.parsers.SAXParser";

    /**
     * The collection of all test case parameters read from the XML 
     * configuration file.
     **/
    private Hashtable testSuiteParams = null;

    /**
     * A reference to the current test case parameter object that is being
     * populated.
     **/
    private HttpTestClientConfig currentHttpTestClientConfig = null;

    /**
     * A boolean flag to indicate if any errors were found while parsing 
     * the configuration file.  This is a type of 'workaround' since some
     * of the SAX callback functions do not allow exceptions to be thrown.
     **/
    private boolean parseErrorFound = false;

    /**
     * The error string associated with a parse error.  This is the final 
     * message that is reported in the Exception thrown from the parse()
     * functions.  This acts as a message buffer used by the 'error()' and 
     * 'fatalError()' callback methods.
     **/
    private String parseErrorMessage = null;

    /**
     * Private constructor.  Prevents object from being instantiated from
     * outside the factory implementation.  The constructor initializes a new
     * hashtable for maintaining test case parameter objects.
     **/
    private HttpTestClientConfigParser() {
	testSuiteParams = new Hashtable();
    }
    
    /**
     * Static factory method for retrieving test case parameters using the 
     * default Xerces SAX parser.  This function just calls 
     * 'parse(inputSource, parserName) with the value of 'DEFAULT_PARSER_NAME'.
     *
     * @param uri
     *        The location of the XML configuration file.  This function will
     *        throw an exception if the uri does not begin with 'file:' or
     *        the file is not found.
     * @return Hashtable
     *         The hashtable of all test case parameters keyed by the test 
     *         case name.
     * @exception Exception
     *            Thrown if an error is encoutered while parsing or the uri
     *            does not meet the above stated specifications.
     **/
    public static Hashtable parse(String uri) 
	throws Exception {	
	if (!uri.startsWith("file:")) {
	    throw new Exception("Only a URI of type \"file:\" is currently supported. URI: \"" + uri + "\"");
	}
	File file = new File(uri.substring(uri.indexOf(':') + 1));
	FileReader fileReader = new FileReader(file);
	return parse(new InputSource(fileReader), DEFAULT_PARSER_NAME);
    }

    /**
     * Static factory method for retrieving test case parameters using the 
     * default Xerces SAX parser.  This function just calls 
     * 'parse(inputSource, parserName) with the value of 'DEFAULT_PARSER_NAME'.
     *
     * @param inputSource
     *        The input source for the XML data to parse.
     * @return Hashtable
     *         The hashtable of all test case parameters keyed by the test 
     *         case name.
     * @exception Exception
     *            Thrown if an error is encoutered while parsing.
     **/
    public static Hashtable parse(InputSource inputSource) 
	throws Exception {
	return parse(inputSource, DEFAULT_PARSER_NAME);
    }

    /**
     * Static factory method for retrieving test case parameters using the 
     * supplied SAX parser.  This function instantiates a 
     * HttpTestClientConfigParser object and registers that instance as the
     * content and error handler for the supplied SAX parser.  Any errors
     * encountered while parsing cause an exception to be thrown.
     *
     * @param inputSource
     *        The input source of the XML data to parse.
     * @param parserName
     *        The supplied class name of a preferred SAX parser.  This will
     *        override the use of the DEFAULT_PARSER_NAME.
     * @return Hashtable
     *         The hashtable of all test case parameters keyed by the test 
     *         case name.
     * @exception Exception
     *            Thrown if an error is encoutered while parsing.
     **/
    public static Hashtable parse(InputSource inputSource, 
				  String parserName) 
	throws Exception {

	HttpTestClientConfigParser httpTestClientConfigParser = 
	    new HttpTestClientConfigParser();

	XMLReader parser = 
	    (XMLReader)Class.forName(parserName).newInstance();
	parser.setContentHandler(httpTestClientConfigParser);
	parser.setErrorHandler(httpTestClientConfigParser);
	parser.setFeature("http://xml.org/sax/features/validation", true);
	
	parser.parse(inputSource);

	if (httpTestClientConfigParser.parseErrorFound) {
	    throw new Exception(httpTestClientConfigParser.parseErrorMessage);
	}

	return httpTestClientConfigParser.testSuiteParams;
    }

    /** 
     * Callback function invoked when a new element is found.  This function
     * implements a switch based on the element type retrieved.  The possible
     * set of elements can be found in HttpTestClientConfiag as static final
     * Strings matching the pattern <elemenet type>_ELEMENT.
     * 
     * @see org.apache.xerces.parsers.SAXParser
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param uri
     *        This argument is not used here.
     * @param local
     *        This argument is not used here.
     * @param elementName
     *        The name of the current element.
     * @param attributes
     *        The attributes associated with this element.
     **/
    public void startElement(String uri, 
			     String local, 
			     String elementName, 
			     Attributes attributes) {

	if (elementName.equals(HttpTestClientConfig.TEST_ELEMENT)) {
	    currentHttpTestClientConfig = new HttpTestClientConfig();
	    addTestToHashtable(attributes);
	} else if (elementName.equals(HttpTestClientConfig.TEST_PARAMS_ELEMENT)) {
	    setTestParamsAttributes(attributes);
	} else if (elementName.equals(HttpTestClientConfig.URL_ELEMENT)) {
	    setUrlAttributes(attributes);
	} else if (elementName.equals(HttpTestClientConfig.CGI_PARAM_ELEMENT)) {
	    setCgiParamAttributes(attributes);
	} else if (elementName.equals(HttpTestClientConfig.CGI_ARG_ELEMENT)) {
	    setCgiArgAttributes(attributes);
	} else if (elementName.equals(HttpTestClientConfig.HEADER_ELEMENT)) {
	    setHeaderAttributes(attributes);
	} else if (elementName.equals(HttpTestClientConfig.COOKIE_ELEMENT)) {
	    setCookieAttributes(attributes);
	} else if (elementName.equals(HttpTestClientConfig.FILE_UPLOAD_ELEMENT)) {
	    setFileUploadAttributes(attributes);
	}
    }

    /**
     * This function adds the current test parameters object into the
     * hashtable keyed by the <test name>_<test name count>  Therefore
     * multiple test parameter groups can be specified for a single test.
     * This function is invoked when an element of type 'Test' is encountered.
     * The list of possible 'Test' element attributes can be found in 
     * HttpTestClientConfig as static final Strings matching the pattern
     * TEST_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void addTestToHashtable(Attributes attributes) {
	for (int i = 0; i < attributes.getLength(); i++) {
	    String attributeName = attributes.getQName(i);
	    if (attributeName.equals(HttpTestClientConfig.TEST_NAME_ATTR)) {
		int cnt = 0;
		while (true) {
		    if (!testSuiteParams.containsKey(attributes.getValue(i) + 
						     "_" + cnt)) {
			break;
		    }
		    cnt++;
		}
		testSuiteParams.put(attributes.getValue(i) + "_" + cnt, 
				    currentHttpTestClientConfig);
	    } else if (attributeName.equals(HttpTestClientConfig.TEST_ID_ATTR)) {
		currentHttpTestClientConfig.id = attributes.getValue(i);
	    }
	}
    }

    /**
     * This function is invoked when an element of type 'TestParams' is 
     * encountered.  The list of possible 'TestParams' element attributes 
     * can be found in HttpTestClientConfig as static final Strings matching 
     * the pattern TEST_PARAMS_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setTestParamsAttributes(Attributes attributes) {
	for (int i = 0; i < attributes.getLength(); i++) {
	    String attributeName = attributes.getQName(i);
	    if (attributeName.equals(HttpTestClientConfig.TEST_PARAMS_NUM_THREADS_ATTR)) {
		int numThreads =
		    getIntegerAttribute(HttpTestClientConfig.TEST_PARAMS_NUM_THREADS_ATTR,
					attributes.getValue(i));
		if (numThreads >= 1) {
		    currentHttpTestClientConfig.numThreads = numThreads;
		}
	    } else if (attributeName.equals(HttpTestClientConfig.TEST_PARAMS_REQUESTS_PER_THREAD_ATTR)) {
		int requestsPerThread = 
		    getIntegerAttribute(HttpTestClientConfig.TEST_PARAMS_NUM_THREADS_ATTR,
					attributes.getValue(i));
		if (requestsPerThread >= 1) {
		    currentHttpTestClientConfig.requestsPerThread = requestsPerThread;
		}
	    } else if (attributeName.equals(HttpTestClientConfig.TEST_PARAMS_DELAY_BETWEEN_REQUESTS_ATTR)) {
		long delayBetweenRequests = 
		    getLongAttribute(HttpTestClientConfig.TEST_PARAMS_DELAY_BETWEEN_REQUESTS_ATTR,
					attributes.getValue(i));
		if (delayBetweenRequests > 0) {
		    currentHttpTestClientConfig.delayBetweenRequests = 
			delayBetweenRequests;
		}
	    } else if (attributeName.equals(HttpTestClientConfig.TEST_PARAMS_SESSION_TRACKING)) {
		currentHttpTestClientConfig.sessionTracking = attributes.getValue(i);
	    }
	}
    }

    /**
     * This function is invoked when an element of type 'Url' is encountered.
     * The list of possible 'Url' element attributes can be found in 
     * HttpTestClientConfig as static final Strings matching the pattern 
     * URL_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setUrlAttributes(Attributes attributes) {
	// Used to determine is the user specified a port number in the
	// configuration file so we can determine if we need to use defaults.
	boolean portSpecified = false;

	for (int i = 0; i < attributes.getLength(); i++) {
	    String attributeName = attributes.getQName(i);
	    if (attributeName.equals(HttpTestClientConfig.URL_PROTOCOL_ATTR)) {
		currentHttpTestClientConfig.protocol = attributes.getValue(i);
		if (!currentHttpTestClientConfig.protocol.equals(HttpTestClientConfig.DEFAULT_PROTOCOL) && portSpecified == false) {
		    // https specified with no port so use default.  http is
		    // initialized by default.
		    currentHttpTestClientConfig.port = 
			HttpTestClientConfig.DEFAULT_HTTPS_PORT;
		}
	    } else if (attributeName.equals(HttpTestClientConfig.URL_KEY_STORE_PASSWORD_ATTR)) {
		currentHttpTestClientConfig.keyStorePassword = attributes.getValue(i);
	    } else if (attributeName.equals(HttpTestClientConfig.URL_KEY_STORE_LOCATION_ATTR)) {
		currentHttpTestClientConfig.keyStoreLocation = attributes.getValue(i);
	    } else if (attributeName.equals(HttpTestClientConfig.URL_HOST_ATTR)) {
		currentHttpTestClientConfig.host = attributes.getValue(i);
	    } else if (attributeName.equals(HttpTestClientConfig.URL_PORT_ATTR)) {
		int port = 
		    getIntegerAttribute(HttpTestClientConfig.URL_PORT_ATTR,
					attributes.getValue(i));
		if (port >= 0 ) {
		    currentHttpTestClientConfig.port = port;
		    portSpecified = true;
		}
	    } else if (attributeName.equals(HttpTestClientConfig.URL_FILE_ATTR)) {
		currentHttpTestClientConfig.file = attributes.getValue(i);
	    } else if (attributeName.equals(HttpTestClientConfig.URL_REDIRECTED_TO_FILE_ATTR)) {
		currentHttpTestClientConfig.redirectedToFile = 
		    attributes.getValue(i);
	    }
	}
    }

    /**
     * This function is invoked when an element of type 'CgiParam' is 
     * encountered.  The list of possible 'CgiParam' element attributes 
     * can be found in HttpTestClientConfig as static final Strings matching 
     * the pattern CGI_PARAM_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setCgiParamAttributes(Attributes attributes) {
	if (attributes.getLength() != 2) {
	    return;
	}
	
	String attributeName = null;
	String attributeValue = null;	
	
	for (int i = 0; i < 2; i++) {
	    if ((attributes.getQName(i)).equals(HttpTestClientConfig.CGI_PARAM_NAME_ATTR)) {
		attributeName = attributes.getValue(i);
	    } else if ((attributes.getQName(i)).equals(HttpTestClientConfig.CGI_PARAM_VALUE_ATTR)) {
		attributeValue = attributes.getValue(i);
	    }
	}
	currentHttpTestClientConfig.cgiParams.put(attributeName, attributeValue);
    }

    /**
     * This function is invoked when an element of type 'CgiArg' is 
     * encountered.  The list of possible 'CgiArg' element attributes 
     * can be found in HttpTestClientConfig as static final Strings matching 
     * the pattern CGI_ARG_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setCgiArgAttributes(Attributes attributes) {
	if (attributes.getLength() != 2) {
	    return;
	}
	
	String attributeName = null;
	String attributeValue = null;	
	
	for (int i = 0; i < 2; i++) {
	    if ((attributes.getQName(i)).equals(HttpTestClientConfig.CGI_ARG_NAME_ATTR)) {
		attributeName = attributes.getValue(i);
	    } else if ((attributes.getQName(i)).equals(HttpTestClientConfig.CGI_ARG_VALUE_ATTR)) {
		attributeValue = attributes.getValue(i);
	    }
	}
	currentHttpTestClientConfig.cgiArgs.put(attributeName, attributeValue);
    }

    /**
     * This function is invoked when an element of type 'Header' is 
     * encountered.  The list of possible 'Header' element attributes 
     * can be found in HttpTestClientConfig as static final Strings matching 
     * the pattern HEADER_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setHeaderAttributes(Attributes attributes) {
	if (attributes.getLength() != 2) {
	    return;
	}
	
	String attributeName = null;
	String attributeValue = null;	
	
	for (int i = 0; i < 2; i++) {
	    if ((attributes.getQName(i)).equals(HttpTestClientConfig.HEADER_NAME_ATTR)) {
		attributeName = attributes.getValue(i);
	    } else if ((attributes.getQName(i)).equals(HttpTestClientConfig.HEADER_VALUE_ATTR)) {
		attributeValue = attributes.getValue(i);
	    }
	}
	currentHttpTestClientConfig.headers.put(attributeName, attributeValue);
    }

    /**
     * This function is invoked when an element of type 'Cookie' is 
     * encountered.  The list of possible 'Cookie' element attributes 
     * can be found in HttpTestClientConfig as static final Strings matching 
     * the pattern COOKIE_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setCookieAttributes(Attributes attributes) {
	if (attributes.getLength() != 2) {
	    return;
	}
	
	String attributeName = null;
	String attributeValue = null;	
	
	for (int i = 0; i < 2; i++) {
	    if ((attributes.getQName(i)).equals(HttpTestClientConfig.COOKIE_NAME_ATTR)) {
		attributeName = attributes.getValue(i);
	    } else if ((attributes.getQName(i)).equals(HttpTestClientConfig.COOKIE_VALUE_ATTR)) {
		attributeValue = attributes.getValue(i);
	    }
	}
	currentHttpTestClientConfig.cookies.put(attributeName, attributeValue);
    }
    
    /**
     * This function is invoked when an element of type 'FileUpload' is 
     * encountered.  The list of possible 'FileUpload' element attributes 
     * can be found in HttpTestClientConfig as static final Strings matching 
     * the pattern FILE_UPLOAD_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setFileUploadAttributes(Attributes attributes) {
	StringBuffer fileNameBuffer = new StringBuffer();
	for (int i = 0; i < attributes.getLength(); i++) {
	    String attributeName = attributes.getQName(i);
	    if (attributeName.equals(HttpTestClientConfig.FILE_UPLOAD_FILE_DIR_ATTR)) {
		fileNameBuffer.insert(0, attributes.getValue(i));
		if (fileNameBuffer.charAt(fileNameBuffer.length() - 1) !=
		    '/') {
		    fileNameBuffer.append('/');
		}
	    } else if (attributeName.equals(HttpTestClientConfig.FILE_UPLOAD_FILE_NAME_ATTR)) {
		fileNameBuffer.append(attributes.getValue(i));
	    }
	}
	currentHttpTestClientConfig.fileUploads.addElement(fileNameBuffer.toString());
    }

    /**
     * This is a convenience function to retrieve an int value from a String.
     * If a number format error is found, the error is printed to standard
     * error and -1 is returned.
     *
     * @param attributeName
     *        This string represents the name of the current attribute, this
     *        if only used for error reporting.
     * @param attributeValue
     *        The string represents the number that will be converted to 
     *        an int value.
     * @return int
     *         The converted String value or -1 if an error occurs.
     **/
    protected int getIntegerAttribute(String attributeName,
				      String attributeValue) {
	try {
	    return Integer.parseInt(attributeValue);
	} catch (NumberFormatException e) {
	    System.err.println("NumberFormatException reading attribute: " + 
			       attributeName + "=\"" + 
			       attributeValue + "\".  Using default.");
	    return -1;
	}
    }

    /**
     * This is a convenience function to retrieve a long value from a String.
     * If a number format error is found, the error is printed to standard
     * error and -1 is returned.
     *
     * @param attributeName
     *        This string represents the name of the current attribute, this
     *        if only used for error reporting.
     * @param attributeValue
     *        The string represents the number that will be converted to 
     *        a long value.
     * @return int
     *         The converted String value or -1 if an error occurs.
     **/
    protected long getLongAttribute(String attributeName,
				    String attributeValue) {
	try {
	    return Long.parseLong(attributeValue);
	} catch (NumberFormatException e) {
	    System.err.println("NumberFormatException reading attribute: " + 
			       attributeName + "=\"" + 
			       attributeValue + "\".  Using default.");
	    return (long)-1;
	}
    }
	
    /** 
     * Warning callback function.  Called by the SAX parser if a warning error
     * is encountered while parsing.  This implementation only prints a 
     * warning to standard error and returns.
     *
     * @see org.apache.xerces.parsers.SAXParser
     *
     * @param e
     *        The SAXParseException that was thrown to cause the warning.
     **/
    public void warning(SAXParseException e) {
        System.err.println("WARNING: " +
                           getLocationString(e)+": " +
                           e.getMessage());
    }

    /** 
     * Error callback function.  Called by the SAX parser if a non-fatal error
     * is encountered while parsing.  This function reports the error to the
     * global class variables 'parseErrorFound' and 'parseErrorMessage' to 
     * cause an exception to be thrown from the 'parse()' function.  This
     * parser implementation considers all errors fatal except warnings.  This
     * is done because a test case can not run with an incomplete set of
     * configuration parameters.
     *
     * @see org.apache.xerces.parsers.SAXParser
     *
     * @param e
     *        The SAXParseException that was thrown to cause the error.
     **/
    public void error(SAXParseException e) {
	parseErrorFound = true;
        parseErrorMessage = "ERROR: " +
	    getLocationString(e)+": " +
	    e.getMessage();
    }

    /** 
     * Fatal Error callback function.  Called by the SAX parser if a fatal 
     * error is encountered while parsing.  This function reports the error 
     * to the global class variables 'parseErrorFound' and 'parseErrorMessage'
     * to cause an exception to be thrown from the 'parse()' function.  This
     * parser implementation considers all errors fatal except warnings.  This
     * is done because a test case can not run with an incomplete set of
     * configuration parameters.
     *
     * @see org.apache.xerces.parsers.SAXParser
     *
     * @param e
     *        The SAXParseException that was thrown to cause the fatal error.
     * @exception SAXException
     *            This is not used in this implementation but is required for
     *            the proper method signiture.  The global error flag and
     *            message buffer is used instead.
     **/
    public void fatalError(SAXParseException e) 
	throws SAXException {
	parseErrorFound = true;
        parseErrorMessage = "FATAL ERROR: " +
	    getLocationString(e)+": " +
	    e.getMessage();
    }

    /** 
     * Returns a string representing the line:column that the exception was
     * encountered.  This is only used from error reporting purposes.
     *
     * @param e
     *        The SAXParseException that was thrown to cause the error.
     * @return String
     *         The line:column that the error occured.
     **/
    private String getLocationString(SAXParseException e) {
        StringBuffer stringBuffer = new StringBuffer();

        String systemId = e.getSystemId();
        if (systemId != null) {
            int index = systemId.lastIndexOf('/');
            if (index != -1) 
                systemId = systemId.substring(index + 1);
            stringBuffer.append(systemId);
        }
        stringBuffer.append(':');
        stringBuffer.append(e.getLineNumber());
        stringBuffer.append(':');
        stringBuffer.append(e.getColumnNumber());

        return stringBuffer.toString();
    }
}
