/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpServerTestUtils.java,v 1.5.2.1.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.common;

import com.lutris.util.Base64Encoder;

import java.io.File;
import java.io.FileInputStream;
import java.security.MessageDigest;

/**
 * This class provides static data and methods that are shared by both the 
 * client and server sides of this test.  See below for details.
 *
 * @author Mike Ward
 **/
public class HttpServerTestUtils {
    /**
     * The string designating a session cookie.
     **/
    public static final String HTTP_COOKIE_SESSION_KEY = "JSESSIONID";

    /**
     * The string designating a session CGI parameter.
     **/
    public static final String CGI_PARAM_SESSION_KEY = "jsessionid";

    /**
     * The string representing an http GET request method.
     **/
    public static final String GET_REQUEST_TAG = "GET";

    /**
     * The string representing an http POST request method.
     **/
    public static final String POST_REQUEST_TAG = "POST";

    /**
     * This string is used by the server side components to indicate the
     * type of post method that was used.  Requests submitted using this
     * header can hava all CGI arguments found in the same manner as a GET
     * request (from the servlet's getParameter*() methods).  Otherwise,
     * the request consists of a multi-part mime and must be decoded.
     **/
    public static final String POST_URL_ENCODED_CONTENT_TYPE = 
	"application/x-www-form-urlencoded";

    /**
     * This is the name used to designate a file upload within a multi-
     * part mime message.
     **/
    public static final String POST_FILE_UPLOAD_PREFIX = "fileUpload";

    /**
     * This is an array of headers commonly set automatically by HTTPClient
     * when issueing request.
     * @see HTTPClient.HTTPConnection
     **/
    public static final String[] defaulHeaderArray = {"Connection", 
						      "Content-type",
						      "Content-length",
						      "Accept-Encoding", 
						      "Cookie", 
						      "Cookie2", 
						      "User-Agent", 
						      "Host", 
						      "TE"};
    
    /**
     * This fuction will loop through the 'defaultHeaderArray' and return
     * true if the supplied header is considered 'default' and false
     * otherwise.
     * 
     * @param headerName
     *        The header name to check against defaults.
     * @return boolean
     *         Returns true if header is found, false otherwise.
     **/
    public static boolean isDefaultHeader(String headerName) {
	for (int i = 0; i < defaulHeaderArray.length; i++) {
	    if (defaulHeaderArray[i].equals(headerName)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * This function will calculate an MD5 hashkey for the given file name.
     * The hashkey is Base64 encoded for safe transport over the connection.
     *
     * @see com.lutris.util.Base64Encoder
     *
     * @param fileName
     *        The file name to compute hashkey for.
     * @return String
     *         The Base64 encoded hashkey for this file.
     * @exception Exception
     *            Thrown if an error ocurs while reading file and/or computing
     *            the file's hashkey value.
     **/
    public static String calculateFileHash(String fileName) 
	throws Exception {
	try {
	    File file = new File(fileName);
	    if (!file.exists() ||
		!file.canRead()) {
		return null;
	    }
	    
	    MessageDigest messageDigest = MessageDigest.getInstance("MD5");

	    FileInputStream input = new FileInputStream(file);
	    byte[] bytes = new byte[1024];
	    int bytesRead = input.read(bytes);
	    while (bytesRead > 0) {
		messageDigest.update(bytes, 0, bytesRead);
		bytesRead = input.read(bytes);
	    }
	    input.close();

	    return Base64Encoder.toBase64SessionKeyString(messageDigest.digest());
	} catch (Exception e) {
	    throw e;
	}
    }
}
