/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BasicTestServlet.java,v 1.2.2.1.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.server.servlet;

import org.enhydra.httpServerTest.common.HttpResponseObject;
import org.enhydra.httpServerTest.common.HttpServerTestUtils;
import org.enhydra.httpServerTest.server.common.FileUploadMangler;

import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.net.URLEncoder;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import HTTPClient.NVPair;
import HTTPClient.Codecs;

import java.util.Enumeration;
import java.util.StringTokenizer;
import java.net.URLEncoder;
import java.io.DataInputStream;
import java.io.File;

/**
 * This class is a generic server side test servlet that client side test 
 * threads talk to.  This class respondes to every request coming in by
 * constructing an HttpResponseObject and returning it in the response as
 * an XML stream.  The response object returned contains information about
 * all data received via the request.  This presentation object properly
 * handles GET and POST method request and multi-part mime encodings.
 *
 * @see org.enhydra.httpServerTest.common.HttpResponseObject
 *
 * @author Mike Ward
 **/
public class BasicTestServlet extends HttpServlet { 

    /**
     * This holds the directory, unique to this request, that any uploaded
     * files are stored.
     **/
    protected File fileCacheDir = null;

    /**
     * This holds a tempoary base directory used to cache uploaded files.
     **/
    protected static final String TEMPORARY_BASE_DIR = File.separator + 
	"tmp" + File.separator + "BasicTestServletFileUploadCache" + 
	File.separator;
    
    /**
     * This is a static counter used by the presentation objects to ensure
     * a unique sub-directory under the specified temporary base directory.
     * The directories created using this count variable are used to cache
     * files uploaded from the client.
     **/
    protected static Long fileCacheDirCnt = new Long(0);

    /**
     * This is a static function to retrieve the next sequence of the conter
     * variable held in this class.  This function synchronizes on the 
     * fileCacheDirCnt and increments this count each time this function is
     * called and then returns the new count.
     *
     * @return long
     *         The current, incremented counter.
     **/
    protected static long getFileCacheDirCnt() {
	synchronized (fileCacheDirCnt) { 
	    fileCacheDirCnt = new Long(fileCacheDirCnt.longValue() + (long)1);
	    return fileCacheDirCnt.longValue();
	}
    }

    /**
     * Overide the service method of HttpServelet.  This method runs through
     * the remainder of the functions held in this class to initialize the
     * response object based on information pulled from the request.  Once
     * the response object is populated, this method calls toString() on it
     * and returns the generated XML structure to the client.
     *
     * @param request 
     *        The request that is sent by a client.
     * @param response 
     *        The response that is sent back to the client.
     **/
    public void service(HttpServletRequest request, 
			HttpServletResponse response) 
	throws ServletException, IOException {

	HttpResponseObject httpResponseObject = new HttpResponseObject();

	populateRequestInfo(httpResponseObject, request);
	if (request.getMethod().equals(HttpServerTestUtils.GET_REQUEST_TAG) ||
	    request.getContentType().equals(HttpServerTestUtils.POST_URL_ENCODED_CONTENT_TYPE)) {
	    // Standard request, url-encoded
	    populateCgiArgsFromGet(httpResponseObject, request);
  	} else {
	    // Multi-part mime request, file uploads expected
  	    populateCgiArgsFromPost(httpResponseObject, request);
  	    populateFileUploadFromPost(httpResponseObject);
	}
	populateCgiParams(httpResponseObject, request);
	populateHeaders(httpResponseObject, request);
	populateCookies(httpResponseObject, request);
	populateSessionInfo(httpResponseObject, request, response);
	
	response.setContentType("text/html");
	byte[] bytes = (httpResponseObject.toString()).getBytes();
	ServletOutputStream outputStream = response.getOutputStream();
	outputStream.write(bytes);
	outputStream.flush();
	return;
    }

    /**
     * This function runs through the HttpServletRequest API and populates
     * the response object with the retrieved information.
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpServletRequest object to pull data from.
     **/
    protected void populateRequestInfo(HttpResponseObject httpResponseObject,
				       HttpServletRequest request) {

	httpResponseObject.authType = request.getAuthType();
	httpResponseObject.method = request.getMethod();
	httpResponseObject.pathInfo = request.getPathInfo();
	httpResponseObject.pathTranslated = request.getPathTranslated();
	String queryString = request.getQueryString();
	if (queryString != null) {
	    httpResponseObject.queryString = 
		URLEncoder.encode(queryString);
	} else {
	    httpResponseObject.queryString = null;
	}
	httpResponseObject.remoteUser = request.getRemoteUser();
	httpResponseObject.requestURI = request.getRequestURI();
	httpResponseObject.servletPath = request.getServletPath();
	if (httpResponseObject.servletPath.length() == 0) {
	    httpResponseObject.servletPath = "/";
	}
    }

    /**
     * This function will populate the response object with all CGI parameters
     * submitted with the request.
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpServletRequest object to pull data from.
     **/
    protected void populateCgiParams(HttpResponseObject httpResponseObject,
				     HttpServletRequest request) {
	StringTokenizer paramTokens = 
	    new StringTokenizer(request.getPathInfo(), ";", false);

	while (paramTokens.hasMoreTokens()) {
	    String currentToken = paramTokens.nextToken();
	    StringTokenizer currentNameValuePair = 
		new StringTokenizer(currentToken, "=", false);
	    if (currentNameValuePair.countTokens() == 2) {
		String paramName = currentNameValuePair.nextToken();
		String paramValue = currentNameValuePair.nextToken();
		httpResponseObject.cgiParams.put(paramName, paramValue);
	    }
	}
    }

    /**
     * This function will populate the response object with all CGI arguments
     * submitted with the request.  This method assumes a stardard GET or
     * url-encoded POST request (ie. no multi-part mime).
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpServletRequest object to pull data from.
     **/
    protected void populateCgiArgsFromGet(HttpResponseObject httpResponseObject,
					  HttpServletRequest request) {

	Enumeration cgiArgs = request.getParameterNames();
	while (cgiArgs.hasMoreElements()) {
	    String cgiArgName = (String)cgiArgs.nextElement();
	    String cgiArgValue = 
		request.getParameter(cgiArgName);
	    httpResponseObject.cgiArgs.put(cgiArgName, cgiArgValue);
	}
    }

    /**
     * This function will populate the response object with all http headers
     * submitted with the request.  This function will not include any
     * 'default' headers applied by HTTPConnection.
     *
     * @see HTTPClient.HTTPConnection
     * @see org.enhydra.httpServerTest.common.HttpServerTestUtils
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpServletRequest object to pull data from.
     **/
    protected void populateHeaders(HttpResponseObject httpResponseObject,
				   HttpServletRequest request) {
	Enumeration headers = request.getHeaderNames();
	while (headers.hasMoreElements()) {
	    String headerName = (String)headers.nextElement();
	    if (!HttpServerTestUtils.isDefaultHeader(headerName)) {
		String headerValue = 
		    request.getHeader(headerName);
		httpResponseObject.headers.put(headerName, headerValue);
	    }
	}
    }

    /**
     * This function will populate the response object with all http cookies
     * submitted with the request.
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpServletRequest object to pull data from.
     **/
    protected void populateCookies(HttpResponseObject httpResponseObject,
				   HttpServletRequest request) {
	Cookie[] cookies = request.getCookies();
	for (int i = 0; i < cookies.length; i++) {
	    httpResponseObject.cookies.put(cookies[i].getName(),
					   cookies[i].getValue());
	}
    }

    /**
     * This function will populate the response object with all CGI arguments
     * submitted with the request.  This method assumes a multi-part mime
     * encoding from a POST request.
     *
     * @see HTTPClient.Codecs
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpServletRequest object to pull data from.
     **/
    protected void populateCgiArgsFromPost(HttpResponseObject httpResponseObject,
					   HttpServletRequest request) {
	try {
	    byte[] body = new byte[request.getContentLength()];
	    new DataInputStream(request.getInputStream()).readFully(body);
	    String contentTypeAndBoundary = request.getContentType();
	
	    initFileCacheDir();

	    NVPair[] cgiArgs = 
		Codecs.mpFormDataDecode(body, 
					contentTypeAndBoundary,
					fileCacheDir.toString(),
					new FileUploadMangler());
	    for (int i = 0; i < cgiArgs.length; i++) {
		if (!cgiArgs[i].getName().startsWith(HttpServerTestUtils.POST_FILE_UPLOAD_PREFIX)) {
		    httpResponseObject.cgiArgs.put(cgiArgs[i].getName(),
						   cgiArgs[i].getValue());
		}
	    }
	} catch (Exception e) {
	}
    }

    /**
     * This function will continually loop until a unique temporary sub-
     * directory is found.  This is accomplished by using the static data 
     * members and methods held in this class.  Once a unique
     * path is found, the directories are created and the global fileCacheDir
     * is initialized.
     **/
    protected void initFileCacheDir() {
	String fileCacheDirString = BasicTestServlet.TEMPORARY_BASE_DIR + 
 	    BasicTestServlet.getFileCacheDirCnt() + File.separator;
	fileCacheDir = new File(fileCacheDirString);
	    
	while (true) {
	    if (fileCacheDir.isDirectory()) {
		fileCacheDirString = BasicTestServlet.TEMPORARY_BASE_DIR + 
		    BasicTestServlet.getFileCacheDirCnt() + File.separator;
		fileCacheDir = new File(fileCacheDirString);
	    } else {
		break;
	    } 
	}
	
	fileCacheDir.mkdirs();
    }

    /**
     * This function will loop through all files found under the unique 
     * directory created by initFileCacheDir() and compute an MD5 hashkey
     * for each individual file.  The response object is then populated with
     * each hashkey.
     *
     * @param httpResponseObject
     *        The response object to populate with file upload data.
     **/
    protected void populateFileUploadFromPost(HttpResponseObject httpResponseObject) {
	String[] uploadedFiles = fileCacheDir.list();
	for (int i = 0; i < uploadedFiles.length; i++) {
	    String currentFile = fileCacheDir.toString() + 
		File.separator + uploadedFiles[i];
	    String hashKey = null;
	    try {
		hashKey = 
		    HttpServerTestUtils.calculateFileHash(currentFile);
	    } catch (Exception e) {
	    }
	    if (!(hashKey == null)) {
		httpResponseObject.md5HashKey.addElement(hashKey);
	    }
	}
	cleanupFiles();
    }

    /**
     * This function is called after all file MD5 hashkeys have been computed.
     * This function will recursively delete all uploaded files and temporary
     * directories created while caching uploaded files.
     **/
    public void cleanupFiles() {
	String[] uploadedFiles = fileCacheDir.list();
	for (int i = 0; i < uploadedFiles.length; i++) {
	    String currentFile = fileCacheDir.toString() + 
		File.separator + uploadedFiles[i];
	    File tmpFile = new File(currentFile);
	    tmpFile.delete();
	}
	fileCacheDir.delete();
    }

    /**
     * This function will populate the response object with all session
     * information submitted with the request.
     *
     * FIX ME: This function does not behave properly.  All initial requests
     *         are initialized as session information comming from a session
     *         cookie.
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpServletRequest object to pull data from.
     **/
    protected void populateSessionInfo(HttpResponseObject httpResponseObject,
				       HttpServletRequest request,
				       HttpServletResponse response) {
	    
	httpResponseObject.sessionIdFromCookie = 
	    request.isRequestedSessionIdFromCookie();
	httpResponseObject.sessionIdFromURL = 
	    request.isRequestedSessionIdFromUrl();

	HttpSession session = request.getSession();
	Cookie sessionCookie = 
	    new Cookie(HttpServerTestUtils.HTTP_COOKIE_SESSION_KEY,
		       session.getId());
	response.addCookie(sessionCookie);

	if (httpResponseObject.sessionIdFromCookie ||
	    httpResponseObject.sessionIdFromURL) {
	    httpResponseObject.sessionIdInvalid = false;
	    httpResponseObject.session = session.getId();
	} else {
	    httpResponseObject.sessionIdInvalid = true;
	    httpResponseObject.session = HttpResponseObject.DEFAULT_SESSION;
	}
    }
}

