/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: JoltToJava.java,v 1.8.12.1 2000/10/19 17:59:08 jasona Exp $
 */





package com.lutris.appserver.server.jolt.joltc;

import com.lutris.appserver.server.jolt.parser.*;
import java.io.*;
import java.util.*;

/**
 * Convert a Jolt source file to a object containing the class.
 * All methods of this class are static.
 *
 * @version	$Revision: 1.8.12.1 $
 * @author	Mark Diekhans
 * @since	2.0
 */
class JoltToJava {
    /*
     * Class names to use for Presentation, Page and pageArgs
     */
    private static String PRESENTATION_BASE_CLASS =
        "JoltPage";
    private static String PAGE_ACCESS =
        "public";

    /*
     * Standard imports to add.
     */
    private static String PRESENTATION_IMPORT =
        "import com.lutris.appserver.server.httpPresentation.*;";
    private static String JOLT_PRESENTATION_IMPORT =
        "import com.lutris.appserver.server.jolt.joltpo.*;";
    private static String KEYWORD_VALUE_IMPORT =
        "import com.lutris.util.*;";

    /*
     * Comment strings to seperate sections of the generated code.
     */
    private static final String sectionSeparator;
    private static final String subSectionSeparator;

    /**
     * Class constructor.
     */
    static {
        StringBuffer strTmp;
        
        /*
         * Generate comment separators that will be used to make generated
         * code more readable.
         */
        strTmp = new StringBuffer ("/*");
        for (int cnt = 0; cnt < 70; cnt++) {
            strTmp.append ('=');
        }
        strTmp.append ("*/");
        sectionSeparator = strTmp.toString ();
        
        strTmp = new StringBuffer ("/*");
        for (int cnt = 0; cnt < 70; cnt++) {
            strTmp.append ('-');
        }
        strTmp.append ("*/");
        subSectionSeparator = strTmp.toString ();
    }

    /*
     * Prevent instanciation.
     */
    private JoltToJava() {
    }
    
    /**
     * Convert a Jolt HTML file to a Java class.
     *
     * @param pagePackage Package name of the class.
     * @param pageClass Name of class for the page.
     * @param srcFile Jolt HTML source file.
     * @return Object containing all of the Java source and a source line
     *  map information.
     * @exception java.io.FileNotFoundException If the source file or
     *  an include file was not found.
     * @exception java.io.IOException If an error occured reading the source
     *  or an included file was not found.
     * @exception JoltCompileException
     */
    protected static FileLines compile(String pagePackage,
                                       String pageClass,
                                       File srcFile,
                                       File javaFile)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {

        JoltCodeGen code = new JoltCodeGen(srcFile);
        
        FileLines javaSrc = new FileLines(javaFile);
        javaSrc.append("package " + pagePackage + ";\n\n" +
                       PRESENTATION_IMPORT + "\n" +
                       JOLT_PRESENTATION_IMPORT + "\n" +
                       KEYWORD_VALUE_IMPORT + "\n\n");
        
        if (code.importDefs.size() > 0) {
            javaSrc.append(sectionSeparator);
            javaSrc.append(code.importDefs);
        }

        javaSrc.append(PAGE_ACCESS + " class " + pageClass +
                       " extends " + PRESENTATION_BASE_CLASS + " {");

        if (code.htmlStrDefs.size () > 0) {
            javaSrc.append(subSectionSeparator);
            javaSrc.append(code.htmlStrDefs);
        }
            
        if (code.classDefs.size() > 0) {
            javaSrc.append(sectionSeparator);
            javaSrc.append(code.classDefs);
        }
        javaSrc.append(sectionSeparator);
        javaSrc.append("protected void runPage()\n" +
                       "      throws Exception {");
            
        javaSrc.incrIndent();
        javaSrc.append("try {");

        javaSrc.incrIndent();
        javaSrc.append(code.runMethod);
            
        javaSrc.decrIndent();
        javaSrc.append("}");
        if (code.pageCatches.size() > 0) {
            javaSrc.append(code.pageCatches);
        } 
            
        javaSrc.append(" finally {");
        javaSrc.append(code.pageFinally);
        javaSrc.append("}");
        javaSrc.decrIndent();
        javaSrc.append("}");
        javaSrc.append("}");

        return javaSrc;
    }
}  
