/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: JoltRequest.java,v 1.18.12.1 2000/10/19 17:59:08 jasona Exp $
 */




package com.lutris.appserver.server.jolt.joltpo;

import java.util.*;
import java.lang.reflect.Array;
import com.lutris.util.*;
import javax.servlet.http.Cookie;
import com.lutris.appserver.server.httpPresentation.*;


/**
 * HTTP arguments, cookies and other headers from the request.
 */
public class JoltRequest {
    /**
     * HTTP request object used to get headers, etc.
     */
    private HttpPresentationRequest httpRequest;

    /**
     * Hash table of named query strings.
     */
    Hashtable queryTable;

    /**
     * Construct a request object.
     * @param request HTTP request information, headers and input stream.
     * @param queryTable Hash table of named query strings.
     */
    protected JoltRequest (HttpPresentationRequest httpRequest,
                           Hashtable queryTable) {
        this.httpRequest = httpRequest;
        this.queryTable = queryTable;
    }

    /**
     * Returns the size of the request entity data, or -1 if not known. Same
     * as the CGI variable CONTENT_LENGTH.
     */
    public int getContentLength() throws HttpPresentationException {
        return httpRequest.getContentLength();
    }

    /*
     * Returns the Internet Media Type of the request entity data, or null if
     * not known. Same as the CGI variable CONTENT_TYPE.
     */
    public String getContentType() throws HttpPresentationException {
        return httpRequest.getContentType();
    }

    /**
     * Returns the IP address of the agent that sent the request. Same as the
     * CGI variable REMOTE_ADDR.
     */
    public String getRemoteAddr() throws HttpPresentationException {
        return httpRequest.getRemoteAddr();
    }

    /**
     * Returns the fully qualified host name of the agent that sent the
     * request. Same as the CGI variable REMOTE_HOST.
     */
    public String getRemoteHost() throws HttpPresentationException {
        return httpRequest.getRemoteHost();
    }

    /**
     * Applies alias rules to the specified virtual path and returns the
     * corresponding real path, or null if the translation can not be
     * performed for any reason.  This resolves the path using the virtual
     * docroot, if virtual hosting is enabled, and with the default docroot
     * otherwise.  Calling this method with the string "/" as an argument
     * returns the document root.
     *
     * @param path the virtual path to be translated to a real path
     */
/*
    String getRealPath(String path) throws HttpPresentationException {
        return httpRequest.getRealPath(path);
    }
*/
    /**
     * Returns a string containing the lone value of the specified query
     * parameter, or null if the parameter does not exist. Presentation
     * writers should use this method only when they are sure that there is
     * only one value for the parameter.  If the parameter has (or could have)
     * multiple values, then use getParameterValues. If a multiple valued
     * parameter name is passed as an argument, the return value is
     * implementation dependent.  This function works for both GET and POST
     * requests (Jolt only, not in like-named HttpPresentationRequest method).
     *
     * @param name the name of the parameter whose value is required.
     */
    public String getParameter(String name) throws HttpPresentationException {
        Object value = queryTable.get(name);
        if (value instanceof String[]) {
            value = ((String[])value)[0];
        }
        return (String) value;
    }

    /**
     * Returns the values of the specified query parameter for the request as
     * an array of strings, or a 0 length array if the named parameter does
     * not exist. This function works for both GET and POST requests (Jolt
     * only, not in like-named HttpPresentationRequest method).
     *
     * @param name the name of the parameter whose value is required.
     */
    public String[] getParameterValues(String name) throws HttpPresentationException {
        Object values = queryTable.get(name);
	if (values == null) {
	    return new String[0];
	}
        if (values instanceof String) {
            String[] valArray = new String[1];
            valArray[0] = (String)values;
            return valArray;
        }
        return (String[])values;
    }

    /**
     * Returns the values of the specified query parameter for the request as
     * an array of strings, or a 0 length array if the named parameter does
     * not exist. This function works for both GET and POST requests (Jolt
     * only, not in like-named HttpPresentationRequest method).
     */
    public Enumeration getParameterNames() throws HttpPresentationException {
        return queryTable.keys();
    }

    /**
     * Returns the method with which the request was made. The returned
     * value can be "GET", "HEAD", "POST", or an extension method. Same
     * as the CGI variable REQUEST_METHOD.
     */
    public String getMethod() throws HttpPresentationException {
        return httpRequest.getMethod();
    }

    /**
     * Returns the request URI as a URL object.
     */
    public String getRequestURI() throws HttpPresentationException {
        return httpRequest.getRequestURI();
    }

    /**
     * Returns the part of the request URI that refers to the presentation
     * object being invoked. Analogous to the CGI variable SCRIPT_NAME.
     *
     * @deprecated This method was named in a confusing manner; it
     *  returns the application, not presentation object path. Use
     *  <A HREF="getApplicationPath.html">getApplicationPath()</A>.
     */
    public String getPresentationPath() throws HttpPresentationException {
        return httpRequest.getPresentationPath();
    }

    /**
     * Returns the part of the request URI that refers to the presentation
     * object being invoked.
     */
    public String getPresentationObjectPath() throws HttpPresentationException {
        return httpRequest.getPresentationObjectPath();
    }

    /*
     * Returns the part of the request URI that refers to the application.
     * Analogous to the CGI variable SCRIPT_NAME.
     */
    public String getApplicationPath() throws HttpPresentationException {
        return httpRequest.getApplicationPath();
    }

    /**
     * Returns optional extra path information following the presentation
     * path, but immediately preceding the query string. Returns null if
     * not specified. Same as the CGI variable PATH_INFO.
     */
    public String getPathInfo() throws HttpPresentationException {
        return httpRequest.getPathInfo();
    }

    /**
     * Returns extra path information translated to a real path. Returns
     * null if no extra path information specified. Same as the CGI variable
     * PATH_TRANSLATED.  
     */
    public String getPathTranslated() throws HttpPresentationException {
        return httpRequest.getPathTranslated();
    }

    /**
     * Returns the query string part of the presentation URI, or null if none.
     * Same as the CGI variable QUERY_STRING.
     */
    public String getQueryString() throws HttpPresentationException {
        return httpRequest.getQueryString();
    }

    /**
     * Returns the name of the user making this request, or null if not
     * known.  The user name is set with HTTP authentication.  Whether
     * the user name will continue to be sent with each subsequent
     * communication is browser-dependent.  Same as the CGI variable
     * REMOTE_USER.
     */
    public String getRemoteUser() throws HttpPresentationException {
        return httpRequest.getRemoteUser();
    }

    /**
     * Returns the authentication scheme of the request, or null if none.
     * Same as the CGI variable AUTH_TYPE.
     */
    public String getAuthType() throws HttpPresentationException {
        return httpRequest.getAuthType();
    }

    /**
     * Gets the array of cookies found in this request.
     *
     * @return The array of cookies found in this request.
     */
    public Cookie[] getCookies() throws HttpPresentationException {
        return httpRequest.getCookies();
    }

    /**
     * Returns the value of a header field, or null if not known.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public String getHeader(String name) throws HttpPresentationException {
        return httpRequest.getHeader(name);
    }

    /**
     * Returns the value of an integer header field, or -1 if not found.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public int getIntHeader(String name) throws HttpPresentationException {
        return httpRequest.getIntHeader(name);
    }

    /**
     * Returns the value of a date header field, or -1 if not found.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public long getDateHeader(String name) throws HttpPresentationException {
        return httpRequest.getDateHeader(name);
    }

    /**
     * Returns an enumeration of strings representing the header names
     * for this request. Some server implementations do not allow headers
     * to be accessed in this way, in which case this method will return null.
     */
    public Enumeration getHeaderNames() throws HttpPresentationException {
        return httpRequest.getHeaderNames();
    }

    /**
     * Get the URI path for a file in the application.  This converts a path
     * to the file part of the URL.  It adds in the reference to application
     * servlet.
     *
     * @param file File with in the application.  Currently this must
     *  be a path relative to the presentation prefix.
     * @return The file path portion of the URL, starting with
     * a <CODE>/</CODE>.
     */
    public String getAppFileURIPath(String file)
	throws HttpPresentationException {
        return httpRequest.getAppFileURIPath(file);
    }
}
