/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: JoltResponse.java,v 1.29.12.1 2000/10/19 17:59:08 jasona Exp $
 */





package com.lutris.appserver.server.jolt.joltpo;

import java.util.*;
import java.io.*;
import java.text.*;
import com.lutris.util.*;
import com.lutris.http.HttpStatus;
import javax.servlet.http.Cookie;
import com.lutris.appserver.server.httpPresentation.*;

/**
 * Cookies and other headers to be passed back to the client.
 */
public class JoltResponse {
    /**
     * Object for formatting dates.
     */
    private static final SimpleDateFormat dateFormatter =
        new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz");

    /**
     * HTTP response object used to return results.
     */
    private HttpPresentationResponse httpResponse;

    /**
     * Buffered response headers.
     */
    private Hashtable headers = null;  // Lazy allocation

    /**
     * Buffered cookies.
     */
    private Vector cookies = null;  // Lazy allocation

    /**
     * Flag indicating if headers have been written and output has begun.
     */
    private boolean startedOutput = false;

    /*
     * Class constructor.
     */
    static {
	dateFormatter.setTimeZone(TimeZone.getTimeZone("GMT"));
    }


    /**
     * Construct a response object.
     * @param httpResponse HTTP response object used to return the response.
     */
    protected JoltResponse (HttpPresentationResponse httpResponse) {
        this.httpResponse = httpResponse;
    }

    /**
     * Verify that output has not begun before setting a header.  If it has,
     * generate an error.
     *
     * @exception HttpPresentationException If output has already begun.
     */
    private void setCheck() throws HttpPresentationException {
        if (startedOutput) {
            throw new HttpPresentationException("output has begun, headers can't be modified");
        }
    }
     

    /**
     * Adds the specified cookie to the response. It can be called multiple
     * times to set more than one cookie.
     * @param cookie - The Cookie to return to the client.
     */
    public void addCookie(Cookie cookie) throws HttpPresentationException {
        setCheck();
        if (cookies == null) {
            cookies = new Vector();
        }
        cookies.addElement(cookie);
    }

    /**
     * Returns true if the response message header has a field with
     * the specified name.
     * @param name the header field name
     */
    public boolean containsHeader(String name) throws HttpPresentationException {
        return (headers != null) && !headers.contains(name.toLowerCase());
    }
    
    /**
     * Adds a field to the response header with a given name and
     * value.  If the field had already been set, the new value
     * overwrites the previous one.  The containsHeader method can be
     * used to test for the presence of a header before setting its
     * value.
     * 
     * @param name the header field name
     * @param value the header field value
     */
    public void setHeader(String name, String value) throws HttpPresentationException {
        setCheck();
        if (headers == null) {
            headers = new Hashtable();
        }
        headers.put(name.toLowerCase(), value);
    }

    /**
     *
     * Adds a field to the response header with a given name and
     * integer value.  If the field had already been set, the new
     * value overwrites the previous one.  The containsHeader method
     * can be used to test for the presence of a header before setting
     * its value.
     *
     * @param name the header field name
     * @param value the header field integer value */
    public void setIntHeader(String name, int value) throws HttpPresentationException {
        setCheck();
        if (headers == null) {
            headers = new Hashtable();
        }
        headers.put(name.toLowerCase(), Integer.toString(value));
    }

    /**
     * 
     * Adds a field to the response header with a given name and
     * date-valued field.  The date is specified in terms of
     * milliseconds since the epoch.  If the date field had already
     * been set, the new value overwrites the previous one.  The
     * containsHeader method can be used to test for the presence of a
     * header before setting its value.
     * 
     * @param name the header field name
     * @param value the header field date value
     */
    public void setDateHeader(String name, long date) throws HttpPresentationException {
        // Date is formatted here so we can keep one hash table of headers.
        setCheck();
        if (headers == null) {
            headers = new Hashtable();
        }
        headers.put(name.toLowerCase(), dateFormatter.format(new Date(date)));
    }

    /**
     * Set the output character encoding.
     *
     * @param enc character encoding 
     */
    public void setEncoding(String enc) {
        httpResponse.setEncoding(enc);
    }
    
    /**
     * Get the output character encoding.
     *
     * @return the current encoding 
     */
    public String getEncoding() {
        return httpResponse.getEncoding();
    }
    
    /**
     * Write buffered headers and cookies to the HTTP client.
     */
    protected void writeHeaders() throws HttpPresentationException {
        boolean setContentType = false;
        boolean setExpires = false;
        if (headers != null) {
            for (Enumeration names = headers.keys(); names.hasMoreElements();) {
                String name = (String)names.nextElement();
                if (name.equalsIgnoreCase("Content-Type")) {
                    setContentType = true;
                }
                if (name.equalsIgnoreCase("Expires")) {
                    setExpires = true;
                }
                httpResponse.setHeader(name, (String)headers.get(name));
            }
        }
        if (cookies != null) {
            for (int idx = 0; idx < cookies.size(); idx++) {
                httpResponse.addCookie((Cookie)cookies.elementAt(idx));
            }
        }

        /*
         * These are not strictly necessary, but let's make sure the browser
	 * knows that this is HTML and that the content normally should not
	 * be cached because it is dynamic.
         */
        if (!setContentType) {
	    // Choose a date way in the past, so that page immediately expires.
            httpResponse.setContentType("text/html");
        }
        if (!setExpires) {
	    // Clients should not cache dynamic pages.

            /*
             * Don't send the Netscape 'Pragma: no-cache' header
             * out as it seemed to contribute to the double refresh problem.
             * Soheil: 8/13/98.
             * httpResponse.setHeader("Pragma", "no-cache");
             */

            httpResponse.setHeader("Cache-Control", "no-cache");
            httpResponse.setHeader("Expires", "0");
        }
        startedOutput = true;
    }

    /**
     * Write a string to the HTTP client.  Set headers if this is
     * the first write.  The output is flushed.
     *
     * @param str The string to write.
     */
    protected void write(String str) throws HttpPresentationException {
        if (!startedOutput) {
            writeHeaders();
        }
        try {
            HttpPresentationOutputStream out = httpResponse.getOutputStream();
            out.print(str);
            out.flush();
        } catch (IOException except) {
            // Want to throw a HttpPresentationIOException to record the fact that
            // the client disconnected, but want to maintain the HttpPresentationException
            // API.
            throw new HttpPresentationException(new HttpPresentationIOException(except));
        }
    }
       
}


