/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: SimComms.java,v 1.14.4.1 2000/10/19 17:59:08 jasona Exp $
 */





package	lutris.jolt.joltsim;
import lutris.jolt.presentation.*;
import lutris.keywordValue.*;
import lutris.http.*;
import java.io.*;
import java.util.*;

/**
 * Simulation comms objects.  Reads HTTP headers from a file and
 * can print out response.  Currently does not handle streaming output.
 */
public class SimComms extends Comms {
    private HttpPageArgs httpPageArgs;
    private String scriptName;
    private int responseStatusCode = 0;
    private String responseStatusReason = "";
    private boolean outputIsStreaming = false;
    private DataOutputStream headerOutputStream;
    private DataOutputStream bodyOutputStream;


    /**
     * Public Constructor
     *
     * @param httpPageArgs Pre-formatted table of HTTP request arguments,
     *  exactly as they will be passed to the presentation.
     * @param headerOutputStream Stream that the HTTP response header is
     *  written to.
     * @param bodyOutputStream Stream that the HTTP response body is
     *  written to.
     */
    public SimComms (HttpPageArgs httpPageArgs,
                     DataOutputStream headerOutputStream,
                     DataOutputStream bodyOutputStream) {
        this.httpPageArgs = httpPageArgs;
        this.headerOutputStream = headerOutputStream;
        this.bodyOutputStream = bodyOutputStream;
    }

    /**
     * Generate an error if data is already streaming
     *
     * @exception IOException 
     */
    private void streamingCheck()
        throws IOException
    {
        if (outputIsStreaming) {
            throw new IOException("Unable to set header;" +
                " Response body data is already streaming.");
        }  
    }


    /**
     * This method sets an output header.
     *
     * @exception IOException If an occurs setting the header.
     */
    public void setHeader(String name,
                          String value)
        throws IOException {
        streamingCheck();
        try {
            httpPageArgs.setHeader(name, value);
        } catch (KeywordValueException e) {
            throw new IOException (e.getMessage());
        }
    }

    /**
     * This method sets the response status header.
     *
     * @exception IOException If an occurs setting the status.
     */
    public void setStatus(int status,
                          String reason)
        throws IOException {
        streamingCheck();
        responseStatusCode = status;
        responseStatusReason = reason;
    }

    /**
     * This method maps a set of cookies into a single "Set-Cookie"
     * header.  Repeated calls to this method will replace the
     * previous "Set-Cookie" settings.
     *
     * @param cookies Table with leaf keys that contain
     *   arrays of Cookie.  All cookies are compined into
     *   a single "Set-Cookie" request.
     */
    public void setCookie(KeywordValueTable cookies)
        throws IOException, KeywordValueException {
        streamingCheck();
        String [] keys = cookies.leafKeys();
        for (int idx1=0; idx1<keys.length; idx1++) {
            Cookie [] c = (Cookie [])cookies.get(keys[idx1]);
            for (int idx2=0; idx2<c.length; idx2++) {
            httpPageArgs.setHttpResponseCookie (c [idx2]);
            }
        }
    }

    /**
     * This method sets a cookie header.
     *
     * @exception IOException If an occurs setting the cookie.
     */
    public void setCookie(String name,
                          String value,
                          String expires,
                          String domain,
                          String path,
                          boolean secure)
        throws IOException {
        streamingCheck();
        try {
            String stag = "";
            if (secure)
                stag = "secure";
            String cookie = name + "=" + value + "; expires=" + expires +
                "; path=" + path + "; domain=" + domain + "; " + stag;
            httpPageArgs.setHeader("Set-Cookie", cookie);
        } catch (KeywordValueException e) {
            throw new IOException (e.getMessage());
        }
    }


    /**
     * This method sets a cookie header localized to
     * the root of the application tree.
     *
     * @exception IOException If an occurs setting the cookie.
     */
    public void setCookie(String name,
                          String value)
        throws IOException {
        streamingCheck();
        setCookie(name, value, "", "", this.scriptName, false);
    }
    


    /**
     * This method expires a cookie.
     *
     * @exception IOException If an occurs expiring the cookie.
     */
    public void expireCookie(String name, String path)
        throws IOException {
        streamingCheck();
        setCookie(name, "expired",
                  "Saturday, 07-Dec-1996 12:00:00 GMT",
                  "",
                  path, false);
    }


    /**
     * This method expires a cookie at the root of
     * the application tree.
     *
     * @exception IOException If an occurs expiring the cookie.
     */
    public void expireCookie(String name)
        throws IOException {
        streamingCheck();
        setCookie(name, "expired",
                  "Saturday, 07-Dec-1996 12:00:00 GMT",
                  "",
                  this.scriptName, false);
    }


    /**
     * This method sets the content length http header.
     *
     * @exception IOException If an occurs setting the header.
     */
    public void setContentLength(int length)
        throws IOException {
        streamingCheck();
        try {
            httpPageArgs.setHttpResponseHeader("ContentLength", Integer.toString (length));
        } catch (KeywordValueException e) {
            throw new IOException (e.getMessage());
        }
    }

    /**
     * This method sets the Contenttype http header.
     *
     * @exception IOException If an occurs setting the header.
     */
    public void setContentType(String contentType)
        throws IOException {
        streamingCheck();
        try {
            httpPageArgs.setHttpResponseHeader("ContentType", contentType);
        } catch (KeywordValueException e) {
            throw new IOException (e.getMessage());
        }
    }

    /**
     * This method returns an array containing all of the values
     * set for the named HTTP request header.
     *
     * @param name The header name.
     * @return An array of strings containing the values set for the
     * header.  The array length is zero if no values were set.
     * @exception InvalidKeywordException if the header name is invalid.
     */
    public String [] getHttpRequestHeader(String name)
        throws Exception {
        return httpPageArgs.getHttpRequestHeader(name);
    }

    /**
     * This method returns all HTTP request cookies that have the
     * specified name.
     *
     * @param name The cookie name.
     * @return Array containing all the matched cookie names.  If
     *   none are matched the array legth is zero.
     */
    public Cookie [] getHttpRequestCookie(String name)
        throws Exception {
        return httpPageArgs.getHttpRequestCookies(name);
    }



    /**
     * This method returns values associated with a HTTP (CGI)
     * request arguments.
     *
     * @return Array containing all the values set for that argument.
     *   The array length is zero if the argument was not set.
     */
    public String [] getHttpRequestArg(String name)
        throws Exception {
        return httpPageArgs.getHttpRequestArg(name);
    }


    /**
     * This method returns values associated with a HTTP (CGI)
     * request arguments.  If no value is set then the default
     * is returned.
     *
     * @param name The name of the HTTP request argument desired.
     * @param defValue The default value to use if no value exists.
     * @return Array containing all the values set for that argument.
     *   The array length is zero if the argument was not set.
     */
    public String [] getHttpRequestArg(String name, String defValue)
        throws Exception {
        return httpPageArgs.getHttpRequestArg(name, defValue);
    }



    /**
     * This method sets an HTTP response header.
     *
     * @param name The header name.
     * @param value The header value.
     */
    public void setHttpResponseHeader(String name, String value)
        throws Exception {
        streamingCheck();
        httpPageArgs.setHttpResponseHeader(name, value);
    }



    /**
     * This method returns an array containing all of the values
     * set for the named HTTP response header.
     *
     * @param name The header name.
     * @return An array of strings containing the values set for the
     * header.  The array length is zero if no values were set.
     * @exception InvalidKeywordException if the header name is invalid.
     */
    public String [] getHttpResponseHeader(String name)
        throws Exception {
        return httpPageArgs.getHttpResponseHeader(name);
    }

    /**
     * This method sets an HTTP response cookie.
     *
     * @param name The cookie name.
     */
    public void setHttpResponseCookie(Cookie cookie)
        throws Exception {
        httpPageArgs.setHttpResponseCookie(cookie);
    }

    /**
     * This method returns all HTTP response cookies that have the
     * specified name.
     *
     * @param name The cookie name.
     * @return Array containing all the matched cookie names.  If
     *   none are matched the array legth is zero.
     */
    public Cookie [] getHttpResponseCookies(String name)
        throws Exception {
        return httpPageArgs.getHttpResponseCookies(name);
    }

    /**
     * This method sets an HTTP redirect header.
     *
     * @param name The header name.
     */
    public void setHttpRedirectHeader(String name, String value)
        throws Exception {
        streamingCheck();
        httpPageArgs.setHttpRedirectHeader(name, value);
    }

    /**
     * This method returns an array containing all of the values
     * set for the named HTTP redirect header.
     *
     * @param name The header name.
     * @return An array of strings containing the values set for the
     * header.  The array length is zero if no values were set.
     * @exception InvalidKeywordException if the header name is invalid.
     */
    public String [] getHttpRedirectHeader(String name)
        throws Exception {
        return httpPageArgs.getHttpRedirectHeader(name);
    }

    /**
     * This method sets an HTTP redirect cookie.
     *
     * @param name The cookie name.
     */
    public void setHttpRedirectCookie(Cookie cookie)
        throws Exception {
        streamingCheck();
        httpPageArgs.setHttpRedirectCookie(cookie);
    }

    /**
     * This method returns all HTTP redirect cookies that have the
     * specified name.
     *
     * @param name The cookie name.
     * @return Array containing all the matched cookie names.  If
     *   none are matched the array legth is zero.
     */
    public Cookie [] getHttpRedirectCookie(String name)
        throws Exception {
        return httpPageArgs.getHttpRedirectCookie(name);
    }

    /**
     * This method sets a HTTP (CGI) arguments to use when redirecting
     * the page.
     *
     * @param name The argument name.
     * @param value The argument value.
     */
    public void setHttpRedirectArg(String name, String value)
        throws Exception {
        streamingCheck();
        httpPageArgs.setHttpRedirectArg(name, value);
    }

    /**
     * This method returns all values associated with a HTTP (CGI)
     * redirect arguments.
     *
     * @return Array containing all the values set for that argument.
     *   The array length is zero if the argument was not set.
     */
    public String [] getHttpRedirectArg(String name)
        throws Exception {
        return httpPageArgs.getHttpRedirectArg(name);
    }


    //FIX: trying to grab entry zero of the returned arrays is really 
    //     a quick hack.  Not sure what the right answer here is.

    /**
     * Return the query string of this request.
     */
    public String getHttpQueryString() {
        String result;
        try {
            result = httpPageArgs.getHttpRequestHeader(HttpPageArgs.QUERY_STRING)[0];
        } catch (Exception e) {
            result = "";
        }
        return result;
    }

    /**
     * Return the url of this request.
     */
    public String getHttpUrl() {
        String result;
        try {
            result = httpPageArgs.getHttpRequestHeader(HttpPageArgs.URL)[0];
        } catch (Exception e) {
            result = "";
        }
        return result;
    }
    
    /**
     * Returns the remote host name.
     */
    public String getHttpRemoteHost() {
        String result;
        try {
            result = httpPageArgs.getHttpRequestHeader(HttpPageArgs.REMOTE_HOST)[0];
        } catch (Exception e) {
            result = "";
        }
        return result;
    }

    /**
     * Return the document root.
     */
    public String getHttpDocumentRoot() {
        String result;
        try {
            result = httpPageArgs.getHttpRequestHeader(HttpPageArgs.DOCUMENT_ROOT)[0];
        } catch (Exception e) {
            result = "";
        }
        return result;
    }

    /**
     * Returns the document root.
     */
    public String getHttpPathInfo() {
        String result;
        try {
            // FIX: This is probably wrong.
            result = httpPageArgs.getHttpRequestHeader(HttpPageArgs.DOCUMENT_URI)[0];
        } catch (Exception e) {
            result = "";
        }
        return result;
    }


    /**
     * This method causes the current request to be redirected
     * to another location.  The redirection is performed on
     * the client side via the "Location:" header.  Any redirect
     * arguments set in this page are appended to the url.  Likewise,
     * any redirect cookies are set.
     *
     * @param url The url which should handle the request.
     * @exception lutris.jolt.presentation.PageRedirectException  This
     *   exception should not be caught as it is handeled by the
     *   presentation manager.
     */
    public Exception clientSideRedirectException(String url) {
        PageRedirectException p = new PageRedirectException(url);
        p.setRedirectionViaClient(true);
        try {
            if (requestArgs.containsKey(requestArgs.REDIRECT_ARGS)) {
                p.setHttpRedirectArgs(requestArgs.
                                      getSection(requestArgs.REDIRECT_ARGS));
            }
            if (requestArgs.containsKey(requestArgs.REDIRECT_COOKIE)) {
                p.setHttpRedirectCookie(requestArgs.
                                        getSection(requestArgs.
                                                   REDIRECT_COOKIE));
            }
        }
        catch (Exception e) {
            System.err.println(e); //FIX:
        }
        return p;
    }

    /**
     * This method causes the current request to be redirected
     * to another location.  The redirection is performed locally 
     * on the server side.  Any redirect
     * arguments set in this page are appended to the url.  Likewise,
     * any redirect cookies are passed on to the new url.
     *
     * <UL>Notes:
     * <LI>cookies are not set on the browser.
     * <LI>internal (server side) redirection should be handeled carefully.
     *     The new url should not contain any relative references to
     *     images and objects since the browser has no way of knowing
     *     that the page was redirected.  If this confuses you then
     *     you are better off using client-side redirection.
     * </UL>
     *
     * @param url The url which should handle the request.
     * @exception lutris.jolt.presentation.PageRedirectException  This
     *   exception should not be caught as it is handeled by the
     *   presentation manager.
     */
    public Exception serverSideRedirectException(String url) {
        PageRedirectException p = new PageRedirectException(url);
        p.setRedirectionViaClient(false);
        try {
            if (requestArgs.containsKey(requestArgs.REDIRECT_ARGS)) {
                p.setHttpRedirectArgs(requestArgs.
                                      getSection(requestArgs.REDIRECT_ARGS));
            }
            if (requestArgs.containsKey(requestArgs.REDIRECT_COOKIE)) {
                p.setHttpRedirectCookie(requestArgs.
                                        getSection(requestArgs.
                                                   REDIRECT_COOKIE));
            }
        }
        catch (Exception e) {
            System.err.println(e);   //FIX:
        }
        return p;
    }

    /**
     * Start writing output if it hasn't already begun.
     *
     * @exception IOException If an HTTP error occurs.
     */
    private void startWrite ()
        throws IOException {
        if (!outputIsStreaming) {
            headerOutputStream.writeBytes ("HTTP/1.0 " + responseStatusCode + " " +
                                           responseStatusReason + "\n");
            try {
                if (httpPageArgs.containsKey (HttpPageArgs.RESPONSE_HEADER)) {
                    KeywordValueTable headers = httpPageArgs.getSection (HttpPageArgs.RESPONSE_HEADER);
                    String [] headerNames = headers.leafKeys();
                    for (int i = 0; i < headerNames.length; i++) {
                        String [] values = httpPageArgs.getHttpResponseHeader(headerNames[i]);
                        for (int j = 0; j < values.length; j++) {
                            headerOutputStream.writeBytes (headerNames[i] + ": " + values[j] + "\n");
                        }
                    }
                }
            } catch (KeywordValueException e) {
                // Hopefully due to no headers being set.
            }

            Cookie [] cookies = requestArgs.getAllHttpResponseCookies();
            for (int idx = 0; idx < cookies.length; idx++) {
                headerOutputStream.writeBytes (cookies[idx].toString());
            }
            headerOutputStream.writeBytes ("\n");
            headerOutputStream.flush();
            outputIsStreaming = true;
        }
    }

    //------------------------------------------
    // Implementation of DataOutput interface
    //------------------------------------------

    public void write(int b)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.write(b);
    }

    public void write(byte[] b)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.write(b);
    }

    public void write(byte[] b, int off, int len)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.write(b, off, len);
    }

    public void writeBoolean(boolean v)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeBoolean(v);
    }

    public void writeByte(int v)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeByte(v);    
    }

    public void writeBytes(String s)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeBytes(s);    
    }

    public void writeChar(int v)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeChar(v);    
    }

    public void writeChars(String s)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeChars(s);    
    }

    public void writeDouble(double v)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeDouble(v);    
    }

    public void writeFloat(float v)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeFloat(v);    
    }

    public void writeInt(int v)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeInt(v);    
    }

    public void writeLong(long v)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeLong(v);    
    }

    public void writeShort(int v)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeShort(v);    
    }

    public void writeUTF(String str)
        throws IOException
    {
        startWrite ();
        bodyOutputStream.writeUTF(str);    
    }

}

