/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.kelp.common.wizard.importer;

// ToolBox imports
import org.enhydra.tool.common.SwingUtil;
import org.enhydra.tool.common.PathHandle;

// Kelp imports
import org.enhydra.kelp.common.ValidationUtil;
import org.enhydra.kelp.common.ValidationException;

// Standard imports
import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.*;
import java.io.File;
import java.util.ResourceBundle;

/**
 * Panel for entering the default options: project name, package
 * and destination.
 *
 * @author Paul Mahar
 */
public class RootPanel extends ImporterPanel {
    static ResourceBundle                 res =
        ResourceBundle.getBundle("org.enhydra.kelp.common.Res");    // nores
    transient private JPanel              panelRoot;
    transient private JTextField          textRoot;
    transient private JButton             buttonRoot;

    //
    transient private JPanel              panelFiller;
    transient private GridBagLayout       layoutMain;
    transient private GridBagLayout       layoutRoot;
    transient private LocalButtonListener buttonListener;
    transient private TitledBorder        borderRoot;

    /**
     * Create default option panel 1 for entering the
     * following default options: project name, package and
     * destination.
     */
    public RootPanel() {
        try {
            jbInit();
            pmInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Read project name, package and destination values from the
     * option set into the swing controls.
     *
     * @exception GeneratorException
     * Thrown if unable to update Swing controls with option set values.
     */
    public void readOptions() {
        String path;

        // project root
        path = getPaths().getRootPath();
        if (path == null) {
            textRoot.setText(new String());
        } else {
            textRoot.setText(path);
        }
    }

    /**
     * Write project name, package and destination values from the
     * swing controls into the option set.
     *
     * @exception GeneratorException
     * Thrown if unable to update option set from Swing control values.
     */
    public void writeOptions() throws ValidationException {
        String path = null;

        if (getPaths() == null) {
            throw new ValidationException(res.getString("Import_paths_not_set1"));
        }

        // root
        path = textRoot.getText();
        getPaths().setRootPath(path);
    }

    /**
     * Validate project name, package and destination values
     * in the swing controls.
     *
     * @exception ValidationException
     * Thrown if swing control values are not valid for the
     * current option set.
     */
    public void validateOptions() throws ValidationException {
        String value = new String();

        // project name
        value = textRoot.getText().trim();
        if (!ValidationUtil.isDirectory(value)) {
            throw new ValidationException(res.getString("Project_directory_is")
                                          + res.getString("Directory_not_found_"));
        }
    }

    /**
     * Get the title to use on the current page.
     *
     * @return
     * A string to place at the top of a CodeGen wizard panel.
     */
    public String getPageTitle() {
        return res.getString("Project_Directory");
    }

    /**
     * Get the instructions for entering option values for the
     * current page.
     *
     * @return
     * A string to place below the page title.
     */
    public String getInstructions() {
        StringBuffer buf = new StringBuffer();

        buf.append(res.getString("RootInstruct1"));
        buf.append(res.getString("RootInstruct2"));
        buf.append(res.getString("RootInstruct3"));
        return buf.toString();
    }

    //
    // PRIVATE
    //

    /**
     * Method declaration
     *
     *
     * @exception Exception
     */
    private void jbInit() throws Exception {
        textRoot = (JTextField) Beans.instantiate(getClass().getClassLoader(),
                                                  JTextField.class.getName());
        panelRoot = (JPanel) Beans.instantiate(getClass().getClassLoader(),
                                               JPanel.class.getName());
        layoutRoot =
            (GridBagLayout) Beans.instantiate(getClass().getClassLoader(),
                                              GridBagLayout.class.getName());
        buttonRoot = (JButton) Beans.instantiate(getClass().getClassLoader(),
                                                 JButton.class.getName());
        layoutMain =
            (GridBagLayout) Beans.instantiate(getClass().getClassLoader(),
                                              GridBagLayout.class.getName());
        panelFiller = (JPanel) Beans.instantiate(getClass().getClassLoader(),
                                                 JPanel.class.getName());
        buttonRoot.setText("...");  // nores
        panelRoot.setLayout(layoutRoot);
        panelRoot.add(textRoot,
                      new GridBagConstraints(0, 0, 1, 1, 0.1, 0.1,
                                             GridBagConstraints.WEST,
                                             GridBagConstraints.HORIZONTAL,
                                             new Insets(5, 5, 5, 5), 0, 0));
        panelRoot.add(buttonRoot,
                      new GridBagConstraints(1, 0, 1, 2, 0.0, 0.0,
                                             GridBagConstraints.CENTER,
                                             GridBagConstraints.NONE,
                                             new Insets(5, 5, 5, 20), 0, 0));
        this.setLayout(layoutMain);
        this.add(panelRoot,
                 new GridBagConstraints(0, 0, 1, 1, 0.1, 0.1,
                                        GridBagConstraints.CENTER,
                                        GridBagConstraints.BOTH,
                                        new Insets(5, 5, 5, 5), 0, 0));
    }

    /**
     */
    private void pmInit() {
        Border border = BorderFactory.createEtchedBorder();

        borderRoot = BorderFactory.createTitledBorder(border);
        borderRoot.setTitle(res.getString("borderRoot_Title"));
        panelRoot.setBorder(borderRoot);
        buttonListener = new LocalButtonListener();
        buttonRoot.addActionListener(buttonListener);
        textRoot.setText(new String());
    }

    /**
     */
    private void browseForDirectory() {
        File dir = null;

        dir =
            SwingUtil.getDirectoryChoice(this, textRoot.getText(),
                                         res.getString("Choose_the_project"));
        if (dir != null) {
            textRoot.setText(PathHandle.createPathString(dir));
        }
    }

    /**
     */
    private class LocalButtonListener implements ActionListener {
        public void actionPerformed(ActionEvent event) {
            browseForDirectory();
        }

    }
}
