/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.kelp.jdeveloper.wizard;

// JDeveloper imports
import borland.jbuilder.addin.Container;
import borland.jbuilder.addin.Context;
import borland.jbuilder.addin.IdeEnvironment;
import borland.jbuilder.addin.VersionedAddin;
import borland.jbuilder.addin.Wizard;
import borland.jbuilder.addin.WizardManager;
import borland.jbuilder.addin.Node;
import borland.jbuilder.addin.HtmlNode;
import borland.jbuilder.addin.Project;

// ToolBox
import org.enhydra.tool.common.SwingUtil;

// Kelp imports
import org.enhydra.kelp.common.Constants;
import org.enhydra.kelp.common.ResUtil;
import org.enhydra.kelp.common.node.OtterProject;
import org.enhydra.kelp.common.node.OtterTemplateNode;
import org.enhydra.kelp.common.swing.IconSet;
import org.enhydra.kelp.common.wizard.deploy.DeployBuildDialog;
import org.enhydra.kelp.jdeveloper.node.OracleProject;
import org.enhydra.kelp.jdeveloper.node.OracleTemplateNode;

// Standard imports
import javax.swing.JOptionPane;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Frame;
import java.awt.Image;
import java.io.File;
import java.io.Serializable;
import java.util.Vector;
import java.lang.System;
import java.util.ResourceBundle;

/**
 * This is the class that gets registered with the IDE through the
 * jbuilder.properties file.
 */
public class DeployWizard implements Serializable, VersionedAddin, Wizard {

    // string not to be resourced
    static ResourceBundle res =
        ResourceBundle.getBundle("org.enhydra.kelp.jdeveloper.Res"); // nores
    private final String  VERSION = "2.0";    // nores

    //
    private String        title = null;
    private int           option = 0;

    /**
     * Constructor declaration
     *
     */
    public DeployWizard() {}

    /**
     * For the VersionAddin Interface.
     */
    public String version() {
        return VERSION;
    }

    /**
     * Entry point exposed by the Wizard interface for early initialization.
     * Registers this Wizard object with the IDE's WizardManager object
     * @$param ide Access to all AddIn interfaces
     */
    public void startup(IdeEnvironment ide) {
        WizardManager e = ide.getWizardManager();

        e.registerWizard(this);
    }

    /**
     * Wizard interface implementation, run the wizard in response to this call.
     * @param ide Access to all AddIn interfaces
     * @param context Current node, project, browser
     */
    public void invoke(IdeEnvironment ide, Context context) {
        Project nativeProject = null;
        Frame   frame = null;

        frame = ide.getIdeFrame();
        nativeProject = context.getProject();
        String ver = System.getProperty(Constants.JAVA_VERSION);

        if (ver.startsWith(Constants.JDK11)) {
            JOptionPane.showMessageDialog(ide.getIdeFrame(),
                                          ResUtil.format(res.getString("Kelp_requires_Java_2"),
                                          ver));
            nativeProject = null;
        }
        if (nativeProject != null
                && isEnhydraAvailable(ide, nativeProject.getClassPath())) {
            nativeProject.save(false);    // save all file changes in project
            DeployBuildDialog dlg = new DeployBuildDialog(frame, getName(),
                    true);
            OtterProject      otterProject = null;

            otterProject = new OracleProject(nativeProject);
            dlg.setProject(otterProject);
            dlg.setNodes(otterProject.getAllInput());
            dlg.show();
            frame.requestFocus();
        }
    }

    /**
     * Wizard interface method, returns information about the type of Wizard this is.
     * @return An integer representing combined flags defined by the
     * Wizard interface.
     */
    public int getInfo() {
        return Wizard.PRODUCES_FILE + Wizard.REQUIRES_PROJECT;
    }

    /**
     * Wizard interface method, returns the name of this wizard.
     * @return Name of the wizard
     */
    public String getName() {
        return res.getString("Enhydra_Deployment");
    }

    /**
     * Wizard interface method, returns the text for the menu label
     * @return Menu label text
     */
    public String getMenuLabel() {
        return ResUtil.format(res.getString("_0_dots"), getName());
    }

    /**
     * Wizard interface method
     * @return Currently returns no image (null)
     */
    public Image getSmallImage() {
        return IconSet.getSmallIcon().getImage();
    }

    /**
     * Wizard interface method
     * @return Currently returns no image (null)
     */
    public Image getBigImage() {
        return IconSet.getLargeIcon().getImage();
    }

    /**
     * Method declaration
     *
     *
     * @return
     */
    public String getTitle() {
        return this.title;
    }

    /**
     * Method declaration
     *
     *
     * @param newTitle
     */
    public void setTitle(String newTitle) {
        this.title = newTitle;
    }

    // read only properties

    /**
     * Method declaration
     *
     *
     * @return
     */
    public int getOption() {
        return option;
    }

    /**
     * Method declaration
     *
     *
     * @param newOption
     */
    private void setOption(int newOption) {
        option = newOption;
    }

    /**
     * Method declaration
     *
     *
     * @param owner
     */
    public void showDialog(Component owner) {
        DeployBuildDialog dialog = null;

        if (owner instanceof Dialog) {
            dialog = new DeployBuildDialog((Dialog) owner, getTitle(), true);
        } else if (owner instanceof Frame) {
            dialog = new DeployBuildDialog((Frame) owner, getTitle(), true);
        }
        if (dialog != null) {
            dialog.show();
            dialog.clearAll();
        }
        if (owner.isVisible()) {
            owner.requestFocus();
        }
    }

    /**
     * Method declaration
     *
     *
     * @param args
     */
    public static void main(String[] args) {
        SwingUtil.setLookAndFeelToSystem();
        DeployWizard wizard;

        wizard = new DeployWizard();
        wizard.showDialog(new Frame());
        System.exit(0);
    }

    /**
     * Method declaration
     *
     *
     * @param ide
     * @param projectClassPath
     *
     * @return
     */
    private boolean isEnhydraAvailable(IdeEnvironment ide,
                                       String projectClassPath) {
        StringBuffer buf = new StringBuffer();

        if (projectClassPath.toLowerCase().indexOf(Constants.FILE_ENHYDRA_JAR)
                < 0) {
            buf.append(res.getString("EnhydraLibErrorD1"));
            buf.append(res.getString("EnhydraLibErrorD2"));
            buf.append(res.getString("EnhydraLibErrorD3"));
            buf.append(res.getString("EnhydraLibErrorD4"));
            buf.append(res.getString("EnhydraLibErrorD5"));
        }
        if (buf.length() > 0) {
            JOptionPane.showMessageDialog(null, buf.toString());
        }
        return (buf.length() == 0);
    }

}
