/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: MultiServer.java,v 1.3.2.1.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package com.lutris.multiServer;
 
import java.io.*;
import java.net.*;
import java.rmi.*;
import java.util.*;
import com.lutris.util.Config;
import com.lutris.util.ConfigFile;
import com.lutris.util.ConfigException;
import com.lutris.util.KeywordValueTable;
import com.lutris.util.KeywordValueException;
import com.lutris.logging.*;
// X22 this is because it is temporarily hardcoded to create ServletContainer
import org.enhydra.servlet.ServletContainer;


/** 
 * A stand-alone server. It starts and stops services like the
 * ServletContainer. <P>
 *
 * At startup time a configuration file is read, and it describes what
 * services to start.
 *
 * After this config file is read, and the server is set up, no further 
 * action is taken by the server. The thread that ran through main() exits.
 * The Services must create threads, and they are responsible for
 * receiving requests from the outside world and processing them. <P>
 *
 * Normally one of the Servlets will be an administration application. This
 * admin app knows about this class, and so it knows how to talk to the
 * current ConnectionMethods and Servlets. It may change the server setup
 * on the fly. For example, it can create and add a new ConnectionMethod.
 * Or it could add a new channel to an already existing ConnectionMethod.
 * Or register a new Servlet. Or add a logging filter to a channel. 
 * Or monitor performance. <P>
 *
 * This class implements main(). It is meant to be run as a java 
 * application. Do not instantiate this class. Use the static variables.<P>
 *
 * The <CODE>main()</CODE> function will initialize and configure the
 * global logging facility. The facility name is "MultiServer", and you
 * may use any of the level names defined in the <CODE>Logger</CODE>
 * class. One set of level names listed in the MultiServer config file
 * have their output directed to the server log file. Another set of
 * level names have their output directed to <CODE>stderr</CODE>.
 * Text written to any other level name not mentioned in the server
 * config file will be ignored.<BR>
 * Any class running in this virtual machine
 * may write to the MultiServer log file, using the following code:
 * <PRE>
 * import com.lutris.logging.*;
 * LogChannel logChan = Logger.getCentralLogger().getChannel("Multiserver");
 * int level = logChan.getLevel("ERROR");
 * logChan.write(level, "An error occured: ....");
 * </PRE>
 * 
 * @author Shawn McMurdo
 * @author Andy John
 */
public class MultiServer {

    /*
     * This is the highwater mark for memory usage by the virtual machine
     */
    private static long maxMemoryUsage = 0;

    /*
     *   Private logging info for use only in this file.
     */
    private static StandardLogger standardLogger;
    private static LogChannel     logChannel;

    // X22 This needs to be properly handled, not an array.
    private static Hashtable services = new Hashtable();

    private static final String defaultConfigFileName = "multiserver.conf";
    private static String configFileName;
    private static ConfigFile configFile;
    private static Config config;

    // Strings used in config file names
    public static final String SERVER = "Server";
    public static final String LOG_FILE = "LogFile";
    public static final String LOG_TO_FILE = "LogToFile";
    public static final String LOG_TO_STDERR = "LogToStderr";
    public static final String SERIAL_NUMBER = "SerialNumber";
    public static final String ACTIVATION_KEY = "ActivationKey";


    /**
     *  Don't ever create an instance of MultiServer. Use the 
     *  static fields, and call the static methods.
     */
    private MultiServer() {}

    /**
     * Do actual work of starting the multiserver.
     */
    private static void start(String argv[]) {
	/*
         *   Very first thing is to establish the central logging
         *   object. It will be the first thing configured by
         *   readConfigFile().
         */
        standardLogger = new StandardLogger(true);
        logChannel = standardLogger.getChannel("Multiserver");

        /*
         *   Read the config file. This is what really gets the
         *   server up and running.
         */
        if (argv.length > 0) {
            readConfigFile(argv[0]); 
        } else {
            readConfigFile(defaultConfigFileName); 
        }

	// start all of the services that were created
	Enumeration serviceEnum = services.elements();
	while (serviceEnum.hasMoreElements()) {
	    Service service = (Service) serviceEnum.nextElement();
	    service.start();
	}

	// This thread will die, but services are now running.
    }

    /**
     * The entry point for the stand-alone server. 
     * If an argument is passed in, it is assumed to be the location
     * of the server config file. If no arguments are passed in, the
     * default filename "./MultiServer.config" is used.
     * @param argv An array of input parameters. None are required.
     */
    public static void main(String argv[]) {
        try {
            start(argv);
        } catch (Throwable err) {
            fatalErr("Failed to start multiServer", err);
        }
    }

    /**
     *  Shut down the server.
     *  This function stops all services.
     */
    public static void shutdown() {
        try {
            Enumeration serviceEnum = services.elements();
            while (serviceEnum.hasMoreElements()) {
                Service service = (Service) serviceEnum.nextElement();
                service.stop();
            }
        } catch (Throwable err) {
            fatalErr("Failed to shutdown multiServer", err);
        }

        Runtime vm = Runtime.getRuntime();
        Runtime.runFinalizersOnExit(true);
        vm.exit(0);
    }

    /**
     * Get the named service.
     *
     * @param the name of the service.
     * @return the Service
     */
    public static Service getService(String name) {
	return (Service) services.get(name);
    }

    /**
     * Parsing and implementing the config file.
     * This sets up the server, according to the config file.
     */
    private static void readConfigFile(String filename) {
	configFileName = filename;
        try {
	    FileInputStream configFIS = new FileInputStream(filename);
	    configFile = new ConfigFile(configFIS);
            config = configFile.getConfig();
            parseServerSection(myGetSection(config, SERVER, false));
            configFIS.close();
        } catch (Throwable e) {
            fatalErr("Unable to read config file " + filename, e);
        }

	// X22 set up services, ServletContainer is hardcoded for now.
	ServletContainer sc = new ServletContainer();
	services.put("ServletContainer", sc);
	sc.init(filename);
	
	/*------------------------------------------------------------------*
	 * This try block calls into the potentially existing ActKeyValidation
	 * class named by validationClassName. Currently only Lutris's
	 * LutrisValidation class is supported. This class, if it is included
	 * in the application, will perform any needed check on the viability
	 * and right-to-run this application.
	 *
	 * If the class is not found in this application, then it's probably
	 * just a standard Enhydra Open Source application that's running,
	 * and that's just dandy and it should run fine.
	 *
	 * Other groups that want to include some sort of key checking or
	 * validation should have the validationClassName variable name a
	 * class that implements the ActKeyValidation interface.
	 *
	 * Note that this isn't the most secure implementation of this schema;
	 * as the wily hacker who is trying to circumvent your checking can
	 * just compile up their own version of MultiServer.class and put it
	 * before yours in the classpath.
         *------------------------------------------------------------------*/

	try {
	    final String validationClassName = 
		"com.lutris.multiServer.LutrisValidator";
	    Object validator = Class.forName(validationClassName).newInstance();
	    if (validator instanceof com.lutris.multiServer.ActKeyValidation) {
		com.lutris.multiServer.ActKeyValidation akv = 
		    (com.lutris.multiServer.ActKeyValidation) validator;
		akv.validate(config, logChannel);
	    } else {
		fatalErr("Poorly formated key validation class: " +
			 validator.getClass());
	    }
	} catch (ClassNotFoundException e) {
	    ;
	} catch (ActKeyValidationException e) {
	    fatalErr("Servlet validation failure: " + e.getMessage(), e);
	} catch (Exception e) {
	    fatalErr("Poor instanciation of key validation class", e);
	}
    }

    /**
     * Retrieve the server parameters from the config file and
     * establish the logging facility.
     *
     * @param server
     *   the config object for the server section.
     */
    private static void parseServerSection(Config server) {
	/*
         * This is bootstrapping code. Until we have logging initialized
         * we can not use the standard helper functions.
         */
        if (server == null)
            fatalErr("No server section specified in config file " +
                     configFileName);

        String logFile = null;
        try {
	    logFile = server.getString(LOG_FILE);
        } catch (ConfigException e) {}
        if (logFile == null)
            fatalErr("No key " + SERVER + "." + LOG_FILE +
                     " found in config file " + configFileName);

        String[] toFile = null;
        try {
            toFile = ((Config) server).getStrings(LOG_TO_FILE);
        } catch (ConfigException e) {}
        if (toFile == null)
            fatalErr("No key " + SERVER + "." + LOG_TO_FILE +
                     " found in config file " + configFileName);

        String[] toStderr = null;
        try {
            toStderr = ((Config) server).getStrings(LOG_TO_STDERR);
        } catch (ConfigException e) {}
        if (toStderr == null)
            fatalErr("No key " + SERVER + "." + LOG_TO_STDERR +
                     " found in config file " + configFileName);

        File theLogFile = new File(logFile);
        try {
            standardLogger.configure(theLogFile, toFile, toStderr);
        } catch (IOException ioe) {
            fatalErr("Unable to begin logging to the file " + logFile, ioe);
        }

        /*
         * Done bootstrapping. From here on out we may use logging.
         */
        logChannel.write(Logger.INFO, "Starting Enhydra Multiserver");
    }

    static private String myGetString(Config cfg, String key) {
        return myGetString(cfg, key, null, true);
    }

    static private String myGetString(Config cfg, String key, String prefix) {
        return myGetString(cfg, key, prefix, true);
    }

    static private String[] myGetStrings(Config t, 
                   String key, String prefix) {
        return myGetStrings(t, key, prefix, true);
    }

    static private String myGetString(Config cfg, String key, 
                                      String prefix, boolean required) {
        String s = null;
        try {
            s = cfg.getString(key);
        } catch (ConfigException e) {}
        if ((s == null) && required) {
            if (prefix != null)
                prefix += ".";
            fatalErr("Error reading config file. Key " + 
                     prefix + key + " not found.");
        }
        return s;
    }

    static private String[] myGetStrings(Config t, 
                             String key, String prefix, boolean required) {
        String[] s = null;
        try {
            s = t.getStrings(key);
        } catch (ConfigException e) {}
        if ((s == null) && required) {
            if (prefix != null)
                prefix += ".";
            fatalErr("Error reading config file. Key " + 
                     prefix + key + " not found.");
        }
        return s;
    }



    static private Config myGetSection(Config cfg, String key) {
        return myGetSection(cfg, key, true);
    }

    static private Config myGetSection(Config cfg,
                                       String key, boolean required) {
        Config cfg2 = null;
        try {
            cfg2 = (Config) cfg.getSection(key);
        } catch (ConfigException e) {
            // Do nothing.
        } catch (Exception e) {
            fatalErr("Internal error parsing config file (section " +
                     key + ").", e);
        }
        if ((cfg2 == null) && required)
            fatalErr("Illegal config file. Section " +
                     key + " not found.");
        return cfg2;
    }



    static private String trim(String s) {
        if (s == null)
            return null;
        int len = s.length();
        while(len > 0) {
            char c = s.charAt(0);
            if ((c == ' ') || (c == '\t')) {
                s = s.substring(1);
                len--;
            } else
                break;
        }
        while(len > 0) {
            char c = s.charAt(len - 1);
            if ((c == ' ') || (c == '\t')) {
                s = s.substring(0, len - 1);
                len--;
            } else
                break;
        }
        return s;
    }

    /**
     *  This function examines the current state of the server and 
     *  writes it out to the same config file that was read when
     *  the server was started up.
     * @param filename The filename of the config file to write out.
     * If null, the same file that was read at startup time will be
     * rewritten.
     */
    public static String writeConfigFile(String filename) {
	// X22
	return null;
    }

    /** 
     * This method if for getting the value that has been stored as 
     * highwater mark for memory usage by the virtual machine.
     */
    static public long getMaxMemoryUsage() {
	return maxMemoryUsage;
    }

    /**
     * This method is used for setting the highwater mark for the 
     * amount of memory that the virtual machine has used.  The
     * method does the appropriate checking to see if the value 
     * passed is greater than the current stored value.
     * 
     * @param used The amount of memory that has been used.
     */
    static public void setMaxMemoryUsage(long used) {
	if (maxMemoryUsage < used)
	    maxMemoryUsage = used;
    }

    /**
     * This method is used to reset the maxMemoryUsage back to 
     * zero.  This is so that new statistics can be gotten without
     * restarting the application.
     */
    static public void resetMaxMemoryUsage() {
	maxMemoryUsage = 0;
    }
     
    /**
     * This method is used to get the current value of the amount of 
     * memory being used by the virtual machine.  
     *
     * @param forceGC A boolean value to tell the method whether or
     *   not to run garbage collection before getting the memory
     *   statistics.  WARNING!!! forcing garbage collection can cause
     *   a major performance impact.
     * @return the current amount of memory being used by the 
     *      virtual machine in bytes.
     */
    static public long currentMemoryUsage(boolean forceGC) {
	Runtime rt = Runtime.getRuntime();
	if (forceGC)
	    rt.gc();
	return rt.totalMemory() - rt.freeMemory();
    }

    /*
     * Log the message, then halt the virtual machine.
     */
    static private void fatalErr(String msg) {
        fatalErr(msg, null);
    }

    /*
     * Either log the message or write it to stderr, then halt the
     * virtual machine. We should always log the messages, unless that
     * is not possible.
     */
    static private void fatalErr(String msg, Throwable err) {
        if (logChannel != null) {
            logChannel.write(Logger.ERROR, msg, err);
            logChannel.write(Logger.ERROR, "Exiting the server.");
        }
        System.err.println(msg);
        if (err != null) {
            err.printStackTrace();
        }
        System.err.println("Exiting the server.");
        Runtime vm = Runtime.getRuntime();
        Runtime.runFinalizersOnExit(true); // For saftey/completeness?
        vm.exit(1);
    }


}

