/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Admin.java,v 1.36.10.1 2000/10/19 17:59:08 jasona Exp $
 */





package com.lutris.appserver.admin;

import java.util.Hashtable;
import com.lutris.util.*;
import com.lutris.http.BasicAuth;
import com.lutris.http.BasicAuthResult;
import com.lutris.util.Config;
import com.lutris.appserver.server.*;
import com.lutris.appserver.server.httpPresentation.*;
import org.enhydra.servlet.debug.DebugManager;


/**
 * Application class for this specific application.  This implements
 * standard startup and shutdown methods and well as application specific
 * ones. The functionality added to StandardApplication in the
 * implementation of HTTP BasicAuth and the default URL implementation.
 */
public class Admin extends StandardApplication {

    // The user who is allowed to run this application. Used with basic auth.
    private String username, password;
    // For the password prompt. Used with basic auth.
    private String relm = "Multiserver Administration";

    /**
     *  The set of all DebugManager, indexed by servlet id.
     *  There needs to be a centeralized set of them in this Application
     *  object so that there will only be at most one manager per
     *  servlet. If the managers are stored in session data, then there
     *  could be one per logged in user. <BR>
     *  The presentation objects manage the contents of this Hashtable.
     */
    public Hashtable debuggers;

    /**
     *  Has the state of the server been modified? If true, then the
     *  MultiServer's config file needs to be re-written. <BR>
     *  The presentaion objects manage this variable.
     */
    public boolean serverDirty;

    /**
     *  How many requests to buffer in the queue when debugging an
     *  application.
     */
    public int debugQueueSize; 

    /**
     *  When debugging, should we keep a copy of the data written to the net?
     *  This is specified in the admin config file, under the key
     *  "Admin.SaveResponseData".
     */
    public boolean saveResponseData;
 
    /**
     * Start the application.
     * 
     * @param appConfig
     *   Configuration object for this application.
     * @exception ApplicationException
     *   If an error occurs starting the application.
     */
    public void startup(Config appConfig)
	throws ApplicationException {

	/*
         * Mandatory config settings.
         */
        try {
	    username = appConfig.getString("Admin.Username");
	    password = appConfig.getString("Admin.Password");
            debugQueueSize = appConfig.getInt("Admin.DebugQueueSize");
	} catch (ConfigException except) {
	    throw new ApplicationException(except);
	}
        if (debugQueueSize < 1) 
            throw new ApplicationException("Admin.DebugQueueSize must " +
                                           "be at least 1.");
        /*
         * Optional config settings.
         */
        saveResponseData = false;
        try {
            saveResponseData = appConfig.getBoolean("Admin.SaveResponseData");
	} catch (ConfigException except) {
            // It's ok.
        }
        debuggers = new Hashtable();
	serverDirty = false;
        super.startup(appConfig);
    }


    /**
     *  This is called for each request, before it is processed.
     *  Ensure that the user has authenticated with HTTP
     *  Basic-Auth before continuing.
     *
     * @param comms
     *   The http communication object through which all aspect of
     *   a request can be obtained.
     * @return
     *   False to proceed with the request.
     * @exception PageUnauthorizedException
     *   if user does not login in correctly.
     * @exception Exception
     *   if any other error occurs.
     */
    public boolean requestPreprocessor(HttpPresentationComms comms)
            throws Exception {

        super.requestPreprocessor(comms);
        if (comms.session == null) {
            // Session not setup; so its not a presentation object.
            return false;
        }

	/*
	 *  Ensure that the username/password are being sent.
         *  If both the username and password are empty strings, then
         *  do not require a login.
	 */
        BasicAuthResult auth = BasicAuth.getAuthentication(comms.request);
        if (((username.length() != 0) || (password.length() != 0)) &&
             ((auth == null) || 
               !username.equals(auth.username) || 
               !password.equals(auth.password))) {
            /*
             * If they click cancel on the username/password dialog box,
             * then the browser will show the HTML sent along with the
             * unauthorized response code. Normally, the user will enter
             * a username and password, the browser will re-issue the request,
             * and the user will never see this.
             */
            throw new PageUnauthorizedException(relm);
        }
        /*
         * Since they sent the correct username and password, we know they
         * are logged in. If there is no user set in the session, then
         * set it now.
         */
        if (comms.session.getUser() == null)
            comms.session.setUser(AdminUser.sharedUser);
	/*
         *  Proceed normally.
	 */
        return false; 
    }


    /**
     *  Since we are a part of Enhydra, we do not need to print out the
     *  copyright header.
     */
    public void printCopyrightHeader() {
    }


}

