/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ConnectionMethodManager.java,v 1.5.2.1.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods;

import com.lutris.util.Config;
import com.lutris.util.ConfigException;
import com.lutris.util.KeywordValueTable;
import com.lutris.util.KeywordValueException;

import com.lutris.logging.LogChannel;
import com.lutris.logging.Logger;

import com.lutris.multiServer.MultiServer;
import org.enhydra.servlet.ServletContainer;

import java.util.Hashtable;
import java.util.Enumeration;


/**
 *   This class maintains a set of ConnectionMethods. Each ConnectionMethod
 *   is associated with an identification string (symbolic name).
 *   This ID string is used to retrieve a ConnectionMethod. <P>
 *
 *   This class does not monitor the creation of ConnectionMethods.
 *   If you create one, you are responsible for adding it. These public
 *   methods are the only things that modify the set of ConnectionMethods
 *   managed by this class. (Note that this is different than the
 *   ServletManager, which is responsible for creating the Servlets
 *   it manages.)
 *
 *   @see org.enhydra.servlet.connectionMethods.ConnectionMethod
 *   @author Andy John
 */
public class ConnectionMethodManager {

    public static final String PORT = "Port";
    public static final String BIND_ADDR = "BindAddress";
    public static final String NUM_THREADS = "NumThreads";
    public static final String QUEUE_SIZE = "QueueSize";
    public static final String THREAD_TIMEOUT = "ThreadTimeout";
    public static final String CLIENT_TIMEOUT = "ClientTimeout";
    public static final String SERVER_HOST = "ServerHost";
    public static final String SERVER_PORT = "ServerPort";

    // used for defining new connection methods in config file in the
    // form of ConnectionMethod.<connection id> = <fully qualified class name>
    public static final String CONNECTION_METHOD = "ConnectionMethod";

    // used for 'default' set of connection methods in the config file
    public static final String CONNECTION = "Connection";
    public static final String TYPE = "Type";

    // Connection type strings used in config file
    public static final String APACHE = "apache";
    public static final String CGI = "cgi";
    public static final String HTTP = "http";
    public static final String HTTPS= "https";
    public static final String IIS = "iis";
    public static final String RMI = "rmi";
    public static final String WAI = "wai";
    public static final String ENHYDRADIRECTOR = "EnhydraDirector";

    // Fully qualified class names of connection methods
//    public static final String APACHE_CONNECTION =
//	"org.enhydra.servlet.connectionMethods.apache.ApacheConnectionMethod";
//    public static final String CGI_CONNECTION =
//	"org.enhydra.servlet.connectionMethods.cgi.CGIConnectionMethod";
    public static final String HTTP_CONNECTION =
	"org.enhydra.servlet.connectionMethods.http.HttpConnectionMethod";
    public static final String HTTPS_CONNECTION =
	"org.enhydra.servlet.connectionMethods.https.HttpsConnectionMethod";
//    public static final String IIS_CONNECTION =
//	"org.enhydra.servlet.connectionMethods.iis.IISConnectionMethod";
//    public static final String RMI_CONNECTION =
//	"org.enhydra.servlet.connectionMethods.rmi.RMIConnectionMethodImpl";
//    public static final String WAI_CONNECTION =
//	"org.enhydra.servlet.connectionMethods.wai.WAIConnectionMethod";
    public static final String ENHYDRADIRECTOR_CONNECTION =
	"org.enhydra.servlet.connectionMethods.EnhydraDirector.EnhydraDirectorConnectionMethod";

    private static final int CONNECTION_TYPE_COUNT = 3;

    private Hashtable connectionTable = null;

    /**
     * The logging channel.
     */
    protected LogChannel logChannel;

    // This is used to generate unique names.
    private int nameCounter = 0;

    private class TableEntry {
        String connectionID;
        ConnectionMethod cm;
    }

    private Hashtable theConnections;

    /**
     *   Create a new, empty, ConnectionMethodManager.
     */
    public ConnectionMethodManager () {
        theConnections = new Hashtable();
	connectionTable = new Hashtable(CONNECTION_TYPE_COUNT);
	//connectionTable.put(APACHE, APACHE_CONNECTION);
	//connectionTable.put(CGI, CGI_CONNECTION);
	connectionTable.put(HTTP, HTTP_CONNECTION);
	connectionTable.put(HTTPS, HTTPS_CONNECTION);
	//connectionTable.put(IIS, IIS_CONNECTION);
	//connectionTable.put(RMI, RMI_CONNECTION);
	//connectionTable.put(WAI, WAI_CONNECTION);
	connectionTable.put(ENHYDRADIRECTOR, ENHYDRADIRECTOR_CONNECTION);
	logChannel = Logger.getCentralLogger().getChannel("Multiserver");
    }

    /**
     *   Parse the Connection section of the config file and attempt to
     *   create all specified connection types.
     *
     *   @param config The config object representating the config file
     *   @exception ConnectionMethodException if a specified connection
     *              method does not exist or the connection method fails
     *              to initialize itself properly.
     **/
    public void initConnectionMethods(Config config)
	throws ConnectionMethodException {
	addAdditionalConnectionMethods(config);
	KeywordValueTable connections = null;
	try {
	    connections = config.getSection(CONNECTION);
        } catch (KeywordValueException e) {
            logChannel.write(Logger.ERROR,
	       "Illegal config file. Section \"Connection\" not found.", e);
	    throw new ConnectionMethodException(e);
        } catch (Exception e) {
            logChannel.write(Logger.ERROR,
	       "Internal error parsing config file (section \"Connection\").", e);
	    throw new ConnectionMethodException(e);
        }

        if (connections!=null) {
    	    String[] connectionKeys = connections.keys();
	    for (int i=0; i<connectionKeys.length; i++) {
	        try {
		    addConnectionMethod(
		        (Config)connections.getSection(connectionKeys[i]),
		        connectionKeys[i]);
	        } catch (KeywordValueException e) {
		    logChannel.write(Logger.ERROR,
		        "Illegal Connection section in config file (key \"" +
		        connectionKeys[i] + "\" not found).", e);
		    throw new ConnectionMethodException(e);
	        }
	    }
        }
    }

    /**
     *  This function will try to retrieve a Class object for the given
     *  connection method.  The first attempt will assume the Type
     *  specified in the config file is an abbreviated name (eg. Http)
     *  and use a translation table to find the appropriate class name.  If
     *  this fails, the Type specified in the config file will be assumed to
     *  be a fully qualified class name and an attempt to retrieve the Class
     *  object will be made again.  If it fails, an exception is thrown.
     *  Otherwise the Class object is instantiated and the connection method
     *  is initiailized and added to the connection pool.
     *
     *  @param connectionConfig The portion of the config file relevant to
     *  this connection method
     *  @param id The id associated with this connection method
     *  @param logChannel Logging utility class
     *  @exception ConnectionMethodException if a specified connection
     *  method does not exist or the connection method fails to initialize
     *  itself properly.
     **/
    protected void addConnectionMethod(Config connectionConfig,
				       String id)
	throws ConnectionMethodException {
	if (connectionConfig == null) {
	    throw new ConnectionMethodException("No Connection Config section found.");
	}

	Class connectionMethodClass = null;
	boolean classNotFound = false;
	String connectionTypeString = null;
	try {
	    connectionTypeString = connectionConfig.getString(TYPE);
	} catch (ConfigException e) {
	    logChannel.write(Logger.ERROR,
	       "Internal error parsing config file (section \"Connection\").", e);
	    throw new ConnectionMethodException(e);
	}

	// Assume abbreviated connection method name and try to get the Class
	// object from translation table.
	try {
	    if (connectionTable.containsKey(connectionTypeString)) {
		connectionMethodClass =
		    Class.forName((String)
				  connectionTable.get(connectionTypeString));
	    } else {
		logChannel.write(Logger.DEBUG,
	            "Abbreviated connection type \"" + connectionTypeString +
		    "\" not found, assuming fully qualified class name.");
		classNotFound = true;
	    }
	} catch (ClassNotFoundException e) {
	    logChannel.write(Logger.DEBUG,
		"Abbreviated connection type \"" + connectionTypeString +
		"\" not found, assuming fully qualified class name.", e);
	    classNotFound = true;
	}
	// If class load failed, assume fully qualified class name in config
	// and try again to retrieve Class object
	if (classNotFound) {
	    try {
		connectionMethodClass =
		    Class.forName(connectionTypeString);
	    } catch (ClassNotFoundException e) {
		logChannel.write(Logger.ERROR,
				 "Unable to create connection method \"" +
				 connectionTypeString +
				 "\" Class not found in CLASSPATH.", e);
		throw new ConnectionMethodException(e);
	    }
	}

	ConnectionMethod connectionMethod = null;
	try {
	    logChannel.write(Logger.DEBUG,
			     "Attempting to instantiate Class object \"" +
			     connectionMethodClass.getName() + "\".");
	    connectionMethod =
		(ConnectionMethod)connectionMethodClass.newInstance();
	    logChannel.write(Logger.DEBUG,
			     "Adding connection method \"" +
			     connectionMethodClass.getName() +
			     "\" to the connection pool.");
	    add(id, connectionMethod);
	    logChannel.write(Logger.DEBUG,
			     "Attempting to initialize connection method \"" +
			     connectionMethodClass.getName() + "\".");
	    ServletContainer sc = (ServletContainer)
		    MultiServer.getService("ServletContainer");
	    connectionMethod.initialize(connectionConfig, id,
					sc.getServletManager(),
					sc.getFilterManager());
	} catch (InstantiationException e) {
	    logChannel.write(Logger.ERROR,
			     "Unable to instantiate connection method \"" +
			     connectionMethodClass.getName() + "\".", e);
	    throw new ConnectionMethodException(e);
	} catch (IllegalAccessException e) {
	    logChannel.write(Logger.ERROR,
			     "Unable to access connection method \"" +
			     connectionMethodClass.getName() + "\".", e);
	    throw new ConnectionMethodException(e);
	}
    }

    /**
     *   Add a ConnectionMethod to the set. It will be referred to by the
     *   given identifer string.
     *
     *   @param connectionID The name to use to refer to the ConnectionMethod.
     *   @param cm The ConnectionMethod to store.
     *   @exception ConnectionMethodException If the ID is already in use.
     */
    public void add(String connectionID, ConnectionMethod cm)
        throws ConnectionMethodException {
        synchronized (theConnections) {
            if (theConnections.containsKey(connectionID))
                throw new ConnectionMethodException("Unable to add. " +
                    "The connection ID " + connectionID +
                    " is already in use.");
            TableEntry t = new TableEntry();
            t.connectionID = connectionID;
            t.cm = cm;
            theConnections.put(connectionID, t);
        }
    }


    /**
     *   Look up a ConnectionMethod from the set, by its identifier string.
     *
     *   @param connectionID The symbolic name to look up.
     *   @return The ConnectionMethod, or null if the name is not found.
     */
    public ConnectionMethod get(String connectionID) {
        synchronized (theConnections) {
            TableEntry t = (TableEntry) theConnections.get(connectionID);
            if (t == null)
                return null;
            return t.cm;
        }
    }


    /**
     *   Get all the names of the ConnectionMethods currently in the set.
     *
     *   @return All the connectionIDs
     */
    public String[] getIDs() {
        synchronized (theConnections) {
            int n = theConnections.size();
            String[] results = new String[n];
            Enumeration e = theConnections.elements();
            int counter = 0;
            while (e.hasMoreElements()) {
                TableEntry t = (TableEntry) e.nextElement();
                results[counter++] = t.connectionID;
            }
            return results;
        }
    }

    /**
     * Add any additional connection methods defined in the config file as
     * ConnectionMethod.<connection id> = <fully qualified class name>
     **/
    private void addAdditionalConnectionMethods(Config config)
	throws ConnectionMethodException {
	KeywordValueTable connectionMethod = null;
	try {
	    connectionMethod = config.getSection(CONNECTION_METHOD);
        } catch (KeywordValueException e) {
            logChannel.write(Logger.DEBUG,
			     "Section \"" + CONNECTION_METHOD +
			     "\" not found, using only default connection methods", e);
        } catch (Exception e) {
            logChannel.write(Logger.ERROR,
			     "Internal error parsing config file (section \"" +
			     CONNECTION_METHOD + "\").", e);
	    throw new ConnectionMethodException(e);
        }

	if (connectionMethod != null) {
	    String[] connectionKeys = connectionMethod.keys();
	    for (int i = 0; i < connectionKeys.length; i++) {
		try {
		    if (!connectionTable.containsKey(connectionKeys[i])) {
			try {
			    Class.forName(connectionMethod.getString(connectionKeys[i]));
			    connectionTable.put(connectionKeys[i],
						connectionMethod.getString(connectionKeys[i]));
			} catch (ClassNotFoundException e) {
			    logChannel.write(Logger.ERROR,
					     "Illegal " + CONNECTION_METHOD +
					     " defined, class \"" +
					     connectionMethod.getString(connectionKeys[i]) +
					     "\" not found.", e);
			    throw new ConnectionMethodException(e);
			}
		    } else {
			logChannel.write(Logger.ERROR,
					 "Illegal " + CONNECTION_METHOD +
					 " section in config file (key \"" +
					 connectionKeys[i] +
					 "\" already defined).");
			throw new ConnectionMethodException(
					 "Illegal " + CONNECTION_METHOD +
					 " section in config file (key \"" +
					 connectionKeys[i] +
					 "\" already defined).");
		    }
		} catch (KeywordValueException e) {
		    logChannel.write(Logger.ERROR,
				     "Illegal " + CONNECTION_METHOD +
				     " section in config file (key \"" +
				     connectionKeys[i] + "\" not found).", e);
		    throw new ConnectionMethodException(e);
		}
	    }
	}
    }

    /**
     * Return the hashtable of 'default' connection methods plus any
     * additional connection methods found in the conf file maked as
     * ConnectionMethod.<connection id> = <fully qualified class name>
     *
     * @return All the 'default' connection methods held in connectionTable
     *         plus those found in the config file
     **/
    public Hashtable getDefaultConnectionMethods() {
	return connectionTable;
    }

    /**
     *   Remove a ConnectionMethod from the set.
     *
     *  @param connectionID The name of the ConnectionMethod to remove.
     *  @exception ConnectionMethodException If the ID is not found.
     */
    public void delete(String connectionID) throws ConnectionMethodException {
        synchronized (theConnections) {
            if (!theConnections.containsKey(connectionID))
                throw new ConnectionMethodException("Unable to delete. " +
                        "Id " + connectionID + " not found.");
            theConnections.remove(connectionID);
        }
    }


    /**
     *   Generate a unique name that is not currently in the table.
     *   This may be used when adding a new connection method and you
     *   don't have a specific ID string you want to use. The names
     *   generated by this function are unique across all calls to this
     *   function, not just unique given the current set of connection
     *   methods. Therefore it is safe to get a name and use it later.
     *   The name will not be given out again, so the name should be
     *   unused when you want to use it to add a connection method.
     *
     *   @return A connection method ID that is not currently in use.
     */
    public String getUniqueName() {
        synchronized (theConnections) {
            while (true) {
                String result = "CM_" + nameCounter++;
                if (!theConnections.containsKey(result))
                    return result;
            }
        }
    }


    /**
     * Asks each connection method to write itself out to the config file
     *
     * @param config The config file
     **/
    public void createConnectionSection(Config config)
	throws ConfigException, KeywordValueException {
	String[] ids = getIDs();
	for (int i=0; i<ids.length; i++) {
	    ConnectionMethod connectionMethod = get(ids[i]);
            String base = CONNECTION + "." + ids[i] + ".";
            if (connectionMethod == null) {
                continue;
	    }
	    connectionMethod.writeToConfig(config, base);
	}
    }

    /**
     *   Returns a multi-line list of the contents of this manager,
     *   connection methods and channels.
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(
                                    "ConnectionMethodManager contents:\n");
        synchronized (theConnections) {
            Enumeration e = theConnections.elements();
            while (e.hasMoreElements()) {
                TableEntry t = (TableEntry) e.nextElement();
                buf.append("  ");
                buf.append(t.connectionID);
                buf.append(" = ");
                if (t.cm != null) {
                    buf.append(t.cm.toString());
                    buf.append("\n");
                } else {
                    buf.append("null\n");
                    continue;
                }
                String[] chIds = t.cm.getChannelIDs();
                for (int i=0; i<chIds.length; i++) {
                    buf.append("    ");
                    buf.append(chIds[i]);
                    buf.append("\n");
                }
            }
        }
        return buf.toString();
    }

}
