/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraDirectorHandler.java,v 1.7.2.4.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.EnhydraDirector;

import java.net.*;
import java.io.*;
import java.util.Hashtable;
import javax.servlet.*;
import javax.servlet.http.*;
import com.lutris.logging.*;
import org.enhydra.servlet.connectionMethods.*;
import org.enhydra.servlet.filter.Filter;
import org.enhydra.servlet.filter.RequestFilter;
import org.enhydra.servlet.filter.ResponseFilter;
import com.lutris.util.ExceptionUtils;
import com.lutris.util.ChainedException;
import com.lutris.util.Base64Encoder;
import com.lutris.appserver.server.session.StandardSessionKeyGen;
import org.enhydra.servlet.ServletIOException;
import com.lutris.multiServer.MultiServer;
import org.enhydra.servlet.ServletContainer;
import org.enhydra.servlet.tomcat.EnhydraRequest;
import org.apache.tomcat.core.Request;
import org.apache.tomcat.core.Response;


/**
 * This class processes servlet requests from the EnhydraDirectorConnectionMethod.
 *
 * @version $Revision: 1.7.2.4.2.1 $
 * @author Mark Sanguinetti
 * @author Shawn McMurdo
 * @author John Marco
 */
class EnhydraDirectorHandler implements Runnable {

    /**
     * The logging channel.
     */
    private static LogChannel logChannel =
	Logger.getCentralLogger().getChannel("Multiserver");

    /**
     * The EnhydraDirector Servlet Request object.
     */
    EnhydraDirectorConnection	request;

    /**
     * The EnhydraDirector Servlet Response object.
     */
    EnhydraDirectorConnection	response;

    /**
     * Reference to the EnhydraDirectorConnectionMethod containing channel data.
     */
    private EnhydraDirectorConnectionMethod connectionMethod;

    /**
     * Connection object.
     */
    EnhydraDirectorConnection	conn;

    /**
     * Constructor.
     *
     * @param acm The EnhydraDirectorConnectionMethod containing channel data.
     */
    EnhydraDirectorHandler(EnhydraDirectorConnectionMethod cm)
    {
        connectionMethod = cm;
    }

    /**
     * Hard-coded debugging flag. MUST BE FALSE IN PRODUCTION!
     */
    private static final boolean debug = false;

    /**
     * Runs the EnhydraDirectorHandler.
     */
    public void run() {
        while ((conn = connectionMethod.getEnhydraDirectorConnection()) != null) {
            try {
		// Enhydra connection context implements both req. and resp.
		request = conn;
		response = conn;
		conn.authenticate();
		conn.activate();

		int listenerPort =
		    connectionMethod.listenerSocket.getLocalPort();
		InetAddress ia = InetAddress.getLocalHost();
		byte[] listenerIP = ia.getAddress();

		logChannel.write(Logger.DEBUG,
			       "EnhydraDirectorHandler setting IP:Port token to \"(" +
			       ia.getHostAddress() + ":" + listenerPort +
			       ")\"");

		// convert ip and port to a base64 encoded 6 byte array
		byte[] listenerIPPort = new byte[6];
		for (int i = 0; i < 4; i++) {
		    listenerIPPort[i] = listenerIP[i];
		}
		listenerIPPort[4] = (byte)((listenerPort >> 8) & 0xff);
		listenerIPPort[5] = (byte)(listenerPort & 0xff);

		logChannel.write(Logger.DEBUG,
			     "EnhydraDirectorHandler base64 encoding IP:Port token " +
			     Base64Encoder.toBase64SessionKeyString(listenerIPPort));
		conn.setIPPortToken(Base64Encoder.toBase64SessionKeyString(listenerIPPort));
		conn.setSessionAffinity(connectionMethod.getSessionAffinity());
                processRequest();
            } catch (Throwable e) {
                logChannel.write(Logger.ERROR, "Error in Enhydra Handler", e);
            }
	    try {
		conn.closeInput();
	    } catch (Throwable t) {
		if (debug) {
		    logChannel.write(Logger.DEBUG,
			"Final CloseInput() Exception: " + t);
		}
	    }
	    try {
		conn.closeOutput(); // Close and recycle to OPEN state
	    } catch (Throwable t) {
		if (debug) {
		    logChannel.write(Logger.DEBUG,
			"Final CloseOutput() Exception: " + t);
		}
	    }
	    try {
		// XXX Right now, just close the connection.
		conn.closeConnection();
	    } catch (Throwable t) {
		if (debug) {
		    logChannel.write(Logger.DEBUG,
			"Final closeConnection() Exception: " + t);
		}
	    }
        }
    }

    /**
     * Handle an error from a servlet service.  This logs the error and reports
     * the problem to the client, if possible.  It also checks if this is an
     * I/O error from the client, indicating that the client has most likely
     * disconnected, in this case, it log the error at the DEBUG level.
     *
     * @param except The exception that occured.
     * @param servletId The servlet that got the error.
     * @param response The response object.
     */
    private
    void handlerServiceError(Throwable except, String servletId,
			     HttpServletResponse response)
    {

        boolean clientIOError = false;

        // Ignore if client I/O error.
        Throwable scan = except;
        while (scan != null) {
            if ((scan instanceof ServletIOException)
                && ((ServletIOException)scan).isTrueIOException()) {
                clientIOError = true;
                break;
            }
            if (scan instanceof ChainedException) {
                scan = ((ChainedException)scan).getCause();
            } else {
                break;
            }
        }

        if (clientIOError) {
            logChannel.write(Logger.DEBUG,
		    "Client appeared to drop connection: " + servletId);
        } else {
            logChannel.write(Logger.ERROR, "Error from servlet: " + servletId,
                             except);
        }

        // Always send error, just in case client didn't really drop.
        try {
            response.sendError(response.SC_NOT_FOUND,
                               ConnectionMethod.errorHtml);
        } catch(Exception except2) {
            // Already logged; so don't report another error if
            // write fails.
        }
    }

    /**
     * Process a request.
     */
    private void processRequest() throws IOException {
        String uri = request.getRequestURI();
        Channel chan = connectionMethod.getChannelByURLPrefix(uri);

        if (chan == null) {
            // Can not serve files.
            response.sendError(response.SC_NOT_FOUND,
                               ConnectionMethod.noChannelHtml);
            logChannel.write(Logger.INFO,
                             "request for files, only servlets supported");
            return;
        }

        // Is a valid request for a servlet.
        if (!chan.getEnabled()) {
            // Servlet not enabled.
            response.sendError(response.SC_MOVED_TEMPORARILY,
                               ConnectionMethod.disabledChannelHtml);
            logChannel.write(Logger.WARNING, "channel is disabled");
            return;
        }

        // Get channel and url data
        String urlPrefix = chan.getURLPrefix();
        int len = uri.length();
        if (urlPrefix.length() < uri.length()) {
            len = urlPrefix.length();
        }

        String servletPath = uri.substring(0, urlPrefix.length());

        // XXX don't have the servlet to get the ServletContext
        request.setServletInfo(servletPath, null);

        try {
            chan.incrRequestCount();
            String[] filterIds = chan.getFilterIDs();
            HttpServletRequest lastReq = request;
	    HttpServletResponse lastRes = response;
            RequestFilter wrapReq = null;
            ResponseFilter wrapRes = null;
            boolean filtered = false;
            if (filterIds.length > 0) {

                for (int i = 0; i < filterIds.length; i++) {
                    Filter filter = connectionMethod.filterManager.get(filterIds[i]);
                    if (filter != null) {
                        wrapReq = filter.wrapRequest(lastReq, lastRes);
                        wrapRes = filter.wrapResponse(lastReq, lastRes);
                        lastReq = wrapReq;
                        lastRes = wrapRes;
                        filtered = true;
                    }
                }
            }

            ServletContainer sc = (ServletContainer)
                    MultiServer.getService("ServletContainer");

            EnhydraRequest treq = new EnhydraRequest();
            org.apache.tomcat.core.Response tres =
                    new org.apache.tomcat.core.Response();
            if (filtered) {
                // X22 set adapter
                treq.setRequestAdapter(wrapReq);
                tres.setResponseAdapter(wrapRes);
            } else {
                // X22 set adapter
                treq.setRequestAdapter(request);
                tres.setResponseAdapter(response);
            }
	    sc.getServletManager().service(chan, treq, tres);

        } catch (Throwable except) {
            handlerServiceError(except, chan.getServletID(), response);
        }
    }
}

