/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraDirectorInputStream.java,v 1.3.2.1.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.EnhydraDirector;

import javax.servlet.ServletInputStream;
import java.util.Enumeration;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.Socket;
import org.enhydra.servlet.ServletIOException;

/**
 * InputStream interface for the Enhydra connection method.
 * <P>
 * Implemention of <code>ServletInputStrem</code> to be passed to 
 * servlets to allow them to receive the incoming HTTP request entity.
 * This input stream and its underlying protocol receiver does not buffer
 * input.  Incoming entity bytes are read directly from the connected
 * socket into the receiving buffer.
 *
 * @author John Marco
 * @version $Revision: 1.3.2.1.2.1 $
 * @see HttpRequest
 * @see javax.servlet.ServletInputStream
 */
class EnhydraDirectorInputStream extends ServletInputStream
{
    /**
     * Enhydra protocol connection instance.
     * <P>
     * Handles the actual communication with the remote client module.
     */
    private EnhydraDirectorConnection conn = null;

    /**
     * Create a new instance with the given connection object.
     * 
     * @param connection EnhydraDirector protocol connection context.
     */
    public
    EnhydraDirectorInputStream(EnhydraDirectorConnection connection)
    {
	conn = connection;
    }

    /**
     * Reads the next byte of incoming entity data.
     * <P>
     * Returns the byte as an <code>int</code> with a value between
     * <code>0</code> and <code>255</code>.  Returns <code>-1</code> on
     * end-of-data.
     *
     * @return     The next byte of data, or <code>-1</code> if the end of the
     *             stream is reached.
     * @exception IOException If an I/O error occurs.
     */
    public
    int read()
    throws IOException
    {
	return conn.read();
    }

    /**
     * Read incoming entity data into part or all of a byte array.
     * <P>
     * This is a more efficient override of the standard <code>
     * InputStream</code> method.  It reads data directly from the
     * socket into the buffer without repeated calls to the single-byte
     * <code>read()</code>.
     * <P>
     * Regardless of the value of <code>off</code> and <code>len</code>,
     * this method will not go beyond the end of <code>buf</code>.
     * 
     * @param buf The byte array to receive the data.
     * @param off The offset within the byte array at which the first
     * 		  data byte is to be placed.
     * @param len The maximum number of bytes to receive.
     * @return    The number of bytes received, or <code>-1</code> if 
     * 		  end-of-data is reached.
     * @exception IOException If an I/O error occurs.
     */
    public
    int read(byte[] buf, int off, int len)
    throws IOException
    {
	return conn.read(buf, off, len);
    }

    /**
     * Read incoming entity data into a byte array.
     * <P>
     * This is an efficient override of the standard <code>
     * InputStream</code> method.  It reads data directly from the
     * socket into the buffer without repeated calls to the single-byte
     * <code>read()</code>.
     * 
     * @param buf The byte array to receive the data.
     * @param off The offset within the byte array at which the first
     * 		  data byte is to be placed.
     * @param len The maximum number of bytes to receive.
     * @return    The number of bytes received, or <code>-1</code> if 
     * 		  end-of-data is reached.
     * @exception IOException If an I/O error occurs.
     */
    public
    int read(byte[] buf)
    throws IOException
    {
	return conn.read(buf, 0, buf.length);
    }

    /**
     * Skip incoming entity data.
     * <P>
     * Reads and discards <code>n</code> bytes of incoming entity data.
     * If end-of-data is reached, may skip fewer than <code>n</code> bytes
     * of data.  If no bytes are skipped before end-of-data is reached,
     * returns <code>-1</code>.
     * 
     * @param n The number of bytes to discard.
     * @return  The number of bytes skipped, or <code>-1</code> if at 
     *          end-of-data.
     * @exception IOException If an I/O error occurs.
     */
    public
    long skip(long n)
    throws IOException
    {
	return conn.skip(n);
    }

    /**
     * Returns the number of bytes that can be read from this input 
     * stream without blocking.
     *
     * @return     The number of bytes that can be read from this input stream
     *             without blocking.
     * @exception  IOException If an I/O error occurs.
     */
    public
    int available()
    throws IOException
    {
	return conn.inputAvailable();
    }

    /**
     * Closes this input stream.
     * <P>
     * Discards any remaining bytes in the incoming request entity
     * and marks the connection's input stream as "closed" so that
     * further read calls will fail.
     *
     * @exception  IOException If an I/O error occurs.
     */
    public
    void close()
    throws IOException
    {
	conn.closeInput();
    }

    /**
     * Resets this input stream for another connection.
     * <P>
     * Does nothing right now, since this low level stream
     * is unbuffered and all state is kept in the <code>conn</code>
     * object, which is always reset for each new response.
     * 
     * @exception  IOException If an I/O error occurs.
     */
    public
    void reset()
    throws IOException
    {
	if (false) throw new IOException();
	return;
    }
}

