/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraDirectorProtoOutputStream.java,v 1.3.2.1.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.EnhydraDirector;

import javax.servlet.ServletOutputStream;
import java.util.Enumeration;
import java.io.OutputStream;
import java.io.IOException;
import java.net.Socket;
import org.enhydra.servlet.ServletIOException;

/**
 * The basic implementation of an output stream for EnhydraDirector.
 * <P>
 * This implementation is intended to be used by the Enhydra
 * connection method's <code>EnhydraDirectorOutputStream</code> class.
 * If buffering is needed, this class can be wrappered in a
 * <code>BufferedOutputStream</code>.
 *
 * @author John Marco
 * @version $Revision: 1.3.2.1.2.1 $
 * @see EnhydraResponse
 * @see javax.servlet.ServletOutputStream
 */
class EnhydraDirectorProtoOutputStream extends OutputStream
{
    /**
     * Internal EnhydraDirector protocol driver instance.
     */
    private EnhydraDirectorConnection conn;

    /**
     * Boolean indicating whether of not bytes have been sent
     * int the request body.
     */
    private boolean noDataSent;

    /**
     * Construct an output stream associate with a response object.
     * 
     * @param conn EnhydraDirector protocol driver instance.
     */
    protected
    EnhydraDirectorProtoOutputStream(EnhydraDirectorConnection conn)
    {
        this.conn = conn;
	noDataSent= true;
    }

    /**
     * Mark the parent EnhydraResponse object as "committed". 
     * <P>
     * That means that data bytes have actually been sent "over the
     * wire" and that headers and status can no longer be set.
     */
    private
    void commit()
    throws IOException
    {
	if (noDataSent) {
	    conn.commit();
	    noDataSent = false;
	}
    }

    /**
     * Write a single byte of response data.
     * <P>
     * This method is extremely inefficient, since it sends a separate
     * packet for each outgoing response byte.  Servlets should pass this
     * output stream to the constructor of a new <code>BufferedOutputStream
     * </code> if the single byte <code>write()</code> call is going 
     * to be used frequently.
     *
     * @param b Integer with a value between 0 and 255, inclusive.
     * @exception IOException If an I/O error occurs.
     */
    public
    void write(int b)
    throws IOException
    {
	if (noDataSent) commit();
	conn.write(b);
    }

    /**
     * Write an array of bytes to this output stream. 
     *
     * @param bytes Array of byte data to write to the stream.
     * @param off Starting offset in the array.
     * @param len Number of bytes to write.
     * @exception  IOException If an I/O error occurs.
     */
    public
    void write(byte[] bytes, int off, int len)
    throws IOException
    {
	if (noDataSent) commit();
	conn.write(bytes, off, len);
    }

    /**
     * Write an array of bytes to this output stream. 
     *
     * @param bytes Array of byte data to write to the stream.
     * @exception  IOException If an I/O error occurs.
     */
    public
    void write(byte[] bytes)
    throws IOException
    {
	if (noDataSent) commit();
	conn.write(bytes, 0, bytes.length);
    }

    /**
     * Closes this output stream and releases any system resources 
     * associated with the stream. 
     *
     * @exception  IOException If an I/O error occurs.
     */
    public
    void close()
    throws IOException
    {
	if (noDataSent) commit();
	conn.closeOutput();
    }

    /**
     * Commits <code>conn</code>'s headers and immediately sends any
     * buffered data.
     *
     * @exception  IOException If an I/O error occurs.
     */
    public
    void flush()
    throws IOException
    {
	if (noDataSent) commit();
	super.flush();
    }
}
