/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpConnectionMethod.java,v 1.4.2.2.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.http;

import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.lutris.logging.*;
import org.enhydra.servlet.connectionMethods.*;
import org.enhydra.servlet.servletManager.ServletManager;
import org.enhydra.servlet.filter.FilterManager;
import org.enhydra.servlet.filter.Filter;
import com.lutris.util.*;



/**
 * This class implements the connection method used to accept request
 * for servlets from an HTTP client.
 *
 * @see org.enhydra.servlet.connectionMethods.BasicConnectionMethod
 * @version 1.8, 6/98
 * @author Mark Sanguinetti
 *
 */
public class HttpConnectionMethod
    extends StandardConnectionMethod implements Runnable {

    public static final String HTTP = ConnectionMethodManager.HTTP;
    public static final String TYPE = ConnectionMethodManager.TYPE;
    public static final String PORT = ConnectionMethodManager.PORT;
    public static final String BIND_ADDR = ConnectionMethodManager.BIND_ADDR;
    public static final String NUM_THREADS =
	ConnectionMethodManager.NUM_THREADS;
    public static final String QUEUE_SIZE = ConnectionMethodManager.QUEUE_SIZE;
    public static final String THREAD_TIMEOUT =
	ConnectionMethodManager.THREAD_TIMEOUT;
    public static final String CLIENT_TIMEOUT =
	ConnectionMethodManager.CLIENT_TIMEOUT;

    /*
     * Default interface to bind to (BindAddress).
     */
    private static final String DEFAULT_BIND_ADDRESS = "(All Interfaces)";

    /*
     * Default port for the connection.
     */
    private static final int DEFAULT_PORT = 80;

    /*
     * Default value for maximum number of threads.
     */
    public static final int DEFAULT_NUM_THREADS = 200;

    /*
     * Default value for maximum quesize.
     */
    public static final int DEFAULT_QUEUE_SIZE = 400;

    /*
     * Default value for threadTimeout, in milliseconds (5 minutes).
     */
    public static final int DEFAULT_THREAD_TIMEOUT = 5*60*1000;

    /*
     * Default value for clientTimeout (socket), in milliseconds (30 seconds).
     */
    public static final int DEFAULT_CLIENT_TIMEOUT = 1*30*1000;

    /*
     * Standard messages for responding to clients
     */
    protected final static String SERVER_BUSY_MSG =
        "<TITLE>Request Denied</TITLE><H1>Request Denied</H1> " +
        "This server can not process your request. Please try again later." ;

    /**
     * This is temporary, to support the CGI Connection
     * method, which is simply another flavor of this class.
     */
    protected String cmType = "HTTP";

    /**
     * The server port.
     */
    protected int port = DEFAULT_PORT;

    /**
     * IP address to bind to.
     */
    protected String bindAddress = DEFAULT_BIND_ADDRESS;

    /**
     * IP address to bind to.
     */
    private InetAddress bindInetAddress = null;


    /**
     * The server socket.
     */
    protected ServerSocket ss;

    /**
     * The logging channel.
     */
    protected LogChannel logChannel;

    /**
     * Thread handling the connections.
     */
    protected Thread dispatcher;

    /**
     * The queue for incoming connections.
     */
    protected Queue connections;

    /**
     * The maximum size of the connection queue.
     */
    protected int maxQueue = HttpConnectionMethod.DEFAULT_QUEUE_SIZE;

    /**
     * The thread group for HttpHandler threads.
     */
    protected ThreadGroup handlers;

    /**
     * The maximum number of HttpHandler threads.
     */
    protected int maxHandlers = HttpConnectionMethod.DEFAULT_NUM_THREADS;

    /**
     * The maximum time for an HttpHandler thread to wait for a connection.
     */
    protected int threadTimeout = HttpConnectionMethod.DEFAULT_THREAD_TIMEOUT;

    /**
     * The maximum time for a thread to keep a socket connection to the client
     * for.
     */
    protected int clientTimeout = HttpConnectionMethod.DEFAULT_CLIENT_TIMEOUT;

    /**
     * The backlog parameter for the server socket.
     */
    protected int backlog = 50;

    /**
     * The total number HttpHandler threads (active and waiting).
     */
    protected int numWaiting = 0;

    /**
     * A list of references to all HttpHandler threads (active and waiting).
     */
    protected Vector liveThreads;

    /**
     * Default constructor used for dynamic connection method creation.
     **/
    public HttpConnectionMethod () {
	logChannel = Logger.getCentralLogger().getChannel("Multiserver");
	logChannel.write(Logger.DEBUG, "HttpConnectionMethod instantiated.");
    }

    /**
     *  Shut down the connection method. Releases the socket, so the
     *  port will be free for others to use.
     *
     *  @exception ConnectionMethodException If an error occurs.
     */
    public void destroy() throws ConnectionMethodException {
        super.destroy();
        dispatcher.interrupt();
        try {
            ss.close();
            logChannel.write(Logger.INFO, cmType + " releasing port " + port);
        } catch (IOException e) {
            logError("Error shutting down",  e);
        }
    }

    /**
     *  Start up. Initially there are no channels.
     *  Parse the config file for connection information
     *  Saves the ServletManager and FilterManager for later use.
     *
     * This initialize is called from the ConnectionMethodManager when
     * the multiserver is started.  The config file contain all relevant
     * connection information.
     *
     *  @param connectionConfig The portion of the config file relevant to
     *  this connection method
     *  @param id The id associated with this connection method
     *  @param servletManager The ServletManager to use to convert servletIDs
     *  into Servlets.
     *  @param filterManager The FilterManager to use to convert filterIDs
     *  into Filters.
     *  @exception ConnectionMethodException If unable to start up.
     *  @see org.enhydra.servlet.filter.Filter
     *  @see org.enhydra.servlet.StandardConnectionMethod
     */
    public void initialize(Config connectionConfig,
			   String id,
			   ServletManager servletManager,
                           FilterManager filterManager)
        throws ConnectionMethodException {

	try {
	    String portStr = connectionConfig.getString(PORT);
	    port = Integer.parseInt(portStr);

	    if (connectionConfig.containsKey(NUM_THREADS)) {
		String nt = connectionConfig.getString(NUM_THREADS);
		maxHandlers = Integer.parseInt(nt);
		if (maxHandlers < 1) {
		    logChannel.write(Logger.ERROR,
				     NUM_THREADS + " must be at least 1 " +
				     "for connection method " + id + ".");
		    throw new ConnectionMethodException(
				     NUM_THREADS + " must be at least 1 " +
				     "for connection method " + id + ".");
		}
	    }

	    if (connectionConfig.containsKey(QUEUE_SIZE)) {
		String qs = connectionConfig.getString(QUEUE_SIZE);
		maxQueue = Integer.parseInt(qs);
		if (maxQueue < 1) {
		    logChannel.write(Logger.ERROR,
				     QUEUE_SIZE + " must be at least 1 " +
				     "for connection method " + id + ".");
		    throw new ConnectionMethodException(
				     QUEUE_SIZE + " must be at least 1 " +
				     "for connection method " + id + ".");
		}
	    }

	    if (connectionConfig.containsKey(CLIENT_TIMEOUT)) {
		String str = connectionConfig.getString(CLIENT_TIMEOUT);
		clientTimeout = Integer.parseInt(str) * 1000;
		if (clientTimeout < 1) {
		    logChannel.write(Logger.ERROR,
				     CLIENT_TIMEOUT + " must be at least 1 " +
				     "for connection method " + id + ".");
		    throw new ConnectionMethodException(
				     CLIENT_TIMEOUT + " must be at least 1 " +
				     "for connection method " + id + ".");
		}
	    }

	    if (connectionConfig.containsKey(THREAD_TIMEOUT)) {
		String str = connectionConfig.getString(THREAD_TIMEOUT);
		threadTimeout = Integer.parseInt(str) * 1000;
		if (threadTimeout < 1) {
		    logChannel.write(Logger.ERROR,
				     THREAD_TIMEOUT + " must be at least 1 " +
				     "for connection method " + id + ".");
		    throw new ConnectionMethodException(
				     THREAD_TIMEOUT + " must be at least 1 " +
				     "for connection method " + id + ".");
		}
	    }
            if (connectionConfig.containsKey(BIND_ADDR)) {
                String str = connectionConfig.getString(BIND_ADDR);
                if ((str != null) && (str.length() > 0)) {
                    setBindAddress(str);
                }
            }
	} catch (NumberFormatException e) {
	    logChannel.write(Logger.ERROR,
			     "Error reading config file, number format exception for connection method " + id + ".");
	    throw new ConnectionMethodException(e);
	} catch (ConfigException e) {
	    logChannel.write(Logger.ERROR,
			     "Error processing config file, config exception for connection method " + id + ".");
	    throw new ConnectionMethodException(e);
	}

        super.initialize(connectionConfig, id, servletManager, filterManager);

        try {
            ss = new ServerSocket(port, backlog,bindInetAddress);
        } catch (BindException be) {
            // A common problem
            logChannel.write(Logger.ERROR, "Bind error on port " + port +
                         " for connection method " + id + ".");
            logError("Bind Exception", be);
            throw new ConnectionMethodException(be);
        } catch (IOException e) {
            logError("Server Socket Error", e);
            throw new ConnectionMethodException(e);
        }

        handlers = new ThreadGroup("HttpHandlers");
        liveThreads = new Vector();
        connections = new Queue(maxQueue);

        logChannel.write(Logger.INFO, cmType + " listening on port: " + port);
        dispatcher = new Thread(this);
        dispatcher.start();
    }

    /**
     *  Saves the ServletManager and FilterManager for later use.
     *
     * This initialize is called from the multiserver admin, when a new
     * connection is being added or modified.  The config file does not
     * necessarly contain all connection information so this method
     * assumes connection attributes are/were set through the get/set
     * method pairs (as part of the bean environment).
     *
     *  @param id The id associated with this connection method
     *  @param servletManager The ServletManager to use to convert servletIDs
     *  into Servlets.
     *  @param filterManager The FilterManager to use to convert filterIDs
     *  into Filters.
     *  @exception ConnectionMethodException If unable to start up.
     *  @see org.enhydra.servlet.filter.Filter
     *  @see org.enhydra.servlet.StandardConnectionMethod
     */
    public void initialize(String id,
			   ServletManager servletManager,
                           FilterManager filterManager)
        throws ConnectionMethodException {
        super.initialize(id, servletManager, filterManager);

        try {
            ss = new ServerSocket(port, backlog);
        } catch (IOException e) {
            logError("Server Socket Error", e);
            throw new ConnectionMethodException(e);
        }

        handlers = new ThreadGroup("HttpHandlers");
        liveThreads = new Vector();
        connections = new Queue(maxQueue);

        logChannel.write(Logger.INFO, cmType + " listening on port: " + port);
        dispatcher = new Thread(this);
        dispatcher.start();
    }

    /**
     * Write out self into config fie
     *
     * @param config The config file
     * @param base The base key string
     **/
    public void writeToConfig(Config connectionConfig,
			      String base)
	throws ConfigException, KeywordValueException {
	connectionConfig.set(base + TYPE, HTTP);
	connectionConfig.set(base + PORT, "" + getPort());
	connectionConfig.set(base + NUM_THREADS, "" + getNumThreads());
	connectionConfig.set(base + QUEUE_SIZE, "" + getQueueSize());
	connectionConfig.set(base + CLIENT_TIMEOUT,
			     "" + getClientTimeout()/1000);
	connectionConfig.set(base + THREAD_TIMEOUT,
			     "" + getThreadTimeout()/1000);
    }

    /**
     * Runs the HttpConnectionMethod.
     */
    public void run() {
        // accept and dispatch connections until interrupted
        while (!dispatcher.isInterrupted()) {
            threadManager();
            // got handler, now accept new connection and add to queue
            Socket s = null;
            try {
                s = ss.accept();
		s.setSoTimeout(clientTimeout);
            } catch (SocketException e) {
                if (dispatcher.isInterrupted()) {
                    // It's ok, it's just destroy() killing us.
                } else {
                    // Must be a real error.
                    logError("SocketException", e);
                    continue;
                }
            } catch (IOException e) {
                logError("Accept Connection", e);
                throw new FatalExceptionError(e);
            }
            if (! putConnection(s)) {
                logError("Connection Queue full");
                try {
                    OutputStream out = s.getOutputStream();
                    writeString(out, SERVER_BUSY_MSG);
                    out.close();
                    s.close();
                } catch (IOException e) {
                    logError("Close Connection", e);
                }
            }
        }
    }

    /*
     * Inventories the active number HttpHandler threads (currently processing
     * and waiting) and creates a new one if the active number is less
     * than maxHandlers
     */
    private void threadManager() {
        synchronized (connections) {
            if (numWaiting < 0 ) {
                throw new RuntimeException("HttpConnectionMethod: Thread Handler");
            }
            if (numWaiting == 0) {
                // Inventory threads
                Enumeration e = liveThreads.elements();
                while (e.hasMoreElements()) {
                    Thread t = (Thread) e.nextElement();
                    if (!t.isAlive()) {
                        liveThreads.removeElement(t);
                    }
                }
                // If no waiting and total thread < max
                if (liveThreads.size() < maxHandlers) {
                    HttpHandler h = new HttpHandler(this);
                    Thread t = new Thread(handlers, h);
                    liveThreads.addElement(t);
                    t.start();
                }
            }
        }
    }

    /**
     * Returns an socket from the connection queue.
     *
     * @return a reference to the object if successful, null otherwise.
     */
    protected Socket getConnection() {
        synchronized (connections) {
            long initialTime = System.currentTimeMillis();
            Socket temp = null;

            while (true) {
                try {
                    temp = (Socket)connections.get();
                } catch (ArrayIndexOutOfBoundsException e) {
                    logError("Connection Queue error", e);
                    throw new FatalExceptionError(e);
                }
                if (temp != null) {
                    // return a socket if queue is not empty
                    return temp;
                } else {
                    // return null as the thread timed out
                    if ((System.currentTimeMillis() - initialTime) > threadTimeout) {
                        return null;
                    }
                    // wait for a connection
                    numWaiting++;
                    try {
                        connections.wait(threadTimeout);
                    } catch (InterruptedException e) {
                        logError("Interrupted", e);
                    }
                    numWaiting--;
                }
            }
        }
    }

    /**
     * Stores a socket in the connections queue.
     *
     * @param s socket to be stored
     * @return True if successful, false otherwise.
     * @exception IllegalMonitorStateException exception.
     */
    private boolean putConnection(Socket s) {
        boolean ret;
        synchronized (connections) {
            try {
                ret = connections.put(s);
                connections.notify();
            } catch (ArrayIndexOutOfBoundsException e) {
                logError("Connection Queue error", e);
                throw new FatalExceptionError(e);
            } catch (IllegalMonitorStateException e) {
                logError("Connection Queue error", e);
                throw new FatalExceptionError(e);
            }
        }
        return (ret == true ? true : false);
    }

    /**
     * Gets Channel data assocated with a given URL Prefix.
     *
     * @param url The URL Prefix.
     * @return The Channel associated with the URL Prefix.
     */
    public Channel getChannelByURLPrefix(String url) {
        Channel c = channelTable.getChannelByURLPrefix(url);
        return c;
    }

    /**
     * Gets a reference to the servlet associated with servletID.
     *
     * @param servletID The Servlet ID.
     * @return Reference to the servlet.
     */
    public Servlet getServlet(String servletID) {
        Servlet s = null;
        try {
            s = servletManager.get(servletID);
        } catch (ServletException e) {
            throw new FatalExceptionError(e);
        }
        return s;
    }

    /**
     * Returns the URL that a user would use to connect to this channel.
     *
     * @return A URL string.
     * @param channelID The channel
     * @exception ConnectionMethodException If an error occurs,
     * for example if the channel does not exist.
     */
    public String getChannelURL(String channelID)
        throws ConnectionMethodException {
        Channel c = channelTable.findID(channelID);
        if (c == null)
            throw new ConnectionMethodException("Unable to get channel URL: " +
                                                "Invalid channel id " + channelID);
        String result = "http://";
        try {
            result += InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            result += "localhost";
        }
        if (port != DEFAULT_PORT)
            result += ":" + port;
        String pre = c.getURLPrefix();
        if ((pre.length() > 0) && (pre.charAt(0) != '/'))
            result += "/";
        result += pre;
        return result;
    }

    /**
     * This connection method returns valid urls.
     *
     * @param channelID The channel
     * @return This function always returns true.
     * @exception ConnectionMethodException If an error occurs,
     *  for example if the channel does not exist.
     */
    public boolean channelURLIsValid(String channelID)
        throws ConnectionMethodException {
        return true;
    }

    /**
     * Compares the port number to determine equivalency
     *
     *  @param compareObject The object to compare this to
     *  @exception ConnectionMethodException If an error occurs.
     **/
    public boolean equivalent(ConnectionMethod compareObject) {
	HttpConnectionMethod httpCompareObject = null;
	if (!(compareObject instanceof HttpConnectionMethod)) {
	    return false;
	} else {
	    httpCompareObject = (HttpConnectionMethod)compareObject;
	}
	if (httpCompareObject.getPort() == getPort()) {
	    return true;
	} else {
	    return false;
	}
    }

    // Get and set methods for bean introspection
    public String getType() {
	return cmType;
    }

    public void setPort(int port) {
	this.port = port;
    }
    public int getPort() {
        return port;
    }

    /**
     * Gets the current Bind address, if any.
     */
    public String getBindAddress() {
        return this.bindAddress;
    }

    /**
     * Sets the current Bind address, if any.
     */
    public void setBindAddress(String addr) {
        String bhost = (addr != null) ? addr.trim().toLowerCase() : "";
        InetAddress baddr = null;
        if ( (bhost.length() > 0) && (!bhost.startsWith("(")) && (!bhost.equals("all"))) {
            try {
                baddr = InetAddress.getByName(bhost);
            } catch (UnknownHostException e) {
                baddr = null;
            }
        }
        if (baddr != null) {
            this.bindAddress = bhost;
            this.bindInetAddress = baddr;
        } else {
            this.bindAddress = DEFAULT_BIND_ADDRESS;
            this.bindInetAddress = null;
        }
    }


    public int getMaxHandlers() {
	return maxHandlers;
    }
    public void setMaxHandlers(int maxHandlers) {
	this.maxHandlers = maxHandlers;
    }

    public int getMaxQueue() {
	return maxQueue;
    }
    public void setMaxQueue(int maxQueue) {
	this.maxQueue = maxQueue;
    }

    /**
     * Gets the maximum number of handler thread associated with this connection method.
     *
     *@return numThreads.
     */
    public int getNumThreads() {
        return getMaxHandlers();
    }

    /**
     * Gets the maximum queue size associated with this connection method.
     *
     *@return queueSize.
     */
    public int getQueueSize() {
        return getMaxQueue();
    }

    /**
     * Gets the client timeout associated with this connection method.
     *
     *@return clientTimeout.
     */
    public int getClientTimeout() {
        return clientTimeout;
    }
    public void setClientTimeout(int clientTimeout) {
	this.clientTimeout = clientTimeout;
    }

    /**
     * Gets the thread timeout associated with this connection method.
     *
     *@return threadTimeout.
     */
    public int getThreadTimeout() {
        return threadTimeout;
    }
    public void setThreadTimeout(int threadTimeout) {
	this.threadTimeout = threadTimeout;
    }

    /*
     * utility for writing to clients before init()
     * used only in once (for notifing of HTTP 0.9 requests)
     */
    private void writeString(OutputStream out, String s)
        throws IOException {
        out.write(s.getBytes());
    }

    /*
     * Utility for logging errors. The logged message will include the port,
     * connection method type and exception thrown.
     *
     * @param s Error message to be logged
     * @param e Exception to be logged
     */
    private void logError(String s, Exception e) {
        String msg = "Error in " + cmType + " Connection method: "
            + "Port: " + port + " : " + s + " : " + e;
        logChannel.write(Logger.ERROR, msg, e);

    }

    /*
     * Utility for logging errors. The logged message will include the port,
     * connection method type and exception thrown.
     *
     * @param s Error message to be logged
     */
    private void logError(String s) {
        String msg = "Error in " + cmType + " Connection method: "
            + "Port: " + port + " : " + s ;
        logChannel.write(Logger.ERROR, msg);

    }
}
