/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpHandler.java,v 1.10.2.2.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.http;

import java.net.*;
import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.lutris.logging.*;
import org.enhydra.servlet.connectionMethods.*;
import org.enhydra.servlet.filter.Filter;
import org.enhydra.servlet.filter.RequestFilter;
import org.enhydra.servlet.filter.ResponseFilter;
import com.lutris.util.ExceptionUtils;
import com.lutris.util.ChainedException;
import org.enhydra.servlet.ServletIOException;
import com.lutris.multiServer.MultiServer;
import org.enhydra.servlet.ServletContainer;
import org.enhydra.servlet.tomcat.EnhydraRequest;
import org.apache.tomcat.core.Request;
import org.apache.tomcat.core.Response;

/**
 * This class proccesses servlet requests from the HttpConnectionMethod.
 *
 * @version $Revision: 1.10.2.2.2.1 $
 * @auther Mark Sanguinetti
 */
class HttpHandler implements Runnable {
    /*
     * Standard messages for responding to clients
     */
    private final static String NOT_SUPPORTED_MSG =
        "<TITLE>Service Not Supported</TITLE><H1>Service Not Supported</H1> " +
        "This service requires HTTP 1.0 requests. " ;

    /*
     * The logging channel.
     */
    private static LogChannel logChannel =
	Logger.getCentralLogger().getChannel("Multiserver");

    /**
     * Request object.
     */
    HttpRequest request = new HttpRequest();

    /**
     * Response object.
     */
    HttpResponse response = new HttpResponse();

    /**
     * Reference to the Object containing channel data.
     */
     private HttpConnectionMethod connectionMethod;

    /**
     * Constructor.
     *
     * @param hcm The Object containing channel data.
     * @param c The Object containing the socket.
     */
    HttpHandler(HttpConnectionMethod cm) {
        connectionMethod = cm;
    }

    /**
     * Runs the HttpHandler.
     */
    public void run() {
        Socket socket;
        while ((socket = connectionMethod.getConnection()) != null) {
            try {
                processRequests(socket);
                socket.close();
            } catch (Throwable e) {
                logChannel.write(Logger.ERROR, "Error in HTTP Handler", e);
            }
        }
    }


    /**
     * Handle an error from a servlet service.  This logs the error and reports
     *
     * the problem to the client, if possible.  It also checks if this is an
     * I/O error from the client, indicating that the client has most likely
     * disconnected, in this case, it log the error at the DEBUG level.
     *
     * @param except The exception that occured.
     * @param servletId The servlet that got the error.
     * @param response The response object.
     */
    private void handlerServiceError(Throwable except, String servletId,
                                     HttpResponse response) {
        boolean clientIOError = false;

        // Ignore if client I/O error.
        Throwable scan = except;
        while (scan != null) {
            if ((scan instanceof ServletIOException)
                && ((ServletIOException)scan).isTrueIOException()) {
                clientIOError = true;
                break;
            }
            if (scan instanceof ChainedException) {
                scan = ((ChainedException)scan).getCause();
            } else {
                break;
            }
        }

        if (clientIOError) {
            logChannel.write(Logger.DEBUG, "Client appeared to drop connection: " + servletId);
        } else {
            logChannel.write(Logger.ERROR, "Error from servlet: " + servletId,
                             except);
        }


        // Always send error, just in case client didn't really drop.
        try {
            response.sendError(response.SC_NOT_FOUND,
                               ConnectionMethod.errorHtml);
        } catch(Exception except2) {
            // Already logged; so don't report another error if
            // write fails.
        }
    }


    /**
     * Process a single request.  Called multiple times for a given connection
     * for keep alive.
     */
    private void doARequest() throws IOException {
        // FIXME servlet runner did: check for HTTP version 1.0 or greater

        String uri = request.getRequestURI();
        Channel chan = connectionMethod.getChannelByURLPrefix(uri);


        if (chan == null) {
            // Can not serve files.
            response.sendError(response.SC_NOT_FOUND,
                               ConnectionMethod.noChannelHtml);
            logChannel.write(Logger.INFO,
                             "request for files, only servlets supported");
            return;
        }

        // Is a valid request for a servlet.
        if (!chan.getEnabled()) {
            // Servlet not enabled.
            response.sendError(response.SC_MOVED_TEMPORARILY,
                               ConnectionMethod.disabledChannelHtml);
            logChannel.write(Logger.WARNING, "channel is disabled");
            return;
        }

        // Get channel and url data
        String urlPrefix = chan.getURLPrefix();
        if (urlPrefix.equals("/"))
            urlPrefix="";

        int len = uri.length();
        if (urlPrefix.length() < uri.length()) {
            len = urlPrefix.length();
        }

        String servletPath = uri.substring(0, urlPrefix.length());

	// XXX don't have the servlet to get the ServletContext
        request.setServletInfo(servletPath, null);

        try {
            chan.incrRequestCount();

	    // Apply filters
            String[] filterIds = chan.getFilterIDs();
	    HttpServletRequest lastReq = request;
	    HttpServletResponse lastRes = response;
	    RequestFilter wrapReq = null;
	    ResponseFilter wrapRes = null;
	    boolean filtered = false;
//System.err.println("number of filters: " + filterIds.length);
//logChannel.write(Logger.INFO, "number of filters: " + filterIds.length);

            if (filterIds.length > 0) {
                for (int i = 0; i < filterIds.length; i++) {
                    Filter filter = connectionMethod.filterManager.get(filterIds[i]);
                    if (filter != null) {
                        wrapReq = filter.wrapRequest(lastReq, lastRes);
                        wrapRes = filter.wrapResponse(lastReq, lastRes);
			lastReq = wrapReq;
			lastRes = wrapRes;
			filtered = true;
                    }
                }
            }

	    ServletContainer sc = (ServletContainer)
		    MultiServer.getService("ServletContainer");

	    EnhydraRequest ereq = new EnhydraRequest();
	    Response tres =  new Response();
	    if (filtered) {
	        // X22 set adapter
	        ereq.setRequestAdapter(wrapReq);
	        tres.setResponseAdapter(wrapRes);
            } else {
	        // X22 set adapter
	        ereq.setRequestAdapter(request);
	        tres.setResponseAdapter(response);
	    }
            // get the session key from the cookies or url and store it.
           // ereq.obtainSessionId();
//            ereq.setContextPath(urlPrefix);
            sc.getServletManager().service(chan, ereq, tres);
        } catch (Throwable except) {
            handlerServiceError(except, chan.getServletID(), response);
        }
    }

    /**
     * Proccess one or more the servlet request on a single connection.
     */
    private void processRequests(Socket socket) throws IOException {
        request.setConnection(socket);
        try {
            response.setConnection(socket);

            while (request.next()) {
                boolean keepAliveRequested = request.keepAliveRequested();
                response.next(keepAliveRequested);
                doARequest();
                request.finish();
                response.finish();
                if (!keepAliveRequested || !response.keepAlivePossible()) {
                    break; // Can't do keep-alive
                }
            }
        } catch (ServletIOException except) {
            // Ignore client I/O errors.
            if (!except.isTrueIOException()) {
                logChannel.write(Logger.ERROR,
                                 "unhandled exception in HTTP connection manager",
                                 except);
            }
        } catch (Throwable except) {
            logChannel.write(Logger.ERROR,
                             "unhandled exception in HTTP connection manager",
                             except);
        } finally {
            try {
                request.close();
            } catch (Throwable except) {
                logChannel.write(Logger.ERROR, "close of HTTP request failed",
                                 except);
            }
            try {
                response.close();
            } catch (Throwable except) {
                logChannel.write(Logger.ERROR, "close of HTTP response failed",
                                 except);
            }
        }
    }
}
