/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpInputStream.java,v 1.4.12.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.http;

import javax.servlet.ServletInputStream;
import java.util.Enumeration;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.Socket;
import org.enhydra.servlet.ServletIOException;

/**
 * Implementation of a <CODE>ServletInputStream</CODE> for the HTTP
 * connection method.  This object is designed to be recycled for several
 * independent requests.
 *
 * @author Mark Diekhans
 * @version $Revision: 1.4.12.1 $
 * @see HttpRequest
 * @see javax.servlet.ServletInputStream
 */
class HttpInputStream extends ServletInputStream {
    /**
     * Size of input buffer.
     */
    private static final int bufferSize = 1024;

    /**
     * Socket associate with a set of requests.
     */
    private Socket socket;

    /**
     * Input stream for the socket.
     */
    private BufferedInputStream socketStream;

    /**
     * Count of bytes transfered.
     */
    long byteCount;

    /**
     * Initial size for the buffer used to read header lines.
     * The buffer will be reset to this size for the next request
     * if it has grown.
     */
    private static final int HEADER_BUF_SIZE = 256;

    /**
     * Maximum size for the buffer used to read header lines.
     */
    private static final int MAX_HEADER_BUF_SIZE = 32768;

    /**
     * Buffer used for header lines.
     */
    private byte[] headerBuf;

    /**
     * Associate a connected socket with this object.  The
     * <CODE>next</CODE> function must be called before any
     * input can be read.
     * @param requestSocket The HTTP connected socket.
     * @exception IOException If an error occured setting up the socket
     *  input stream.
     */
    protected void setConnection(Socket requestSocket) throws IOException {
        socket = requestSocket;
        try {
            socketStream = new BufferedInputStream(socket.getInputStream(),
                                                   bufferSize);
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Reset state to read the next request off the socket.
     */
    protected void next() {
        byteCount = 0L;
        if (headerBuf == null) {
            headerBuf = new byte[HEADER_BUF_SIZE];
        }
    }

    /**
     * Reads the next byte of data from this input stream. 
     *
     * @return     The next byte of data, or <code>-1</code> if the end of the
     *             stream is reached.
     * @exception  IOException If an I/O error occurs.
     */
    public int read() throws IOException {
        try {
            int sbyte = socketStream.read();
            if (sbyte != -1) {
                byteCount++;
            }
            return sbyte;
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Expand the header buffer.
     * @exception IOException If the buffer has reached the max size.
     */
    private void expandHeaderBuf() throws IOException {
        if (headerBuf.length == MAX_HEADER_BUF_SIZE) {
            throw new ServletIOException("HTTP request line exceeds maximum of "
                                         + MAX_HEADER_BUF_SIZE + " bytes");
        }
        int newLen = headerBuf.length * 2;
        if (newLen > MAX_HEADER_BUF_SIZE) {
            newLen = MAX_HEADER_BUF_SIZE;
        }
        byte[] newBuf = new byte[newLen];
        System.arraycopy(headerBuf, 0, newBuf, 0, headerBuf.length);
        headerBuf = newBuf;
    }

    /**
     * Read an ASCII HTTP header line.
     *
     * @return The header as a string or <CODE>null</CODE> if an end of
     *  header or file was reached.
     * @exception IOException If an I/O error occurs or if a legal
     *  HTTP header can not be read.
     */
    protected String readHeader() throws IOException {
        // Read into buffer until CR-LF
        int charCount = 0;
        int hbyte;
        int prevHbyte = 'X';  // Anything but CR.
        try {
            while ((hbyte = read()) != -1) {
                if ((prevHbyte == '\r') && (hbyte == '\n')) {
                    break;  // EOH
                }
                if (charCount == headerBuf.length) {
                    expandHeaderBuf();
                }
                headerBuf[charCount++] = (byte)(hbyte & 0xFF);
                prevHbyte = hbyte; 
            }
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }

        /*
         * EOH/EOF check.  A EOF with not bytes read or an line of only 
         * CRLF are the only legal EOH indicators.  Any other EOF indicates
         * that the CRLF was not read.
         */
        if (((hbyte == - 1) && (charCount == 0)) || (charCount == 1)) {
            return null;
        }
        if (hbyte == - 1) {
            throw new ServletIOException("Premature EOF in request HTTP headers");
        }
        return new String(headerBuf, 0, charCount);
    }

    /**
     * Returns the number of bytes that can be read from this input 
     * stream without blocking.
     *
     * @return     The number of bytes that can be read from this input stream
     *             without blocking.
     * @exception  IOException If an I/O error occurs.
     */
    public int available() throws IOException {
        try {
            return socketStream.available();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Closes this input stream and releases any system resources 
     * associated with the stream. 
     *
     * @exception  IOException If an I/O error occurs.
     */
    public void close() throws IOException {
        try {
            socketStream.close();
            socket.close();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
        // Don't keep around large buffers.
        if (headerBuf.length > HEADER_BUF_SIZE) {
            headerBuf = null;
        }
    }

    /**
     * Get the number of bytes that have been transfered since the last
     * <CODE>next()</CODE>.
     */
    protected long getByteCount() {
        return byteCount;
    }
}
