/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpResponse.java,v 1.4.12.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.http;

import javax.servlet.http.*;
import javax.servlet.ServletResponse;
import javax.servlet.ServletOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.Socket;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.TimeZone;
import java.util.Date;
import java.util.Enumeration;
import java.util.Locale;
import java.text.SimpleDateFormat;
import com.lutris.http.HttpCookieParser;
import com.lutris.http.HttpStatus;
import org.enhydra.servlet.ServletIOException;
import com.lutris.Lutris;
import org.apache.tomcat.core.ResponseAdapter;
import org.apache.tomcat.util.MimeHeaders;
import org.apache.tomcat.util.MimeHeaderField;

/**
 * Implementation of a HTTP servlet response for the HTTP connection
 * method.
 *
 * @author Mark Diekhans
 * @version $Revision: 1.4.12.1 $
 * @see javax.servlet.http.HttpServletResponse
 * @see javax.servlet.ServletResponse
 */
public class HttpResponse implements HttpServletResponse, ResponseAdapter {
    /**
     * Status code.
     */
    private int statusCode;

    /**
     * Status message to used to  generate the HTML page.
     */
    private String statusMsg;

    /**
     * Content length.
     */
    private int contentLength;

    /**
     * Table of headers (except cookies).
     */
    private HttpHeaders headers = new HttpHeaders();

    /**
     * Table of cookies
     */
    private Vector cookies = new Vector();

    /**
     * The socket associated with the request.
     */
    private Socket socket;

    /**
     * Output stream assoicated with the socket.
     */
    private HttpOutputStream outputStream = new HttpOutputStream(this);

    /**
     * Format for dates.
     */
    private static final SimpleDateFormat dateFormatter;

    /**
     * Has getOutputStream() been called on the current request?
     */
    private boolean getOutputStreamCalled;

    /**
     * Print writer for the stream
     */
    private PrintWriter writer;

    /**
     * Flag to indicate headers have been written.  Set *before* the
     * write and used so that the outputStream doesn't recursivelt call
     * us to write the headers again.
     */
    private boolean headersWritten;

    /**
     * Number of header bytes written.
     */
    private int headerBytesWritten;

    /**
     * Local host name; for defaulting Server: header.
     */
    private static final String hostName;

    /**
     * Class initializer.
     */
    static {
        dateFormatter = new SimpleDateFormat("EEE, dd MMM yyyy kk:mm:ss zzz");
        dateFormatter.setTimeZone(TimeZone.getTimeZone("GMT"));

        InetAddress inet;
        String localHostName;
        try {
            inet = InetAddress.getLocalHost();
            localHostName = inet.getHostName();
            if (localHostName == null) {
                localHostName = inet.getHostAddress();
            }
        } catch (UnknownHostException except) {
            localHostName = "server.host.unknown";
        }
        hostName = localHostName;
    }

    /**
     * Associate a connected socket with this object.  The
     * <CODE>next</CODE> function must be called before a
     * response can be processed.
     * @param responseSocket The HTTP connected socket.
     * @exception IOException If an error occured setting up the socket
     *  input stream.
     */
    protected void setConnection(Socket responseSocket) throws IOException {
        socket = responseSocket;
        outputStream.setConnection(responseSocket);
    }

    /**
     * Close the socket associate with this object.
     * @exception IOException If an error occured setting up the socket
     *  input stream.
     */
    protected void close() throws IOException {
        try {
            if (writer != null) {
                writer.close();
            }
            outputStream.close();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Sets the content length for this response.
     *
     * @param len The content length.
     */
    public void setContentLength(int len) {
        checkOutputBegun();
        contentLength = len;
        headers.set("Content-Length", Integer.toString(len));
    }

    /**
     * Sets the content type for this response.
     *
     * @param Type the content's MIME type.
     */
    public void setContentType(String type) {
        headers.set("Content-Type", type);
    }

    /**
     * Returns an output stream for writing binary response data.
     *
     * @exception IllegalStateException if getWriter has been
     *	called on this same request.
     * @exception IOException if an I/O exception has occurred.
     */
    public ServletOutputStream getOutputStream() throws IOException {
        if (writer != null) {
            throw new IllegalStateException("getWriter() has been called on this request; can't use getOutputStream()");
        }
        getOutputStreamCalled = true;
        return outputStream;
    }

    /**
     * Generate an error if output has already begun and headers can't
     * be modified.
     * @exception IllegalStateException If output has already been started.
     */
    private void checkOutputBegun() {
        if (outputStream.getByteCount() > 0) {
            throw new IllegalStateException("Output has already been written to HTTP response channel; can't modify headers or response status");
        }
    }

    /**
     * Returns a print writer for writing formatted text responses.
     *
     * @exception UnsupportedEncodingException if no such encoding can
     * be provided.
     * @exception IllegalStateException if getOutputStream has been
     *	called on this same request.
     * @exception IOException on other errors.
     */
    public PrintWriter getWriter() throws IOException {
        if (getOutputStreamCalled) {
            throw new IllegalStateException("getOutputStream() has been called on this request; can't use getWriter()");
        }
        if (writer == null) {
            writer = new PrintWriter(outputStream);
        }
        //FIXME.  Need to set encoding
        return writer;
    }

    /**
     * Returns the character set encoding used for this MIME body.
     */
    public String getCharacterEncoding () {
        return null; //FIXME
    }

    /**
     * Adds the specified cookie to the response. 
     *
     * @param cookie The Cookie to return to the client.
     */
    public void addCookie(Cookie cookie) {
        cookies.addElement(cookie);
    }

    /**
     * Checks whether the response message header has a field with
     * the specified name.
     * 
     * @param name The header field name.
     * @return True if the response message header has a field with
     * the specified name; false otherwise.
     */
    public boolean containsHeader(String name) {
        return headers.exists(name);
    }

    /**
     * Sets the status code and message for this response.
     * 
     * @param sc The status code.
     * @param sm The status message.
     */
    public void setStatus(int sc, String sm) {
        checkOutputBegun();
        statusCode = sc;
        statusMsg = sm;
    }

    /**
     * Sets the status code for this response.
     *
     * @param sc the status code
     */
    public void setStatus(int sc) {
        setStatus(sc, HttpStatus.getStatusMsg(sc));
    }

    /**
     * Adds a field to the response header with the given name and value.
     * 
     * @param name The name of the header field.
     * @param value The header field's value.
     */
    public void setHeader(String name, String value) {
        checkOutputBegun();
        headers.set(name, value);
    }

    /**
     * Adds a field to the response header with the given name and
     * integer value.
     *
     * @param name The name of the header field.
     * @param value The header field's integer value.
     */
    public void setIntHeader(String name, int value) {
        checkOutputBegun();
        headers.set(name, Integer.toString(value));
    }

    /**
     * Adds a field to the response header with the given name and
     * date-valued field.
     * 
     * @param name The name of the header field.
     * @param value The header field's date value.
     */
    public void setDateHeader(String name, long date) {
        checkOutputBegun();
        headers.set(name, dateFormatter.format(new Date(date)));
    }

    /**
     * Generate an HTML error page for sendError.
     */
    private String getErrorPage(int sc, String msg) {
        StringBuffer buf = new StringBuffer();

        //FIXME: Need to figure who generates <TITLE>  Should it be in msg or here.
        buf.append("<HTML><HEAD><TITLE>");
        buf.append(HttpStatus.getStatusMsg(sc));
        buf.append("</TITLE><CENTER><H1>");
        buf.append(HttpStatus.getStatusMsg(sc));
        buf.append("</H1></CENTER>");
        if (msg != null) {
            buf.append(msg);
            buf.append("<P>");
        }
        buf.append("</BODY></HTML>");
        return buf.toString();
    }

    /**
     * Sends an error response to the client using the specified status
     * code and descriptive message.
     * 
     * @param sc The status code
     * @param msg The generate a HTML response with this message.
     * @exception IOException If an I/O error has occurred. 
     */
    public void sendError(int sc, String msg) throws IOException {
        String page =  getErrorPage(sc, msg);

        // If output has not begun; do it right with headers; otherwise
        // just send HTML.
        try {
            if (!headersWritten) {
                setStatus(sc);
                setContentType("text/html");
                setContentLength(page.length());
                writeHeaders();
            }
            outputStream.println(page);
            outputStream.flush();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Sends an error response to the client using the specified
     * status code and a default message.
     * @param sc The status code.
     * @exception IOException If an I/O error has occurred.
     */
    public void sendError(int sc) throws IOException {
        sendError(sc, HttpStatus.getStatusMsg(sc));
    }

    /**
     * Generate an HTML page for sendRedirect.
     */
    private String getRedirectPage(String url) {
        StringBuffer buf = new StringBuffer();
        
        buf.append("<HTML><HEAD><TITLE>Document Has Moved</TITLE>");
        buf.append("<BODY><CENTER><H1>Document Has Moved</H1></CENTER>");
        buf.append("This document has moved to <A HREF=\"");
        buf.append(url);
        buf.append("\">");
        buf.append(url);
        buf.append("</A><P>");
        buf.append("</BODY></HTML>");
        return buf.toString();
    }

    /**
     * Sends a temporary redirect response to the client using the
     * specified redirect location URL.
     *
     * @param location the redirect location URL.
     * @exception IOException If an I/O error has occurred.
     */
    public void sendRedirect(String location) throws IOException {
        checkOutputBegun();
        
        try {
            String page = getRedirectPage(location);
            setStatus(HttpStatus.movedTemporarily);
            setContentType("text/html");
            setHeader("Location", location);
            setContentLength(page.length());
            writeHeaders();
            outputStream.println(page);
            outputStream.flush();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Encodes the specified URL by including the session ID in it,
     * or, if encoding is not needed, returns the URL unchanged.
     *
     * @param url The url to be encoded.
     * @return The encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     */
    public String encodeUrl (String url) {
	// X22 see encodeURL
        return url;  // Sessions not implemented.
    }

    /**
     * Encodes the specified URL for use in the
     * <code>sendRedirect</code> method.
     * 
     * @param url The url to be encoded.
     * @return The encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     */
    public String encodeRedirectUrl (String url) {
	// X22 see encodeRedirectURL
        return url;  // Sessions not implemented.
    }

    /**
     * Initialize for the next response on a socket. With keep-alive, multiple
     * responses can be sent on a single connection. 
     *
     * @param keepAliveRequested Indicates if a keep-alive connection has
     *  been requested by the client.
     * @exception IOException If an I/O error occurs.
     */
    protected void next(boolean keepAliveRequested) throws IOException {
        statusCode = HttpStatus.ok;
        statusMsg = HttpStatus.getStatusMsg(HttpStatus.ok);
        contentLength = -1;
        headers.clear();
        cookies.setSize(0);
        outputStream.next();
        getOutputStreamCalled = false;
        writer = null;
        headersWritten = false;
        headerBytesWritten = 0;
        if (keepAliveRequested) {
            setHeader("Connection", "Keep-Alive");
        }
    }

    /**
     * Default headers that don't exists.
     */
    private void defaultHeaders() {
        if (!headers.exists("Server")) {
            headers.add("Server", "Enhydra-MultiServer/"
                        + Lutris.getEnhydraVersion());

        }
        if (!headers.exists("Date")) {
            headers.add("Date", dateFormatter.format(new Date()));
        }
    }

    /**
     * Add cookies as header table.  We add this as a single header,
     * since some browsers can't handle multiple Set-Cookie headers
     * in a single response.
     */
    private void addCookieHeaders() {
        if (cookies.size() > 0) {
            StringBuffer value = new StringBuffer();
            for (int i = 0; i < cookies.size(); i++) {
                if (i > 0) {
                    value.append(", ");
                }
                value.append(HttpCookieParser.formatResponseCookie((Cookie)cookies.elementAt(i)));
            }
            headers.add("Set-Cookie", value.toString());
        }
    }
    

    /**
     * Write out the headers.  Does nothing if they have already been
     * written.
     *
     * @exception IOException If an I/O error occurs writing the headers.
     */
    protected void writeHeaders() throws IOException {
        if (outputStream.getByteCount() > 0) {
            throw new IllegalStateException("internal error: writeheaders() called after output has begun");
        }

        defaultHeaders();
        addCookieHeaders();

        // *MUST* set before any writes, so output stream can see it.
        headersWritten = true;

        try {
            outputStream.print("HTTP/1.0 ");
            outputStream.print(statusCode);
            outputStream.print(" ");
            outputStream.println(statusMsg);

            headers.write(outputStream);

            outputStream.println(); // End of header line

            // Remember what we wrote.
            headerBytesWritten = (int)outputStream.getByteCount();

            /*
             * WARNING: At one time we thought that writting the header and
             * body as a single write caused some version of some browsers
             * to double-load pages (netscape 4.05 on windozes and mac).
             * We had a flush here, which made things better, but not 
             * completely.  It turned out that the double-load problem
             * was to to a meta-tag contain a character-set specification.
             * In a later version of netscape (4.07 on windozes), having
             * a flush here caused `transfered interrupted' to occasionally
             * occur.  So we took it back out.  If your are about to add
             * a flush() here, beware, the world of browsers in an ugly
             * place.
             */
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }
        
    /**
     * Indicate the end of handing a request.  Force data to be written if
     * it hasn't already happened.
     *
     * @exception IOException If an I/O error occurs writing the headers.
     */
    protected void finish() throws IOException {
        try {
            if (headerBytesWritten == 0) {
                writeHeaders();
            }
            if (writer != null) {
                writer.flush();
            }
            outputStream.flush();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    
    /**
     * Determine if keep alive is possible.  That is content length
     * was set and is correct.  The finish() method must be called before
     * this is called.
     *
     * @return True if a keep alive is possible.
     */
    protected boolean keepAlivePossible() {
        String contentLenStr = (String)headers.getValue("Content-Length");
        if (contentLenStr == null) {
            return false;
        }
        int contentLength;
        try {
            contentLength = Integer.parseInt(contentLenStr);
        } catch (NumberFormatException except) {
            return false;  // bad number format
        }
        return (contentLength ==
                (outputStream.getByteCount() - headerBytesWritten));
    }

    /**
     * Check if headers have been written to the socket.
     * @return True if they have and no more can be written; false otherwise.
     */
    protected boolean haveHeadersBeenWritten() {
        return headersWritten;
    }
    

    public Locale getLocale() {
	// X22
	throw new Error("not implemented");
    }

    public void flushBuffer() {
	// X22
	throw new Error("not implemented");
    }

    public String encodeURL(String url) {
	// X22
	throw new Error("not implemented");
    }

    public boolean isCommitted() {
	// X22
	throw new Error("not implemented");
    }

    public int getBufferSize() {
	// X22
	throw new Error("not implemented");
    }

    public void reset() {
	// X22
	throw new Error("not implemented");
    }

    public String encodeRedirectURL(String url) {
	// X22
	throw new Error("not implemented");
    }

    public void addIntHeader(String name, int value) {
	// X22
	throw new Error("not implemented");
    }

    public void addDateHeader(String name, long value) {
	// X22
	throw new Error("not implemented");
    }

    public void addHeader(String name, String value) {
	headers.add(name, value);
    }

    public void setBufferSize(int size) {
	// X22
	throw new Error("not implemented");
    }

    public void setLocale(Locale locale) {
	// X22
	throw new Error("not implemented");
    }

    // X22 ResponseAdapter
    // XXX This one or multiple addHeader?
    // Probably not a big deal - but an adapter may have
    // an optimized version for this one ( one round-trip only )
    public void addMimeHeaders(MimeHeaders headers) throws IOException {
	int size = headers.size();
        for (int i = 0; i < size; i++) {
            MimeHeaderField h = headers.getField(i);
            addHeader(h.getName(), h.getValue());
        }
    }

    // X22 ResponseAdapter
    /** Signal that we're done with a particular request, the
	server can go on and read more requests or close the socket
    */
    public void endResponse() throws IOException {
	// X22 should we call finish here?
	finish();
    }

    private static final byte[] CRLF = { (byte)'\r', (byte)'\n' };

    // X22 ResponseAdapter
    /** Signal that we're done with the headers, and body will follow.
	The adapter doesn't have to maintain state, it's done inside the engine
    */
    public void endHeaders() throws IOException {
	// X22 need to send CRLF here?
	// outputStream.write(CRLF, 0, 2);
    }

    // X22 ResponseAdapter
    public ServletOutputStream getServletOutputStream() throws IOException {
	return outputStream;
    }

    // X22 ResponseAdapter
    public void recycle() {
	// X22 should this be keep alive or not?
	try {
	    next(true);
	} catch (IOException ioe) {
	    // don't care
	}
    }
    
    /** Write a chunk of bytes. Should be called only from ServletOutputStream implementations,
     *	No need to implement it if your adapter implements ServletOutputStream.
     *  Headers and status will be written before this method is exceuted.
     */
    public void doWrite(byte buffer[], int pos, int count) throws IOException {
	// X22 is it right to flush() here?
	outputStream.write(buffer, pos, count);
	outputStream.flush();
    }

}

