/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpsConnectionMethod.java,v 1.4.2.7.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.https;
import java.text.*;
import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.lutris.logging.*;
import org.enhydra.servlet.connectionMethods.*;
import org.enhydra.servlet.servletManager.ServletManager;
import org.enhydra.servlet.filter.FilterManager;
import org.enhydra.servlet.filter.Filter;
import com.lutris.util.*;

import javax.net.*;
import javax.net.ssl.*;
import javax.security.cert.*;
import com.sun.net.ssl.*;
import java.security.*;

/**
 * This class implements the connection method used to accept request
 * for servlets from an HTTP client.
 *
 * @see org.enhydra.servlet.connectionMethods.BasicConnectionMethod
 * @version 1.8, 6/98
 * @author Mark Sanguinetti
 * @author SSL added by Steve Latif
 */
public class HttpsConnectionMethod
    extends StandardConnectionMethod implements Runnable {

    public static final String HTTPS = ConnectionMethodManager.HTTPS;
    public static final String TYPE = ConnectionMethodManager.TYPE;
    public static final String PORT = ConnectionMethodManager.PORT;
    public static final String BIND_ADDR = ConnectionMethodManager.BIND_ADDR;
    public static final String NUM_THREADS = ConnectionMethodManager.NUM_THREADS;
    public static final String QUEUE_SIZE = ConnectionMethodManager.QUEUE_SIZE;
    public static final String THREAD_TIMEOUT = ConnectionMethodManager.THREAD_TIMEOUT;
    public static final String CLIENT_TIMEOUT = ConnectionMethodManager.CLIENT_TIMEOUT;

    //https stuff
    public static final String SSL = "SSL";
    public static final String SERVER_CERTIFICATE = "ServerCertificate";
    public static final String SECURE_RANDOM_ALGORITHM ="SecureRandomAlgorithm";
    public static final String SECURE_RANDOM_PROVIDER ="SecureRandomProvider";
    public static final String SSL_CONTEXT_PROVIDER = "SSLContextProvider";
    public static final String SSL_CONTEXT_PROTOCOL = "SSLContextProtocol";
    public static final String KEY_STORE_LOCATION ="KeyStoreLocation";
    public static final String KEY_STORE_PROVIDER = "KeyStoreProvider";
    public static final String KEY_MANAGER_ALGORITHM ="KeyManagerAlgorithm";
    public static final String KEY_MANAGER_PROVIDER = "KeyManagerProvider";
    public static final String TRUST_MANAGER = "TrustManager";
    public static final String PASSWORD = "Password";
    public static final String CLIENT_AUTHENTICATION = "ClientAuthentication";


    /*
     * Default interface to bind to (BindAddress).
     */
    private static final String DEFAULT_BIND_ADDRESS = "(All Interfaces)";

    /*
     * Default port for the connection.
     */
    private static final int DEFAULT_PORT = 443;

    /*
     * Default value for maximum number of threads.
     */
    public static final int DEFAULT_NUM_THREADS = 200;

    /*
     * Default value for maximum quesize.
     */
    public static final int DEFAULT_QUEUE_SIZE = 400;

    /*
     * Default value for threadTimeout, in milliseconds (5 minutes).
     */
    public static final int DEFAULT_THREAD_TIMEOUT = 5*60*1000;

    /*
     * Default value for clientTimeout (socket), in milliseconds (30 seconds).
     */
    public static final int DEFAULT_CLIENT_TIMEOUT = 1*30*1000;


    // SSL stuff

    /*
     * Default value for the secure Random Algorithm
     */
    public static final String DEFAULT_SECURE_RANDOM_ALGORITHM = "SHA1PRNG";


    /*
     * Default value for the Secure Random provider
     */
    public static final String DEFAULT_SECURE_RANDOM_PROVIDER = "SUN";

    /*
     * Default value for the SSL Context provider
     */
    public static final String DEFAULT_SSL_CONTEXT_PROVIDER = "SunJSSE";

    /*
     * Default value for the SSL Context protocol
     */
    public static final String DEFAULT_SSL_CONTEXT_PROTOCOL = "TLS";

    /*
     * Default value for the key store locaction
     */
    public static final String DEFAULT_KEY_STORE_LOCATION = ".";

    /*
     * Default value for the keystore location
     */
    public static final String DEFAULT_KEY_STORE_PROVIDER = "JKS";

    /*
     * Default value for the key manager algorithm
     */

    public static final String DEFAULT_KEY_MANAGER_ALGORITHM = "SUNX509";

    /*
     * Default value for the Key Manager provider
     */
    public static final String DEFAULT_KEY_MANAGER_PROVIDER = "SunJSSE";

    /*
     * Default value for the Trust Manager
     */
    public static final String DEFAULT_TRUST_MANAGER = "SUNX509";

    /*
     * Default value for the password
     */
    public static final String DEFAULT_PASSWORD = "";

    /*
     * Default value for Client Authentication
     */
    public static final boolean DEFAULT_CLIENT_AUTHENTICATION = true;





    /*
     * Standard messages for responding to clients
     */
    protected final static String SERVER_BUSY_MSG =
        "<TITLE>Request Denied</TITLE><H1>Request Denied</H1> " +
        "This server can not process your request. Please try again later." ;

    /**
     * This is temporary, to support the CGI Connection
     * method, which is simply another flavor of this class.
     */
    protected String cmType = "HTTPS";

    /**
     * The server port.
     */
    protected int port = DEFAULT_PORT;

    //----------------------

    //SSL stuff

    /**
     * The secure random algorithm
     */
    protected String secureRandomAlgorithm = DEFAULT_SECURE_RANDOM_ALGORITHM;

    /**
     * The secure Random provider
     */
    protected String secureRandomProvider = DEFAULT_SECURE_RANDOM_PROVIDER;

    /**
     * The SSL Context provider
     */
    protected String SSLContextProvider = DEFAULT_SSL_CONTEXT_PROVIDER;

    /**
     * The SSL Context protocol
     */
    protected String SSLContextProtocol = DEFAULT_SSL_CONTEXT_PROTOCOL;

    /**
     * The key store location
     */
    protected String keyStoreLocation = DEFAULT_KEY_STORE_LOCATION;

    /**
     * The key store provider
     */
    protected String keyStoreProvider = DEFAULT_KEY_STORE_PROVIDER;

    /**
     * The key manager algorithm
     */
    protected String keyManagerAlgorithm = DEFAULT_KEY_MANAGER_ALGORITHM;

    /**
     * The key manager provider
     */
    protected String keyManagerProvider = DEFAULT_KEY_MANAGER_PROVIDER;

    /**
     * The trust manager
     */
    protected String trustManager = DEFAULT_TRUST_MANAGER;

    /**
     * The password
     */
    protected String password = DEFAULT_PASSWORD;

    /**
     * Client Authentication
     */
    protected boolean clientAuthentication = DEFAULT_CLIENT_AUTHENTICATION;

    //---------------------


    /**
     * IP address to bind to.
     */
    protected String bindAddress = DEFAULT_BIND_ADDRESS;

    /**
     * IP address to bind to.
     */
    private InetAddress bindInetAddress = null;

    /**
     * The server socket.
     */
    //protected ServerSocket ss;

    /**
     * The logging channel.
     */
    protected LogChannel logChannel;

    /**
     * Thread handling the connections.
     */
    protected Thread dispatcher;

    /**
     * The queue for incoming connections.
     */
    protected Queue connections;

    /**
     * The maximum size of the connection queue.
     */
    protected int maxQueue = HttpsConnectionMethod.DEFAULT_QUEUE_SIZE;

    /**
     * The thread group for HttpHandler threads.
     */
    protected ThreadGroup handlers;

    /**
     * The maximum number of HttpHandler threads.
     */
    protected int maxHandlers = HttpsConnectionMethod.DEFAULT_NUM_THREADS;

    /**
     * The maximum time for an HttpHandler thread to wait for a connection.
     */
    protected int threadTimeout = HttpsConnectionMethod.DEFAULT_THREAD_TIMEOUT;

    /**
     * The maximum time for a thread to keep a socket connection to the client
     * for.
     */
    protected int clientTimeout = HttpsConnectionMethod.DEFAULT_CLIENT_TIMEOUT;

    /**
     * The backlog parameter for the server socket.
     */
    protected int backlog = 50;

    /**
     * The total number HttpHandler threads (active and waiting).
     */
    protected int numWaiting = 0;

    /**
     * A list of references to all HttpHandler threads (active and waiting).
     */
    protected Vector liveThreads;






    //SSL stuff

    /**
     * Secure random
     */
    protected SecureRandom secureRandom;

    /**
     * The SSL context for deriving the SSL Socket Factory
     */
    protected SSLContext context;

    /**
     * The keystore
     */
    protected KeyStore keyStore;

    /**
     * The key manager factory
     */
    protected KeyManagerFactory keyManagerFactory;

    /**
     * The list of key managers
     */
    protected KeyManager [] keyManager;

    /**
     * The trust manager factory
     */
    protected TrustManagerFactory trustManagerFactory;

    /**
     * List of trust managers
     */
    protected TrustManager[] trustMan;

    /**
     * the SSL server socket factories
     */
    protected SSLServerSocketFactory sslServerSocketFactory;

    /**
     * The SSL server socket
     */
    protected SSLServerSocket ss;



    /**
     * Default constructor used for dynamic connection method creation.
     **/
    public HttpsConnectionMethod () {
	logChannel = Logger.getCentralLogger().getChannel("Multiserver");
        addSecurityProviders();
    }

    /**
     *  Shut down the connection method. Releases the socket, so the
     *  port will be free for others to use.
     *
     *  @exception ConnectionMethodException If an error occurs.
     */
    public void destroy() throws ConnectionMethodException {
        super.destroy();
        dispatcher.interrupt();
        try {
            ss.close();
            logChannel.write(Logger.INFO, cmType + " releasing port " + port);
        } catch (IOException e) {
            logError("Error shutting down",  e);
        }
    }

    /**
     * Add the security provider required to use SSL.
     */
    private void addSecurityProviders() {
        // Global synchronized
        synchronized (Security.class) {
            if (Security.getProvider("sun.security.provider.Sun") == null) {
                Security.addProvider(new sun.security.provider.Sun());
            }
            if (Security.getProvider("com.sun.net.ssl.internal.ssl.Provider") == null) {
                Security.addProvider(new com.sun.net.ssl.internal.ssl.Provider());
            }
        }
    }

    /**
     *  Start up. Initially there are no channels.
     *  Parse the config file for connection information
     *  Saves the ServletManager and FilterManager for later use.
     *
     * This initialize is called from the ConnectionMethodManager when
     * the multiserver is started.  The config file contain all relevant
     * connection information.
     *
     *  @param connectionConfig The portion of the config file relevant to
     *  this connection method
     *  @param id The id associated with this connection method
     *  @param servletManager The ServletManager to use to convert servletIDs
     *  into Servlets.
     *  @param filterManager The FilterManager to use to convert filterIDs
     *  into Filters.
     *  @exception ConnectionMethodException If unable to start up.
     *  @see org.enhydra.servlet.filter.Filter
     *  @see org.enhydra.servlet.StandardConnectionMethod
     */
    public void initialize(Config connectionConfig,
			   String id,
			   ServletManager servletManager,
                           FilterManager filterManager)
        throws ConnectionMethodException {

	try {
	    String portStr = connectionConfig.getString(PORT);
	    port = Integer.parseInt(portStr);

	    if (connectionConfig.containsKey(NUM_THREADS)) {
		String nt = connectionConfig.getString(NUM_THREADS);
		maxHandlers = Integer.parseInt(nt);
		if (maxHandlers < 1) {
		    logChannel.write(Logger.ERROR,
				     NUM_THREADS + " must be at least 1 " +
				     "for connection method " + id + ".");
		    throw new ConnectionMethodException(
				     NUM_THREADS + " must be at least 1 " +
				     "for connection method " + id + ".");
		}
	    }

	    if (connectionConfig.containsKey(QUEUE_SIZE)) {
		String qs = connectionConfig.getString(QUEUE_SIZE);
		maxQueue = Integer.parseInt(qs);
		if (maxQueue < 1) {
		    logChannel.write(Logger.ERROR,
				     QUEUE_SIZE + " must be at least 1 " +
				     "for connection method " + id + ".");
		    throw new ConnectionMethodException(
				     QUEUE_SIZE + " must be at least 1 " +
				     "for connection method " + id + ".");
		}
	    }

	    if (connectionConfig.containsKey(CLIENT_TIMEOUT)) {
		String str = connectionConfig.getString(CLIENT_TIMEOUT);
		clientTimeout = Integer.parseInt(str) * 1000;
		if (clientTimeout < 1) {
		    logChannel.write(Logger.ERROR,
				     CLIENT_TIMEOUT + " must be at least 1 " +
				     "for connection method " + id + ".");
		    throw new ConnectionMethodException(
				     CLIENT_TIMEOUT + " must be at least 1 " +
				     "for connection method " + id + ".");
		}
	    }

	    if (connectionConfig.containsKey(THREAD_TIMEOUT)) {
		String str = connectionConfig.getString(THREAD_TIMEOUT);
		threadTimeout = Integer.parseInt(str) * 1000;
		if (threadTimeout < 1) {
		    logChannel.write(Logger.ERROR,
				     THREAD_TIMEOUT + " must be at least 1 " +
				     "for connection method " + id + ".");
		    throw new ConnectionMethodException(
				     THREAD_TIMEOUT + " must be at least 1 " +
				     "for connection method " + id + ".");
		}
	    }

	    //--------------------------

	    // Get all the https and ssl specific stuff and pass it to the
	    // connection method

	    //Server Certifiacte stuff not implemented yet
	    if (connectionConfig.containsKey(SERVER_CERTIFICATE)) {
	    }

	    if (connectionConfig.containsKey(SECURE_RANDOM_ALGORITHM)) {
		String str =
		    connectionConfig.getString(SECURE_RANDOM_ALGORITHM);
		secureRandomAlgorithm = str;

	    }

	    if (connectionConfig.containsKey(SECURE_RANDOM_PROVIDER)) {
		String str = connectionConfig.getString(SECURE_RANDOM_PROVIDER);
		secureRandomProvider = str;

	    }

	    if (connectionConfig.containsKey(SSL_CONTEXT_PROVIDER)) {
		String str = connectionConfig.getString(SSL_CONTEXT_PROVIDER);
		SSLContextProvider = str;


	    }

	    if (connectionConfig.containsKey(SSL_CONTEXT_PROTOCOL)) {
		String str = connectionConfig.getString(SSL_CONTEXT_PROTOCOL);
		SSLContextProtocol=str;

	    }

	    if (connectionConfig.containsKey(KEY_STORE_LOCATION)) {
		String str = connectionConfig.getString(KEY_STORE_LOCATION);
		keyStoreLocation = str;

	    }

	    if (connectionConfig.containsKey(KEY_STORE_PROVIDER)) {
		String str = connectionConfig.getString(KEY_STORE_PROVIDER);
		keyStoreProvider = str;

	    }

	    if (connectionConfig.containsKey(KEY_MANAGER_ALGORITHM)) {
		String str = connectionConfig.getString(KEY_MANAGER_ALGORITHM);
		keyManagerAlgorithm = str;

	    }

	    if (connectionConfig.containsKey(KEY_MANAGER_PROVIDER)) {
		String str = connectionConfig.getString(KEY_MANAGER_PROVIDER);
		keyManagerProvider = str;

	    }

	    if (connectionConfig.containsKey(TRUST_MANAGER)) {
		String str = connectionConfig.getString(TRUST_MANAGER);
		trustManager = str;

	    }

	    if (connectionConfig.containsKey(PASSWORD)) {
		String str = connectionConfig.getString(PASSWORD);
		password = str;

	    }

	    if (connectionConfig.containsKey(CLIENT_AUTHENTICATION)) {
		String str = connectionConfig.getString(CLIENT_AUTHENTICATION);
		if (str.equalsIgnoreCase("false")) {
		    clientAuthentication=false;

		}
	    }
	    //----------------------------

            if (connectionConfig.containsKey(BIND_ADDR)) {
                String str = connectionConfig.getString(BIND_ADDR);
                if ((str != null) && (str.length() > 0)) {
                    setBindAddress(str);
                }
            }
	} catch (NumberFormatException e) {
	    logChannel.write(Logger.ERROR,
			     "Error reading config file, number format exception for connection method " + id + ".");
	    throw new ConnectionMethodException(e);
	} catch (ConfigException e) {
	    logChannel.write(Logger.ERROR,
			     "Error processing config file, config exception for connection method " + id + ".");
	    throw new ConnectionMethodException(e);
	}

        super.initialize(connectionConfig, id, servletManager, filterManager);
   try {

	    secureRandom= SecureRandom.getInstance(secureRandomAlgorithm,
						   secureRandomProvider);


	    context=SSLContext.getInstance(SSLContextProtocol, SSLContextProvider);

	    FileInputStream fis = new FileInputStream(keyStoreLocation);


	    keyStore = KeyStore.getInstance(keyStoreProvider);

	    keyStore.load(fis, null);

	    keyManagerFactory = KeyManagerFactory.getInstance(keyManagerAlgorithm,
							      keyManagerProvider);


	    // Convert the password string to a character array
	    char[] pass =stringToArray(password);

	    keyManagerFactory.init(keyStore, pass);

	    keyManager = keyManagerFactory.getKeyManagers();

	    trustManagerFactory = TrustManagerFactory.getInstance(keyManagerAlgorithm);
	    trustManagerFactory.init(keyStore);

	    trustMan =trustManagerFactory.getTrustManagers();
	    context.init(keyManager, trustMan, secureRandom);

	    sslServerSocketFactory = context.getServerSocketFactory();
            ServerSocket socket;
            if (bindInetAddress != null) {
                socket = sslServerSocketFactory.createServerSocket(port, backlog,
                    bindInetAddress);

            } else {
                socket = sslServerSocketFactory.createServerSocket(port, backlog);
            }
            ss = (SSLServerSocket)socket;

            logChannel.write(Logger.DEBUG, "Client authentication is " +
                (clientAuthentication ? "true" : "false") );

	    ss.setNeedClientAuth(clientAuthentication);
	    ss.setEnableSessionCreation(true);


        } catch (IOException e) {
            logError("Server Socket Error", e);
            throw new ConnectionMethodException(e);
        } catch (NoSuchProviderException ns) {
	    logError ("SSLContext No Such Provider Error", ns);
	} catch (NoSuchAlgorithmException na) {
	    logError("SSLContext No such Algorithm Error" , na);
	} catch (KeyStoreException k) {
	    logError("Key Store Exception", k);
	} catch (java.security.cert.CertificateException c) {
	    logError ("java security cert CertificateException" , c);
	} catch (UnrecoverableKeyException uke) {
	    logError("Key Exception ", uke) ;
	} catch (KeyManagementException kme) {
	    logError("Key Management", kme);
	}

        handlers = new ThreadGroup("HttpsHandlers");
        liveThreads = new Vector();
        connections = new Queue(maxQueue);


        logChannel.write(Logger.INFO, cmType + " listening on port: " + port);
        dispatcher = new Thread(this);

        dispatcher.start();
    }


    public void initialize(String id,
			   ServletManager servletManager,
                           FilterManager filterManager)
        throws ConnectionMethodException {
	super.initialize(id, servletManager, filterManager);

        try {
	    secureRandom= SecureRandom.getInstance(secureRandomAlgorithm,
						   secureRandomProvider);

	    context=SSLContext.getInstance(SSLContextProtocol, SSLContextProvider);

	    FileInputStream fis = new FileInputStream(keyStoreLocation);

	    keyStore = KeyStore.getInstance(keyStoreProvider);
	    keyStore.load(fis, null);

	    keyManagerFactory = KeyManagerFactory.getInstance(keyManagerAlgorithm,
							      keyManagerProvider);

	    // Convert the password string to a character array
	    char[] pass =stringToArray(password);

	    keyManagerFactory.init(keyStore, pass);

	    keyManager = keyManagerFactory.getKeyManagers();

	    trustManagerFactory = TrustManagerFactory.getInstance(keyManagerAlgorithm);
	    trustManagerFactory.init(keyStore);

	    trustMan =trustManagerFactory.getTrustManagers();
	    context.init(keyManager, trustMan, secureRandom);

	    sslServerSocketFactory = context.getServerSocketFactory();
	    ServerSocket socket = sslServerSocketFactory.createServerSocket(port, backlog);
	    ss = (SSLServerSocket)socket;
	    if(clientAuthentication)
		System.out.println("Needs client authentication");
	    else{System.out.println("No such thing");}
	    ss.setNeedClientAuth(clientAuthentication);
	    ss.setEnableSessionCreation(true);


        } catch (IOException e) {
            logError("Server Socket Error", e);
            throw new ConnectionMethodException(e);
        } catch (NoSuchProviderException ns) {
	    logError ("SSLContext No Such Provider Error", ns);
	} catch (NoSuchAlgorithmException na) {
	    logError("SSLContext No such Algorithm Error" , na);
	} catch (KeyStoreException k) {
	    logError("Key Store Exception", k);
	} catch (java.security.cert.CertificateException c) {
	    logError ("java security cert CertificateException" , c);
	} catch (UnrecoverableKeyException uke) {
	    logError("Key Exception ", uke) ;
	} catch (KeyManagementException kme) {
	    logError("Key Management", kme);
	}

        handlers = new ThreadGroup("HttpsHandlers");
        liveThreads = new Vector();
        connections = new Queue(maxQueue);

        logChannel.write(Logger.INFO, cmType + " listening on port: " + port);
        dispatcher = new Thread(this);
        dispatcher.start();
    }

    /**
     * Compares the port number to determine equivalency
     *
     *  @param compareObject The object to compare this to
     *  @exception ConnectionMethodException If an error occurs.
     **/
    public boolean equivalent(ConnectionMethod compareObject) {
	HttpsConnectionMethod httpsCompareObject = null;
	if (!(compareObject instanceof HttpsConnectionMethod)) {
	    return false;
	} else {
	    httpsCompareObject = (HttpsConnectionMethod)compareObject;
	}
	if (httpsCompareObject.getPort() == getPort()) {
	    return true;
	} else {
	    return false;
	}
    }



    /**
     * Write out self into config fie
     *
     * @param config The config file
     * @param base The base key string
     **/
    public void writeToConfig(Config connectionConfig,
			      String base)
	throws ConfigException, KeywordValueException {
	connectionConfig.set(base + TYPE, HTTPS);
	connectionConfig.set(base + PORT, "" + getPort());
	connectionConfig.set(base + NUM_THREADS, "" + getNumThreads());
	connectionConfig.set(base + QUEUE_SIZE, "" + getQueueSize());
	connectionConfig.set(base + CLIENT_TIMEOUT,
			     "" + getClientTimeout()/1000);
	connectionConfig.set(base + THREAD_TIMEOUT,
			     "" + getThreadTimeout()/1000);
	// Add https parameters

	connectionConfig.set(base + SECURE_RANDOM_ALGORITHM, "" +
			     getSecureRandomAlgorithm());
	connectionConfig.set(base + SECURE_RANDOM_PROVIDER, "" +
			     getSecureRandomProvider());
	connectionConfig.set(base + SSL_CONTEXT_PROVIDER, "" +
			     getSSLContextProvider());
	connectionConfig.set(base + SSL_CONTEXT_PROTOCOL, "" +
			     getSSLContextProtocol());
	connectionConfig.set(base + KEY_STORE_LOCATION, "" +
			     getKeyStoreLocation());
	connectionConfig.set(base + KEY_STORE_PROVIDER, "" +
			     getKeyStoreProvider());
	connectionConfig.set(base + KEY_MANAGER_ALGORITHM, "" +
			     getKeyManagerAlgorithm());
	connectionConfig.set(base + KEY_MANAGER_PROVIDER, "" +
			     getKeyManagerProvider());
	connectionConfig.set(base + TRUST_MANAGER, "" +
			     getTrustManager());
	connectionConfig.set(base + PASSWORD, "" +
			     getPassword());
	connectionConfig.set(base + CLIENT_AUTHENTICATION, "" +
			     getClientAuthentication());
    }



    /**
     * Runs the HttpConnectionMethod.
     */
    public void run() {
        // accept and dispatch connections until interrupted
        while (!dispatcher.isInterrupted()) {
            threadManager();
            // got handler, now accept new connection and add to queue
            Socket s = null;
            try {
                s = ss.accept();
		s.setSoTimeout(clientTimeout);
            } catch (SocketException e) {
                if (dispatcher.isInterrupted()) {
                    // It's ok, it's just destroy() killing us.
                } else {
                    // Must be a real error.
                    logError("SocketException", e);
                    continue;
                }
            } catch (IOException e) {
                logError("Accept Connection", e);
                throw new FatalExceptionError(e);
            }
            if (! putConnection(s)) {
                logError("Connection Queue full");
                try {
                    OutputStream out = s.getOutputStream();
                    writeString(out, SERVER_BUSY_MSG);
                    out.close();
                    s.close();
                } catch (IOException e) {
                    logError("Close Connection", e);
                }
            }
        }
    }

    /*
     * Inventories the active number HttpHandler threads (currently processing
     * and waiting) and creates a new one if the active number is less
     * than maxHandlers
     */
    private void threadManager() {
        synchronized (connections) {
            if (numWaiting < 0 ) {
                throw new RuntimeException("HttpConnectionMethod: Thread Handler");
            }
            if (numWaiting == 0) {
                // Inventory threads
                Enumeration e = liveThreads.elements();
                while (e.hasMoreElements()) {
                    Thread t = (Thread) e.nextElement();
                    if (!t.isAlive()) {
                        liveThreads.removeElement(t);
                    }
                }
                // If no waiting and total thread < max
                if (liveThreads.size() < maxHandlers) {
                    HttpsHandler h = new HttpsHandler(this);
                    Thread t = new Thread(handlers, h);
                    liveThreads.addElement(t);
                    t.start();
                }
            }
        }
    }

    /**
     * Returns an socket from the connection queue.
     *
     * @return a reference to the object if successful, null otherwise.
     */
    protected Socket getConnection() {
        synchronized (connections) {
            long initialTime = System.currentTimeMillis();
            Socket temp = null;

            while (true) {
                try {
                    temp = (Socket)connections.get();
                } catch (ArrayIndexOutOfBoundsException e) {
                    logError("Connection Queue error", e);
                    throw new FatalExceptionError(e);
                }
                if (temp != null) {
                    // return a socket if queue is not empty
                    return temp;
                } else {
                    // return null as the thread timed out
                    if ((System.currentTimeMillis() - initialTime) > threadTimeout) {
                        return null;
                    }
                    // wait for a connection
                    numWaiting++;
                    try {
                        connections.wait(threadTimeout);
                    } catch (InterruptedException e) {
                        logError("Interrupted", e);
                    }
                    numWaiting--;
                }
            }
        }
    }

    /**
     * Stores a socket in the connections queue.
     *
     * @param s socket to be stored
     * @return True if successful, false otherwise.
     * @exception IllegalMonitorStateException exception.
     */
    private boolean putConnection(Socket s) {
        boolean ret;
        synchronized (connections) {
            try {
                ret = connections.put(s);
                connections.notify();
            } catch (ArrayIndexOutOfBoundsException e) {
                logError("Connection Queue error", e);
                throw new FatalExceptionError(e);
            } catch (IllegalMonitorStateException e) {
                logError("Connection Queue error", e);
                throw new FatalExceptionError(e);
            }
        }
        return (ret == true ? true : false);
    }

    /**
     * Gets Channel data assocated with a given URL Prefix.
     *
     * @param url The URL Prefix.
     * @return The Channel associated with the URL Prefix.
     */
    public Channel getChannelByURLPrefix(String url) {
        Channel c = channelTable.getChannelByURLPrefix(url);
        return c;
    }

    /**
     * Gets a reference to the servlet associated with servletID.
     *
     * @param servletID The Servlet ID.
     * @return Reference to the servlet.
     */
    public Servlet getServlet(String servletID) {
        Servlet s = null;
        try {
            s = servletManager.get(servletID);
        } catch (ServletException e) {
            throw new FatalExceptionError(e);
        }
        return s;
    }

    /**
     * Returns the URL that a user would use to connect to this channel.
     *
     * @return A URL string.
     * @param channelID The channel
     * @exception ConnectionMethodException If an error occurs,
     * for example if the channel does not exist.
     */
    public String getChannelURL(String channelID)
        throws ConnectionMethodException {
        Channel c = channelTable.findID(channelID);
        if (c == null)
            throw new ConnectionMethodException("Unable to get channel URL: " +
                                                "Invalid channel id " + channelID);
        String result = "https://";
        try {
            result += InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            result += "localhost";
        }
        if (port != DEFAULT_PORT)
            result += ":" + port;
        String pre = c.getURLPrefix();
        if ((pre.length() > 0) && (pre.charAt(0) != '/'))
            result += "/";
        result += pre;
        return result;
    }

    /**
     * This connection method returns valid urls.
     *
     * @param channelID The channel
     * @return This function always returns true.
     * @exception ConnectionMethodException If an error occurs,
     *  for example if the channel does not exist.
     */
    public boolean channelURLIsValid(String channelID)
        throws ConnectionMethodException {
        return true;
    }


    // Get and set methods for bean introspection
    public String getType() {
	return cmType;
    }

    public void setPort(int port) {
	this.port = port;
    }
    public int getPort() {
        return port;
    }

    /**
     * Gets the current Bind address, if any.
     */
    public String getBindAddress() {
        return this.bindAddress;
    }

    /**
     * Sets the current Bind address, if any.
     */
    public void setBindAddress(String addr) {
        String bhost = (addr != null) ? addr.trim().toLowerCase() : "";
        InetAddress baddr = null;
        if ( (bhost.length() > 0) && (!bhost.startsWith("(")) && (!bhost.equals("all"))) {
            try {
                baddr = InetAddress.getByName(bhost);
            } catch (UnknownHostException e) {
                baddr = null;
            }
        }
        if (baddr != null) {
            this.bindAddress = bhost;
            this.bindInetAddress = baddr;
        } else {
            this.bindAddress = DEFAULT_BIND_ADDRESS;
            this.bindInetAddress = null;
        }
    }


    public int getMaxHandlers() {
	return maxHandlers;
    }
    public void setMaxHandlers(int maxHandlers) {
	this.maxHandlers = maxHandlers;
    }

    public int getMaxQueue() {
	return maxQueue;
    }
    public void setMaxQueue(int maxQueue) {
	this.maxQueue = maxQueue;
    }

    public String getServerCertificate() {
	// not implemented yet
	return "not implemented yet";
    }
    public void setServerCertificate(String serverCertificate) {
	// not implemented yet
    }

    public String getSecureRandomAlgorithm() {
	return secureRandomAlgorithm;
    }
    public void setSecureRandomAlgorithm(String secureRandomAlgorithm) {
	this.secureRandomAlgorithm = secureRandomAlgorithm;
    }

    public String getSecureRandomProvider() {
	return secureRandomProvider;
    }
    public void setSecureRandomProvider(String secureRandomProvider) {
	this.secureRandomProvider = secureRandomProvider;
    }

    public String getSSLContextProvider() {
	return SSLContextProvider;
    }
    public void setSSLContextProvider(String SSLContextProvider) {
	this.SSLContextProvider = SSLContextProvider;
    }

    public String getSSLContextProtocol() {
	return SSLContextProtocol;
    }
    public void setSSLContextProtocol(String SSLContextProtocol) {
	this.SSLContextProtocol = SSLContextProtocol;
    }

    public String getKeyStoreLocation() {
	return keyStoreLocation;
    }
    public void setKeyStoreLocation(String keyStoreLocation) {
	this.keyStoreLocation = keyStoreLocation;
    }

    public String getKeyStoreProvider() {
	return keyStoreProvider;
    }
    public void setKeyStoreProvider(String keyStoreProvider) {
	this.keyStoreProvider = keyStoreProvider;
    }

    public String getKeyManagerAlgorithm() {
	return keyManagerAlgorithm;
    }
    public void setKeyManagerAlgorithm(String keyManagerAlgorithm) {
	this.keyManagerAlgorithm = keyManagerAlgorithm;
    }

    public String getKeyManagerProvider() {
	return keyManagerProvider;
    }
    public void setKeyManagerProvider(String keyManagerProvider) {
	this.keyManagerProvider = keyManagerProvider;
    }

    public String getTrustManager() {
	return trustManager;
    }
    public void setTrustManager(String trustManager) {
	this.trustManager = trustManager;
    }

    public String getPassword() {
	return password;
    }
    public void setPassword(String password) {
	this.password = password;
    }

    public boolean getClientAuthentication() {
	return clientAuthentication;
    }
    public void setClientAuthentication(boolean clientAuthentication) {
	this.clientAuthentication = clientAuthentication;
    }




    /**
     * Gets the maximum number of handler thread associated with this connection method.
     *
     *@return numThreads.
     */
    public int getNumThreads() {
        return getMaxHandlers();
    }

    /**
     * Gets the maximum queue size associated with this connection method.
     *
     *@return queueSize.
     */
    public int getQueueSize() {
        return getMaxQueue();
    }

    /**
     * Gets the client timeout associated with this connection method.
     *
     *@return clientTimeout.
     */
    public int getClientTimeout() {
        return clientTimeout;
    }
    public void setClientTimeout(int clientTimeout) {
	this.clientTimeout = clientTimeout;
    }

    /**
     * Gets the thread timeout associated with this connection method.
     *
     *@return threadTimeout.
     */
    public int getThreadTimeout() {
        return threadTimeout;
    }
    public void setThreadTimeout(int threadTimeout) {
	this.threadTimeout = threadTimeout;
    }

    /*
     * utility for writing to clients before init()
     * used only in once (for notifing of HTTP 0.9 requests)
     */
    private void writeString(OutputStream out, String s)
        throws IOException {
        out.write(s.getBytes());
    }

    /*
     * Utility for logging errors. The logged message will include the port,
     * connection method type and exception thrown.
     *
     * @param s Error message to be logged
     * @param e Exception to be logged
     */
    private void logError(String s, Exception e) {
        String msg = "Error in " + cmType + " Connection method: "
            + "Port: " + port + " : " + s + " : " + e;
        logChannel.write(Logger.ERROR, msg, e);

    }

    /*
     * Utility for logging errors. The logged message will include the port,
     * connection method type and exception thrown.
     *
     * @param s Error message to be logged
     */
    private void logError(String s) {
        String msg = "Error in " + cmType + " Connection method: "
            + "Port: " + port + " : " + s ;
        logChannel.write(Logger.ERROR, msg);

    }

    /*
     * Utility for converting a string to a character array
     *
     * @param s String to be converted to a character array
     *
     */
    private char[] stringToArray(String s) {
	int stringLength = s.length();
	char [] a = new char[stringLength];
	StringCharacterIterator it = new StringCharacterIterator(s);

	for (int i= it.first();i != it.DONE ;i= it.next())
	    a[it.getIndex()]=(char)i;

	return a;
    }


}
