/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpsOutputStream.java,v 1.4.12.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.https;

import javax.servlet.ServletOutputStream;
import java.util.Enumeration;
import java.io.OutputStream;
import java.io.IOException;
import java.net.Socket;
import org.enhydra.servlet.ServletIOException;

/**
 * Implementation of a <CODE>ServletOutputStream</CODE> for the HTTP
 * connection method.  This object is designed to be recycled for several
 * independent requests.  For speed; it uses an internal buffer instead of
 *
 * @author Mark Diekhans
 * @version $Revision: 1.4.12.1 $
 * @see HttpsRequest
 * @see javax.servlet.ServletOutputStream
 */
class HttpsOutputStream extends ServletOutputStream {
    /**
     * Size of output buffer.
     */
    private static final int bufferSize = 4096;

    /**
     * Output buffer;
     */
    private byte[] buffer = new byte[bufferSize];

    /**
     * Index of next byte to store in buffer.
     */
    private int bufferIndex;

    /**
     * Request object. Used to write headers before first data write.
     */
    private HttpsResponse response;

    /**
     * Socket associate with a set of requests.
     */
    private Socket socket;

    /**
     * Output stream for the socket.
     */
    private OutputStream socketStream;

    /**
     * Count of bytes transfered.
     */
    long byteCount;

    /**
     * Construct an output stream associate with a response object.
     */
    protected HttpsOutputStream(HttpsResponse containingResponse) {
        response = containingResponse;
    }

    /**
     * Associate a connected socket with this object.  The
     * <CODE>next</CODE> function must be called before any
     * output can be written.
     * @param requestSocket The HTTP connected socket.
     * @exception IOException If an error occured setting up the socket
     *  output stream.
     */
    protected void setConnection(Socket requestSocket) throws IOException {
        socket = requestSocket;
        try {
            socketStream = socket.getOutputStream();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Reset state to write the next request socket.
     */
    protected void next() {
        bufferIndex = 0;
        byteCount = 0L;
    }

    /**
     * Write the buffer to the output stream.
     *
     * @exception  IOException If an I/O error occurs.
     */
    private void writeBuffer() throws IOException {
        try {
            socketStream.write(buffer, 0, bufferIndex);
            bufferIndex = 0;
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Write a byte of data to this output stream. 
     *
     * @param dbyte Byte of data to write to the stream.
     * @exception  IOException If an I/O error occurs.
     */
    public synchronized void write(int dbyte) throws IOException {
        // Force headers to be written if its hasn't happened
        if ((byteCount == 0) && !response.haveHeadersBeenWritten()) {
            response.writeHeaders();
        }
        if (bufferIndex == bufferSize) {
            writeBuffer();
        }
        buffer[bufferIndex++] = (byte)dbyte;
        byteCount++;
    }

    /**
     * Write an array of bytes to this output stream. 
     *
     * @param bytes Array of byte data to write to the stream.
     * @param off Starting offset in the array.
     * @param len Number of bytes to write.
     * @exception  IOException If an I/O error occurs.
     */
    public synchronized void write(byte[] bytes, int off, int len) throws IOException {
        // Force headers to be written if its hasn't happened

        if ((byteCount == 0) && !response.haveHeadersBeenWritten()) {
            response.writeHeaders();
        }

        while (len > 0) {
            if (bufferIndex == bufferSize) {
                writeBuffer();
            }
            int moveCnt = ((bufferSize - bufferIndex) < len) ?
                (bufferSize - bufferIndex) : len;
            System.arraycopy(bytes, off, buffer, bufferIndex, moveCnt);
            off += moveCnt;
            len -= moveCnt;
            bufferIndex += moveCnt;
            byteCount += moveCnt;
        }
    }

    /**
     * Prints the string provided.   Overrides the default version
     * to avoid slow, character-at-a-time conversions.
     *
     * @exception IOException if an I/O error has occurred
     * @see javax.servlet.ServletOutputStream#print
     */
    public synchronized void print(String s) throws IOException {
        // Force headers to be written if its hasn't happened
        if ((byteCount == 0) && !response.haveHeadersBeenWritten()) {
            response.writeHeaders();
        }

        int off = 0;
        int len = s.length();
        while (len > 0) {
            if (bufferIndex == bufferSize) {
                 writeBuffer();
            }
            int moveCnt = ((bufferSize - bufferIndex) < len) ?
                (bufferSize - bufferIndex) : len;

            // N.B. This version of getBytes() is deprecated.  However using the
            // version that takes an encoding is a serious performance impact in
            // Sun JDK 1.1.*.
            s.getBytes(off, off+moveCnt, buffer, bufferIndex);
            off += moveCnt;
            len -= moveCnt;
            bufferIndex += moveCnt;
            byteCount += moveCnt;
        }
    }

    /**
     * Flushes this output stream and forces any buffered output bytes 
     * to be written out. 
     *
     * @exception IOException  if an I/O error occurs.
     */
    public synchronized void flush() throws IOException {
        try {
            if ((byteCount == 0) && !response.haveHeadersBeenWritten()) {
                response.writeHeaders();
            }
            if (bufferIndex > 0) {
                socketStream.write(buffer, 0, bufferIndex);
            }
            socketStream.flush();
            bufferIndex = 0;
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Closes this output stream and releases any system resources 
     * associated with the stream. 
     *
     * @exception  IOException If an I/O error occurs.
     */
    public synchronized void close() throws IOException {
        try {

            socketStream.flush();
            socketStream.close();

            socket.close();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Get the number of bytes that have been transfered since the last
     * <CODE>next()</CODE>.
     */
    protected long getByteCount() {
        return byteCount;
    }
}
