/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DebugManager.java,v 1.4.4.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.debug;

import org.enhydra.servlet.connectionMethods.*;
import com.lutris.multiServer.*;
import org.enhydra.servlet.ServletContainer;
import org.enhydra.servlet.filter.*;


/**
 * Servlet debug manager for monitoring servlet I/O.  An instance of this
 * object monitors a single servlet instance.  A fixed-sized queue of servlet
 * transaction records is maintained.  As a transaction is completed by the
 * servlet, a record is added to the end of the queue.  If the queue has
 * reached its maximum size, an element is removed from the head of the
 * queue.  Applications wishing to monitor servlet transactions registers
 * with this object and is called when objects are added or removed from
 * the queue.  Calling on removal is necessary as one of the main goals of
 * this class is to support presentations that display a list of transaction
 * records. A unique id number is associated with each transaction record.
 * The transaction record ids are sequencially allocated, one with a large
 * number occured after one with a smaller number.
 * 
 * @version  $Revision: 1.4.4.1 $
 * @author   Mark Diekhans
 * @since    2.0
 */
public class DebugManager {

    /**
     * Servlet being managed.
     */
    private String servletId;

    /**
     * Current monitoring state.
     */
    private boolean enabled = false;

    /**
     * Queue of transaction records. This is protected so that subclasses
     * may access the queue.
     */
    protected ServletRecordQueue recordQueue;

    /**
     * List of objects to call on servlet debug events.  Kept as an array
     * since addition/removal of callbacks is infrequence and the expected
     * number is small.  The array may not be full.
     */
    private ServletRecordCallback[] callbacks =
            new ServletRecordCallback[1];

    /**
     * Number of valid entries in callbacks.
     */
    private int numCallbacks = 0;

    /**
     * Filter that is used by this DebugManager.
     */
    private ServletRecorderFilter filter;
    private String filterID; 

    /**
     * Construct an new servlet debug manager and associate it with a servlet.
     * Servlet monitoring is initially disabled. 
     *
     * @param servletId symbolic servlet identifier of the servlet to manage.
     * @param queueSize maximum number of servlet transaction records that
     *  can be queued at a given time.
     * @param saveResponseData Should we save a copy of the data written
     *  to the socket for each transaction? Could use up alot of memory.
     */
    public DebugManager(String servletId, 
                        int queueSize,
                        boolean saveResponseData) {
        this.servletId = servletId;
        recordQueue = new ServletRecordQueue(queueSize);
        filter = new ServletRecorderFilter(this, saveResponseData, servletId);
    }

    

    /**
     * Enable recording servlet transactions. This adds a filter to the
     * filter manager, and then adds the filter to all channels that
     * map to the servlet.
     *
     *  @exception ConnectionMethodException If there is an error getting
     *   the connection methods and channels, or adding the filter.
     */
    public synchronized void enable() throws ConnectionMethodException {
	ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	ConnectionMethodManager cmm = sc.getConnectionMethodManager();
	FilterManager fm = sc.getFilterManager();
        String[] connectionIds = cmm.getIDs();
        /*
         *   Add to thr filter manager. We must delete it later when
         *   debugging is disabled. Because the name starts with an
         *   underscore, it will not be displayed in the admin app.
         */
        filterID = fm.getUniqueID("_debug_" + servletId);
        fm.add(filterID, filter, "internal debugging");
        for (int cmi = 0; cmi < connectionIds.length; cmi++) {
            ConnectionMethod method = cmm.get(connectionIds[cmi]);
            String[] channelIds = method.getChannelIDs();
            for (int ch = 0; ch < channelIds.length; ch++) {
                String chanServletId = 
                        method.getChannelStatus(channelIds[ch]).servletID;
                if (servletId.equals(chanServletId)) {
                    method.addTransactionFilter(channelIds[ch], filterID);
                }
            }
        }
	enabled = true;
    }

    /**
     * Disable recording servlet transactions. This removes the filter from
     * all the channels, then removes the filter from the filter manager.
     *
     *  @exception ConnectionMethodException If there is an error getting
     *   the connection methods and channels, or adding the filter.
     */
    public synchronized void disable() throws ConnectionMethodException {
	ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	ConnectionMethodManager cmm = sc.getConnectionMethodManager();
        String[] connectionIds = cmm.getIDs();

        for (int cmi = 0; cmi < connectionIds.length; cmi++) {
            ConnectionMethod method = cmm.get(connectionIds[cmi]);
            String[] channelIds = method.getChannelIDs();
            for (int ch = 0; ch < channelIds.length; ch++) {
                String[] filterIds =
                        method.getChannelStatus(channelIds[ch]).filterIDs;
                for (int fi = 0; fi < filterIds.length; fi++) {
                    if (filterIds[fi].equals(filterID)) {
                        method.removeTransactionFilter(channelIds[ch],
                                                       filterID);
                    }
                }
            }
        }
        /*
         *   Delete the filter from the filter manager.
         */
        sc.getFilterManager().delete(filterID);
	enabled = false;
    }

    /**
     * Determine if servlets monitoring is enabled.
     *
     * @return <CODE>true</CODE> if enabled, <CODE>false</CODE> if disabled.
     */
    public boolean isEnabled() {
        return enabled;
    }

    /**
     * Get the servlet associated with this object.
     * 
     * @return the servlet id.
     */
    public String getServlet() {
        return servletId;
    }

    /**
     * Registered to be called when a servlet transaction is queued.
     *
     * @param callback The object to invoke when a transaction object
     * is available.
     * @param catchUp Pass in true if you want an initial series of
     * <CODE>transactionRecordAddEvent()</CODE> calls, reflecting the
     * current set of transactions stored in the queue. 
     * Wanring: if this is set to true, and there are transactions in the
     * queue, then <CODE>transactionRecordAddEvent()</CODE> will be
     * called before this function returns. 
     */
    public synchronized void addTransactionCallback(
            ServletRecordCallback callback, boolean catchUp) {
        if (numCallbacks == callbacks.length) {
            ServletRecordCallback[] newTbl =
                new ServletRecordCallback[callbacks.length * 2];
            System.arraycopy(callbacks, 0, newTbl, 0, callbacks.length);
            callbacks = newTbl;
        }
        callbacks[numCallbacks] = callback;
        numCallbacks++;

        /*
         * Should we inform the callee of the current set of transactions?
         */
        if (catchUp)
            recordQueue.dumpContents(callback);
    }

    /**
     * Remove transaction callback that was created with
     * <CODE>addTransactionCallback</CODE>.
     *
     * @param callback object to delete from callback list.
     */
    public synchronized void removeTranactionCallback(
            ServletRecordCallback callback) {
        int i;
        for (i = 0; i < numCallbacks; i++) {
            if (callbacks[i] == callback) {
                break;
            }
        }
        // Compact.  Array bounds violation nails bad callback.
        callbacks[i] = callbacks[numCallbacks - 1];
        numCallbacks--;
    }

    /**
     * Find a servlet transaction record by id.
     *
     * @param transactionId unique number identifying the desired transaction
     *  record.
     * @return a transaction record or <CODE>null</CODE> if the transaction
     *  is no longer in the queue.
     */
    public synchronized ServletTransactionRecord getTransactionRecord(
            int transactionId) {
        return recordQueue.findById(transactionId);
    }

    /**
     * Add a servlet transaction to the end of the queue.  Use by the
     * servlet filters.
     * 
     * @param transactionRecord the record to add to the queue.
     */
    protected synchronized void addTransactionRecord(
             ServletTransactionRecord transactionRecord) {
        ServletTransactionRecord deletedRecord =
            recordQueue.addRecord(transactionRecord);
        
        for (int i = 0; i < numCallbacks; i++) {
            callbacks[i].transactionRecordAddEvent(transactionRecord,
                                                   deletedRecord);
        }
    }

    /**
     * Clear all records from the queue.
     */
    public synchronized void clearRecordQueue() {
        recordQueue.clear();
        for (int i = 0; i < numCallbacks; i++) {
            callbacks[i].clearRecordsEvent();
        }
    }
}
