/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: InputStreamRecord.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.debug;

import java.io.PrintWriter;
import java.io.IOException;
import javax.servlet.ServletInputStream;
import com.lutris.util.BytesToString;

/**
 *  All calls are added to the trace log in the associated request
 *  ojbect. There is no option to save the data, this will be added in
 *  the future. <P>
 *  
 *  All input is passed through unmodified.
 */
class InputStreamRecord extends ServletInputStream {

    private RequestRecord requestRecord;
    private ServletInputStream  realStream;
   
    /**
     *  Wrap around the ServletInputStream realStream, and log all the 
     *  calls made to it to the trace string in requestRecord.
     *  The data is passed through unmodified.
     *
     * @param requestRecord The request associated with this input stream.
     * <CODE>addTrace()</CODE> will be called on this object to log the
     * interaction between the servlet and this object.
     * @param realStream The real SerlvetInputStream to wrap around.
     */
    InputStreamRecord(RequestRecord requestRecord, 
                      ServletInputStream realStream) {
        this.realStream = realStream;
        this.requestRecord = requestRecord;
    }


    public int readLine(byte b[], int off, int len) throws IOException {
        int numBytes;
        try {
            numBytes = realStream.readLine(b, off, len);
        } catch (IOException e) {
            addTrace("readLine(byte[], " + off + ", " + len + ") threw " + e);
            throw(e);
        }
        addTrace("readLine(byte[], " + off + ", " + len + ") = " + numBytes);
        return numBytes;
    }


    public int read() throws IOException {
        int n;
        try {
            n = realStream.read();
        } catch (IOException e) {
            addTrace("read() threw " + e);
            throw(e);
        }
        addTrace("read() = " + n);
        return n;
    }


    public int read(byte b[]) throws IOException {
        int n;
        try {
            n = realStream.read(b);
        } catch (IOException e) {
            addTrace("read(byte[" + b.length + "]) threw " + e);
            throw(e);
        }
        addTrace("read(byte[" + b.length + "]) = " + n);
        return n;
    }


    public long skip(long n) throws IOException {
        long l;
        try {
            l = realStream.skip(n);
        } catch (IOException e) {
            addTrace("skip(" + n + ") threw " + e);
            throw(e);
        }
        addTrace("skip(" + n + ") = " + l);
        return l;
    }


    public int available() throws IOException {
        int n;
        try {
            n = realStream.available();
        } catch (IOException e) {
            addTrace("available() threw " + e);
            throw(e);
        }
        addTrace("available() = " + n);
        return n;
    }

    
    public void close() throws IOException {
        try {
            realStream.close();
        } catch (IOException e) {
            addTrace("close() threw " + e);
            throw(e);
        }
        addTrace("close()");
    }

    
    public synchronized void mark(int readlimit) {
        realStream.mark(readlimit);
        addTrace("mark(" + readlimit + ")");
    }


    public synchronized void reset() throws IOException {
        try {
            realStream.reset();
        } catch (IOException e) {
            addTrace("reset() threw " + e);
            throw(e);
        }
        addTrace("reset()");
    }


    public boolean markSupported() {
        boolean b = realStream.markSupported();
        addTrace("markSupported() = " + b);
        return b;
    }


    /*
     * How many times have we added to the trace log?
     */
    private int counter = 0;
 
    /*
     * What is the maxmimum number of times to write to the trace log?
     * There could be hundreds of calls to read()...
     */
    private static final int maxTraces = 16;
 
    /*
     * If we havn't logged too much, log a message to the trace string.
     */
    private void addTrace(String msg) {
        counter++;
        if (counter > maxTraces)
            return;
        if (requestRecord == null)
            return;
        msg = "ServletInputStream: " + msg;
        if (counter == maxTraces) {
            requestRecord.addTrace(msg);
            requestRecord.addTrace("(No more ServletInputStream calls " +
                                   "will be shown)");
            return;
        }
        requestRecord.addTrace(msg);
    }

}
    

