/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PrintWriterRecord.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.debug;

import java.io.PrintWriter;
import java.io.OutputStream;
import java.io.IOException;
import javax.servlet.ServletOutputStream;

/**
 *  This class is used to trace the calls to the output methods
 *  of a servlet's PrintWriter and, optionally, to save a copy of the data
 *  written to the output stream. The tracing and data-copying are 
 *  accomplished via an associated ResponseRecord object.
 */
class PrintWriterRecord extends PrintWriter {

    private ResponseRecord          responseRecord;
    private PrintWriter             realWriter;
    private StringBuffer            responseData; // Use if not null.

    /**
     *  All calls will also be logged to the trace string via the
     *  <CODE>ResponseRecord</CODE> passed in. If saveResponseData
     *  is true, the data written will also be saved to the
     *  ResponseData's StringBuffer called responseData. <P>
     *
     *  This class passes all data through unmodified. <P>
     *
     * @param responseRecord The response associated with this PrintWriter.
     * It's <CODE>addTrace()</CODE> method will be called to log the 
     * calls to the output methods.
     * If saveResponseData is true, then it's StringBuffer
     * responseData will be appened to, saving a copy of the data written
     * to the output stream.
     *
     * @param realWriter After saving a copy, all the data is passed
     * through unmodified to this output stream.
     */
    PrintWriterRecord (ResponseRecord responseRecord,
                       boolean saveResponseData,
                       PrintWriter realWriter) {
        /*
         * This works for Java 1.1, but not 1.2. We don't use our own
         * output methods, so we don't need to be a real PrintWriter.
         *     super((OutputStream)null); 
         */
        super(realWriter);  // Our own methods are not called.
        this.responseRecord = responseRecord;
        this.realWriter = realWriter;
        if (saveResponseData)
            responseData = responseRecord.responseData;
        else
            responseData = null;
     }


    public void print(boolean b) {
        realWriter.print(b);
        if (responseData != null)
            responseData.append(b);
        addTrace("print(" + b + ")");
    }

    public void print(char c) {
        realWriter.print(c);
        if (responseData != null)
            responseData.append(c);
        addTrace("print(char " + (int)c + ")");
    }

    public void print(int i) {
        realWriter.print(i);
       if (responseData != null)
            responseData.append(i);
        addTrace("print(int " + i + ")");
    }

    public void print(long l) {
        realWriter.print(l);
        if (responseData != null)
            responseData.append(l);
        addTrace("print(long " + l + ")");
    }

    public void print(float f) {
        realWriter.print(f);
        if (responseData != null)
            responseData.append(f);
        addTrace("print(float " + f + ")");
    }

    public void print(double d) {
        realWriter.print(d);
        if (responseData != null)
            responseData.append(d);
        addTrace("print(double " + d + ")");
    }

    public void print(char s[]) {
        realWriter.print(s);
        if (responseData != null)
            responseData.append(s);
        addTrace("print(char[ " + s.length + "])");
    }

    public void print(String s) {
        realWriter.print(s);
        if (responseData != null)
            responseData.append(s);
        addTrace("print(String, " + s.length() + " chars)");
    }

    public void print(Object obj) {
        realWriter.print(obj);
        if (responseData != null)
            responseData.append(obj.toString());
        addTrace("print(Object " + obj + ")");
    }


    public void println() {
        realWriter.println();
        if (responseData != null)
            responseData.append('\n');
        addTrace("println()");
    }




    public void println(boolean b) {
        realWriter.println(b);
        if (responseData != null) {
            responseData.append(b);
            responseData.append('\n');
        }
        addTrace("println(" + b + ")");
    }

    public void println(char c) {
        realWriter.println(c);
        if (responseData != null) {
            responseData.append(c);
            responseData.append('\n');
        }
        addTrace("println(char " + (int)c + ")");
    }

    public void println(int i) {
        realWriter.println(i);
       if (responseData != null) {
            responseData.append(i);
            responseData.append('\n');
        }
        addTrace("println(int " + i + ")");
    }

    public void println(long l) {
        realWriter.println(l);
        if (responseData != null) {
            responseData.append(l);
            responseData.append('\n');
        }
        addTrace("println(long " + l + ")");
    }

    public void println(float f) {
        realWriter.println(f);
        if (responseData != null) {
            responseData.append(f);
            responseData.append('\n');
        }
        addTrace("println(float " + f + ")");
    }

    public void println(double d) {
        realWriter.println(d);
        if (responseData != null) {
            responseData.append(d);
            responseData.append('\n');
        }
        addTrace("println(double " + d + ")");
    }

    public void println(char s[]) {
        realWriter.println(s);
        if (responseData != null) {
            responseData.append(s);
            responseData.append('\n');
        }
        addTrace("println(char[ " + s.length + "])");
    }


    public void println(String s) {
        realWriter.println(s);
        if (responseData != null) {
            responseData.append(s);
            responseData.append('\n');
        }
        addTrace("println(String, " + s.length() + " chars)");
    }

    public void println(Object obj) {
        realWriter.println(obj);
        if (responseData != null) {
            responseData.append(obj.toString());
            responseData.append('\n');
        }
        addTrace("println(Object " + obj + ")");
    }



    public void write(int c) {
        realWriter.write(c);
        if (responseData != null)
            responseData.append(c);
        addTrace("write(int " + c + ")");
    }

    public void write(char b[]) {
        realWriter.write(b);
        if (responseData != null)
            responseData.append(b);
        addTrace("write(char[" + b.length + "])"); 
    }


    public void write(char b[], int off, int len) {
        realWriter.write(b, off, len);
        if (responseData != null) {
            String temp = new String(b, off, len);
            responseData.append(temp.toCharArray());
        }
        addTrace("write(char[" + b.length + "], " + off + ", " + len + ")"); 
    }

    public void write(String s) {
        realWriter.write(s);
        if (responseData != null)
            responseData.append(s);
        addTrace("write(String, " + s.length() + " chars)");
    }

    public void write(String s, int off, int len) {
        realWriter.write(s, off, len);
        if (responseData != null) {
            String temp = s.substring(off, off + len + 1);
            responseData.append(temp.toCharArray());
        }
        addTrace("write(String " + s.length() + " chars, " + off + ", " + len + ")"); 
    }

    public void flush() {
        realWriter.flush();
        addTrace("flush()");
    }

    public boolean checkError() {
        boolean result = realWriter.checkError();
        addTrace("checkError() = " + result);
        return result;
    }

    public void close() {
        realWriter.close();
        addTrace("close()");
    }

    /*
     * How many times have we added to the trace log?
     */
    private int counter = 0;

    /*
     * What is the maxmimum number of times to write to the trace log?
     * There could be hundreds of calls to write()...
     */
    private static final int maxTraces = 64;

    /*
     * If we havn't logged too much, log a message to the trace string.
     */
    private void addTrace(String msg) {
        counter++;
        if (counter > maxTraces)
            return;
        if (responseRecord == null)
            return;
        msg = "PrintWriter: " + msg;
        if (counter == maxTraces) {
            responseRecord.addTrace(msg); 
            responseRecord.addTrace("(No more PrintWriter calls " +
                                    "will be shown)");
            return;
        }
        responseRecord.addTrace(msg); 
    }
}


