/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: RecorderServiceInterceptor.java,v 1.4.2.7.2.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.debug;

import java.io.PrintWriter;
import java.io.IOException;
import java.util.Enumeration;
import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.apache.tomcat.core.ServiceInterceptor;
import org.apache.tomcat.core.Context;
import org.apache.tomcat.core.InvokerServlet;
import com.lutris.appserver.server.session.Session;
import com.lutris.appserver.server.session.SessionData;
import com.lutris.appserver.server.httpPresentation.servlet.HttpPresentationServlet;
import com.lutris.appserver.server.httpPresentation.servlet.HttpPresentationServletGlue;

import javax.servlet.http.HttpSession;


/**
 * Servlet filter for recording servlet transactions for debug display.
 * On each request, a request and response filter is used to record
 * the state of the request/response object.  These objects are then
 * passed to the DebugManager queue.
 *
 * @version  $Revision: 1.4.2.7.2.1 $
 * @author   Mark Diekhans
 * @see com.lutris.filter.TransactionFilter
 * @see javax.servlet.http.HttpServlet
 */
public class RecorderServiceInterceptor implements ServiceInterceptor {

    /**
     *  The name of the underlying real servlet.
     */
    private String servletName;

    /**
     * The debug manager object this instance is associated with.
     */
    private DebugManager debugManager;

    /**
     *  Should we save a copy of all the data written to the socket?
     */
    private boolean saveResponseData;

    /**
     *  If the servlet we are monitoring is a Enhydra application (i.e.
     *  HttpPresentationServlet), then we save a pointer to it so that
     *  we can get the session data before and after the call to
     *  service(). If we are being used by the Multiserver, then
     *  it handles the filters, and we will be given the
     *  HttpPresentationServlet directly. If we are running in an
     *  unknown server (perhaps Apache) and using the Enhydra Debugger,
     *  then the applications themselves manage a list of filters, and
     *  as a side effect we will be given a HttpPresentationServletGlue
     *  servlet, which is simply a thin wrapper around the real
     *  HttpPresentationServlet.
     */

    /**
     * Construct a transaction recording filter servlet.
     *
     * @param originalServlet The real Servlet that all the calls will
     * be passed on through to.
     */
    protected RecorderServiceInterceptor(DebugManager debugManager,
                              boolean saveResponseData,
                              String servletName,
                              Servlet originalServlet) {

        this.saveResponseData = saveResponseData;
        this.debugManager = debugManager;
        this.servletName = servletName;
    }

    /**
     * Service a request, recording the request input and output.
     *
     * @param request The request object.
     * @param response The response object.
     * @exception ServletException See the Servlet documentation.
     * @exception IOException See the Servlet documentation.
     * @see javax.servlet.Servlet#service
     * @see javax.servlet.ServletException
     */
    public void preInvoke(Context context, Servlet servlet,
	    HttpServletRequest request, HttpServletResponse response) {

        if (servlet instanceof InvokerServlet) {
            // do not debug the tomcat serverlet invoker
            return;
        }

        RequestRecord requestRecord = (RequestRecord) request.getAttribute(
		"Enhydra.RecorderServiceInterceptor.requestRecord");
        ResponseRecord responseRecord = (ResponseRecord) request.getAttribute(
		"Enhydra.RecorderServiceInterceptor.responseRecord");
        ServletTransactionRecord record =
            new ServletTransactionRecord(servletName,
                                         requestRecord,
                                         responseRecord);
        requestRecord.transactionRecord = record;
        responseRecord.transactionRecord = record;

	request.setAttribute("Enhydra.RecorderServiceInterceptor.record",
		record);


        /*
         *  If we are monitoring an Enhydra servlet, we can reach into it and
         *  get the session data before the request is processed.
         *  If it's not an Enhydra servlet, then skip this step.
         *  If there is no valid incomming cookie, the session will not be
         *  found.
         *  Once we get the session, we keep a pointer to it because the
         *  app will call pasivate() on it, so it will no longer be
         *  accessable. By keeping the reference to it we can read the
         *  session data after the request is done, even if the session
         *  has been written to the database and removed from the
         *  session manager's hashtable.
         */
	HttpPresentationServlet lutrisServlet = null;

        /*
         *  If we are monitoring an Enhydra servlet we can get the
         *  session data before and after the request is processed.
         */
        if (servlet instanceof HttpPresentationServlet) {
            /*
             *  We are running in the Multiserver. It handles the filters,
             *  so we get the real Enhydra application.
             */
            lutrisServlet = (HttpPresentationServlet) servlet;
        } else if (servlet instanceof HttpPresentationServletGlue) {
            /*
             *  We are running the Enhydra Debugger in some other server, so
             *  the applications themselves handle the filters. Therefore
             *  we get the glue servlet. Look inside for the real application.
             */
            HttpPresentationServletGlue glue =
                             (HttpPresentationServletGlue) servlet;
            lutrisServlet = (HttpPresentationServlet) glue.getServlet();
        }


        HttpSession httpSession = null;
        Session session = null;

        javax.servlet.http.HttpSession hs;
        hs = request.getSession(false);
        if (hs instanceof Session) {
            session = (Session) hs;
        }

        //if (lutrisServlet != null) {
            //session = lutrisServlet.getSession(request);
            if (session == null) {
                record.setPreSessionString("<I>No Session</I>");
            } else {
                SessionData sd = session.getSessionData();
                String html = sd == null ? "" : sd.toHtml();
                if (html == null) {
                    html = "";
                }

                StringBuffer attributes = getAttribuestAsStringBuffer(hs);
                html+=attributes.toString();
                if (html.length() == 0) {
                    html = "<I>Empty</I>";
	        }
                record.setPreSessionString(html);
            }
        //} else {
          //  record.setPreSessionString("<I>N/A <font size=-1>(not an <a href=\"http://www.enhydra.org\" target=eWindow>Enhydra</a> application)</font></I>");
	//}

	if (session != null) {
	    request.setAttribute("Enhydra.RecorderServiceInterceptor.session",
		    session);
	}

        long startTime = System.currentTimeMillis();

	request.setAttribute("Enhydra.RecorderServiceInterceptor.startTime",
		new Long(startTime));
    }

    public void postInvoke(Context context, Servlet servlet,
	    HttpServletRequest request, HttpServletResponse response) {

        if (servlet instanceof InvokerServlet) {
            // do not debug the tomcat serverlet invoker
            return;
        }

        long stopTime = System.currentTimeMillis();

        Long startTimeLong = (Long) request.getAttribute(
		"Enhydra.RecorderServiceInterceptor.startTime");
	long startTime = 0L;
	if (startTimeLong != null) {
	    startTime = startTimeLong.longValue();
	}
        Session session = (Session) request.getAttribute(
		"Enhydra.RecorderServiceInterceptor.session");

        if (session == null) {
            javax.servlet.http.HttpSession hs;
            hs = request.getSession(false);
            if (hs instanceof Session) {
                session = (Session) hs;
            }
        }

        RequestRecord requestRecord = (RequestRecord) request.getAttribute(
		"Enhydra.RecorderServiceInterceptor.requestRecord");
        ResponseRecord responseRecord = (ResponseRecord) request.getAttribute(
		"Enhydra.RecorderServiceInterceptor.responseRecord");
        ServletTransactionRecord record = (ServletTransactionRecord)
		request.getAttribute(
		"Enhydra.RecorderServiceInterceptor.record");
	if (record == null) {
	    // There is no record!
	    // We must have had an exception in preInvoke.
	    return;
	}

        record.addTrace("Elapsed time: " + (stopTime - startTime) +
                " milliseconds.");
        record.stopTracing();

        /*
         *  If we are monitoring an Enhydra servlet, we can reach into
         *  it and get the session data after the request is processed.
         *  If it's not an Enhydra servlet, then skip this step.
         *  If there is no valid incomming cookie, the session will not
         *  be found. In the future check the outgoing cookies and use
         *  that if set. The first request will give you a null post
         *  session because it's using the old/wrong/no cookie. As a
         *  workaround, look at the pre-request session on the second
         *  request.
         *  We keep the session pointer around in case the session
         *  was paged out (see passivate).
         *  TO BE FIXED: We really should get the session via getSession()
         *  then passivate() it when done. But this would require
         *  additional API for the HttpPresentationServlet, which would
         *  break compatability with Enhydra 2.1. Once Enhydra 2.2 is
         *  widespread, then add it.
         */
        //if (lutrisServlet != null) {

            if (session == null) {
                record.setPostSessionString("<I>No Session</I>");
            } else {
                SessionData sd = session.getSessionData();
                if (sd == null) {
                    record.setPostSessionString("<I>No Session Data</I>");
                } else {
                    String html = sd.toHtml();
                    if (html == null) {
                        html = "";
                    }
                    HttpSession hs = request.getSession(false);
                    StringBuffer attributes = getAttribuestAsStringBuffer(hs);
                    html+=attributes.toString();
                    if (html.length() == 0) {
                        html = "<I>Empty</I>";
	            }
                    record.setPostSessionString(html);
                }
            }
        //} else {
        //    record.setPostSessionString("<I>N/A <font size=-1>(not an <a href=http://www.enhydra.org target=eWindow>Enhydra</a> application)</font></I>");
	//}

        debugManager.addTransactionRecord(record);
    }

    StringBuffer getAttribuestAsStringBuffer(HttpSession hs) {
        StringBuffer attributes = new StringBuffer();
        Enumeration enum = hs.getAttributeNames();
        if (hs != null) {
            attributes.append("<UL>\n");
            while (enum.hasMoreElements()) {
                String name = (String) enum.nextElement();
                Object value = hs.getAttribute(name);
                attributes.append("<LI> <TT>");
                attributes.append(name);
                attributes.append(": </TT>");
                attributes.append(value.toString());
                attributes.append ("\n");
            }
            attributes.append("</UL>\n");
        }
        return attributes;
    }
}
