/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Filter.java,v 1.4.2.1.2.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.filter;

import javax.servlet.Servlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.apache.tomcat.core.LifecycleInterceptor;
import org.apache.tomcat.core.ServiceInterceptor;
import org.apache.tomcat.core.Context;
import com.lutris.multiServer.MultiServer;
import org.enhydra.servlet.ServletContainer;
import org.enhydra.servlet.servletManager.ServletManager;
import org.enhydra.servlet.connectionMethods.Channel;

/**
 * This replaces the BasicTransactionFilter.
 * Also note that the BasicTransactionServlet is no longer used
 * and that it is functionally replaced by implementing the
 * org.apache.tomcat.core.LifecycleInterceptor and
 * org.apache.tomcat.core.ServiceInterceptor.
 * There is no interface defined for this, since this is a temporary
 * halfway point between the old Enhydra TransactionFilter and the
 * future Tomcat Interceptors.
 * Extend this class to create a filter with additional attributes.
 * The object model is not currently very clean because we need to
 * bridge between the Interceptor model of Tomcat for service(), init(),
 * and destroy(), while maintaining the Enhydra request and response
 * wrapping (Valve model).
 *
 * @author Shawn McMurdo
 */
abstract public class Filter {

    private ServiceInterceptor serviceInterceptor;
    private LifecycleInterceptor initInterceptor;
    private LifecycleInterceptor destroyInterceptor;
    private String description;

    /**
     * Default constructor.
     * Other filter elements must be set with the set methods.
     */
    public Filter() {
	this.description = "Standard Filter";
    }

    /**
     * Constructor given a description only.
     * Other filter elements must be set with the set methods.
     *
     * @param description a String describing the filter, suitable for
     *		a select list.
     */
    public Filter(String description) {
	this.description = description;
    }

    /**
     * Adds the service, init, and destroy interceptors to the
     * Tomcat Context for the given servlet.
     *
     * @param id The servlet id used to find the Context.
     */
    public void registerInterceptors(String id) {
	ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	ServletManager sm = sc.getServletManager();
	Context context = sm.getContextByServletID(id);
	if (serviceInterceptor != null) {
	    context.addServiceInterceptor(serviceInterceptor);
	}
	if (initInterceptor != null) {
	    context.addInitInterceptor(initInterceptor);
	}
	if (destroyInterceptor != null) {
	    context.addDestroyInterceptor(destroyInterceptor);
	}
    }

    /**
     * Removes the service, init, and destroy interceptors from the
     * Tomcat Context for the given servlet.
     *
     * @param id The servlet id used to find the Context.
     */
    public void unregisterInterceptors(String id) {
	ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	ServletManager sm = sc.getServletManager();
	Context context = sm.getContextByServletID(id);
	// Currently no way to remove Interceptors in Tomcat 3.0.
	if (serviceInterceptor != null) {
	    context.removeServiceInterceptor(serviceInterceptor);
	}
	if (initInterceptor != null) {
	    context.removeInitInterceptor(initInterceptor);
	}
	if (destroyInterceptor != null) {
	    context.removeDestroyInterceptor(destroyInterceptor);
	}
    }

    /**
     * Set the String description for this Filter.
     * This should be suitable for display in a select list.
     *
     * @param description is a String describing the filter.
     */
    public void setDescription(String description) {
	this.description = description;
    }

    /**
     * Get the String description for this Filter.
     *
     * @return a String describing the filter.
     */
    public String getDescription() {
	return description;
    }

    /**
     * Override toString since that is what the Admin used to use
     * to get the description.
     *
     * @return a String describing the filter.
     */
    public String toString() {
	return description;
    }

    /**
     * Get the RequestFilter for this Filter by wrapping the given Request.
     *
     * @param request an HttpServletRequest to be wrapped by the filter.
     * @param response an HttpServletResponse that is generally not used, but
     *		is available for request-response data coordination.
     * @return a RequestFilter
     */
    public RequestFilter wrapRequest(HttpServletRequest request,
	    HttpServletResponse response) {
	return new RequestFilter(request);
    }

    /**
     * Get the ResponseFilter for this Filter by wrapping the given Response.
     *
     * @param request an HttpServletRequest that is generally not used, but
     *		is available for request-response data coordination.
     * @param response an HttpServletResponse to be wrapped by the filter.
     * @return a ResponseFilter
     */
    abstract public ResponseFilter wrapResponse(HttpServletRequest request,
                                                HttpServletResponse response);

    /**
     * Set the ServiceInterceptor for this Filter.
     *
     * @param serviceInterceptor is a ServiceInterceptor
     */
    public void setServiceInterceptor(ServiceInterceptor serviceInterceptor) {
	this.serviceInterceptor = serviceInterceptor;
    }

    /**
     * Get the ServiceInterceptor for this Filter.
     *
     * @return a ServiceInterceptor
     */
    public ServiceInterceptor getServiceInterceptor() {
	return serviceInterceptor;
    }

    /**
     * Set the init() LifecycleInterceptor for this Filter.
     *
     * @param initInterceptor is a LifecycleInterceptor for init()
     */
    public void setInitInterceptor(LifecycleInterceptor initInterceptor) {
	this.initInterceptor = initInterceptor;
    }

    /**
     * Get the init() LifecycleInterceptor for this Filter.
     *
     * @return a LifecycleInterceptor for init()
     */
    public LifecycleInterceptor getInitInterceptor() {
	return initInterceptor;
    }

    /**
     * Set the destroy() LifecycleInterceptor for this Filter.
     *
     * @param destroyInterceptor is a LifecycleInterceptor for destroy()
     */
    public void setDestroyInterceptor(LifecycleInterceptor destroyInterceptor) {
	this.destroyInterceptor = destroyInterceptor;
    }

    /**
     * Get the destroy() LifecycleInterceptor for this Filter.
     *
     * @return a LifecycleInterceptor for destroy()
     */
    public LifecycleInterceptor getDestroyInterceptor() {
	return destroyInterceptor;
    }

}

