/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: MultiServerFilter.java,v 1.4.2.1.2.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.filter;

import com.lutris.util.KeywordValueTable;


/**
    This is the base class for filters specified by the administrator in the
    MultiServer config file. <P>

    Any filter specified in the MultiServer config file must have two
    propeties: it must implement the TransactionFilter interface, and
    it must have a constructor that takes a string and a KeywordValueTable.
    This abstract class provides these for you, which is why it is recommended
    that you extend this class if you need to write your own filter for
    the MultiServer. However, it is possible that a filter absolutly must
    extend some other class (for example, the RMI connection method can
    not extend StandardConnectionMethod because it must extend an object from
    the RMI library). Unless there is a specific reason, filters intended
    to be listed in the MultiServer config file should extend this class. <P>

    This is an abstract class; if you extend it you must implement the
    <CODE>wrap()</CODE> function. When the MultiServer reads it's conifig
    file, the server administrator may list filters to be applied to a
    channel. The classname of the filters is specified. It is the 
    responsibility of the MultiServer to create these filter objects
    and to install them into the correct channels. In order to be able
    to instantiate the filters, they must all have a well-known 
    constructor. That is what this class ensures. <P>

    The constructor, which will most likely be overridden, takes a
    description string and a KeywordValueTable (KVT). The KVT is used
    to allow a flexible way for initial arguments to be passed in to
    the filter. In the MultiServer config file, an entire hierarchy
    of key=value pairs can be specified as the initial arguments to
    the filter. This sub-tree is extracted to it's own KVT, and passed
    in to the filter. Therefore, even though all the filters must use
    the same constructor, the set of initial arguments is completely
    up to the developer of the filter.

    @see org.enhydra.servlet.filter.TransactionFilter
    @see com.lutris.util.KeywordValueTable
    @author Andy John
    @author Shawn McMurdo
*/
abstract public class MultiServerFilter extends Filter {

    /**
     *  Create a new filter. You will probably want to override this
     *  to process your initial arguments. The description is saved
     *  and will be returned by <CODE>toString()</CODE>. 
     *  The initial arguments are ignored. <P>
     *
     *  Note: if you define other constructors, which take other
     *  arguments, they will not be called. The MultiServer will be
     *  instantiating instances of this class, and it will only use
     *  this constructor.
     * 
     *  @param description A human readable description of this filter.
     *  @param initArgs A KeywordValueTable, populated with whatever
     *  key=value paris were specified as the initial arguments in the
     *  MultiServer config file. If no initial arguments were given
     *  in the config file, this will be null.
     */
    public MultiServerFilter(String description, KeywordValueTable initArgs) {
	super(description);
    }

}

