/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PrintTransactionFilter.java,v 1.4.2.1.2.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.filter;

import java.io.OutputStream;
import java.io.PrintWriter;
import javax.servlet.Servlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;


/**
    An implementation of TransactionFilter that logs everything the Servlet
    does to a given OutputStream. After logging, the calls are passed through. 

    @see org.enhydra.servlet.filter.TransactionFilter
    @see org.enhydra.servlet.filter.WrapResult
    @see java.io.OutputStream
    @author Andy John
*/
public class PrintTransactionFilter extends Filter {

    private String      description;
    private String      servletPrefix, requestPrefix, responsePrefix;
    private boolean     logIO;
    private PrintWriter out;
    private static int  tagCounter = 0;
    private static Object tagLock = new Object();
    private int         myTag;
    private boolean     servletUsedTag, reqUsedTag, respUsedTag;

    /**
        Create a new filter. 
    
        @param description A human readable description of this filter.
        @param servletPrefix A String to prepend to the beginning of each
        logging statement generated by calls to the Servlet.
        @param requestPrefix A String to prepend to the beginning of each
        logging statement generated by calls from the Servlet to the
        request object.
        @param responsePrefix A String to prepend to the beginning of each
        logging statement generated by calls from the Servlet to the
        response object.
        @param out Where to write the logging information to.
        @param logIO Should the data read from and written to the net
        be logged also? Can genereate large quantities of junk if images
        are being served.
    */
    public PrintTransactionFilter (String description, String servletPrefix,
            String requestPrefix, String responsePrefix, OutputStream out,
            boolean logIO) {
        this.description = description;
        this.requestPrefix = requestPrefix;
        this.responsePrefix = responsePrefix;
        this.servletPrefix = servletPrefix;
        this.out = new PrintWriter(out, true);
        this.logIO = logIO;
        reqUsedTag = respUsedTag = servletUsedTag = true;
    }


    /**
        Apply this filter to a Servlet, a request and a response.
        Each one is wrapped with a wrapper that passes all calls through,
        however a log of each call made will be written to the
        OutputStream passed in to the constructor. The method name, the
        input parameters, and the results will be logged. <P>

        All logging lines will start with the appropriate prefix passed in to
        the constructor, with an integer appeneded. Each time 
        <CODE>wrap()</CODE> is called a unique integer is appended to the 
        prefixes, so you can tell which request's output matches with 
        which result's output.

        @param originalObjects The real Servlet, request and response.
        All calls will be passed through to the origional object unchanged,
        after logging the call.
        @return A new Servlet, request and response, stored in a WrapResult
        object. Each object "wraps around" the original object.
        @see org.enhydra.servlet.filter.WrapResult
        @see javax.servlet.Servlet
        @see javax.servlet.http.HttpServletRequest
        @see javax.servlet.http.HttpServletResponse
    */
/* XXX
    public WrapResult wrap(WrapResult originalObjects) {
        int myTag = 0;
        synchronized (tagLock) {
            myTag = tagCounter++;
        }
        String prefix = servletPrefix + " " + myTag + ": ";
        PrintTransactionServlet srv = 
            new PrintTransactionServlet(originalObjects.servlet, prefix, out);
        prefix = requestPrefix + " " + myTag + ": ";
        PrintTransactionRequest req = 
            new PrintTransactionRequest(originalObjects.request, 
                            prefix, out, logIO);
        prefix = responsePrefix + " " + myTag + ": ";
        PrintTransactionResponse resp =
            new PrintTransactionResponse(originalObjects.response, 
                            prefix, out, logIO);
        return new WrapResult(srv, req, resp);
    }
XXX */

    public RequestFilter wrapRequest(HttpServletRequest request,
	    HttpServletResponse response) {
	int myTag = 0;
        synchronized (tagLock) {
            myTag = tagCounter++;
        }
	request.setAttribute("Enhydra.PrintTransactionFilter.myTag",
		new Integer(myTag));
        String prefix = requestPrefix + " " + myTag + ": ";
        PrintTransactionRequest req = 
            new PrintTransactionRequest(request, prefix, out, logIO);
	return req;
    }

    public ResponseFilter wrapResponse(HttpServletRequest request,
	    HttpServletResponse response) {
	int myTag = 0;
	Integer myTagInt = (Integer)
		request.getAttribute("Enhydra.PrintTransactionFilter.myTag");
	if (myTagInt != null) {
	    myTag = myTagInt.intValue();
	}
        String prefix = responsePrefix + " " + myTag + ": ";
        PrintTransactionResponse resp =
            new PrintTransactionResponse(response, prefix, out, logIO);
	return resp;
    }

    /**
        Returns a human-readable description string. It returns the same
        string passed in to the constructor.

        @return A description of this filter.
    */
    public String toString() {
        return description;
    }

}

