/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardLoggingResponse.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.filter;

import java.io.IOException;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletOutputStream;
import com.lutris.http.HttpStatus;

/**
 *
 *  @see org.enhydra.servlet.StandardLoggingFilter
 *  @see org.enhydra.servlet.StandardLoggingServlet
 *  @author Andy John
 */
public class StandardLoggingResponse extends ResponseFilter {

    // Data we pick up along the way.
    private String type;
    private int statusCode;
    private String action;
    private int contentLength;
    private boolean contentLengthSet;

    /**
     *  Constructor. You must pass in the original response, an id number,
     *  the log channel to log to, and the loc level to use.
     *  This class passes all calls through unmodified. It makes note of
     *  certain calls, storing the parameters. When the method
     *  <CODE>logData()</CODE> is called, it writes a line to the log
     *  channel, describing the
     *    
     *  @param originalResponse The real HttpServletResponse to route the calls
     *  to.
     *  @see javax.servlet.ServletResponse
     *  @see javax.servlet.http.HttpServletResponse
     */
    public StandardLoggingResponse (HttpServletResponse originalResponse) {
        super(originalResponse);
        contentLengthSet = false;
        statusCode = HttpStatus.ok; // If not set, the default is OK.   
	type = "text/html";
    }

    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  The type is stored, and will be logged when <CODE>logData()</CODE>
     *  is called.
     *
     *  @see javax.servlet.ServletResponse
     */
    public void setContentType (String type) {
        this.type = type;
        super.setContentType(type);
    }

    /**
     * This method allows servlets of this package, like the 
     * StandardLoggingServlet, to get the information that is grabbed as the
     * setContentType method is being called.
     *
     * @return the contentType of the response
     */
    public String getContentType() {
	return type;
    }

    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  The status is stored, and will be logged when <CODE>logData()</CODE>
     *  is called.
     *
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void setStatus(int sc, String sm) {
        statusCode = sc;
        super.setStatus(sc, sm);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  The status is stored, and will be logged when <CODE>logData()</CODE>
     *  is called.
     *
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void setStatus(int sc) {
        statusCode = sc;
        super.setStatus(sc);
    }

    /**
     * This method allows servlets of this package, like the 
     * StandardLoggingServlet, to get the information that is grabbed as the
     * setStatus method is being called.
     *
     * @return the status code of the response
     */
    public int getStatus() {
	return statusCode;
    }

    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  The status is stored, and will be logged when <CODE>logData()</CODE>
     *  is called.
     *
     *  @see javax.servlet.HttpServletResponse
     */
    public void setContentLength(int len) {
        contentLength = len;
        contentLengthSet = true;
        super.setContentLength(len);
    }

    /**
     * This method allows servlets of this package, like the 
     * StandardLoggingServlet, to get the information that is grabbed as the
     * setContentLength method is being called.
     *
     * @return the contentLength of the response
     */
    public int getContentLength() {
	return contentLength;
    }

    /**
     * Check to see if the contentLength was set
     *
     * @return the boolean value marking if the contentLength was set
     */
    public boolean wasContentLengthSet() {
	return contentLengthSet;
    }

    /**
     * This method is to return that action that was taken, if any.  There
     * will only be an action if sendError or sendRedirect was called
     *
     * @return the action that was taken if sendError was called and NULL if
     *    if sendError or sendRedirect was never called.
     */
    public String getAction() {
	return action;
    }

    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  The error is stored, and will be logged when <CODE>logData()</CODE>
     *  is called.
     *
     *  @exception IOException See HttpServletResponse documentation
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void sendError(int sc, String msg) throws IOException {
        action = "send error " + sc;
	statusCode = sc;
        super.sendError(sc, msg);
    }

    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  The error is stored, and will be logged when <CODE>logData()</CODE>
     *  is called.
     *
     *  @exception IOException See HttpServletResponse documentation
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void sendError(int sc) throws IOException {
        action = "send error " + sc;
	statusCode = sc;
        super.sendError(sc);
    }

    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  The redirect location is stored, and will be logged when 
     *  <CODE>logData()</CODE>
     *  is called.
     *
     *  @see javax.servlet.http.HttpServletResponse
     *  @exception IOException See HttpServletResponse documentation
     */
    public void sendRedirect(String location) throws IOException {
        action = "redirect " + location;
	statusCode = 302;
        super.sendRedirect(location);
    }
}





