/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraServletContext.java,v 1.5.2.2.2.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.servletManager;

import java.io.File;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;
import java.net.URL;
import java.net.MalformedURLException;
import javax.servlet.ServletContext;
import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.RequestDispatcher;
import com.lutris.logging.Logger;
import com.lutris.logging.LogChannel;
import com.lutris.classloader.MultiClassLoader;
import org.enhydra.servlet.ServletContainer;
import com.lutris.multiServer.MultiServer;


public class EnhydraServletContext implements ServletContext {

    private String           name;
    private LogChannel       logChannel;
    private ServletManager   servletManager;
    private Properties       initArgs;
    private String           docRoot;
    private MultiClassLoader classLoader;


    protected EnhydraServletContext(String name,
                                    LogChannel logChannel,
                                    ServletManager servletManager,
                                    Properties initArgs,
                                    String docRoot,
                                    MultiClassLoader classLoader) {
        this.name = name;
        this.logChannel = logChannel;
        this.servletManager = servletManager;
        if (initArgs == null)
            initArgs = new Properties();
        this.initArgs = initArgs;
        this.classLoader = classLoader;
        /*
         *  Be sure that doc root ends with a slash. This is important
         *  so that any string can be appended onto the end and the
         *  results will be a legal path.
         */
        if (docRoot == null)
            docRoot = "/";
        else if (!docRoot.endsWith("/"))
            docRoot += "/";
        this.docRoot = docRoot;
    }


    public Servlet getServlet(String name) throws ServletException {
        return servletManager.get(name);
    }

    // This is .deprecated
    public Enumeration getServlets() {
        return new Vector().elements();
    }

    public void log(String msg) {
        if (logChannel != null)
            logChannel.write(Logger.INFO, msg);
    }

    public void log(Exception except, String msg) {
        if (logChannel != null) {
            logChannel.write(Logger.INFO, msg, except);
        }
    }

    /**
     *  Transate a partial path to the real location on disk. This depends
     *  on what the servlet's docRoot was set to. For example, suppose a
     *  servlet is configured with the docRoot "/tmp/images/". Then
     *  "/tmp/images" will be passed into this object's constructor.
     *  If the servlet then asks for the path "/foo/bar.gif" to be translated,
     *  it will recieve "/tmp/images/foo/bar.gif".
     *
     * @param path The filname (probably the tail end of a URL) to translate.
     * @return A filename on the hard disk.
     */
    public String getRealPath(String path) {
        if (path == null)
            return docRoot;
        if (path.startsWith(File.separator))
            /*
             * We know docroot ends with a slash. Avoid two in a row.
             */
            return docRoot + path.substring(1);
        else
            return docRoot + path;
    }

    public String getMimeType(String file) {
        return servletManager.getMimeType(file);
    }

    public String getServerInfo() {
        return servletManager.getServerInfoString();
    }

    /**
     * Returns the major version of the Java Servlet API that this
     * Web server supports.
     *
     * @return 		2
     */
    public int getMajorVersion() {
	return 2;
    }

    /**
     * Returns the minor version of the Servlet API that this
     * Web server supports.
     *
     * @return 		2
     */
    public int getMinorVersion() {
	return 2;
    }

    /**
     * Returns a <code>ServletContext</code> object that
     * corresponds to a specified URL on the server.
     *
     * <p>This method allows servlets to gain
     * access to the resources located at a specified URL and obtain
     * {@link RequestDispatcher} objects from it.
     *
     * <p>In security conscious environments, the servlet engine can
     * return <code>null</code> for a given URL.
     *
     * @param uripath 	a <code>String</code> specifying the URL for
     *			which you are requesting a <code>ServletContext</code>
     *			object
     *
     * @return		the <code>ServletContext</code> object that
     *			corresponds to the named URL
     *
     * @see 		RequestDispatcher
     */
    public ServletContext getContext(String uripath) {
	// X22
	return null;
    }

    /**
     * Returns the resource that is mapped to a specified
     * path. The path must be in the form
     * <code>/dir/dir/file.ext</code>.
     *
     * <p>This method allows the Web
     * server to make a resource available to a servlet from
     * any source. Resources
     * can be located on a local or remote
     * file system, in a database, or on a remote network site.
     *
     * <p>This method can return <code>null</code>
     * if no resource is mapped to the pathname.
     *
     * <p>The servlet engine must implement the URL handlers
     * and <code>URLConnection</code> objects that are necessary
     * to access the resource.
     *
     * <p>This method has a different purpose than
     * <code>java.lang.Class.getResource</code>,
     * which looks up resources based on a class loader. This
     * method does not use class loaders.
     *
     * @param path 		a <code>String</code> specifying
     *				the path to the resource,
     * 				in the form <code>/dir/dir/file.ext</code>
     *
     * @return 			the resource located at the named path,
     * 				or <code>null</code> if there is no resource
     *				at that path
     *
     * @exception MalformedURLException 	if the pathname is not given in
     * 						the correct form
     *
     */
    public URL getResource(String path) throws MalformedURLException {
	// X22
	throw new Error("not implemented");
    }

    /**
     * Returns the resource located at the named path as
     * an <code>InputStream</code> object.
     *
     * <p>The data in the <code>InputStream</code> can be
     * of any type or length. The path must be of
     * the form <code>/dir/dir/file.ext</code>. This method
     * returns <code>null</code> if no resource exists at
     * the specified path.
     *
     * <p>Metainformation such as content length and content type
     * that is available when you use the <code>getResource</code>
     * method is lost when you use this method.
     *
     * <p>The servlet engine must implement the URL handlers
     * and <code>URLConnection</code> objects necessary to access
     * the resource.
     *
     * <p>This method is different from
     * <code>java.lang.Class.getResourceAsStream</code>,
     * which uses a class loader. This method allows servlet engines
     * to make a resource available
     * to a servlet from any location, without using a class loader.
     *
     * @param name 	a <code>String</code> specifying the path
     *			to the resource,
     * 			in the form <code>/dir/dir/file.ext</code>
     *
     * @return 		the <code>InputStream</code> returned to the
     *			servlet, or <code>null</code> if no resource
     *			exists at the specified path
     */
    public InputStream getResourceAsStream(String path) {
	// X22
	throw new Error("not implemented");
    }

    /**
     * Returns a {@link RequestDispatcher} object that acts
     * as a wrapper for the resource located at the named path.
     * You can use a <code>RequestDispatcher</code> object to forward
     * a request to the resource or include a resource in a response.
     *
     * <p>The pathname must be in the form <code>/dir/dir/file.ext</code>.
     * This method returns <code>null</code> if the <code>ServletContext</code>
     * cannot return a <code>RequestDispatcher</code>.
     *
     * <p>The servlet engine is responsible for wrapping the resource
     * with a <code>RequestDispatcher</code> object.
     *
     * @param urlpath 	a <code>String</code> specifying the pathname
     *			to the resource
     *
     * @return 		a <code>RequestDispatcher</code> object
     *			that acts as a wrapper for the resource
     *			at the path you specify
     *
     * @see 		RequestDispatcher
     */
    public RequestDispatcher getRequestDispatcher(String urlpath) {
	// X22
	throw new Error("not implemented");
    }

    /**
     * Writes the stack trace and an explanatory message
     * for a given <code>Throwable</code> exception
     * to the servlet log file. The stack trace is
     * part of the <code>Throwable</code> object, and
     * the message is the one you specify in the <code>message</code>
     * parameter. The name and type of the servlet log file is specific to
     * the servlet engine, but it is usually an event log.
     *
     * @param message 		a <code>String</code> that
     *				describes the error or exception
     *
     * @param throwable 	the <code>Throwable</code> error
     *				or exception
     */
    public void log(String message, Throwable throwable) {
	// X22
        if (logChannel != null) {
            logChannel.write(Logger.INFO, message, throwable);
        }
    }

    public Object getAttribute(String name) {
	// X22 shouldn't be using initArgs Properties
        return initArgs.getProperty(name);
    }

    /**
     * Returns an <code>Enumeration</code> containing the
     * attribute names available
     * within this servlet context. You can use the
     * {@link #getAttribute} method with an attribute name
     * to get the value of an attribute.
     *
     * @return 		an <code>Enumeration</code> of attribute
     *			names
     *
     * @see		#getAttribute
     */
    public Enumeration getAttributeNames() {
	// X22
	return initArgs.propertyNames();
    }

    /**
     * Gives an attribute a name in this servlet context. If
     * the name specified is already used for an attribute, this
     * method will overwrite the old attribute and bind the name
     * to the new attribute.
     *
     * <p>Attribute names should follow the same convention as package
     * names. The Java Servlet API specification reserves names
     * matching <code>java.*</code>, <code>javax.*</code>, and
     * <code>sun.*</code>.
     *
     * @param name 	a <code>String</code> specifying the name
     *			of the attribute
     *
     * @param object 	an <code>Object</code> representing the
     *			attribute to be given the name
     */
    public void setAttribute(String name, Object object) {
	// X22
	initArgs.put(name, object.toString());
    }

    /**
     * Removes the attribute with the given name from
     * the servlet context. If you remove an attribute, and
     * then use {@link #getAttribute} to retrieve the
     * attribute's value, <code>getAttribute</code> returns <code>null</code>.
     *
     * @param name	a <code>String</code> specifying the name
     * 			of the attribute to be removed
     */
    public void removeAttribute(String name) {
	// X22
	initArgs.remove(name);
    }

    /**
     * Returns a dispatcher for the given resource name.
     *
     * @param name
     *		A String resource name to use to find a RequestDispatcher.
     * @return a RequestDispatcher for the given resource name.
     */
    public RequestDispatcher getNamedDispatcher(String name) {
	// X22
	throw new Error("not implemented");
    }

    /**
     * Return the value for the given context parameter name.
     *
     * @param name
     *		A String context parameter name.
     * @return a String context parameter value.
     */
    public String getInitParameter(String name) {
	// X22
	return initArgs.getProperty(name);
    }

    /**
     * Return an Enumeration of all of the context parameter names.
     *
     * @return an Enumeration of String parameter names.
     */
    public Enumeration getInitParameterNames() {
	return initArgs.propertyNames();
    }


    /*******************************************************************
     *
     *  Extra functions that servlets may use if they realize that they
     *  are being passed a ServletManagerContext object rather than a
     *  generic ServletContext object. Enhydra applications use these.
     *
     ********************************************************************/

    /**
     * Returns the name of this servlet. This is the id string used
     * to identify the servlet to the ServletManager.
     *
     * @return The servlet's name.
     */
    public String getName() {
        return name;
    }


    /**
     * Return the <CODE>LogChannel</CODE> for the servlet. The servlet my
     * write directly to this channel, or use the <CODE>log()</CODE>
     * method, which will write to the log channel at the <CODE>INFO</CODE>
     * level.
     *
     * @return The servlet's LogChannel.
     */
    public LogChannel getLogChannel() {
        return logChannel;
    }

    /**
     * Return the class loader used to load the servlet. Enhydra applications
     * will use this to load the presentation objects, avoiding a second
     * redundant class loader.
     *
     * @return The class loaded used to load the servlet.
     */
    public MultiClassLoader getClassLoader() {
        return classLoader;
    }

    /**
     * Returns an enumeration of the Servlet object names in this server.
     */
    public Enumeration getServletNames() {
	ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
        return sc.getServletManager().getServletNames();
    }

}

