package com.simscomputing.testbed;

import com.simscomputing.SoftwareFaultException;
import com.simscomputing.util.Debugger;

import java.io.ByteArrayOutputStream;

/**
For each test, a TestResult object indicates whether the test passed and
how long the test took to execute in milliseconds.

@author David Sims
@author Dave Lamy
@version $Revision: 1.1.1.1 $ $Date: 2000/02/21 21:22:34 $
*/
public class TestResult implements Cloneable {
  /** The name of the test case collection. */
// fixme  private TestCaseCollection testCaseCollection;
//  private String testCaseCollectionName;
  private String testCaseCollectionClassName;

  /** The name of the test case. */
  private String testCaseName;

  /** Whether the test passed */
  private boolean passed;
  /** How long the test took to run in milliseconds. */
  private long executionTimeInMilliseconds;
  private Throwable abnormalFailureException;
  private String testFailureDescription;

  /**!!! 10-04-99 Begin D Lamy code modification !!!**/

  /** Stored stream of system.out data **/
  private ByteArrayOutputStream outStream;
  /** Stored stream of system.err data **/
  private ByteArrayOutputStream errStream;

  private boolean didNotRun = false;

  public TestResult(TestCaseCollection testCaseCollection,
                    String testCaseName) {
    didNotRun = true;
    this.testCaseCollectionClassName = testCaseCollection.getClass().getName();
    this.testCaseName = testCaseName;
  } // constructor

  public boolean didTestCaseRun() {
    return !didNotRun;
  } // didTestCaseRun

  /**
  Saves the results of running a test case.

  @param String testName
  @param boolean passed
  @param long executionTimeInMilliseconds
  @param outStream - a ByteArrayOutputStream that has all system.out data buffered in it
  @param errStream - a ByteArrayOutputStream that has all system.err data buffered in it
  @precondition passed ||
    (failureException != null &&
      failureException instanceof TestFailedException => failureException.getMessage() != null)
  */
  public TestResult(TestCaseCollection testCaseCollection,
                    String testCaseName,
                    boolean passed,
                    long executionTimeInMilliseconds,
                    Throwable failureException,
                    ByteArrayOutputStream outStream,
                    ByteArrayOutputStream errStream) {

    Debugger.checkPrecondition(passed || failureException != null,
			       "if the test failed, failureException must not be null");

//fixme    this.testCaseCollection = testCaseCollection;
    this.testCaseCollectionClassName = testCaseCollection.getClass().getName();
    this.testCaseName = testCaseName;
    this.passed = passed;
    this.executionTimeInMilliseconds = executionTimeInMilliseconds;
    this.outStream = outStream;
    this.errStream = errStream;

    if (!passed) {
      Debugger.checkPrecondition(failureException != null, "failureException must not be null");

      if (failureException instanceof com.simscomputing.testbed.TestFailedException) {
	testFailureDescription = ((TestFailedException) failureException).getMessage();
	Debugger.checkPrecondition(testFailureDescription != null,
				   "failureDescription must not be null");
      } // if
      else {
	this.abnormalFailureException = failureException;
      } // else
    } // if
  } // constructor

  /**!!! 10-04-99 End D Lamy code modification !!!**/



  ///////////////
  // accessors //
  ///////////////

  /**
  Return the testName attribute.

  @return String - testName
  */
  public String getTestCaseName() {
    return testCaseName;
  } // getTestName()

//  public String getTestCaseCollectionName() {
//    return testCaseCollectionName; // fixme: testCaseCollection.getClass().getName();
//  } // getTestName()

  public String getTestCaseCollectionClassName() {
    return testCaseCollectionClassName;
  } // getTestCaseCollectionClassName()

  // fixme
//  public TestCaseCollection hi() {
//    return testCaseCollection;
//  } // getTestCaseCollection()

//  public Class getTestCaseCollection() {
//    return testCaseCollection.getClass();
//  } // getTestCaseCollection()

  public String getLongName() {
    return getTestCaseCollectionClassName() + "." + getTestCaseName();
  } // getLongName()

  /**
  Return the passed attribute.

  @return boolean - passed
  */
  public boolean didTestPass() {
    return passed;
  } // getTestResult()

  public boolean didTestFail() {
    return !didTestPass();
  } // didTestFail()

  public boolean didTestFailAbnormally() {
    return abnormalFailureException != null;
  } // didTestFailAbnormally()

  /**
  Accesses the failure exception, if any.

  @return the exception that caused the test to fail, if any
  */
  public Throwable getTestFailureException() {
    Debugger.checkPrecondition(didTestFailAbnormally(), "test must have failed abnormally");
    return abnormalFailureException;
  } // getFailureException()

  public String getTestFailureDescription() {
    Debugger.checkPrecondition(didTestFail() || !didTestFailAbnormally(),
			       "test must have failed but not abnormally");
    return testFailureDescription;
  } // getTestFailureDescription()

  /**
  Return the executionTimeInMilliseconds attribute.

  @return long - executionTimeInMilliseconds
  */
  public long getExecutionTimeInMilliseconds() {
    return executionTimeInMilliseconds;
  } // getExecutionTimeInMilliseconds()

  public Object clone() {
    try {
      return (TestResult) super.clone();
    } // try
    catch (CloneNotSupportedException e) {
      throw new SoftwareFaultException("clone should work", e);
    } // catch
  } // clone()

  /** !!! 10-04-99 Begin D Lamy code block !!! **/
  public String getOutStreamData() {
    return outStream.toString();
  } // getOutStreamData

  public String getErrStreamData() {
    return errStream.toString();
  } // getOutStreamData

  /** !!! 10-04-99 End D Lamy code block !!! **/  

} // class TestResult
