package com.simscomputing.testbed;

import com.simscomputing.SoftwareFaultException;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.SortedMap;
import java.util.TreeMap;

/**
Provides methods that operate against a collection of TestResult objects.

@author David Sims
@version $Revision: 1.1.1.1 $ $Date: 2000/02/21 21:22:34 $
*/
public class TestResultCollection implements Cloneable {
  TreeMap /* TestResult */ testResults = new TreeMap();

  public Object clone() {
    try {
      TestResultCollection result = (TestResultCollection) super.clone();
      result.testResults = (TreeMap) testResults.clone();
      return result;
    } // try
    catch (CloneNotSupportedException e) {
      throw new SoftwareFaultException("should be able to clone", e);
    } // catch
  } // clone()

  public void add(TestResult testResult) {
    this.testResults.put(testResult.getLongName(), testResult);
  } // add()

  public void add(TestResultCollection testResultCollection) {
    final Iterator iterator = testResultCollection.getIterator();
    TestResult testResult;

    while (iterator.hasNext()) {
      testResult = (TestResult) iterator.next();
      add(testResult);
    } // while
  } // add()

  public Iterator getIterator() {
    return testResults.values().iterator();
  } // getIterator()

  public boolean didAllTestsPass() {
    Iterator iterator = getIterator();
    boolean result = true;

    while (iterator.hasNext()) {
      TestResult testResult = (TestResult) iterator.next();
      if (!testResult.didTestPass()) {
        result = false;
        break;
      } // if
    } // while

    return result;
  } // didAllTestsPass()

  /**
  Returns the number of test case results encapsulated in this object.

  @returns int the number of test cases
  */
  public int getNumberOfTests() {
    return testResults.size();
  } // getNumberOfTests()

  /**
  Returns the number of tests that passed

  @returns int - the number of tests that passed
  */
  public int getNumberOfTestsThatPassed() {
    Iterator iterator = getIterator();
    int result = 0;

    while (iterator.hasNext()) {
      TestResult testResult = (TestResult) iterator.next();
      if (testResult.didTestPass()) {
	++result;
      } // if
    } // while

    return result;
  } // getNumberOfTestsThatPassed()

  public boolean getTestResult(final String testName) throws NoSuchTestException {
    TestResult testResult = (TestResult) testResults.get(testName);
    if (testResult == null) {
      throw new NoSuchTestException();
    } // if

    return testResult.didTestPass();
  } // getTestResult()

  /**
  */
  public TestResult[] getClonedTestResults() {
    Object[] results = testResults.values().toArray();
    TestResult[] resultArray = new TestResult[results.length];
    for (int index = 0; index < results.length; ++index) {
      resultArray[index] = ((TestResult) ((TestResult) results[index]).clone());
    } // for index

    return resultArray;
  } // getTestResults()

  /**
  Returns the time it took to execute a sub-test in Test.
  */
  public long getTestExecutionTimeInMilliseconds(final String testName)
    throws NoSuchTestException {
    TestResult testResult = (TestResult) testResults.get(testName);
    if (testResult == null) {
      throw new NoSuchTestException();
    } // if

    return testResult.getExecutionTimeInMilliseconds();
  } // getTestExecutionTimeInMilliseconds()

  /**
  Returns the time it took to execute all sub-tests in Test.

  @return long - cumulative test time in milliseconds
  */
  public long getCumulativeTestExecutionTimeInMilliseconds() {
    long result = 0;
    Iterator iterator = testResults.values().iterator();
    TestResult testResult;

    while (iterator.hasNext()) {
      testResult = (TestResult) iterator.next();
      result += testResult.getExecutionTimeInMilliseconds();
    } // while

    return result;
  } // getCollectiveTestExecutionTimeInMilliseconds()

  /**
  Returns all the test results.

  @param SortedMap - all test results
  */
//    public SortedMap getAllTestResults() {
//      return testResults;
//    } // getTestResults()

// fixme
  public TestResult[] findTestResults(final String /*Class*/ testCaseCollectionClassName) {
    final LinkedList result = new LinkedList();
    final Iterator iterator = testResults.values().iterator();
    TestResult testResult;

//      System.out.println("top of iterator: " + testCaseCollection.getName());
    while (iterator.hasNext()) {
      testResult = (TestResult) iterator.next();
//        System.out.println("comparing " + testCaseCollection.getName() +
//                           " with " + testResult.getTestCaseCollection().getName());
      if (testCaseCollectionClassName.equals(testResult.getTestCaseCollectionClassName())) {
        result.add(testResult);
      } // if
    } // while
//      System.out.println("bottom of iterator: " + testCaseCollection.getName());
//      System.out.println("ready to return array of size " + result.size());

    final TestResult[] arrayResult = new TestResult[result.size()];

    for (int index = 0; index < arrayResult.length; ++index) {
      arrayResult[index] = (TestResult) result.get(index);
    } // for index

    return arrayResult;
  } // findTestResults()
} // class TestResultCollection
