// Common: classes common to many projects
// Copyright 1999 Sims Computing
// This software is licensed under the GNU General Public License,
// available at http://www.gnu.org/copyleft/gpl.html.

package com.simscomputing.util;

/**
Supports debugging operations.

@author <a href="http://www.simscomputing.com/">Sims Computing</a>
@version $Revision: 1.1.1.1 $ $Date: 2000/02/21 21:22:36 $
*/
public final class Debugger {
  /** Indicates whether debugging is enabled. */
  public static final boolean DEBUGGING_ON = true;

  /** Indicates whether assertions are checked. */
  public static final boolean ASSERTIONS_ON = true && DEBUGGING_ON;

  /** Indicates whether preconditions are checked. */
  public static final boolean PRECONDITIONS_ON = true && DEBUGGING_ON;

  /** Indicates whether postconditions are checked. */
  public static final boolean POSTCONDITIONS_ON = true && DEBUGGING_ON;

  /** Indicates whether class invariants are checked. */
  public static final boolean INVARIANTS_ON = true && DEBUGGING_ON;

  /**
  Do not allow this class to be instantiated.
  */
  private Debugger() {
    // do nothing
  } // constructor

  /**
  For debugging purposes, ensures that the boolean condition is satisfied. The
  boolean condition passed in should not have any side effects.

  @param boolean - the condition to check
  @throw AssertionException - if 'condition' is false
  */
  public static final void assert(boolean condition) {
    if (ASSERTIONS_ON && !condition) {
      throw new AssertionException("assertion failed");
    } // if
  } // assert()

  /**
  For debugging purposes, ensures that the boolean condition is satisfied. The
  boolean condition passed in should not have any side effects.

  @param boolean the condition to check
  @param String the assertion message.
  @throw AssertionException if 'condition' is false
  */
  public static final void assert(boolean condition, String message) {
    if (ASSERTIONS_ON && !condition) {
      throw new AssertionException(message);
    } // if
  } // assert()

  /**
  For debugging purposes, ensures that the precondition is satisfied. The
  precondition passed in should not have any side effects.

  @param boolean - the condition to check
  @throw BrokenContractException - if 'condition' is false
  */
  public static final void checkPrecondition(boolean condition, String message) {
    if (PRECONDITIONS_ON && !condition) {
      throw new UnsatisfiedPreconditionException(message);
    } // if
  } // checkPrecondition()

  /**
  For debugging purposes, ensures that the postcondition is satisfied. The
  postcondition passed in should not have any side effects.

  @param boolean - the condition to check
  @throw BrokenContractException - if 'condition' is false
  */
  public static final void checkPostcondition(boolean condition, String message) {
    if (POSTCONDITIONS_ON && !condition) {
      throw new UnsatisfiedPostconditionException(message);
    } // if
  } // checkPostcondition()

  /**
  For debugging purposes, ensures that the class invariant is satisfied. The
  invariant passed in should not have any side effects.

  @param boolean - the invariant to check
  @throw BrokenContractException - if 'invariant' is false
  */
  public static final void checkInvariant(boolean invariant, String message) {
    if (INVARIANTS_ON && !invariant) {
      throw new UnsatisfiedInvariantException(message);
    } // if
  } // checkInvariant()
} // Debugger
