package com.simscomputing.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.File;
import java.io.IOException;

/**
Handy utilities for manipulating files and directories in a file system.

@author David Sims
@version $Revision: 1.1.1.1 $ $Date: 2000/02/21 21:22:36 $
*/
public class FileUtilities {
  /**
  Recursively deletes all files and directories in the specified
  directory. The specified directory is also removed.
  */
  public static void deleteDirectory(final String directoryPath) {
    final File directory = new File(directoryPath);
    if (directory.exists() && !directory.isDirectory()) {
      throw new IllegalArgumentException(directoryPath + " is not a directory");
    } // if

    // if the directory doesn't exist, our work is already done
    if (!directory.exists()) {
      return;
    } // if

    // first, find all files in this directory and delete them
    final String[] entries = directory.list();
    final int size = entries.length;
    File entry;

    for (int index = 0; index < size; ++index) {
      entry = new File(directoryPath + File.separatorChar + entries[index]);

      // if this entry is a file, delete it
      if (entry.isFile()) {
        entry.delete();
      } // if

      // if this entry is a directory, delete it and all its contents
      if (entry.isDirectory()) {
        deleteDirectory(directoryPath + File.separatorChar + entries[index]);
      } // if
    } // for index

    // finally, delete this empty directory
    directory.delete();
  } // deleteDirectory()

  

  /**
  Copies the specified file to the specified directory.
  */
  public static void copyFile(final String filePath, final String targetDirectory) 
      throws IOException {
    BufferedInputStream in;
    BufferedOutputStream out;

    in = new BufferedInputStream(new FileInputStream(filePath));
    final String outPath = targetDirectory + File.separatorChar + getBasename(filePath);
    out = new BufferedOutputStream(new FileOutputStream(outPath));

    final int BUFFER_SIZE = 1024 * 8;
    final byte[] buffer = new byte[BUFFER_SIZE];
    int bytes;
    final long fileSize = new File(filePath).length();
    long length;

    // first, calculate the number of bytes to be copied. Then keep
    // copying in blocks until it's done.
    for (length = fileSize; length > 0; ) {

      // read in some bytes
      bytes = (length > BUFFER_SIZE ? BUFFER_SIZE : (int) length);
      bytes = in.read(buffer, 0, bytes);
      if (bytes < 0) {
        break;
      } // if
      
      // reduce the number of bytes to be copied
      length -= bytes;

      // write out those same bytes
      out.write(buffer, 0, bytes);
    } // for

    in.close();
    out.close();
  } // copyFile()

  /**
  Returns the base name of the specified file path. Eg,
  getBasename("hello/there/world") returns
  "world". getBasename("hello") returns "hello".
  */
  public static String getBasename(final String filePath) {
    // work backwards looking for a file separator character

    // throw away the empty string case
    final int size = filePath.length();
    if (size == 0) {
      return filePath;
    } // if

    // throw away the case that has a trailing separator character
    if (filePath.charAt(size - 1) == File.separatorChar) {
      return "";
    } // if

    // find the substring from the last separator character to the end
    // of the string
    for (int index = size - 1; index >= 0; --index) {
      if (filePath.charAt(index) == File.separatorChar) {
        return filePath.substring(index, size);
      } // if
    } // for index
    
    // no separator character found. The filePath *is* the basename.
    return filePath;
  } // getBasename()
} // class FileUtilities
