// Copyright 2000 Sims Computing, Inc.
// Licensed under the GNU Lesser General Public License.

package com.simscomputing.util;

import com.simscomputing.SoftwareFaultException;
import com.simscomputing.util.Debugger;

import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Locale;

/**
Implements a mutable string class. Implements all constructors and methods from
the String and StringBuffer classes. Some additional constructors and methods
are also implemented. This class is not thread safe.

<p>

<strong>Note:</strong> it has been discovered that the Sun implementation of
String and StringBuffer is written such that a StringBuffer can be converted
into a String without copying data. Consequently, the String API can
be used on a StringBuffer. This revelation may make this MutableString
class unnecessary.

@author David Sims
@version $Revision: 1.1.1.1 $ $Date: 2000/02/21 21:22:36 $
*/
public class MutableString implements Serializable {

  /**************/
  /* attributes */
  /**************/

  /** The mutable string is stored internally as an ArrayList. */
  ArrayList stringArray = new ArrayList();

  /****************/
  /* constructors */
  /****************/

  /**
  Constructs an empty MutableString.
  */
  public MutableString() {}

  /**
  Constructs an empty FastString with capacity equal to 'length'.

  @throws NegativeArraySizeException if the length argument is less than 0.
  */
  public MutableString(int length) {
    // make sure the string is big enough to hold 'length' elements
    stringArray.ensureCapacity(length);
  } // constructor

  /**
  Constructs a MutableString that is initialized to 'string'.
  */
  public MutableString(String string) {
    for (int index = 0; index < string.length(); ++index) {
      stringArray.add(new Character(string.charAt(index)));
    } // for index
  } // constructor

  /**
  Constructs a MutableString that is initialized to 'string'.
  */
  /*
  public MutableString(StringBuffer string) {
    for (int index = 0; index < string.length(); ++index) {
      string.add(new Character(string.charAt(index)));
    } // for index
  } // constructor
  */

  /**
  Constructs a MutableString that is initialized to 'string'.
  */
  public MutableString(MutableString string) {
    // deep copy the string
    stringArray.ensureCapacity(string.length());
    for (int index = 0; index < string.length(); ++index) {
      stringArray.add(string.stringArray.get(index));
    } // for index
  } // constructor

  /**
  Constructs a MutableString that is initialized to 'value'.

  @throws NullPointerException - if 'value' is null
  */
  public MutableString(char[] value) {
    this(value, 0, value.length);
  } // constructor

  /**
  Constructs a MutableString that is initialized to
  the string 'value[offset]..value[offset+count-1]'.

  @throws IndexOutOfBoundsException if the offset and count arguments index
  characters outside the bounds of the value string.
  @throws NullPointerException if value is null.
  */
  public MutableString(char[] value,
                       int offset,
                       int count) {
    stringArray.ensureCapacity(count);
    for (int index = offset; index < offset + count; ++index) {
      stringArray.add(new Character(value[index]));
    } // for index
  } // constructor

  /**
  Constructs a MutableString that is initialized to
  the string 'bytes[offset]..bytes[offset+length-1]'
  using the specified character encoding.

  @throws UnsupportedEncodingException if the named encoding is not supported
  @throws IndexOutOfBoundsException if the offset and count arguments index characters
  outside the bounds of the value string.
  */
  public MutableString(byte[] bytes,
                       int offset,
                       int length,
                       String enc)
                       throws UnsupportedEncodingException {
    constructor(bytes, offset, length, enc);
  } // constructor

  /**
  Constructs a MutableString that is initialized to 'bytes'
  using the specified character encoding.

  @throws UnsupportedEncodingException if the named encoding is not supported
  */
  public MutableString(byte[] bytes,
                       String enc)
                       throws UnsupportedEncodingException {
    this(bytes, 0, bytes.length, enc);
  } // constructor

  /**
  Constructs a MutableString that is initialized to
  the string 'bytes[offset]..bytes[offset+length-1]'.
  */
  public MutableString(byte[] bytes,
                       int offset,
                       int length) {
    try {
      constructor(bytes, offset, length, null);
    } // try
    catch (UnsupportedEncodingException e) {
      throw new SoftwareFaultException("statement should not be reached");
    } // catch
  } // constructor

  /**
  Constructs a MutableString that is initialized to 'bytes'.

  @param byte[] - the MutableString is initialized from these bytes
  */
  public MutableString(byte[] bytes) {
    try {
      constructor(bytes, 0, bytes.length, null);
    } // try
    catch (UnsupportedEncodingException e) {
      throw new SoftwareFaultException("statement should not be reached");
    } // catch
  } // constructor

  /**
  A method used in constructing the object.

  @param byte[] - initialize the object from these bytes
  @param int - begin initializing from index 'offset'
  @param int - take 'length' bytes from 'bytes'
  @param String - the encoding used
  */
  private void constructor(byte[] bytes,
                           int offset,
                           int length,
                           String enc)
               throws UnsupportedEncodingException {
    // in case a non-constructor method calls this method
    stringArray.clear();
    
    if (enc == null) {
      // use default encoding
      stringArray.ensureCapacity(length);
      for (int index = offset; index < offset + length; ++index) {
        stringArray.add(new Character( (char) bytes[index]));
      } // for index
    } // if
    else {
      // use supplied encoding via the String class
      // note: is there a way we can do encodings directly?
      String encodedString = new String(bytes, offset, length, enc);
      for (int index = 0; index < encodedString.length(); ++index) {
        stringArray.add(new Character( (char) encodedString.charAt(index)));
      } // for index
    } // else
  } // constructor()

  /***********/
  /* methods */
  /***********/

  /**
  Returns the size of the string.

  @return int - the size of the string
  */
  public int length() {
    return stringArray.size();
  } // length()

  /**
  Returns string[index].

  @return char - the character at position 'index'
  */
  public char charAt(int index) {
    if (index < 0 || index >= stringArray.size()) {
      throw new StringIndexOutOfBoundsException();
    } // if

    return ((Character) stringArray.get(index)).charValue();
  } // charAt()

  /**
  Copies MutableString[srcBegin..srcEnd - 1] into 'dst', beginning at offset
  'dstBegin'.

  @param int - the beginning offset into this object
  @param int - the ending offset into this object
  @param char[] - the container for the results
  @param int - the offset into 'dst' to begin inserting the results

  IndexOutOfBoundsException - If any of the following is true:
  srcBegin is negative.
  srcBegin is greater than srcEnd
  srcEnd is greater than the length of this string
  dstBegin is negative
  dstBegin+(srcEnd-srcBegin) is larger than dst.length
  NullPointerException - if dst is null
  */
  public void getChars(int srcBegin,
                       int srcEnd,
                       char[] dst,
                       int dstBegin) {
    // note: the following 'if' statement would throw a NullPointerException
    // if 'dst' were null. Checked here anyway; the code is easier to
    // understand.
    if (dst == null) {
      throw new NullPointerException();
    } // if

    if (srcBegin < 0 || srcBegin > srcEnd || srcEnd > stringArray.size() ||
        dstBegin < 0 || dstBegin + (srcEnd - srcBegin) > dst.length) {
      throw new IndexOutOfBoundsException();
    } // if

    int dstIndex = dstBegin;
    for (int index = srcBegin; index < srcEnd; ++index) {
      dst[dstIndex] = ((Character) stringArray.get(index)).charValue();
      ++dstIndex;
    } // for index
  } // getChars()

  /**
  Returns the MutableString in byte[] form.

  @param String - the encoding to use
  @return byte[] - the string as bytes using encoding 'enc'
  */
  public byte[] getBytes(String enc)
                         throws UnsupportedEncodingException {
    byte[] result = new byte[stringArray.size()];
    for (int index = 0; index < stringArray.size(); ++index) {
      result[index] = (byte) ((Character) stringArray.get(index)).charValue();
    } // for index

    if (enc != null) {
      String s = new String(result);
      return s.getBytes(enc);
    } // else

    return result;
  } // getBytes()

  /**
  Returns the MutableString in byte[] form.

  @return byte[] - the string as bytes
  */
  public byte[] getBytes() {
    try {
      return getBytes(null);
    } // try
    catch (UnsupportedEncodingException e) {
      throw new SoftwareFaultException("statement should not be reached");
    } // catch
  } // getBytes()

  /**
  Indicates whether 'anObject' is equivalent to this object. 'anObject' must
  be an instance of MutableString.

  @param Object - the object to compare to this object
  @return boolean - whether the objects match
  */
  public boolean equals(Object anObject) {
    /*
    if (anObject instanceof String) {
      return equals(new MutableString( (String) anObject), true);
    } // if
    */

    if (!(anObject instanceof MutableString)) {
      return false;
    } // if

    // 'anObject' is a MutableString
    MutableString aString = (MutableString) anObject;
    return equals(aString, true);
  } // equals()

  /**
  Indicates whether 'anObject' is equivalent to this object.

  @param String - the object to compare to this object
  @return boolean - whether the objects match
  */
  public boolean equals(String anObject) {
    return equals(new MutableString( (String) anObject), true);
  } // equals()

  /**
  Indicates whether 'anotherString' equals this object. Case is ignored.

  @param MutableString - the MutableString to compare to this object
  @return boolean - whether the MutableStrings match
  */
  public boolean equalsIgnoreCase(MutableString anotherString) {
    return equals(anotherString, false);
  } // equalsIgnoreCase()

  /**
  Indicates whether 'anotherString' equals this object. Case is ignored.

  @param String - the string to compare to this object
  @return boolean - whether the objects match
  */
  public boolean equalsIgnoreCase(String anotherString) {
    return equalsIgnoreCase(new MutableString(anotherString));
  } // equalsIgnoreCase()

  /**
  Indicates whether 'anotherString' equals this object.

  @param MutableString - the string to compare to this object
  @param boolean - whether the comparison should match case
  @return boolean - whether the objects match
  */
  private boolean equals(MutableString anotherString, boolean matchCase) {
    if (anotherString.stringArray.size() != stringArray.size()) {
      return false;
    } // if

    // check if all the characters match
    char leftChar;
    char rightChar;

    for (int index = 0; index < stringArray.size(); ++index) {
      if (matchCase) {
        leftChar = ((Character) stringArray.get(index)).charValue();
        rightChar = ((Character) anotherString.stringArray.get(index)).charValue();
      } // if
      else {
        leftChar = Character.toLowerCase( ((Character) stringArray.get(index)).charValue());
        rightChar = Character.toLowerCase( ((Character) anotherString.stringArray.get(index)).charValue());
      } // else

      if (leftChar != rightChar) {
        // at least one pair of characters don't match
        return false;
      } // if
    } // for index

    // all the characters match
    return true;
  } // equals()

  /**
  Returns an integer that represents how the MutableString compares
  to 'anotherString'. The comparison result is computed as follows:

  If length() != anotherString.length(), result is
  length() - anotherString.length().
  Otherwise:
  If this object exactly matches 'anotherString', result is 0.
  Otherwise:
  For the first character position where the characters from the two strings
  do not match, result is 'string[position] - anotherString[position]'.

  @param MutableString - the string to compare to this object
  @return int - the comparison result
  @throws NullPointerException - if anotherString is null.
  */
  public int compareTo(MutableString anotherString) {
    return compareTo(anotherString, true);
  } // compareTo()

  /**
  Returns an integer that represents how the MutableString compares
  to 'anotherString'. The comparison result is computed as follows:

  If length() != anotherString.length(), result is
  length() - anotherString.length().
  Otherwise:
  If this object exactly matches 'anotherString', result is 0.
  Otherwise:
  For the first character position where the characters from the two strings
  do not match, result is 'string[position] - anotherString[position]'.

  @param String - the string to compare to this object
  @return int - the comparison result
  @throws NullPointerException - if anotherString is null.
  */
  public int compareTo(String anotherString) {
    return compareTo(new MutableString(anotherString), true);
  } // compareTo()

  /**
  Returns an integer that represents how the MutableString compares
  to 'anotherString'. The comparison result is computed as follows:

  If 'anotherString' is not a String, result is 0.
  Otherwise:
  If length() != anotherString.length(), result is
  length() - anotherString.length().
  Otherwise:
  If this object exactly matches 'anotherString', result is 0.
  Otherwise:
  For the first character position where the characters from the two strings
  do not match, result is 'string[position] - anotherString[position]'.

  @param MutableString - the string to compare to this object
  @return int - the comparison result
  @throws NullPointerException - if anotherString is null.
  */
  public int compareTo(Object anotherString) {
    if (!(anotherString instanceof MutableString)) {
      return 0;
    } // if

    MutableString string = (MutableString) anotherString;
    return compareTo(string, true);
  } // compareTo()

  /**
  Returns an integer that represents how the MutableString compares
  to 'anotherString'. Caes is ignored. The comparison result is computed as
  follows:

  If length() != anotherString.length(), result is
  length() - anotherString.length().
  Otherwise:
  If this object exactly matches 'anotherString', result is 0.
  Otherwise:
  For the first character position where the characters from the two strings
  do not match, result is 'string[position] - anotherString[position]'.

  @param MutableString - the string to compare to this object
  @return int - the comparison result
  @throws NullPointerException - if anotherString is null.
  */
  public int compareToIgnoreCase(MutableString string) {
    return compareTo(string, false);
  } // compareToIgnoreCase()

  /**
  Returns an integer that represents how the MutableString compares
  to 'anotherString'. Case is ignored. The comparison result is computed as
  follows:

  If length() != anotherString.length(), result is
  length() - anotherString.length().
  Otherwise:
  If this object exactly matches 'anotherString', result is 0.
  Otherwise:
  For the first character position where the characters from the two strings
  do not match, result is 'string[position] - anotherString[position]'.

  @param String - the string to compare to this object
  @return int - the comparison result
  @throws NullPointerException - if anotherString is null.
  */
  public int compareToIgnoreCase(String string) {
    return compareTo(new MutableString(string), false);
  } // compareToIgnoreCase()

  /**
  Returns an integer that represents how the MutableString compares
  to 'anotherString'. The comparison result is computed as follows:

  If length() != anotherString.length(), result is
  length() - anotherString.length().
  Otherwise:
  If this object exactly matches 'anotherString', result is 0.
  Otherwise:
  For the first character position where the characters from the two strings
  do not match, result is 'string[position] - anotherString[position]'.

  @param MutableString - the string to compare to this object
  @return int - the comparison result
  @throws NullPointerException - if anotherString is null.
  */
  private int compareTo(MutableString anotherString, boolean matchCase) {
    if (stringArray.size() != anotherString.length()) {
      return stringArray.size() - anotherString.length();
    } // if

    // check if all the characters match
    char leftChar;
    char rightChar;

    for (int index = 0; index < stringArray.size(); ++index) {
      if (matchCase) {
        leftChar = ((Character) stringArray.get(index)).charValue();
        rightChar = ((Character) anotherString.stringArray.get(index)).charValue();
      } // if
      else {
        leftChar = Character.toLowerCase( ((Character) stringArray.get(index)).charValue());
        rightChar = Character.toLowerCase( ((Character) anotherString.stringArray.get(index)).charValue());
      } // else

      if (leftChar != rightChar) {
        // at least one pair of characters don't match
        if (matchCase) {
          return ((Character) stringArray.get(index)).charValue() -
                 ((Character) anotherString.stringArray.get(index)).charValue();
        } // if
        else {
          return leftChar - rightChar;
        } // else
      } // if
    } // for index

    // all the characters match
    return 0;
  } // compareTo()

  /**
  Indicates whether the MutableString[toffset..toffset + len] matches
  other[ooffset..ooffset + len], ignoring case.

  @param int - the offset into this object
  @param MutableString - the string to compare against this object
  @param int - the offset into 'other'
  @param int - the number of characters to compare
  @returns boolean - whether the two strings match
  @throws NullPointerException - if 'other' is null
  */
  public boolean regionMatches(int toffset,
                               String other,
                               int ooffset,
                               int len) {
    return regionMatches(toffset, new MutableString(other), ooffset, len);
  } // regionMatches()

  /**
  Indicates whether the MutableString[toffset..toffset + len] matches
  other[ooffset..ooffset + len], ignoring case.

  @param int - the offset into this object
  @param MutableString - the string to compare against this object
  @param int - the offset into 'other'
  @param int - the number of characters to compare
  @returns boolean - whether the two strings match
  @throws NullPointerException - if 'other' is null
  */
  public boolean regionMatches(int toffset,
                               MutableString other,
                               int ooffset,
                               int len) {
    return regionMatches(false, toffset, other, ooffset, len);
  } // regionMatches()

  /**
  Indicates whether the MutableString[toffset..toffset + len] matches
  other[ooffset..ooffset + len], ignoring case.

  @param boolean - whether to ignore case
  @param int - the offset into this object
  @param String - the string to compare against this object
  @param int - the offset into 'other'
  @param int - the number of characters to compare
  @returns boolean - whether the two strings match
  @throws NullPointerException - if 'other' is null
  */
  public boolean regionMatches(boolean ignoreCase,
                               int toffset,
                               String other,
                               int ooffset,
                               int len) {
    return regionMatches(ignoreCase, toffset, new MutableString(other),
                         ooffset, len);
  } // regionMatches()

  /**
  Indicates whether the MutableString[toffset..toffset + len] matches
  'other[ooffset..ooffset + len]', ignoring case.

  @param boolean - whether to ignore case
  @param int - the offset into this object
  @param MutableString - the string to compare against this object
  @param int - the offset into 'other'
  @param int - the number of characters to compare
  @returns boolean - whether the two strings match
  @throws NullPointerException - if 'other' is null
  */
  public boolean regionMatches(boolean ignoreCase,
                               int toffset,
                               MutableString other,
                               int ooffset,
                               int len) {
    if (!(toffset + len <= stringArray.size() && ooffset + len <= other.stringArray.size())) {
      return false;
    } // if

    // check if all the characters match
    char leftChar;
    char rightChar;

    for (int count = 0; count < len; ++count) {
      if (!ignoreCase) {
        leftChar = ((Character) stringArray.get(toffset + count)).charValue();
        rightChar = ((Character) other.stringArray.get(ooffset + count)).charValue();
      } // if
      else {
        leftChar = Character.toLowerCase( ((Character) stringArray.get(toffset + count)).charValue());
        rightChar = Character.toLowerCase( ((Character) other.stringArray.get(ooffset +
                                                                        count)).charValue());;
      } // else

      if (leftChar != rightChar) {
        // at least one pair of characters don't match
        return false;
      } // if
    } // for index

    // all the characters match
    return true;

  } // regionMatches()

  /**
  Indicates whether the MutableString begins with
  'prefix[toffset..prefix.length() - 1]'.

  @param MutableString - the prefix to check
  @returns boolean - whether the string begins with 'prefix'
  @throws NullPointerException - if prefix is null
  */
  public boolean startsWith(String prefix,
                            int toffset) {
    return startsWith(new MutableString(prefix), toffset);
  } // startsWith()

  /**
  Indicates whether the MutableString begins with
  'prefix[toffset..prefix.length() - 1]'.

  @param MutableString - the prefix to check
  @returns boolean - whether the string begins with 'prefix'
  @throws NullPointerException - if prefix is null
  */
  public boolean startsWith(MutableString prefix,
                            int toffset) {
    if (stringArray.size() < prefix.stringArray.size()) {
      return false;
    } // if

    return regionMatches(0, prefix, toffset, prefix.stringArray.size());
  } // startsWith()

  /**
  Indicates whether the MutableString begins with 'prefix'.

  @param String - the prefix to check
  @returns boolean - whether the string begins with 'prefix'
  @throws NullPointerException - if prefix is null
  */
  public boolean startsWith(String prefix) {
    return startsWith(new MutableString(prefix));
  } // startsWith()

  /**
  Indicates whether the MutableString begins with 'prefix'.

  @param MutableString - the prefix to check
  @returns boolean - whether the string begins with 'prefix'
  @throws NullPointerException - if prefix is null
  */
  public boolean startsWith(MutableString prefix) {
    return startsWith(prefix, 0);
  } // startsWith()

  /**
  Indicates whether the MutableString ends with 'suffix'.

  @param MutableString - the suffix to check
  @returns boolean - whether the string ends with 'suffix'
  @throws NullPointerException - if suffix is null.
  */
  public boolean endsWith(MutableString suffix) {
    if (stringArray.size() < suffix.stringArray.size()) {
      return false;
    } // if

    // regionMatches() performs the comparison
    return regionMatches(stringArray.size() - suffix.stringArray.size(),
                         suffix,
                         0,
                         suffix.stringArray.size());
  } // endsWith()

  /**
  Indicates whether the string ends with 'suffix'.

  @param String - the suffix to check
  @returns boolean - whether the string ends with 'suffix'
  */
  public boolean endsWith(String suffix) {
    return endsWith(new MutableString(suffix));
  } // endsWith()

  /**
  Computes a hash code on the string.
  
  @returns int - s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1]
  */
  public int hashCode() {
    // s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1]

    // note: check if 'return new String(ms.toString()).hashCode()' is faster
    int result = 0;
    int power;

    for (int index = 0; index < stringArray.size(); ++index) {
      // compute 31^(length() - index - 1)
      power = 1;
      for (int j = 0; j < stringArray.size() - index - 1; ++j) {
        power *= 31;
      } // for j

      // result += ((Character) stringArray.get(index)).charValue() * ((int) Math.pow(31.0, ));
      result += ((Character) stringArray.get(index)).charValue() * power;
    } // index

    return result;
  } // hashCode()

  /**
  Returns the index, if it exists, of the first occurence of 'ch' in
  the string.

  @param int - the character to search for
  @return int - the index, if it exists, of the first occurence of 'ch'. If the
  character does not exist, -1 is returned
  */
  public int indexOf(int ch) {
    return indexOf(ch, 0);
  } // indexOf()

  /**
  Returns the index, if it exists, of the first occurence of 'ch' in
  the string starting at position 'fromIndex' in the string.

  @param int - the character to search for
  @return int - the index, if it exists, of the first occurence of 'ch'. If the
  character does not exist, -1 is returned
  */
  public int indexOf(int ch,
                     int fromIndex) {
    for (int index = fromIndex; index < stringArray.size(); ++index) {
      if ( ((Character) stringArray.get(index)).charValue() == ch) {
        return index;
      } // if
    } // for index

    return -1;
  } // indexOf()

  /**
  Returns the index, if it exists, of the last occurence of 'ch' in
  the string.

  @param ch - the character to search for
  @return int - the index, if it exists, of the last occurence of 'ch'. If the
  character does not exist, -1 is returned
  */
  public int lastIndexOf(int ch) {
    return lastIndexOf(ch, 0);
  } // lastIndexOf()

  /**
  Returns the index, if it exists, of the last occurence of 'ch' in
  the string starting at position 'fromIndex' in the string.

  @param char - the character to search for
  @return int - the index, if it exists, of the last occurence of 'ch'. If the
  character does not exist, -1 is returned
  */
  public int lastIndexOf(int ch,
                         int fromIndex) {
    int result = -1;

    for (int index = fromIndex; index < stringArray.size(); ++index) {
      if ( ((Character) stringArray.get(index)).charValue() == ch) {
        result = index;
      } // if
    } // for index

    return result;
  } // lastIndexOf()

  /**
  Returns the index, if it exists, of the first occurence of 'str' in
  the string.

  @param MutableString - the string to search for
  @return int - the index, if it exists, of the first occurence of 'str'. If the
  character does not exist, -1 is returned
  */
  public int indexOf(MutableString str) {
    return indexOf(str.toString());
  } // indexOf()

  /**
  Returns the index, if it exists, of the first occurence of 'str' in
  the string.

  @param String - the string to search for
  @return int - the index, if it exists, of the first occurence of 'str'. If the
  character does not exist, -1 is returned
  */
  public int indexOf(String str) {
    return indexOf(str, 0, false);
  } // indexOf()

  /**
  Returns the index, if it exists, of the first occurence of 'str' in
  the string beginning at offset 'fromIndex'.

  @param MutableString - the string to search for
  @return int - the index, if it exists, of the first occurence of 'str'
  beginning at offset 'fromIndex'. If the character does not exist, -1 is
  returned
  */
  public int indexOf(MutableString str, int fromIndex) {
    return indexOf(str.toString(), fromIndex);
  } // indexOf()

  /**
  Returns the index, if it exists, of the first occurence of 'str' in
  the string beginning at offset 'fromIndex'.

  @param String - the string to search for
  @return int - the index, if it exists, of the first occurence of 'str'
  beginning at offset 'fromIndex'. If the character does not exist, -1 is
  returned
  */
  public int indexOf(String str,
                     int fromIndex) {
    return indexOf(str, fromIndex, false);
  } // indexOf()

  /**
  Returns the index, if it exists, of the last occurence of 'str' in
  the string.

  @param MutableString - the string to search for
  @return int - the index, if it exists, of the last occurence of 'str'. If the
  character does not exist, -1 is returned
  */
  public int lastIndexOf(MutableString str) {
    return lastIndexOf(str.toString());
  } // lastIndexOf()

  /**
  Returns the index, if it exists, of the last occurence of 'str' in
  the string.

  @param String - the string to search for
  @return int - the index, if it exists, of the last occurence of 'str'. If the
  character does not exist, -1 is returned
  */
  public int lastIndexOf(String str) {
    return indexOf(str, 0, true);
  } // lastIndexOf()

  /**
  Returns the index, if it exists, of the last occurence of 'str' in
  the string beginning at offset 'fromIndex'.

  @param MutableString - the string to search for
  @return int - the index, if it exists, of the last occurence of 'str'
  beginning at offset 'fromIndex'. If the character does not exist, -1 is
  returned
  */
  public int lastIndexOf(MutableString str,
                         int fromIndex) {
    return lastIndexOf(str.toString(), fromIndex);
  } // lastIndexOf()

  /**
  Returns the index, if it exists, of the last occurence of 'str' in
  the string beginning at offset 'fromIndex'.

  @param String - the string to search for
  @return int - the index, if it exists, of the last occurence of 'str'
  beginning at offset 'fromIndex'. If the character does not exist, -1 is
  returned
  */
  public int lastIndexOf(String str,
                         int fromIndex) {
    return indexOf(str, fromIndex, true);
  } // lastIndexOf()

  /**
  Returns the index, if it exists, of the first or last occurence of 'str' in
  the string beginning at offset 'fromIndex'.

  @param String - the string to search for
  @param int - the index to begin searching from
  @param boolean - whether to look for the first or last occurence
  @return int - the index, if it exists, of the first or last occurence of 'str'
  beginning at offset 'fromIndex'. If the character does not exist, -1 is
  returned
  */
  private int indexOf(String str, int fromIndex, boolean last) {
    // note: use a better string matching algorithm

    int matchIndex = -1;

    for (int index = fromIndex; index < stringArray.size(); ++index) {
      boolean matched = true;

      // check if there's enough room left in 'string' to match 'str'
      if (index + str.length() < stringArray.size()) {
        int count = 0;
        for (int j = 0; j < str.length(); ++j) {
          if ( ((Character) stringArray.get(index + j)).charValue() !=
               str.charAt(j)) {
            matched = false;
          } // if
        } // for j
      } // if
      else {
        matched = false;
      } // else

      if (matched) {
        if (last) {
          matchIndex = index;
        } // if
        else {
          return index;
        } // else
      } // if
    } // index

    return matchIndex;
  } // indexOf()

  /**
  Returns string[beginIndex..length() - 1] as a new MutableString.

  @returns MutableString - a new MutableString
  @throws IndexOutOfBoundsException - if beginIndex is negative or larger than
  the length of this String object.
  */
  public MutableString substring(int beginIndex) {
    return substring(beginIndex, stringArray.size());
  } // substring()

  /**
  Returns string[beginIndex..endIndex - 1] as a new MutableString.

  @returns MutableString - a new MutableString 
  @throws IndexOutOfBoundsException - if the beginIndex is negative, or
  endIndex is larger than the length of this String object, or  beginIndex is
  larger than endIndex.
  */
  public MutableString substring(int beginIndex,
                                 int endIndex) {
    char[] stringArray = new char[endIndex - beginIndex];

    int j = 0;
    for (int index = beginIndex; index < endIndex; ++index) {
      stringArray[j] = ((Character) this.stringArray.get(index)).charValue();
      ++j;
    } // for index
    return new MutableString(stringArray);
  } // substring()

  /**
  Concatenates 'str' to the MutableString.

  @param MutableString - the string to concatenate
  @returns MutableString - a reference to this object
  @throws NullPointerException - if str is null
  */
  public MutableString concat(MutableString str) {
    return concat(str.toString());
  } // concat()

  /**
  Concatenates 'str' to the MutableString.

  @param String - the string to concatenate
  @returns MutableString - a reference to this object
  @throws NullPointerException - if str is null
  */
  public MutableString concat(String str) {
    // make the string big enough to accommodate the new string
    stringArray.ensureCapacity(stringArray.size() + str.length());

    // concatenate the new string
    for (int index = 0; index < str.length(); ++index) {
      stringArray.add(new Character(str.charAt(index)));
    } // for index

    return this;
  } // concat()

  /**
  Replaces all instances of 'oldChar' with 'newChar'.

  @param char - the old character
  @param char - the new character that replaces the old character
  @returns MutableString - a reference to this object
  */
  public MutableString replace(char oldChar,
                               char newChar) {
    for (int index = 0; index < stringArray.size(); ++index) {
      if ( ((Character) stringArray.get(index)).charValue() == oldChar) {
        stringArray.set(index, new Character(newChar));
      } // if
    } // for index

    return this;
  } // replace()

  /**
  Converts all characters in the string to upper case.

  @param Locale - convert characters using the specified locale
  @returns MutableString - a reference to the string
  */
  public MutableString toLowerCase(Locale locale) {
    if (locale == null) {
      // use the default locale

      char ch;
      Character character;
      for (int index = 0; index < stringArray.size(); ++index) {
        ch = ((Character) stringArray.get(index)).charValue();
        character = new Character(Character.toLowerCase(ch));
        stringArray.set(index, character);
      } // for index
    } // if
    else {
      // use the specified locale

      String s = new String(toString());

      // convert to lower case using the specified locale
      s.toLowerCase(locale);
      // make sure the length of the string didn't change
      Debugger.assert(s.length() == length());

      Character character;
      for (int index = 0; index < stringArray.size(); ++index) {
        character = new Character(s.charAt(index));
        stringArray.set(index, character);
      } // for index
    } // else

    return this;
  } // toLowerCase()

  /**
  Converts all characters in the string to lower case.

  @returns MutableString - a reference to the string
  */
  public MutableString toLowerCase() {
    return toLowerCase(null);
  } // toLowerCase()

  /**
  Converts all characters in the string to upper case.

  @param Locale - convert characters using the specified locale
  @returns MutableString - a reference to the string
  */
  public MutableString toUpperCase(Locale locale) {
    if (locale == null) {
      // use the default locale

      char ch;
      Character character;
      for (int index = 0; index < stringArray.size(); ++index) {
        ch = ((Character) stringArray.get(index)).charValue();
        character = new Character(Character.toUpperCase(ch));
        stringArray.set(index, character);
      } // for index
    } // if
    else {
      // use the specified locale

      String s = new String(toString());

      // convert to lower case using the specified locale
      s.toUpperCase(locale);
      // make sure the length of the string didn't change
      Debugger.assert(s.length() == length());

      Character character;
      for (int index = 0; index < stringArray.size(); ++index) {
        character = new Character(s.charAt(index));
        stringArray.set(index, character);
      } // for index
    } // else

    return this;
  } // toUpperCase()

  /**
  Converts all characters in the string to upper case.

  @returns MutableString - a reference to the string
  */
  public MutableString toUpperCase() {
    return toUpperCase(null);
  } // toUpperCase()

  /**
  Removes leading and trailing whitespace.

  @returns a reference to the object
  */
  public MutableString trim() {
    boolean removed;

    // remove leading whitespace
    do {
      removed = false;
      if ( ((Character) stringArray.get(0)).charValue() <= '\u0020') {
        stringArray.remove(0);
        removed = true;
      } // if
    } while (removed);

    // remove trailing whitespace
    removed = true;
    while (removed && stringArray.size() > 0) {
      removed = false;
      if ( ((Character) stringArray.get(stringArray.size() - 1)).charValue() <= '\u0020') {
        stringArray.remove(stringArray.size() - 1);
        removed = true;
      } // if
    } // while

    return this;
  } // trim()

  /**
  Converts the MutableString to a String

  @returns String - the String representation of the MutableString
  */
  public String toString() {
    return new String(toCharArray());
  } // toString()

  /**
  Converts the MutableString to a character string

  @returns String - the char[] representation of the MutableString
  */
  public char[] toCharArray() {
    Object[] characters = stringArray.toArray();

    char[] chars = new char[characters.length];

    for (int index = 0; index < chars.length; ++index) {
      chars[index] = ((Character) characters[index]).charValue();
    } // for index

    return chars;
  } // toCharArray()

  /**
  Creates a MutableString from the string representation of 'obj'.

  @param Object - the object to create a string from
  @returns MutableString - the MutableString representation of 'obj'
  */
  public static MutableString valueOf(Object obj) {
    if (obj == null) {
      return new MutableString("null");
    } // if

    return new MutableString(obj.toString());
  } // valueOf()

  /**
  Creates a MutableString from the string representation of 'data'.

  @param char[] - the object to create a string from
  @returns MutableString - the MutableString representation of 'data'
  */
  public static MutableString valueOf(char[] data) {
    return new MutableString(data);
  } // valueOf()

  /**
  Creates a MutableString from the string representation of
  'data[offset..offset + count - 1]'.

  @param char[] - the object to create a string from
  @param int - the offset into 'data' to begin copying from
  @param int - the amount of characters to extract from 'data'
  @returns MutableString - the MutableString representation of 'data'
  @throws NullPointerException - if data is null.
  @throws IndexOutOfBoundsException - if offset is negative, or count is
  negative, or offset+count is larger than data.length.
  */
  public static MutableString valueOf(char[] data,
                                      int offset,
                                      int count) {
    if (offset < 0 || count < 0 || offset + count >= data.length) {
      throw new IndexOutOfBoundsException();
    } // if
    
    char[] destination = new char[count];
    System.arraycopy(data, offset, destination, 0, count);
    return new MutableString(destination);
  } // valueOf()

  /**
  Creates a MutableString from the string representation of
  'data[offset..offset + count - 1]'.

  @param char[] - the object to create a string from
  @param int - the offset into 'data' to begin copying from
  @param int - the amount of characters to extract from 'data'
  @returns MutableString - the MutableString representation of 'data'
  @throws NullPointerException - if data is null.
  @throws IndexOutOfBoundsException - if offset is negative, or count is
  negative, or offset+count is larger than data.length.
  */
  public static MutableString copyValueOf(char[] data,
                                          int offset,
                                          int count) {
    // how is this different than valueOf(char[], int, int)?
    return valueOf(data, offset, count);
  } // copyValueOf()

  /**
  Creates a MutableString from the string representation of 'data'

  @param char[] - the object to create a string from
  @returns MutableString - the MutableString representation of 'data'
  @throws NullPointerException - if data is null.
  @throws IndexOutOfBoundsException - if offset is negative, or count is
  negative, or offset+count is larger than data.length.
  */
  public static MutableString copyValueOf(char[] data) {
    return valueOf(data, 0, data.length);
  } // copyValueOf()

  /**
  Creates a MutableString from the string representation of 'b'.

  @param char[] - the object to create a string from
  @returns MutableString - the MutableString representation of 'b'
  */
  public static MutableString valueOf(boolean b) {
    if (b) {
      return new MutableString("true");
    } // if
    else {
      return new MutableString("false");
    } // else
  } // valueOf()

  /**
  Creates a MutableString from the string representation of 'c'.

  @param char[] - the object to create a string from
  @returns MutableString - the MutableString representation of 'c'
  */
  public static MutableString valueOf(char c) {
    return new MutableString((new Character(c)).toString());
  } // valueOf()

  /**
  Creates a MutableString from the string representation of 'i'.

  @param char[] - the object to create a string from
  @returns MutableString - the MutableString representation of 'i'
  */
  public static MutableString valueOf(int i) {
    return new MutableString(Integer.toString(i));
  } // valueOf()

  /**
  Creates a MutableString from the string representation of 'l'.

  @param char[] - the object to create a string from
  @returns MutableString - the MutableString representation of 'l'
  */
  public static MutableString valueOf(long l) {
    return new MutableString(Long.toString(l));
  } // valueOf()

  /**
  Creates a MutableString from the string representation of 'f'.

  @param char[] - the object to create a string from
  @returns MutableString - the MutableString representation of 'f'
  */
  public static MutableString valueOf(float f) {
    return new MutableString(Float.toString(f));
  } // valueOf()

  /**
  Creates a MutableString from the string representation of 'd'.

  @param char[] - the object to create a string from
  @returns MutableString - the MutableString representation of 'd'
  */
  public static MutableString valueOf(double d) {
    return new MutableString(Double.toString(d));
  } // valueOf()

  /***********************/
  /* end of String class */
  /***********************/

  /**********************/
  /* StringBuffer class */
  /**********************/

  /**
  Makes sure the capacity of the object is at least 'minimumCapacity'.

  @param int - the minimum capacity
  */
  public void ensureCapacity(int minimumCapacity) {
    stringArray.ensureCapacity(minimumCapacity);
  } // ensureCapacity()

  /**
  Resets the string to string[0..newLength - 1]. If the old string is
  shorter, characters are truncated. If the old string is longer, null
  characters are inserted.

  @param int - the new length of the string
  @throws IndexOutOfBoundsException - if 'newLength' is negative
  */
  public void setLength(int newLength) {
    if (newLength < 0) {
      throw new IndexOutOfBoundsException();
    } // if

    if (newLength >= stringArray.size()) {
      // string is becoming longer

      // note: there's got to be a faster way to do this      
      for (int index = 0; index < newLength - stringArray.size() + 1; ++index) {
        stringArray.add(new Character('\u0000'));
      } // for index
    } // if
    else {
      // string is becoming shorter

      // note: there's got to be a faster way to do this
      for (int index = stringArray.size() - 1; index >= newLength; --index) {
        stringArray.remove(index);
      } // for index
    } // else
  } // setLength()

  /**
  Sets the character at offset 'index' to 'ch'.

  @throws IndexOutOfBoundsException - if index is negative or greater than or
  equal to length()
  */
  public void setCharAt(int index,
                        char ch) {
    if (index < 0 || index >= stringArray.size()) {
      throw new IndexOutOfBoundsException();
    } // if
    
    stringArray.set(index, new Character(ch));
  } // setCharAt()

  /**
  Appends the string representation of 'obj' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(Object obj) {
    append(obj.toString());

    return this;
  } // append()

  /**
  Appends 'str' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(MutableString str) {
    concat(str);

    return this;
  } // append()

  /**
  Appends 'str' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(String str) {
    concat(str);

    return this;
  } // append()

  /**
  Appends 'str' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(char[] str) {
    append(new String(str));

    return this;
  } // append()

  /**
  Appends 'str[offset]..str[len-1]' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(char[] str,
                              int offset,
                              int len) {
    MutableString ms = MutableString.valueOf(str, offset, len);
    append(ms);

    return this;
  } // append()

  /**
  Appends the string representation of 'b' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(boolean b) {
    if (b) {
      append("true");
    } // if
    else {
      append("false");
    } // else

    return this;
  } // append()

  /**
  Appends 'c' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(char c) {
    append(new Character(c).toString());

    return this;
  } // append()

  /**
  Appends the string representation of 'i' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(int i) {
    append(Integer.toString(i));

    return this;
  } // append()

  /**
  Appends the string representation of 'l' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(long l) {
    append(Long.toString(l));

    return this;
  } // append()

  /**
  Appends the string representation of 'f' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(float f) {
    append(Float.toString(f));

    return this;
  } // append()

  /**
  Appends the string representation of 'd' to the object.

  @return MutableString - the object after appending
  */
  public MutableString append(double d) {
    append(Double.toString(d));

    return this;
  } // append()

  /**
  Deletes all characters from string[start..end-1].

  @param int - the starting index
  @param int - one greater than the ending index
  @throws StringIndexOutOfBoundsException - if start is negative, greater than
  length(), or greater than end.
  */
  public MutableString delete(int start,
                              int end) {
    if (start < 0 || start >= stringArray.size() || start > end) {
      throw new StringIndexOutOfBoundsException();
    } // if
    
    // note: where's my min() method?
    int last = end - 1;
    if (stringArray.size() - 1 < last) {
      last = stringArray.size() - 1;
    } // if

    for (int index = last; index >= start; --index) {
      stringArray.remove(index);
    } // for index

    return this;
  } // delete()

  /**
  Deletes the character at offset 'index'.
  
  @throws StringIndexOutOfBoundsException - if start is negative or greater
  than or equal to length()
  */
  public MutableString deleteCharAt(int index) {
    if (index < 0 || index >= stringArray.size()) {
      throw new StringIndexOutOfBoundsException();
    } // if
    
    stringArray.remove(index);

    return this;
  } // deleteCharAt()

  /**
  The characters at string[start..end-1] are deleted. Then 'str' is
  inserted into the string starting at the 'start' position.

  @param int - the starting index
  @param int - one position past the ending index, ie, the ending index is
  end - 1
  @param MutableString - the string that replaces string[start..end-1]
  @returns MutableString - a reference to this object
  @throws StringIndexOutOfBoundsException - if start is negative, greater than
  or equal to length(), or greater than end.
  */
  public MutableString replace(int start,
                               int end,
                               MutableString str) {
    return replace(start, end, str.toString());
  } // replace()


  /**
  The characters at string[start..end-1] are deleted. Then 'str' is
  inserted into the string starting at the 'start' position.

  @param int - the starting index
  @param int - one position past the ending index, ie, the ending index is
               end - 1
  @param String - the string that replaces string[start..end-1]
  @returns MutableString - a reference to this object
  @throws StringIndexOutOfBoundsException - if start is negative, greater than
  or equal to length(), or greater than end.
  */
  public MutableString replace(int start,
                               int end,
                               String str) {
    if (start < 0 || start >= stringArray.size() || start > end) {
      throw new StringIndexOutOfBoundsException();
    } // if

    // note: where's my min() method?
    int last = end - 1;
    if (last >= stringArray.size()) {
      last = stringArray.size() - 1;
    } // if
    int charsToReplace = last - start + 1;
    delete(start, last + 1);
    insert(start, str.substring(0, charsToReplace));

    return this;
  } // replace()

  /**
  Inserts 'str[offset..offset + len]' into the string beginning at position
  index.

  @param int - the position to begin inserting into the string
  @param char[] - the string to insert
  @param int - the offset into the string to insert
  @param int - the number of characters to extract from 'str' and
  subsequently insert into the string
  @throws StringIndexOutOfBoundsException - if index is negative or greater than
  length(), or offset or len are negative, or (offset+len) is greater than
  str.length.
  */
  public MutableString insert(int index,
                              char[] str,
                              int offset,
                              int len) {
    char[] destination = new char[len];
    System.arraycopy(str, offset, destination, 0, len);
    insert(index, new String(destination));

    return this;
  } // insert()

  /**
  Inserts 'str[offset..offset + len]' into the string beginning at position
  index.

  @param int - the position to begin inserting into the string
  @param char[] - the string to insert
  @param int - the offset into the string to insert
  @param int - the number of characters to extract from 'str' and
  subsequently insert into the string
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              Object obj) {
    insert(offset, obj.toString());

    return this;
  } // insert()

  /**
  Inserts 'str[offset..offset + len]' into the string beginning at position
  index.

  @param int - the position to begin inserting into the string
  @param MutableString - the string to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              MutableString string) {
    return insert(offset, string.toString());
  } // insert()

  /**
  Inserts 'str[offset..offset + len]' into the string beginning at position
  index.

  @param int - the position to begin inserting into the string
  @param String - the string to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              String string) {
    try {
      for (int index = 0; index < string.length(); ++index) {
        stringArray.add(offset + index, new Character(string.charAt(index)));
      } // for index

      return this;
    } // try
    catch (IndexOutOfBoundsException e) {
      throw new StringIndexOutOfBoundsException();
    } // catch
  } // insert()

  /**
  Inserts 'str[offset..length() - 1]' into the string beginning at position
  index.

  @param int - the position to begin inserting into the string
  @param char[] - the string to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              char[] str) {
    insert(offset, new String(str));

    return this;
  } // insert()

  /**
  Inserts 'true' or 'false' intothe string beginning at index 'offset'.

  @param int - the position to begin inserting into the string
  @param boolean - what to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              boolean b) {
    if (b) {
      insert(offset, "true");
    } // if
    else {
      insert(offset, "false");
    } // else

    return this;
  } // insert()

  /**
  Inserts 'c' into the string at index 'offset'.

  @param int - the position to begin inserting into the string
  @param char - what to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              char c) {
    String string = new Character(c).toString();
    insert(offset, string);

    return this;
  } // insert()

  /**
  Inserts the string representation of 'i' into the string beginning at index
  'offset'.

  @param int - the position to begin inserting into the string
  @param int - what to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              int i) {
    String string = Integer.toString(i);
    insert(offset, string);

    return this;
  } // insert()

  /**
  Inserts the string representation of 'l' into the string beginning at index
  'offset'.

  @param int - the position to begin inserting into the string
  @param long - what to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              long l) {
    String string = Long.toString(l);
    insert(offset, string);

    return this;
  } // insert()

  /**
  Inserts the string representation of 'f' into the string beginning at index
  'offset'.

  @param int - the position to begin inserting into the string
  @param float - what to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              float f) {
    String string = Float.toString(f);
    insert(offset, string);

    return this;
  } // insert()

  /**
  Inserts the string representation of 'd' into the string beginning at index
  'offset'.

  @param int - the position to begin inserting into the string
  @param double - what to insert
  @throws StringIndexOutOfBoundsException - if the offset is invalid
  */
  public MutableString insert(int offset,
                              double d) {
    String string = Double.toString(d);
    insert(offset, string);

    return this;
  } // insert()

  /**
  Reverses the MutableString in place.

  @returns MutableString - a reference to this object
  */
  public MutableString reverse() {
    Object temporaryObject;

    for (int index = 0; index < stringArray.size() / 2; ++index) {
      temporaryObject = stringArray.get(index);
      stringArray.set(index, stringArray.get(stringArray.size() - index - 1));
      stringArray.set(stringArray.size() - index - 1, temporaryObject);
    } // for index

    return this;
  } // reverse()
} // MutableString
