/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.tool.codegen;

// ToolBox imports

import org.enhydra.tool.ToolBoxInfo;
import org.enhydra.tool.codegen.wizard.ConfigDialog;
import org.enhydra.tool.codegen.wizard.CodeGenDialog;
import org.enhydra.tool.codegen.wizard.CodeGenPage;
import org.enhydra.tool.common.ToolException;
import org.enhydra.tool.common.SwingUtil;
import org.enhydra.tool.common.ResUtil;

// Standard imports

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Properties;
import java.util.Vector;
import java.beans.Beans;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Frame;
import javax.swing.JOptionPane;
import java.util.ResourceBundle;

/**
 * Main class for CodeGen wizard and command line interface.
 */
public class CodeGen implements Constants {

    /**
     * When the debug flag is true, stack traces are printed
     * for all exceptions.
     */
    static ResourceBundle       res =
        ResourceBundle.getBundle ("org.enhydra.tool.codegen.Res");
    public static boolean       debug = false;

    /**
     * When the verbose flag is true, file names stream out
     * during file operations.
     */
    public static boolean       verbose = false;

    //

    private static final String GENERATOR = "codegen.generator";        // nores
    private static final String CLASS_WEB_GEN =
        "org.enhydra.tool.codegen.webapp.WebAppGenerator";      // nores
    private static final String CLASS_EN3_GEN =
        "org.enhydra.tool.codegen.en3app.Enhydra3AppGenerator";         // nores

    //

    private Generator[]         generators = new Generator[0];
    private Generator           selection = null;
    private Properties          properties = new Properties ();
    private int                 choice = CodeGenDialog.CHOICE_CANCEL;
    private boolean             swing = false;

    // /
    // / PUBLIC CONSTRUCTORS
    // /

    /**
     * Create an Code Generator in command line mode so that no
     * swing components are instantiated.
     *
     * @throws GeneratorException
     * Thrown if no generators could be instantiated. Warnings
     * print out for each generators that could not be instantiated.
     */
    public CodeGen () throws GeneratorException {
        this (false);
    }

    /**
     * Create an Code Generator and read in the
     * codegen.properties file. Use Beans instantiate to
     * create instances of each generator listed in
     * codegen.properties. Also pass the properties object
     * into each generator.
     *
     * @throws GeneratorException
     * Thrown if no generators could be instantiated. Warnings
     * print out for each generators that could not be instantiated.
     *
     * @param swing
     * If true, swing components will be used for user interaction.
     */
    public CodeGen (boolean useSwing) throws GeneratorException {
        setSwing (useSwing);
        initProperties ();
        initGenerators ();

        if (generators.length == 0) {
            throw new GeneratorException (res.getString ("No_generators"));
        }
    }

    /**
     * Create a Code Generator using a specified implementation
     * of the Generator class. This constructor is used by Kelp
     * to create seperate wizards for each generator.
     *
     * @throws GeneratorException
     * Thrown if properties fail to initialize.
     *
     * @param selection
     * The generator implementation to use for all
     * generation operations.
     */
    public CodeGen (Generator selection) throws GeneratorException {
        setSwing (true);
        initProperties ();      // this is only used when preselecting for swing

        generators = new Generator[1];

        selection.setProperties (properties);

        generators[0] = selection;
        this.selection = selection;
    }

    // /
    // / PUBLIC METHODS
    // /

    public boolean isSwing () {
        return swing;
    }

    public void setSwing (boolean b) {
        swing = b;
    }

    /**
     * Get an integer representing the button selected
     * by the use to close the wizard dialog.
     *
     * @return
     * CodeGenDialog.CHOICE_FINISH or CodeGenDialog.CHOICE_CANCEL.
     * The CHOICE_FINISH value indicates that code generation was
     * attempted.
     */
    public int getChoice () {
        return choice;
    }

    /**
     * Get generator implementation to use for all generation
     * operations.
     *
     * @return
     * A specific implementation of the generator interface.
     */
    public Generator getSelection () {
        return selection;
    }

    /**
     * Open the wizard dialog where a user can select what generator
     * to use, specify generation options and generate the project.
     *
     * @param host
     * A component that can act as the owner of the modal dialog. May be
     * null, Frame or Dialog.
     *
     * @return
     * Returns an array of generated file references.
     */
    public File[] invokeWizard (Component host) {
        SwingUtil.setLookAndFeelToSystem ();

        CodeGenDialog   dialog = null;
        File[]          files = new File[0];

        if (host == null) {
            dialog = new CodeGenDialog ();
        } else if (host instanceof Frame) {
            dialog = new CodeGenDialog ((Frame) host);
        } else if (host instanceof Dialog) {
            dialog = new CodeGenDialog ((Dialog) host);
        }

        if (dialog != null) {
            dialog.setCodeGen (this);

            if (selection != null) {
                try {
                    dialog.setSelection (selection);
                } catch (GeneratorException e) {
                    JOptionPane.showMessageDialog (host, e.getMessage (),
                                                   res.getString ("Unable_to_set"),
                                                   JOptionPane.ERROR_MESSAGE);

                    if (CodeGen.debug) {
                        e.printStackTrace ();
                    }
                }
            }

            dialog.show ();

            choice = dialog.getChoice ();
            files = dialog.getGeneratedFiles ();

            if (choice == CodeGenDialog.CHOICE_CANCEL) {

                // done

            } else {
                if (files == null || files.length == 0) {
                    JOptionPane.showMessageDialog (host,
                                                   res.getString ("No_files_generated_"),
                                                   res.getString ("AppWizard_Complete"),
                                                   JOptionPane.WARNING_MESSAGE);
                } else {
                    String      parentPath = files[0].getParent ();

                    for (int i = 0; i < files.length; i++) {
                        if (files[i].getParent ().length ()
                                < parentPath.length ()) {
                            parentPath = files[i].getParent ();
                        }
                    }

                    String      mess =
                        ResUtil.format (res.getString ("Generated_0_files_in"),
                                        (new String ()) + files.length,
                                        parentPath);

                    JOptionPane.showMessageDialog (host, mess,
                                                   res.getString ("AppWizard_Complete"),
                                                   JOptionPane.INFORMATION_MESSAGE);
                }
            }

            dialog.clearAll ();
            dialog.dispose ();
        }

        return files;
    }

    // /
    // / PRIVATE STATIC
    // /

    public void setEnhydraHome (File f) {
        if (f != null && f.isDirectory ()) {
            properties.setProperty (ToolBoxInfo.ENHYDRA_HOME,
                                    f.getAbsolutePath ());
        }
    }

    public File getEnhydraHome () {
        String  homePath = new String ();
        File    home = null;

        homePath = properties.getProperty (ToolBoxInfo.ENHYDRA_HOME);

        if (homePath != null) {
            home = new File (homePath);
        }

        return home;
    }

    /**
     * Method declaration
     *
     *
     * @param enhydraHome
     *
     * @exception GeneratorException
     */
    public static void config () throws GeneratorException {
        File    home = new File ((new String ()) + '.');

        home = new File (home.getAbsolutePath ());

        if (home.getParentFile () != null) {
            home = home.getParentFile ();

            if (home.getParentFile () != null) {
                home = home.getParentFile ();
            }
        }

        CodeGen.config (home.getParent ());
    }

    public static void config (String enhydraHome) throws GeneratorException {

        // create default property set

        Properties      prop = new Properties ();
        File            file = null;

        file = new File (enhydraHome);

        prop.setProperty (ToolBoxInfo.ENHYDRA_HOME, file.getAbsolutePath ());
        prop.setProperty (GENERATOR + '.' + '0', CLASS_WEB_GEN);
        prop.setProperty (GENERATOR + '.' + '1', CLASS_EN3_GEN);

        try {
            file = ToolBoxInfo.storeProperties (prop);

            System.out.println (ResUtil.format (res.getString ("Configuration"),
                                                file));
        } catch (Exception e) {
            throw new GeneratorException (e,
                                          ResUtil.format (res.getString ("Unable_to_write_file"),
                                          ToolBoxInfo.getPropertyFilename ()));
        }
    }

    // /
    // / PRIVATE
    // /

    public void setGenerators (Generator[] g) {
        generators = g;
    }

    public Generator[] getGenerators () {
        return generators;
    }

    /**
     * Method declaration
     *
     */
    public void initGenerators () throws GeneratorException {
        String  className;
        Object  newObject;
        Vector  genVector = new Vector ();
        int     validCount = 0;

        for (int i = 0; i < 100; i++) {
            className = properties.getProperty (GENERATOR + '.' + i);
            newObject = null;

            if (className != null && className.trim ().length () > 0) {
                try {
                    newObject =
                        Beans.instantiate (getClass ().getClassLoader (),
                                           className);
                } catch (Exception e) {
                    newObject = null;

                    System.out.println (ResUtil.format (res.getString ("AppWizard_Warning"),
                                                        className));

                    if (CodeGen.debug) {
                        System.out.println (TAB4 + e.toString ());
                    }
                }

                if (newObject != null) {
                    if (newObject instanceof Generator) {
                        genVector.addElement ((Generator) newObject);
                    } else {
                        System.out.println (ResUtil.format (res.getString ("AppWizard_Warning_Not"),
                                                            className));
                    }
                }
            }
        }

        for (int i = 0; i < genVector.size (); i++) {
            Generator   current = null;

            current = (Generator) genVector.elementAt (i);

            try {
                current.setProperties (properties);
                current.setSwing (isSwing ());

                validCount++;
            } catch (GeneratorException e) {
                System.out.println (ResUtil.format (res.getString ("Invalid_generator_0_"),
                                                    current.getCommandName ()));
                System.out.println (e.getMessage ());
                genVector.setElementAt (new Object (), i);
            }
        }

        if (validCount == 0) {
            throw new GeneratorException (res.getString ("No_generators1"));
        } else {
            generators = new Generator[validCount];
        }

        validCount = 0;

        for (int i = 0; i < genVector.size (); i++) {
            if (genVector.elementAt (i) instanceof Generator) {
                generators[validCount] = (Generator) genVector.elementAt (i);
                validCount++;
            }
        }
    }

    /**
     * Method declaration
     *
     */
    protected void list () {
        int     count = generators.length;

        System.out.println (new String ());

        if (count < 1) {
            System.out.println (res.getString ("No_generators_to_list"));
        } else {
            System.out.println (res.getString ("Available_generators_"));

            for (int i = 0; i < count; i++) {
                System.out.println (TAB4 + generators[i].getCommandName ());
            }

            System.out.println (new String ());
            System.out.println (res.getString ("Generator"));

            for (int i = 0; i < count; i++) {
                StringBuffer    buf = new StringBuffer ();

                buf.append (generators[i].getCommandName ());
                buf.append (TAB4);
                buf.append (generators[i].getDisplayName ());
                buf.append ('\n');
                buf.append (generators[i].getDescription ());
                buf.append ('\n');
                System.out.println (buf.toString ());
            }
        }

        System.out.println (new String ());
    }

    /**
     * Method declaration
     *
     *
     * @param args
     */
    protected void invokeGenerator (String[] args) throws GeneratorException {
        File[]  files = new File[0];

        initSelection (args[0]);

        if (selection == null){}

        // do nothing

        else {
            processGeneratorOptions (args);
            System.out.println (ResUtil.format (res.getString ("Generator_0_"),
                                                selection.getCommandName ()));

            GeneratorOption[]   options =
                selection.getOptionSet ().toArray ();

            for (int i = 0; i < options.length; i++) {
                if (options[i].isRequired () || CodeGen.verbose) {
                    System.out.println (options[i].getDisplayName () + ':'
                                        + ' ' + options[i].getValue ());
                }
            }

            files = selection.generate ();

            System.out.println (ResUtil.format (res.getString ("File_count_0_"),
                                                files.length));
        }
    }

    /**
     * Method declaration
     *
     * @param args
     * @exception GeneratorException
     */
    private void processGeneratorOptions (String[] args)
            throws GeneratorException {

        // Do basic validation and update option set
        // with argument values.

        GeneratorOption[]       options =
            selection.getOptionSet ().toArray ();

        // Clear out values for required items.

        for (int i = 0; i < options.length; i++) {
            if (options[i].isRequired ()) {
                options[i].clearValue ();
            }
        }

        // Report any arguments we don't recognize and will ignore.

        for (int i = 1; i < args.length; i++) {
            if ((args[i].length () > 0) && (args[i].charAt (0) == '-')) {
                boolean found = false;

                // it is a name.

                for (int j = 0; j < options.length; j++) {
                    String      paramName = '-' + options[j].getName ();

                    if (paramName.equalsIgnoreCase (args[i])) {
                        found = true;

                        if (options[j].isBoolean ()) {
                            options[j].setValue ((new Boolean (true)).toString ());
                        } else if (i + 1 < args.length) {
                            options[j].setValue (args[i + 1]);
                        }
                    }
                }

                if (!found) {
                    System.out.println (ResUtil.format (res.getString ("Ignoring_unrecognized"),
                                                        args[i]));
                }
            }
        }

        StringBuffer    buf = new StringBuffer ();
        boolean         missing = false;

        for (int j = 0; j < options.length; j++) {
            if (options[j].isRequired ()) {
                if (options[j].isEmpty ()) {
                    if (!missing) {
                        missing = true;

                        buf.append (res.getString ("Required_value"));
                        buf.append ('\n');
                    }

                    buf.append (TAB4);
                    buf.append ('-');
                    buf.append (options[j].getName ());
                    buf.append ('\n');
                }
            }
        }

        if (buf.length () > 0) {
            buf.append ('\n');
            buf.append (res.getString ("For_more_information"));
            buf.append (' ');
            buf.append (selection.getCommandName ());

            throw new GeneratorException (buf.toString ());
        }

        // do something wonderful

    }

    /**
     * Method declaration
     *
     * @param commandName
     */
    private void initSelection (String commandName)
            throws GeneratorException {
        selection = null;

        for (int i = 0; i < generators.length; i++) {
            if (generators[i].getCommandName ().equalsIgnoreCase (commandName)) {
                selection = generators[i];

                break;
            }
        }

        if (selection == null) {
            System.out.println (ResUtil.format (res.getString ("_0_is_not_a_valid"),
                                                commandName));
        }
    }

    /**
     * Method declaration
     *
     * @exception GeneratorException
     */
    private void initProperties () throws GeneratorException {
        try {
            readProperties ();
        } catch (GeneratorException e) {
            if (isSwing ()) {
                swingConfig ();
                initProperties ();
            } else {
                throw e;
            }
        }
    }

    /**
     * Method declaration
     *
     *
     * @exception GeneratorException
     */
    private void readProperties () throws GeneratorException {
        String  fn = ToolBoxInfo.getPropertyFilename ();
        String  recreateMsg = res.getString ("You_can_use_config");
        File    file = new File (fn);

        if (!file.exists ()) {
            CodeGen.config (ToolBoxInfo.getEnhydraPath ());
        }

        try {
            properties = ToolBoxInfo.loadProperties ();
        } catch (ToolException e) {
            throw new GeneratorException (e,
                                          res.getString ("Unable_to_read"));
        }

        // validate properties

        String          value =
            properties.getProperty (ToolBoxInfo.ENHYDRA_HOME);
        StringBuffer    message = new StringBuffer ();

        if (value == null) {
            message.append (ResUtil.format (res.getString ("Unable_to_read1"),
                                            fn));
            message.append ('\n');
            message.append (recreateMsg);

            throw new GeneratorException (message.toString ());
        }

        file = new File (value + File.separator + DIR_LIB + File.separator
                         + FILE_ENHYDRA_JAR);

        if (!file.exists ()) {
            message.append (ResUtil.format (res.getString ("Invalid_enhydra_home"),
                                            value));
            message.append ('\n');
            message.append (recreateMsg);

            throw new GeneratorException (message.toString ());
        }

        for (int i = 0; i < 101; i++) {
            value = properties.getProperty (GENERATOR + '.' + i);

            if (value != null) {
                break;
            }

            if (i == 100) {
                message.append (ResUtil.format (res.getString ("No_generator_entries"),
                                                fn));
                message.append ('\n');
                message.append (recreateMsg);

                throw new GeneratorException (message.toString ());
            }
        }
    }

    /**
     * Method declaration
     *
     * @exception GeneratorException
     */
    private void swingConfig () throws GeneratorException {
        SwingUtil.setLookAndFeelToSystem ();

        ConfigDialog    dialog = new ConfigDialog ();

        dialog.show ();

        if (dialog.getChoice () == ConfigDialog.CHOICE_CANCEL) {
            throw new GeneratorException (res.getString ("Process_cancelled_"));
        } else {
            config (dialog.getEnhydraHome ().getAbsolutePath ());
        }
    }

    /**
     * Method declaration
     *
     * @param commandName
     */
    protected void showGeneratorOptionHelp (String commandName) {
        try {
            initSelection (commandName);
        } catch (GeneratorException e) {
            e.printStackTrace ();
        }

        if (selection == null) {
            System.out.println (ResUtil.format (res.getString ("Unable_to_show"),
                                                commandName));
        } else {
            GeneratorOption[]   options =
                selection.getOptionSet ().toArray ();

            if (options.length == 0) {
                System.out.println (ResUtil.format (res.getString ("No_options_available"),
                                                    commandName));
            } else {
                for (int i = 0; i < options.length; i++) {
                    StringBuffer        buf = new StringBuffer ();

                    buf.append (' ');
                    buf.append ('-');
                    buf.append (options[i].getName ());

                    if (options[i].isBoolean ()) {

                        // skip placeholder

                    } else {
                        buf.append (' ');
                        buf.append (options[i].getDisplayName ());
                    }

                    buf.append (TAB4);
                    buf.append (options[i].getDescription ());
                    System.out.println (buf.toString ());
                }
            }
        }
    }

}

