/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.tool.codegen.wizard;

//
// ToolBox imports
//
import org.enhydra.tool.common.SwingUtil;
import org.enhydra.tool.common.TBWizardDialog;
import org.enhydra.tool.codegen.CodeGen;
import org.enhydra.tool.codegen.Generator;
import org.enhydra.tool.codegen.GeneratorException;
import org.enhydra.tool.codegen.ValidationException;

//
// Standard imports
//
import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.beans.*;
import java.util.Vector;
import java.util.ResourceBundle;

/**
 * The CodeGenDialog defines a default container for presenting
 * wizards. To use this class you need to a create one or more
 * CodeGenPage objects and add them to this dialog. This dialog
 * supplies action buttons for the Previous, Next, Finish,
 * Cancel and Help.
 * <P><I>
 * This class is used when creating a standalone wizard. Use the
 * OpenTools BasicWizard when developing JBuilder add-ins.
 * </I></P>
 */
public class CodeGenDialog extends TBWizardDialog {

    //
    private static ResourceBundle cgRes = ResourceBundle.getBundle("org.enhydra.tool.codegen.Res"); // nores
    private final static String         DEFAULT_TITLE =
        cgRes.getString("Enhydra_Application");

    //
    transient private File[]            files = new File[0];
    transient private GenSelectionPanel selectionPanel;
    transient private CodeGen           codeGen = null;

    /**
     * Create a wizard dialog as a child of another dialog.
     *
     * @param owner
     * The dialog that is the parent of this wizard.
     */
    public CodeGenDialog(Dialog owner) {
        super(owner);
        setWizardTitle(CodeGenDialog.DEFAULT_TITLE);
        setModal(true);
        construct();
    }

    /**
     * Create a wizard dialog as a child of a frame.
     *
     * @param owner
     * The frame that is the parent of this wizard.
     */
    public CodeGenDialog(Frame owner) {
        super(owner);
        setWizardTitle(CodeGenDialog.DEFAULT_TITLE);
        setModal(true);
        construct();
    }

    /**
     * Create a wizard dialog without a parent.
     */
    public CodeGenDialog() {
        super();
        setModal(true);
        setTitle(CodeGenDialog.DEFAULT_TITLE);
        construct();
    }

    /**
     * Get an array of references to generated files.
     *
     * @return
     * An array of generated files. May have a zero length or
     * be null if no files have been generated.
     *
     */
    public File[] getGeneratedFiles() {
        return files;
    }

    /**
     * Get the list of available generators.
     *
     * @return
     * An array of generators from which a user can make a selection.
     */
    public Generator[] getGenerators() {
        return codeGen.getGenerators();
    }

    /**
     * Set the list of available generators.
     *
     * @param generators
     * An array of generators from which a user can make a selection.
     */
    public void setCodeGen(CodeGen cd) {
        codeGen = cd;
        selectionPanel.setGenerators(codeGen.getGenerators());
        selectionPanel.setEnhydraHome(codeGen.getEnhydraHome());
    }

    public CodeGen getCodeGen() {
        return codeGen;
    }

    /**
     * Get the generator selected by the user.
     *
     * @return
     * The generator to show options for and to use for generating a project.
     */
    public Generator getSelection() {
        return getCodeGenDeck().getSelection();
    }

    /**
     * Set the selected generator for a user and by pass the
     * selection page.
     *
     * @param selection
     * A generator containing option pages.
     *
     * @exception GeneratorException
     * Thrown if unable to create option pages for the selection.
     *
     */
    public void setSelection(Generator selection) throws GeneratorException {
        getCodeGenDeck().setSelection(selection);
        if (selection != null) {
            getCodeGenDeck().setSelectionCardHidden(true);
            setWizardTitle(selection.getWizardTitle());
        }
    }

    // /
    // / PRIVATE METHODS
    // /

    /**
     * Construct the object for either constructor.
     */
    protected void construct() {
        super.construct();
        try {
            codeGenInit();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Method declaration
     *
     *
     * @see
     */
    protected void next() {
        if (getCodeGenDeck().isSelectionCardShowing()) {
            try {
                getCodeGenDeck().setSelection(selectionPanel.getSelection());
                getCodeGenDeck().next();
            } catch (GeneratorException e) {
                JOptionPane.showMessageDialog(this, e.getMessage(),
                                              cgRes.getString("Generator_Error"),
                                              JOptionPane.ERROR_MESSAGE);
                if (CodeGen.debug) {
                    e.printStackTrace();
                }
            }
        } else {
            super.next();
        }
    }

    /**
     * Method declaration
     *
     *
     * @see
     */
    protected void finish() {
        Generator selection;

        try {
            getCodeGenDeck().writeOptionSet();
            selection = selectionPanel.getSelection();
            files = selection.generate();
            super.finish();
        } catch (ValidationException e) {
            JOptionPane.showMessageDialog(this, e.getValidationMessage(),
                                          cgRes.getString("Validation_Error"),
                                          JOptionPane.ERROR_MESSAGE);
        } catch (GeneratorException e) {
            JOptionPane.showMessageDialog(this, e.getMessage(),
                                          cgRes.getString("Generation_Error"),
                                          JOptionPane.ERROR_MESSAGE);
        }
    }

    /**
     * Add listeners and other items that are not generated by
     * the JBuilder designer.
     */
    private void codeGenInit() throws Exception {
        CodeGenDeck newDeck = null;
        CodeGenPage page = null;
        newDeck =
            (CodeGenDeck) Beans.instantiate(getClass().getClassLoader(),
                                                  CodeGenDeck.class.getName());
        selectionPanel =
            (GenSelectionPanel) Beans.instantiate(getClass().getClassLoader(),
                                                  GenSelectionPanel.class.getName());
        setDeck(newDeck);
        page = new CodeGenPage();
        page.setPageTitle(selectionPanel.getPageTitle());
        page.setInstructions(selectionPanel.getInstructions());
        page.add(selectionPanel);
        addWizardPage(page);
    }

    private CodeGenDeck getCodeGenDeck() {
      return (CodeGenDeck) getDeck();
    }



}

