/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.tool.codegen.wizard;

// ToolBox imports
import org.enhydra.tool.common.SwingUtil;
import org.enhydra.tool.common.ResUtil;

// Standard imports
import javax.swing.*;
import java.io.File;
import java.awt.*;
import java.awt.event.*;
import java.util.ResourceBundle;

/**
 * The ConfigDialog provides a Swing interface for specifying the
 * Enhydra home directory and creating a codegen.properties file. Note
 * that the dialog does not do the actual configuration. The caller
 * must retreive the selected Enhydra home directory to complete
 * the configuration.
 *
 * @author Paul Mahar
 */
public class ConfigDialog extends JDialog {
    static ResourceBundle res = ResourceBundle.getBundle("org.enhydra.tool.codegen.Res"); // nores
    //
    private final String FILE_ENHYDRA_JAR = "enhydra.jar";  // nores
    private final String DIR_LIB = "lib";  // nores

    /**
     * Indicates user choose to close the dialog with the Configure button.
     */
    public static final int     CHOICE_CONFIG = 0;

    /**
     * Indicates user choose to close the dialog with the Cancel button.
     */
    public static final int     CHOICE_CANCEL = 1;

    //
    transient private int                 choice = ConfigDialog.CHOICE_CANCEL;
    transient private JPanel              panelMain = new JPanel();
    transient private BorderLayout        layoutMain = new BorderLayout();
    transient private ConfigPanel         configPanel = new ConfigPanel();
    transient private LocalButtonListener buttonListener;

    /**
     * Open the configuration dialog with no parent.
     *
     * @param args
     * Arguments are ignored.
     */
    public static void main(String[] args) {
        ConfigDialog dialog = new ConfigDialog();

        dialog.show();
    }

    /**
     * Create a configuration dialog with a specified owner, title
     * and modality.
     *
     * @param frame
     * The owner for modal dialogs.
     * @param title
     * The dialog caption.
     * @param modal
     * True to make dialog modal.
     */
    public ConfigDialog(Frame frame, String title, boolean modal) {
        super(frame, title, modal);
        try {
            jbInit();
            pmInit();
            pack();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Create a modal configuration dialog with no parent or caption.
     */
    public ConfigDialog() {
        this(null, new String(), true);
    }

    /**
     * Overrides JDialog.show() to open the dialog in
     * the center of the screen.
     */
    public void show() {

        // setResizable(false); // crashes JVM on Win98 (NT/Win95 OK).
        Point cPoint;

        invalidate();
        doLayout();
        pack();
        cPoint = SwingUtil.getCenteringPoint(getSize());
        setLocation(cPoint);
        super.show();
    }

    /**
     * Remove references to swing components.
     */
    public void clearAll() {
        removeAll();
        dispose();
    }

    /**
     * Get a file reference to the selected Enhydra home directory.
     *
     * @return
     * The Enhydra home directory.
     *
     */
    public File getEnhydraHome() {
        return configPanel.getEnhydraHome();
    }

    public void setEnhydraHome(File f) {
      configPanel.setEnhydraHome(f);
    }

    /**
     * Get the button choice that was used to close the dialog.
     *
     * @return
     * CHOICE_CONFIG or CHOICE_CANCEL
     *
     */
    public int getChoice() {
        return choice;
    }

    // /
    // / PRIVATE
    // /

    /**
     * Method declaration
     *
     *
     * @exception Exception
     *
     * @see
     */
    private void jbInit() throws Exception {
        panelMain.setLayout(layoutMain);
        panelMain.add(configPanel, BorderLayout.CENTER);
        getContentPane().add(panelMain);
    }

    /**
     * Method declaration
     *
     *
     * @see
     */
    private void pmInit() {
        buttonListener = new LocalButtonListener();
        configPanel.buttonCancel.addActionListener(buttonListener);
        configPanel.buttonConfig.addActionListener(buttonListener);
        setTitle(res.getString("Configure_Enhydra"));
    }

    /**
     * Method declaration
     *
     *
     * @see
     */
    private void config() {
        if (isEnhydraHomeValid()) {
            closeDialog();
        } else {
            JOptionPane.showMessageDialog(this,
               ResUtil.format(
                           res.getString("Cannot_find_enhydra"), getEnhydraJar()),
                            res.getString("Unable_to_configure_"),
                            JOptionPane.ERROR_MESSAGE);
        }
    }

    /**
     * Method declaration
     *
     *
     * @see
     */
    private void closeDialog() {
        setVisible(false);
        clearAll();
    }

    /**
     * Method declaration
     *
     *
     * @return
     *
     * @see
     */
    private File getEnhydraJar() {
        File jar = null;
        StringBuffer buf = new StringBuffer();
        buf.append(getEnhydraHome().getAbsolutePath());
        buf.append(File.separator);
        buf.append(DIR_LIB);
        buf.append(File.separator);
        buf.append(FILE_ENHYDRA_JAR);
        jar = new File(buf.toString());
        return jar;
    }

    /**
     * Method declaration
     *
     *
     * @return
     *
     * @see
     */
    private boolean isEnhydraHomeValid() {
        File file = getEnhydraJar();

        return file.isFile();
    }

    /**
     * This inner class defines the actions for buttons in the wizard.
     *
     * @author
     * Paul Mahar
     * @version %I%, %G%
     */
    private class LocalButtonListener implements ActionListener {

        /**
         * ActionListener implementation.
         *
         * @param event
         * Event that triggered this listener.
         *
         */
        public void actionPerformed(ActionEvent event) {
            Object source = event.getSource();

            if (source == configPanel.buttonCancel) {
                choice = ConfigDialog.CHOICE_CANCEL;
                closeDialog();
            } else if (source == configPanel.buttonConfig) {
                choice = ConfigDialog.CHOICE_CONFIG;
                config();
            }
        }

    }
}
