/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.tool.common;

//
// Standard imports
//

import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.beans.*;
import java.util.Vector;
import java.util.ResourceBundle;

/**
 * The CodeGenDialog defines a default container for presenting
 * wizards. To use this class you need to a create one or more
 * CodeGenPage objects and add them to this dialog. This dialog
 * supplies action buttons for the Previous, Next, Finish,
 * Cancel and Help.
 * <P><I>
 * This class is used when creating a standalone wizard. Use the
 * OpenTools BasicWizard when developing JBuilder add-ins.
 * </I></P>
 */
public class TBWizardDialog extends JDialog {
    static ResourceBundle                       res =
        ResourceBundle.getBundle ("org.enhydra.tool.common.Res");       // nores

    //

    private final String                        FILE_WIZARD_JPG =
        "org/enhydra/tool/common/media/wizard.jpg";     // nores

    /**
     * Indicates user choose to close the dialog with the Finish button.
     */
    public final static int                     CHOICE_FINISH = 0;

    /**
     * Indicates user choose to close the dialog with the Cancel button.
     */
    public final static int                     CHOICE_CANCEL = 1;

    //

    private final static String                 TITLE =
        res.getString ("Enhydra_Wizard");

    //

    transient private Border                    border;
    transient private JButton                   buttonBack;
    transient private JButton                   buttonCancel;
    transient private JButton                   buttonFinish;
    transient private JButton                   buttonHelp;
    transient private JButton                   buttonNext;
    transient private TBWizardDeck              deck;
    transient private BorderLayout              layoutImage;
    transient private GridBagLayout             layoutConstant;
    transient private GridBagLayout             layoutMain;
    transient private JLabel                    labelImage;
    transient private JPanel                    panelConstant;
    transient private JPanel                    panelFiller;
    transient private JPanel                    panelImage;
    transient private JPanel                    panelLine;
    transient private JPanel                    panelMain;
    transient private LocalButtonListener       buttonListener;
    transient private int                       choice =
        TBWizardDialog.CHOICE_CANCEL;

    /**
     * Create a wizard dialog as a child of another dialog.
     *
     * @param owner
     * The dialog that is the parent of this wizard.
     */
    public TBWizardDialog (Dialog owner) {
        super (owner, TBWizardDialog.TITLE, true);

        construct ();
    }

    /**
     * Create a wizard dialog as a child of a frame.
     *
     * @param owner
     * The frame that is the parent of this wizard.
     */
    public TBWizardDialog (Frame owner) {
        super (owner, TBWizardDialog.TITLE, true);

        construct ();
    }

    /**
     * Create a wizard dialog without a parent.
     */
    public TBWizardDialog () {
        setModal (true);
        setWizardTitle (TBWizardDialog.TITLE);
        construct ();
    }

    /**
     * Overrides JDialog.show() to open the dialog in
     * the center of the screen.
     */
    public void show () {

        // setResizable(false); // crashes JVM on Win98 (NT/Win95 OK).

        Point   cPoint;

        invalidate ();
        doLayout ();
        pack ();

        cPoint = SwingUtil.getCenteringPoint (getSize ());

        setLocation (cPoint);
        super.show ();
    }

    /**
     * Clear all object references.
     */
    public void clearAll () {
        this.removeAll ();
        this.dispose ();
    }

    /**
     * Get the button choice that was used to close the dialog.
     *
     * @return
     * CHOICE_FINISH or CHOICE_CANCEL
     *
     */
    public int getChoice () {
        return choice;
    }

    public TBWizardDeck getDeck () {
        return deck;
    }

    public void setDeck (TBWizardDeck d) {
        Component[]     comps = new Component[0];

        deck = d;

        deck.setMinimumSize (new Dimension (100, 100));
        deck.setPreferredSize (new Dimension (100, 100));

        comps = panelMain.getComponents ();

        for (int i = 0; i < comps.length; i++) {
            if (comps[i] instanceof TBWizardDeck) {
                panelMain.remove (comps[i]);
            }
        }

        panelMain.add (deck,
                       new GridBagConstraints (1, 0, 1, 1, 0.9, 0.9,
                                               GridBagConstraints.CENTER,
                                               GridBagConstraints.BOTH,
                                               new Insets (2, 2, 2, 2), 0,
                                               0));
    }

    /**
     * Get the title used for the caption bar.
     */
    public String getWizardTitle () {
        return this.getTitle ();
    }

    /**
     * Set the title to use in the caption bar.
     */
    public void setWizardTitle (String s) {
        this.setTitle (s);
    }

    /**
     * Add a page to the wizard.
     *
     * @param page
     * A panel that contains swing controls that
     * are specific to the current wizard.
     */
    public void addWizardPage (TBWizardPage page) {
        deck.addWizardPage (page);
    }

    // / PROTECTED

    /**
     * Construct the object for either constructor.
     */
    protected void construct () {
        try {
            jbInit ();
            pmInit ();
        } catch (Exception e) {
            e.printStackTrace ();
        }
    }

    /**
     * Method declaration
     *
     *
     */
    protected void finish () {
        closeDialog ();
    }

    /**
     * Method declaration
     *
     *
     */
    protected void next () {
        deck.next ();
    }

    // /
    // / PRIVATE METHODS
    // /

    /**
     * Close the dialog and release swing references.
     */
    private void closeDialog () {
        setVisible (false);
        clearAll ();
    }

    /**
     * Add listeners and other items that are not generated by
     * the JBuilder designer.
     */
    private void pmInit () throws Exception {
        ImageIcon       image;

        deck =
            (TBWizardDeck) Beans.instantiate (getClass ().getClassLoader (),
                                              TBWizardDeck.class.getName ());

        setDeck (deck);

        //

        try {
            image =
                new ImageIcon (ClassLoader.getSystemResource (FILE_WIZARD_JPG));
        } catch (NullPointerException e) {
            image = null;
        }

        if (image == null) {
            panelImage.setVisible (false);
        } else {
            panelImage.setVisible (true);
            labelImage.setText (new String ());
            labelImage.setIcon (image);
        }

        //

        buttonListener = new LocalButtonListener ();

        buttonBack.setEnabled (false);
        buttonFinish.setEnabled (false);
        buttonBack.addActionListener (buttonListener);
        buttonNext.addActionListener (buttonListener);
        buttonFinish.addActionListener (buttonListener);
        buttonCancel.addActionListener (buttonListener);
        buttonHelp.addActionListener (buttonListener);
    }

    /**
     * UI code generated by JBuilder.
     * <P><I>Do not modify the signature of this method.</I></P>
     */
    private void jbInit () throws Exception {
        panelConstant =
            (JPanel) Beans.instantiate (getClass ().getClassLoader (),
                                        JPanel.class.getName ());
        panelLine = (JPanel) Beans.instantiate (getClass ().getClassLoader (),
                                                JPanel.class.getName ());
        panelMain = (JPanel) Beans.instantiate (getClass ().getClassLoader (),
                                                JPanel.class.getName ());
        layoutMain =
            (GridBagLayout) Beans.instantiate (getClass ().getClassLoader (),
                                               GridBagLayout.class.getName ());
        border = BorderFactory.createEtchedBorder ();
        layoutConstant =
            (GridBagLayout) Beans.instantiate (getClass ().getClassLoader (),
                                               GridBagLayout.class.getName ());
        buttonBack =
            (JButton) Beans.instantiate (getClass ().getClassLoader (),
                                         JButton.class.getName ());
        buttonNext =
            (JButton) Beans.instantiate (getClass ().getClassLoader (),
                                         JButton.class.getName ());
        buttonFinish =
            (JButton) Beans.instantiate (getClass ().getClassLoader (),
                                         JButton.class.getName ());
        buttonCancel =
            (JButton) Beans.instantiate (getClass ().getClassLoader (),
                                         JButton.class.getName ());
        buttonHelp =
            (JButton) Beans.instantiate (getClass ().getClassLoader (),
                                         JButton.class.getName ());
        panelFiller =
            (JPanel) Beans.instantiate (getClass ().getClassLoader (),
                                        JPanel.class.getName ());
        panelImage =
            (JPanel) Beans.instantiate (getClass ().getClassLoader (),
                                        JPanel.class.getName ());
        layoutImage =
            (BorderLayout) Beans.instantiate (getClass ().getClassLoader (),
                                              BorderLayout.class.getName ());
        labelImage =
            (JLabel) Beans.instantiate (getClass ().getClassLoader (),
                                        JLabel.class.getName ());

        panelMain.setLayout (layoutMain);
        panelLine.setBorder (border);
        panelLine.setMinimumSize (new Dimension (5, 5));
        panelLine.setPreferredSize (new Dimension (5, 5));
        buttonBack.setText (res.getString ("_Back"));
        buttonBack.setMnemonic ('B');
        buttonNext.setText (res.getString ("Next_"));
        buttonNext.setMnemonic ('N');
        buttonFinish.setText (res.getString ("Finish"));
        buttonCancel.setText (res.getString ("Cancel"));
        buttonHelp.setText (res.getString ("Help"));
        buttonHelp.setMnemonic ('H');
        buttonHelp.setEnabled (false);
        panelConstant.setLayout (layoutConstant);
        panelImage.setLayout (layoutImage);
        panelImage.add (labelImage, BorderLayout.CENTER);
        panelConstant.add (panelFiller,
                           new GridBagConstraints (0, 0, 1, 1, 0.1, 0.1,
                                                   GridBagConstraints.CENTER,
                                                   GridBagConstraints.BOTH,
                                                   new Insets (0, 0, 0, 0),
                                                   0, 0));
        panelConstant.add (buttonBack,
                           new GridBagConstraints (1, 0, 1, 1, 0.0, 0.0,
                                                   GridBagConstraints.CENTER,
                                                   GridBagConstraints.NONE,
                                                   new Insets (3, 3, 3, 3),
                                                   0, 0));
        panelConstant.add (buttonNext,
                           new GridBagConstraints (2, 0, 1, 1, 0.0, 0.0,
                                                   GridBagConstraints.CENTER,
                                                   GridBagConstraints.NONE,
                                                   new Insets (3, 3, 3, 3),
                                                   0, 0));
        panelConstant.add (buttonFinish,
                           new GridBagConstraints (3, 0, 1, 1, 0.0, 0.0,
                                                   GridBagConstraints.CENTER,
                                                   GridBagConstraints.NONE,
                                                   new Insets (3, 3, 3, 3),
                                                   0, 0));
        panelConstant.add (buttonCancel,
                           new GridBagConstraints (4, 0, 1, 1, 0.0, 0.0,
                                                   GridBagConstraints.CENTER,
                                                   GridBagConstraints.NONE,
                                                   new Insets (3, 3, 3, 3),
                                                   0, 0));
        panelConstant.add (buttonHelp,
                           new GridBagConstraints (5, 0, 1, 1, 0.0, 0.0,
                                                   GridBagConstraints.CENTER,
                                                   GridBagConstraints.NONE,
                                                   new Insets (3, 3, 3, 3),
                                                   0, 0));
        panelMain.setMinimumSize (new Dimension (542, 372));
        panelMain.setPreferredSize (new Dimension (542, 372));
        panelMain.add (panelImage,
                       new GridBagConstraints (0, 0, 1, 1, 0.1, 0.1,
                                               GridBagConstraints.CENTER,
                                               GridBagConstraints.NONE,
                                               new Insets (2, 2, 2, 2), 0,
                                               0));
        panelMain.add (panelLine,
                       new GridBagConstraints (0, 1, 2, 1, 0.0, 0.0,
                                               GridBagConstraints.CENTER,
                                               GridBagConstraints.HORIZONTAL,
                                               new Insets (1, 1, 1, 1), 0,
                                               0));
        panelMain.add (panelConstant,
                       new GridBagConstraints (0, 2, 2, 1, 0.0, 0.0,
                                               GridBagConstraints.CENTER,
                                               GridBagConstraints.BOTH,
                                               new Insets (1, 1, 1, 1), 0,
                                               0));
        this.getContentPane ().add (panelMain);
    }

    /**
     * This inner class defines the actions for buttons in the wizard.
     *
     * @author
     * Paul Mahar
     * @version %I%, %G%
     */
    class LocalButtonListener implements ActionListener {

        /**
         * ActionListener implementation.
         *
         * @param event
         * Event that triggered this listener.
         *
         */
        public void actionPerformed (ActionEvent event) {
            Object      source = event.getSource ();

            if (source == buttonNext) {
                next ();
            } else if (source == buttonBack) {
                deck.previous ();
            } else if (source == buttonFinish) {
                choice = TBWizardDialog.CHOICE_FINISH;

                finish ();
            } else if (source == buttonCancel) {
                choice = TBWizardDialog.CHOICE_CANCEL;

                closeDialog ();
            }

            buttonNext.setEnabled (!deck.isLastCardShowing ());
            buttonBack.setEnabled (!deck.isFirstCardShowing ());
            buttonFinish.setEnabled (!deck.isFirstCardShowing ());
        }

    }

}

