/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.io;

import org.enhydra.xml.dom.DOMTraversal;
import org.enhydra.xml.dom.DOMError;
import org.enhydra.xml.dom.DOMOps;
import org.enhydra.xml.dom.DOMAccess;
import java.util.HashMap;
import java.io.Writer;
import java.io.StringWriter;
import java.io.IOException;
import org.w3c.dom.*;

/**
 * Base class with common functionally XML and HTML DOM formatting
 * functionality.
 */
abstract class BaseDOMFormatter implements Formatter, DOMTraversal.Handler {
    // N.B. A lot of fields are made final to aid in optimization.

    /**
     * Maximum character value in an entity quick-check table.
     */
    protected static final int MAX_ENTITY_QUICK_CHECK_CHAR = 0x7f;

    /**
     * Quote character to use for attribute values.
     */
    protected static final char ATTR_QUOTE_CHAR = '"';

    /**
     * Entity reference for attribute value quote character.
     */
    protected static final String ATTR_QUOTE_CHAR_ENTITY_REF = "#34";

    /**
     * The output options.
     */
    protected final OutputOptions fOptions;

    /**
     * Output writer.
     */
    protected Writer fOut;

    /**
     * DOM traverser.
     */
    protected final DOMTraversal fTraverser;

    /**
     * Should character entity references be omitted from attribute values?
     */
    private final boolean fOmitAttributeCharEntityRefs;

    /**
     * Should character entity references be subtituted in the current
     * context.  This is modified when outputting an attribute value if
     * to implement fOmitAttributeCharEntityRefs.
     */
    private boolean fSubstitutingCharEntities = true;

    /**
     * Are we doing pre-formatting?  
     */
    protected final boolean fPreFormatMode;

    /**
     * Should pre-formatted text be written.  This is only enabled if the
     * document implements PreFormattedTextDocument and the preformatted
     * encoding matches the output encoding.
     */
    protected final boolean fUsePreFormattedText;

    /**
     * Should pre-formatted Elements be written.  This is the same as
     * fUsePreFormattedText, unless a URLRewriter is present, it which
     * cases its false, since we need to check the attributes for URLs.
     */
    protected final boolean fUsePreFormattedElements;

    /**
     * Count of preformatted text nodes that were written.
     */
    private int fPreFormattedTextCount;

    /**
     * Count of text nodes that were written not using preformatted text.
     */
    private int fDynamicFormattedTextCount;

    /**
     * Count of preformatted element open tags that were written.
     */
    protected int fPreFormattedElementCount;

    /**
     * Count of element open tags that were written not using preformatted
     * text.
     */
    protected int fDynamicFormattedElementCount;

    /**
     * Are we pretty-printing?
     */
    protected final boolean fPrettyPrinting;
    
    /**
     * Pretty-printing indent size.
     */
    private final int fIndentSize;

    /**
     * Java encoding for the current document.
     */
    private final String fJavaEncoding;

    /**
     * XML encoding for the current document.
     */
    private final String fXmlEncoding;

    /**
     * The maximum value of an unicode character in the document's encoding.
     */
    private final int fMaxCharacterValue;

    /**
     * Object to do URL value rewriting, or null if URL rewriting is
     * not being done.
     */
    private final URLRewriter fURLRewriter;

    /**
     * Document being formatted.
     */
    protected final Document fDocument;
    
    /**
     * DocumentType from document.
     */
    protected final DocumentType fDocType;
    
    /**
     * Public id to use for DOCTYPE, or null if none.
     */
    protected final String fPublicId;

    /**
     * System id to use for DOCTYPE, or null if none.
     */
    protected final String fSystemId;

    /**
     * If the document implements DocumentInfo, this is set.  It is
     * used by URL rewriting.
     */
    private final DocumentInfo fDocInfo;

    /**
     * Entity quick-check table, indexed by character value, indicating if the
     * character must be represented as an character entity reference.
     */
    private final boolean[] fEntityQuickCheck;

    /**
     * Platform line separator.
     */
    private static final String fNewLine;

    /**
     * Static string that is used as a quick way to get a certain indent
     * level.  Indent strings are generated by taking substrings of this
     * string.  This string is grown as needed.
     */
    private static String fIndentSource
        = "                                                               ";

    /**
     * Static initializer.
     */
    static {
        fNewLine = System.getProperty("line.separator");
        if (fNewLine == null) {
            throw new XMLIOError("System property line.separator not found");
        }
    }

    /**
     * Check if preformatted text might be used for this document.
     * 
     * @return The output options that were used for preformatting or null
     *  if preformatted text can be used for some reason.
     */
    private static OutputOptions checkPreformattedText(Document doc,
                                                       String defaultXmlEncoding,
                                                       int maxCharacterValue) {
        // Document must implement PreFormattedTextDocument
        if (!(doc instanceof PreFormattedTextDocument)) {
            return null;
        }
        
        // OutputOptions used for preformatting must be available
        OutputOptions pfOptions = ((PreFormattedTextDocument)doc).getPreFormatOutputOptions();
        if (pfOptions == null) {
            return null;
        }

        // Maximum character size must match (to prevent numeric entity problems)
        String preFormatXmlEncoding = pfOptions.getXmlEncoding();
        if (preFormatXmlEncoding == null) {
            preFormatXmlEncoding = defaultXmlEncoding;
        }
        if (maxCharacterValue != Encoding.getMaxCharacterValue(preFormatXmlEncoding)) {
            return null;
        }
        return pfOptions;
    }

    /**
     * Determine if a DocumentInfo object is available.  Node is passed in as
     * to support getting DocumentInfo from an XMLObject.
     * FIXME:  The node thing is a hack.
     */
    private static DocumentInfo findDocumentInfo(Node node,
                                                 Document document) {
        if (node instanceof DocumentInfo) {
            // This will handle the XMLC Document object
           return (DocumentInfo)node;
        } else if (document instanceof DocumentInfo) {
            return (DocumentInfo)document;
        } else {
            return null;
        }
    }

    /**
     * Constructor.
     *
     * @param node Any node of the document that this formatter will be
     *  associated with.  This can also be an XMLC Document object (XMLObject).
     * @param options The output options.
     * @param defaultXmlEncoding The default XML encoding for this format.
     * @param forPreFormatting Is this going to be used for preformatting?
     * @param entityQuickCheck Document-type specific table that provides
     *  a quick check of the need to encode that character as a character
     *  entity reference.  This table MUST include the double-quote character,
     *  as it it used to quote attribute values.
     */
    protected BaseDOMFormatter(Node node,
                               OutputOptions outputOptions,
                               boolean forPreFormatting,
                               String defaultXmlEncoding,
                               boolean[] entityQuickCheck) {
        fDocument = DOMOps.getDocument(node);
        fDocType = DOMAccess.accessDocumentType(fDocument);
        fOptions = outputOptions;
        fTraverser = DOMTraversal.getTraverser(this, 0, node);

        // Get DOCTYPE information (and overrides from OutputOptions)
        String publicId = fOptions.getPublicId();
        String systemId = fOptions.getSystemId();
        if (fDocType != null) {
            if (fDocType.getPublicId() != null) {
                publicId = fDocType.getPublicId();
            }
            if (fDocType.getSystemId() != null) {
                systemId = fDocType.getSystemId();
            }
        }
        fPublicId = publicId;
        fSystemId = systemId;

        // Initialize the encoding information.
        String xmlEncoding = fOptions.getXmlEncoding();
        fXmlEncoding = (xmlEncoding != null) ? xmlEncoding : defaultXmlEncoding;
        fJavaEncoding = Encoding.xml2JavaEncoding(fXmlEncoding);
        fMaxCharacterValue = Encoding.getMaxCharacterValue(fXmlEncoding);

        // Setup URL rewriting, which needs DocumentInfo
        fDocInfo = findDocumentInfo(node, fDocument);
        fURLRewriter = (fDocInfo != null) ? fOptions.getURLRewriter() : null;
        
        // Various options from output options.
        fIndentSize = fOptions.getIndentSize();
        fPrettyPrinting = (fIndentSize > 0) && fOptions.getPrettyPrinting();
        fOmitAttributeCharEntityRefs = fOptions.getOmitAttributeCharEntityRefs();

        // Initialize preformatted text options
        OutputOptions pfOptions = checkPreformattedText(fDocument, defaultXmlEncoding,
                                                        fMaxCharacterValue);
        fUsePreFormattedText = (pfOptions != null) && (!forPreFormatting)
            && (!fPrettyPrinting);
        fUsePreFormattedElements = (pfOptions != null) && (!forPreFormatting)
            && (pfOptions.getDropHtmlSpanIds() == fOptions.getDropHtmlSpanIds())
            && (pfOptions.getOmitAttributeCharEntityRefs() == fOptions.getOmitAttributeCharEntityRefs())
            && (fURLRewriter == null);
        fPreFormatMode = forPreFormatting;

        // Entity handling
        fEntityQuickCheck = entityQuickCheck;

        // initialize for pre-formatting.
        if (forPreFormatting) {
            fOut = new StringWriter(4096); // bigger than default
        }
    }

    /**
     * @see Formatter#getJavaEncoding
     */
    public final String getJavaEncoding() {
        return fJavaEncoding;
    }

    /**
     * @see Formatter#getXmlEncoding
     */
    public final String getXmlEncoding() {
        return fXmlEncoding;
    }

    /**
     * @see Formatter#usedPreFormattedText
     */
    public boolean usedPreFormattedText() {
        return fUsePreFormattedText;
    }

    /**
     * Get the count of preformatted text nodes that were written.
     */
    public final int getPreFormattedTextCount() {
        return fPreFormattedTextCount;
    }

    /**
     * Get the count of text nodes that were written not using
     * preformatted text.
     */
    public final int getDynamicFormattedTextCount() {
        return fDynamicFormattedTextCount;
    }

    /**
     * @see Formatter#usedPreFormattedElements
     */
    public final boolean usedPreFormattedElements() {
        return fUsePreFormattedElements;
    }

    /**
     * Get the count of preformatted element open tags that were written.
     */
    public final int getPreFormattedElementCount() {
        return fPreFormattedElementCount;
    }

    /**
     * Get the count of element open tags that were written not using
     * preformatted text.
     */
    public final int getDynamicFormattedElementCount() {
        return fDynamicFormattedElementCount;
    }

    /**
     * Write a newline
     */
    protected final void writeln() throws IOException {
        fOut.write(fNewLine);
    }

    /**
     * Grow indentation source string to at least the specified size
     * if needed.
     */
    private static void ensureIndentSource(int numChars) {
        // Grow as needed; coded to avoid sync
        while (fIndentSource.length() < numChars) {
            fIndentSource += fIndentSource;
        }
    }

    /**
     * Print indentation to the current level.
     */
    protected final void printIndent() throws IOException {
        int indent = fTraverser.getDepth() * fIndentSize;
        ensureIndentSource(indent);
        fOut.write(fIndentSource, 0, indent);
    }

    /**
     * Get a character entity name for a character.
     * This is the slow-path, so its ok this is an abstract method call.
     * @return The character entity name, or null if this character doesn't
     *  have one.
     */
    abstract protected String getCharacterEntity(char textChar);

    /**
     * Output a character, possibly substituting a character entity reference
     * or a numeric entity. This is the slow-path method.
     */
    private void writeCharacter(char textChar) throws IOException {
        // Determine the entity to substitute; always do quote, even if substitute disabled,
        // as this is only implemented for attribute values.
        String entity;
        if (fSubstitutingCharEntities) {
            entity = getCharacterEntity(textChar);
        } else {
            entity = (textChar == ATTR_QUOTE_CHAR) ? ATTR_QUOTE_CHAR_ENTITY_REF : null;
        }
        if (entity != null) {
            fOut.write('&');
            fOut.write(entity);
            fOut.write(';');
        } else if (textChar > fMaxCharacterValue) {
            fOut.write("&#");
            fOut.write(Integer.toString(textChar));
            fOut.write(';');
        } else {
            fOut.write(textChar);
        }
    }

    /**
     * Write a text string, encoding document type-specific character entities.
     * This is an expensive procedure and has been carefully hand optimized.
     */
    protected final void writeString(String text) throws IOException {
        int len = text.length();
        char ch;

        for (int idx = 0; idx < len; idx++) {
            ch = text.charAt(idx);
            if ((ch <= fMaxCharacterValue)
                && (ch <= MAX_ENTITY_QUICK_CHECK_CHAR)
                && (!fEntityQuickCheck[ch])) {
                fOut.write(ch);  // Fast path.
            } else {
                writeCharacter(ch);
            }
        }
    }

    /**
     * Handler called for Entity nodes; should never be called.
     * @see DOMTraversal.Handler#handleEntity
     */
    public final void handleEntity(Entity entity) {
        throw new XMLIOError("Unexpected call to handleEntity");
    }

    /**
     * Handler called for Notation nodes; should never be called.
     * @see DOMTraversal.Handler#handleNotation
     */
    public final void handleNotation(Notation notation) {
        throw new XMLIOError("Unexpected call to handleNotation");
    }

    /**
     * Handler called for EntityReference nodes.
     * @see DOMTraversal.Handler#handleEntityReference
     */
    public final void handleEntityReference(EntityReference entityRef) throws IOException {
        fOut.write('&');
        fOut.write(entityRef.getNodeName());
        fOut.write(';');
    }

    /**
     * Handler called for Comment nodes.
     * @see DOMTraversal.Handler#handleComment
     */
    public final void handleComment(Comment comment) throws IOException {
        fOut.write("<!--");
        fOut.write(comment.getData());
        fOut.write("-->");
    }

    /**
     * Handler called for Text nodes.
     * @see DOMTraversal.Handler#handleText
     */
    public void handleText(Text text) throws IOException {
        if (fUsePreFormattedText && (text instanceof PreFormattedText)) {
            String data = ((PreFormattedText)text).getPreFormattedText();
            if (data != null) {
                fOut.write(data);
                fPreFormattedTextCount++;
                return;
            }
        }
        // No formatted text.
        writeString(text.getData());
        fDynamicFormattedTextCount++;
    }

    /**
     * Write an attribute value.  Convert characters to character entity
     * references as needed.
     */
    protected final void writeAttributeValue(Attr attr) throws IOException {
        fOut.write('=');
        fOut.write(ATTR_QUOTE_CHAR);
        
        if (fOmitAttributeCharEntityRefs) {
            fSubstitutingCharEntities = false;
        }

        // Are we doing URL rewritting and is this a URL attr?
        if ((fURLRewriter != null)
            && fDocInfo.isURLAttribute(attr.getOwnerElement(),
                                       attr.getName())) {
            writeString(fURLRewriter.rewriteURL(attr.getValue()));
        } else {
            fTraverser.processChildren(attr);
        }
        if (fOmitAttributeCharEntityRefs) {
            fSubstitutingCharEntities = true;
        }
        fOut.write(ATTR_QUOTE_CHAR);
    }

    /**
     * Preformat an element.
     */
    private String preFormatElement(Element element) {
        fTraverser.traverse(element);
        return ((StringWriter)fOut).getBuffer().toString();
    }

    /**
     * Preformat a text node.  The preformatted text is returned even if
     * its the same.  We assume the compiler will not duplicate identical
     * string constants.
     */
    private String preFormatText(Text text) {
        fTraverser.traverse(text);
        return ((StringWriter)fOut).getBuffer().toString();
    }

    /**
     * @see Formatter.preFormatNode
     */
    public final String preFormatNode(Node node) {
        ((StringWriter)fOut).getBuffer().setLength(0); // Clear buffer

        switch (node.getNodeType()) {
        case Node.ELEMENT_NODE:
            return preFormatElement((Element)node);
        case Node.TEXT_NODE:
            return preFormatText((Text)node);
        }
        return null;
    }

    /**
     * Format a Node and children to the specified writer.
     * @see Formatter#write
     */
    public final void write(Node node,
                            Writer writer) throws IOException {
        try {
            fOut = writer;
            fTraverser.traverse(node);
            writeln();
        } catch (DOMError error) {
            // Rethrow IOExceptions
            Throwable cause = error.getCause();
            if (cause instanceof IOException) {
                throw (IOException)cause;
            } else {
                throw error;
            }
        }
    }
}
