/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XMLC.java,v 1.8.2.2.4.2 2000/10/13 18:28:01 markd Exp $
 */

package org.enhydra.xml.xmlc.commands.xmlc;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.compiler.Compiler;
import org.enhydra.xml.xmlc.commands.ErrorHandling;
import org.enhydra.xml.xmlc.commands.options.*;
import org.enhydra.xml.xmlc.metadata.*;
import org.enhydra.xml.io.ErrorReporter;
import java.io.PrintWriter;
import java.io.IOException;

/**
 * XMLC compiler program.  See the user manual for details on the command
 * line.  Will look for a system property `org.enhydra.xml.xmlc.javac' to find
 * the compiler to use if -javac is not specified.
 */
public class XMLC {
    /**
     * System property to pass default javac.
     */
    private static final String JAVAC_PROPERTY = "org.enhydra.xml.xmlc.javac";

    /**
     * Command line parsing object.
     */
    private XMLCOptionsParser optionsParser = new XMLCOptionsParser();

    /**
     * Enable verbose output.
     */
    private boolean verbose = false;

    /*
     * Error output.
     */
    ErrorReporter errorReporter = new ErrorReporter(new PrintWriter(System.err, true));

    /*
     * Trace and verbose output.
     */
    private PrintWriter traceOut = new PrintWriter(System.err, true);

    /*
     * Print the XMLC version number.
     */
    private void printVersion() {
        System.out.println("Enhydra XMLC version " + XMLCVersion.VERSION);
        System.out.println("See http://xmlc.enhydra.org for latest distribution");
    }

    /**
     * Parse arguments.
     */
    private MetaData parseArgs(String[] args) throws XMLCException, IOException {
        MetaData metaData = optionsParser.parse(args, errorReporter);
        
        // Default javac if supplied by a property.
        String defaultJavac = System.getProperty(JAVAC_PROPERTY);
        if (defaultJavac != null) {
            JavaCompilerSection compilerSection = metaData.getJavaCompilerSection();
            if (!compilerSection.isJavacSpecified()) {
                compilerSection.setJavac(defaultJavac);
            }
        }
        return metaData;
    }

    /**
     * Parse arguments and compile the page.
     */
    public void compile(String[] args) throws XMLCException, IOException {
        
        // Setup and document parsing.
        MetaData metaData = parseArgs(args);
        CompileOptions compileOptions = metaData.getCompileOptions();
        verbose = compileOptions.getVerbose();

        if (compileOptions.getPrintVersion()) {
            printVersion();
        }

        if (compileOptions.getInputDocument() != null) {
            Compiler compiler = new Compiler(errorReporter,
                                             traceOut);
            compiler.compile(metaData);
        }
    }

    /*
     * Parse arguments and compile the page.  Print errors that
     * are thrown.
     */
    private void compileHandleErrors(String[] args) {
        try {
            compile(args);
        } catch (Exception except) {
            ErrorHandling.handleException(except, verbose);
        }
    }

    /*
     * Program entry point.
     */
    public static void main(String[] args) {
        XMLC xmlc = new XMLC();
        xmlc.compileHandleErrors(args);
    }
}
