/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.xmlc.compiler;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.metadata.*;
import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.io.InputSourceOps;
import com.lutris.util.ChainedThrowable;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;
import java.io.*;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.DOMFormatter;

/**
 * XMLC compiler class.  This class implements the steps in compiling a
 * document.  This is a programatic interface and does not have a main or
 * handle command line options.
 */
public class Compiler {
    /*
     * Document metadata.
     */
    private MetaData metaData;

    /*
     * Compile options from metadata.
     */
    private CompileOptions compileOptions;

    /**
     * Enable verbose output.
     */
    private boolean verbose = false;

    /**
     * Input source being compiled.
     */
    private InputSource inputSource;

    /*
     * Document being compiled.
     */
    private Document document;

    /*
     * Error output.
     */
    ErrorReporter errorReporter;

    /*
     * Trace and verbose output.
     */
    private PrintWriter traceOut;

    /*
     * XML document management object.
     */
    private XMLCDocument xmlcDoc;

    /**
     * Constructor.
     * @param errorReporter Object called to handle and output errors.
     * @param traceOut Object to write trace and document information to.
     */
    public Compiler(ErrorReporter errorReporter,
                    PrintWriter traceOut) {
        this.errorReporter = errorReporter;
        this.traceOut = traceOut;
    }

    /*
     * Parse the page into the DOM and perform various checks and edits.
     */
    private void parsePage() throws XMLCException, IOException {
        if (verbose) {
            traceOut.println(">>> parsing " + InputSourceOps.getName(inputSource));
        }

        Parse parser = new Parse(errorReporter, traceOut);
        xmlcDoc = parser.parse(metaData);
        document = xmlcDoc.getDocument();

        if (compileOptions.getPrintDocumentInfo()) {
            if (verbose) {
                traceOut.println(">>> printing document info");
            }
            PrintInfo info = new PrintInfo(document, xmlcDoc);
            info.printInfo(traceOut);
        }

        EditDOM domEditor = new EditDOM(metaData);
        domEditor.edit(xmlcDoc);

        if (compileOptions.getPrintDOM()) {
            if (verbose) {
                traceOut.println(">>> dumping document hierarchy");
            }
            DOMInfo.printTree("DOM hierarchy", document, 
                              traceOut);
        }
    }

    /*
     * Generate the Java source files.
     */
    private void generateJavaSource() throws XMLCException, IOException {
        if (verbose) {
            traceOut.println(">>> generating code");
        }
        ClassGenerator codeGen =
            new ClassGenerator(metaData, xmlcDoc,
                               (compileOptions.getPrintAccessorInfo() ? traceOut : null));
        codeGen.generateJavaSource(verbose ? traceOut : null);
    }

    /*
     * Compile the generate source.
     */
    private void compileJavaSource() throws XMLCException {
        if (verbose) {
            traceOut.println(">>> compiling code");
        }
        Javac javac = new Javac();
        javac.compile(metaData,
                      errorReporter,
                      verbose ? traceOut : null);
    }

    /*
     * Write the DOM to a file.
     */
    private void writeDOM() throws IOException {
        File docOut = new File(compileOptions.getDocumentOutput());
        String dir = docOut.getParent();
        if (dir != null) {
            new File(dir).mkdirs();
        }
        OutputOptions options = new OutputOptions();
        options.setXmlEncoding(xmlcDoc.getEncoding());
        DOMFormatter formatter = new DOMFormatter(options);
        formatter.write(xmlcDoc.getDocument(),
                        docOut);
    }

    /*
     * Generate the Jave source and compile the page.
     */
    private void compileDocument()
        throws XMLCException, IOException {
        DocumentClass documentClass = metaData.getDocumentClass();
        try {
            generateJavaSource();
            if (compileOptions.getCompileSource()) {
                compileJavaSource();
                if (documentClass.getRecompilation()) {
                    metaData.getDocument().serialize();
                }
            }
        } finally {
            if (!compileOptions.getKeepGeneratedSource()) {
                if (documentClass.getJavaClassSource() != null) {
                    documentClass.getJavaClassSource().delete();
                }
                if (documentClass.getJavaInterfaceSource() != null) {
                    documentClass.getJavaInterfaceSource().delete();
                }
            }
        }
    }

    /*
     * Parse arguments and compile the page.
     */
    public void compile(MetaData metaData)
            throws XMLCException, IOException {
        
        this.metaData = metaData;
        compileOptions = metaData.getCompileOptions();
        verbose = compileOptions.getVerbose();
        errorReporter.setPrintWarnings(compileOptions.getWarnings());
        
        String inputDocument = compileOptions.getInputDocument();
        if (inputDocument == null) {
            throw new XMLCException("input document not specified");
        }
        inputSource = new InputSource(inputDocument);

        parsePage();

        if (compileOptions.getCreateSource()) {
            compileDocument();
        }

        // Write document after edits, if requested.
        if (compileOptions.getDocumentOutput() != null) {
            writeDOM();
        }
        if (verbose) {
            traceOut.println(">>> completed");
        }
    }
}
