/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DocBuilder.java,v 1.5.2.1.2.1 2000/10/11 10:20:42 markd Exp $
 */

package org.enhydra.xml.xmlc.dom;

import org.enhydra.xml.xmlc.*;
import org.w3c.dom.*;
import java.util.ArrayList;

//FIXME:  Delete this, HTML supprot move to HTML; use native DOM parser
//for Xerces.

/**
 * Class for building a Document DOM.  Used by XMLC parsers to build a document
 * DOM.
 * <P>
 * The document builder functions assume they are being called in the order the
 * document is parsed.  They keep a current node where new child nodes are
 * appended.
 */
public class DocBuilder {
    /**
     * Factory for creating the document.
     */
    private XMLCDomFactory fDomFactory;

    /**
     * Object being used to build the DocumentType object. If no
     * DOCTYPE or  HTML, this is null.
     */
    private DocTypeBuilder fDocTypeBuilder;
    
    /**
     * Information from XML header.
     */
    private String fDocXMLVersion;
    private String fDocEncoding;
    private boolean fDocStandalone;

    /**
     * The document.
     */
    private Document fDocument;

    /**
     * Root element name
     */
    private String fRootTagName;

    /**
     * The current node that is being constructed.  This functions
     * as a stack during document construction.
     */
    protected Node fCurrentNode;

    /**
     * Holds comments before the first element until the first element is
     * encountered.
     */
    private ArrayList fPendingComments;

    /**
     * Constructor.
     * 
     * @param domFactory Factory class for Documents.
     * @param docTypeBuilder Object being used to construct the DocumentType.
     *  It will be called to get the document type when first needed.
     *  Should be null to indicate an HTML document.
     */
    public DocBuilder(XMLCDomFactory domFactory,
                      DocTypeBuilder docTypeBuilder) {
        fDomFactory = domFactory;
        fDocTypeBuilder = docTypeBuilder;
    }

    /**
     * Generate error about a method being called that should
     * be called before the document is created.
     */
    private void docNotCreatedError() {
        throw new XMLCError("Bug: parser event on document contents occured before document is created");
    }

    /**
     * Create the document.  Called only when needed to allow DocumentType
     * information to be collected.
     */
    protected Document createDocument() {
        return fDomFactory.createDocument(fRootTagName,
                                          fDocTypeBuilder.getCreateDocType());
    }

    /**
     * Hold a comment until the first element is encountered.
     */
    private void holdComment(String data) {
        if (fPendingComments == null) {
            fPendingComments = new ArrayList();
        }
        fPendingComments.add(data);
    }

    /**
     * Add in comments that occured before the first element.
     */
    private void addPendingComments() {
        // Add comments before document element.
        Node first = fDocument.getFirstChild();
        for (int idx = fPendingComments.size()-1; idx >= 0; idx--) {
            Comment com = fDocument.createComment((String)fPendingComments.get(idx));
            fDocument.insertBefore(com, first);
        }
    }

    /**
     * Create the Document object when the first element of the document is
     * found.  This is delayed until the the first element, since all
     * information needed to build the document is not available until that
     * time.
     */
    private void makeDocument(String rootTagName) {
        fRootTagName = rootTagName;

        // Create the document, which also creates the DocumentElement.
        fDocument = createDocument();
        fCurrentNode = fDocument.createElement(rootTagName);
        fDocument.appendChild(fCurrentNode);

        if (fPendingComments != null) {
            addPendingComments();
        }
    }

    /**
     * Get the document associated with this object.
     */
    public Document getDocument() {
        if (fDocument == null) {
            throw new XMLCError("XMLC bug: Attempt to get document that hasn;t been created");
        }
        return fDocument;
    }

    /**
     * Get the DOM factory associated with this object.
     */
    public XMLCDomFactory getDomFactory() {
        return fDomFactory;
    }

    /**
     * Get the node on the top of the stack during parsing.
     * FIXME: To support the broken swing parser.
     */
    public Node getCurrentNode() {
        return fCurrentNode;
    }

    /**
     * Pop the current node off of the stack.  This is *only* used
     * during error recover from a broken parser.
     * FIXME: This and getCurrentNode() should go away when the
     * with the broken swing parser.
     */
    public void popCurrentNode() {
        fCurrentNode = fCurrentNode.getParentNode();
    }

    /**
     * Set the XML version.
     *
     * @param xmlVersion XML version string.
     */
    public void setXMLVersion(String xmlVersion) {
        fDocXMLVersion = xmlVersion;
    }

    /**
     * Get the XML version.
     */
    public String getXMLVersion() {
        return fDocXMLVersion;
    }

    /**
     * Set the standalone flag.
     *
     * @param standalone Is standalone specified?
     */
    public void setStandalone(boolean standalone) {
        fDocStandalone = standalone;
    }

    /**
     * Get the standalone flag.
     */
    public boolean getStandalone() {
        return fDocStandalone;
    }

    /**
     * Get the publicId.
     */
    public String getPublicId() {
        return (fDocTypeBuilder != null) ? fDocTypeBuilder.getPublicId() : null;
    }

    /**
     * Get the systemId.
     */
    public String getSystemId() {
        return (fDocTypeBuilder != null) ? fDocTypeBuilder.getSystemId() : null;
    }

    /**
     * Set the encoding for the document.
     *
     * @param encoding The encoding for the document
     */
    public void setEncoding(String encoding) {
        fDocEncoding = encoding;
    }

    /**
     * Get the encoding specified in the document.
     */
    public String getEncoding() {
        return fDocEncoding;
    }

    /**
     * Start a new Element.
     */
    public void startElement(String tagName) {
        if (fDocument == null) {
            // root (first) element
            makeDocument(tagName);
        } else {
            // non-root element
            Element element = fDocument.createElement(tagName);
            fCurrentNode.appendChild(element);
            fCurrentNode = element;
        }
    }
    
    /**
     * Add an attribute to the element on the top of the
     * stack.
     */
    public void addAttribute(String name, String value) {
        ((Element)fCurrentNode).setAttribute(name, value);
    }

    /**
     * Finish the element being constructed. 
     */
    public void finishElement() {
        if (fCurrentNode == null) {
            throw new XMLCError("node stack underflow; malformed document");
        }
        if (!(fCurrentNode instanceof Element)) {
            throw new XMLCError("DOM node top of stack not a element for end tag");
        }
        fCurrentNode = fCurrentNode.getParentNode();
    }

    /**
     * Start an entity reference in the document (not DTD).
     */
    public void startEntityReference(String entityName) {
        if (fDocument == null) {
            docNotCreatedError();
        }

        EntityReference entityRef = fDocument.createEntityReference(entityName);
        fCurrentNode.appendChild(entityRef);
        fCurrentNode = entityRef;
    }

    /**
     * End an entity reference.
     */
    public void endEntityReference() {
        if (fCurrentNode == null) {
            throw new XMLCError("node stack underflow; malformed document");
        }
        if (!(fCurrentNode instanceof EntityReference)) {
            throw new XMLCError("DOM node top of stack not a EntityReference for end tag");
        }
        fCurrentNode = fCurrentNode.getParentNode();
    }

    
    /**
     * Add a <code>Text</code> node.
     */
    public void addTextNode(String data) {
        if (fDocument == null) {
            docNotCreatedError();
        }
        fCurrentNode.appendChild(fDocument.createTextNode(data));
    }

    /**
     * Add a <code>Comment</code> node.
     */
    public void addComment(String data) {
        if (fDocument == null) {
            holdComment(data);  // before first element..
        } else {
            fCurrentNode.appendChild(fDocument.createComment(data));
        }
    }

    /**
     * Add a <code>CDATASection</code> node.
     */
    public void addCDATASection(String data) {
        if (fDocument == null) {
            docNotCreatedError();
        }
        fCurrentNode.appendChild(fDocument.createCDATASection(data));
    }

    /**
     * Add a <code>ProcessingInstruction</code> node.
     */
    public void addProcessingInstruction(String target, 
                                         String data) {
        if (fDocument == null) {
            docNotCreatedError();
        }
        fCurrentNode.appendChild(fDocument.createProcessingInstruction(target, data));
    }

    /**
     * Add an EntityReference object.
     */
    public void addEntityReference(String name) {
        if (fDocument == null) {
            docNotCreatedError();
        }
        fCurrentNode.appendChild(fDocument.createEntityReference(name));
    }
}
