/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.xmlc.dom.generic;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.codegen.*;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.xmlc.compiler.ElementInfo;
import org.enhydra.xml.xmlc.compiler.ElementTable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;

/**
 * Class to generate code to build the document tree using only the W3C DOM
 * interface.  The document building code is generated in several functions to
 * avoid the limit on the maximum size of a method.
 */
public class GenericDocBuilderGenerator implements DocBuilderGenerator {
    /*
     * Maximum creation-cost per build method.  One unit is approximately the
     * cost to create one node.
     */
    private static final int MAX_CREATE_COST_PER_BUILD_METHOD = 512;

    /*
     * Name of variable in init function containing Document.
     */
    private static final String DOCUMENT_VAR = "document";

    /*
     * Name of variable in init function that holds document type object.
     */
    private static final String DOC_TYPE_VAR = "documentType";

    /**
     * Name of argument to method containing parent node.
     */
    private static String PARENT_NODE_ARG = "parentNode";

    /**
     * Base name of DOM building methods.
     */
    private static final String SUBDOCUMENT_BUILD_METHOD = "buildSubDocument";

    /**
     * Doc creater generator object.
     */
    private GenericDocCreateGenerator fDocCreateGenerator;

    /**
     * Constructor.
     * @param docCreateGenerator Object that creates the document.  A
     *  subclassed object can be provided to handle specific DOM
     *  implementations.
     */
    protected GenericDocBuilderGenerator(GenericDocCreateGenerator docCreateGenerator) {
        fDocCreateGenerator = docCreateGenerator;
    }

    /**
     * Generate code to create the document.
     */
    private void createDocumentCreation(XMLCDocument xmlcDoc,
                                        JavaMethod buildDocumentMethod) {
        JavaCode body = buildDocumentMethod.getCode();
        
        fDocCreateGenerator.generateDocCreate(xmlcDoc.getDocument(), true,
                                              DOCUMENT_VAR, body);
        // Save document information.
        body.addln("setDocument(" + DOCUMENT_VAR + ","
                   + JavaLang.createStringConst(xmlcDoc.getDomFactory().getMIMEType()) + ", "
                   + JavaLang.createStringConst(xmlcDoc.getEncoding()) + ");");
        body.addln();
    }

    //FIXME. Xerces should extends and add the enable/disable error checking
    // Just the information neede3d to build the document.  In the
    // meantime, we pass in XMLCDoc (this was the original intent
    // of this class.

    /**
     * Generate the document builder method.
     * @see DocBuilderGenerator#createBuildDocumentMethod
     */
    public void createBuildDocumentMethod(XMLCDocument xmlcDoc,
                                          AccessorGenerator accessorGenerator,
                                          ElementTable elementTable,
                                          JavaClass docClass,
                                          JavaMethod buildDocumentMethod) throws XMLCException {
        // Create code to create DOM.
        DOMBuilderGenerator domBuilderGenerator
            = new DOMBuilderGenerator(SUBDOCUMENT_BUILD_METHOD,
                                      xmlcDoc.getDocument(),
                                      Document.class.getName(),
                                      new NodeCreateGenerator(xmlcDoc),
                                      accessorGenerator,
                                      elementTable,
                                      docClass,
                                      MAX_CREATE_COST_PER_BUILD_METHOD,
                                      false);
        // Generate the document building method.
        createDocumentCreation(xmlcDoc, buildDocumentMethod);
        domBuilderGenerator.createMethodCall(DOCUMENT_VAR,
                                             buildDocumentMethod.getCode());
    }
}
