/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.xmlc.dom.generic;

import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import org.enhydra.xml.xmlc.codegen.JavaCode;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;

/**
 * Class used create a DOM nodes.  This maybe extended by a class needing
 * to create nodes in a different
 */
public class NodeCreateGenerator {
    /** Document object being compiled */
    private XMLCDocument fXmlcDoc;
    
    /**
     * Constructor.
     */
    public NodeCreateGenerator(XMLCDocument xmlcDoc) {
        fXmlcDoc = xmlcDoc;
    }

    /**
     * Check if this is a node that is create when the document
     * is created.
     */
    public boolean isPreCreatedNode(Node node) {
        return fXmlcDoc.isPreCreatedNode(node);
    }

    /**
     * Generate Attr creation.
     */
    private void genAttrCreate(Attr attr,
                               String docVarName,
                               JavaCode body) {
        if (attr.getNamespaceURI() == null) {
            body.addln(docVarName + ".createAttribute("
                       + JavaLang.createStringConst(attr.getName()) + ");");
        } else {
            body.addln(docVarName + ".createAttributeNS("
                       + JavaLang.createStringConst(attr.getNamespaceURI()) + ", "
                       + JavaLang.createStringConst(attr.getName()) + ");");
        }
    }

    /**
     * Generate Element creation.
     */
    private void genElementCreate(Element element,
                                  String docVarName,
                                  JavaCode body) {
        if (isPreCreatedNode(element)) {
            body.addln(docVarName + ".getDocumentElement();");
        } if (element.getNamespaceURI() == null) {
            body.addln(docVarName + ".createElement("
                       + JavaLang.createStringConst(element.getTagName()) + ");");
        } else {
            body.addln(docVarName + ".createElementNS("
                       + JavaLang.createStringConst(element.getNamespaceURI()) + ", "
                       + JavaLang.createStringConst(element.getTagName()) + ");");
        }
    }

    /**
     * Generate Notation creation.
     */
    private void genNotationCreate(Notation notation,
                                   String docVarName,
                                   JavaCode body) {
        body.addln(docVarName + ".createNotation("
                   + JavaLang.createStringConst(notation.getNodeName()) + ", "
                   + JavaLang.createStringConst(notation.getPublicId()) + ", "
                   + JavaLang.createStringConst(notation.getSystemId()) + ");");
    }

    /**
     * Generate ProcessingInstruction creation.
     */
    private void genProcessInstructionCreate(ProcessingInstruction procInstr,
                                             String docVarName,
                                             JavaCode body) {
        body.addln(docVarName + ".createProcessingInstruction("
                   + JavaLang.createStringConst(procInstr.getNodeName()) + ", "
                   + JavaLang.createStringConst(procInstr.getData()) + ");");
    }

    /**
     * Generate code to construct a node, assigning it to the supplied
     * variable. Can't construct top-level documents.  Can be overriden by
     * derived class
     */
    public void genNodeCreate(String docVarName,
                              String varName,
                              Node node,
                              JavaCode body) throws XMLCException {
        body.add(varName + " = ");
        
        switch (node.getNodeType()) {
        case Node.ATTRIBUTE_NODE:
            genAttrCreate((Attr)node, docVarName, body);
            break;

        case Node.CDATA_SECTION_NODE:
            body.addln(docVarName + ".createCDATASection("
                       + JavaLang.createStringConst(((CDATASection)node).getData()) + ");");
            break;

        case Node.COMMENT_NODE:
            body.addln(docVarName + ".createComment("
                       + JavaLang.createStringConst(((Comment)node).getData()) + ");");
            break;

        case Node.ELEMENT_NODE:
            genElementCreate((Element)node, docVarName, body);
            break;

        case Node.ENTITY_NODE:
            //FIXME: need abstract method, not part of DOM.
            throw new XMLCException("XMLC bug: entity node found in document");

        case Node.ENTITY_REFERENCE_NODE:
            body.addln(docVarName + ".createEntityReference("
                       + JavaLang.createStringConst(((EntityReference)node).getNodeName()) + ");");
            break;

        case Node.NOTATION_NODE:
            genNotationCreate((Notation)node, docVarName, body);
            break;

        case Node.PROCESSING_INSTRUCTION_NODE:
            genProcessInstructionCreate((ProcessingInstruction)node,
                                        docVarName, body);
            break;

        case Node.TEXT_NODE:
            body.addln(docVarName + ".createTextNode("
                       + JavaLang.createStringConst(((Text)node).getData()) + ");");
            break;

        case Node.DOCUMENT_FRAGMENT_NODE:
        case Node.DOCUMENT_NODE:
        case Node.DOCUMENT_TYPE_NODE:
            throw new XMLCError("Creating node of this type not supported: "
                                + node.getClass().getName());

        default:
            throw new XMLCError("Unknown node type: " + node.getNodeType());
        }
    }
}
