/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TidyErrorHandler.java,v 1.2.16.2 2000/10/29 00:34:21 markd Exp $
 */

package org.enhydra.xml.xmlc.html.parsers.tidy;

import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.io.ErrorReporter;
import java.io.Writer;
import java.io.PrintWriter;
import java.io.IOException;
import gnu.regexp.*;


/**
 * Class for intercepting Tidy error output and doing custom error handing.
 * This is a PrintWrite that overrides all public methods to do our own
 * handling.  It parses output looking for errors an warnings and passes them
 * to the XMLC error handler.  This is hack, hopefully Tidy.parse will get
 * real error callbacks.  It will break if tidy is ever localized.
 */
class TidyErrorHandler extends PrintWriter {
    /**
     * Actual output stream.
     */
    private ErrorReporter reporter;

    /**
     * Name of file being parsed.
     */
    private String fileName;

    /**
     * Line buffer.  Use to hold message until end of line.
     */
    private StringBuffer outbuf = new StringBuffer();

    /**
     * Regular expression for parsing warning/error lines.
     * Result subexpressions are:
     *  1 - line number
     *  2 - "Warning" or "Error"
     *  3 - message
     * If these change, processWarnError must change.
     */
    private static final String warnErrorRegExp = 
        "^line ([0-9,]+) column [0-9,]+ - (Warning|Error): (.*)$";

    /**
     * Compiler warning/error regexp.
     */
    private static final RE warnErrorRE;

    /**
     * Regular expressions defining lines to discard.
     */
    private static final String[] discardRegExps = {
        ".*Tidy \\(vers.*\\) Parsing.*",
        ".*\\\".*\\\" appears to be.*",
        ".*warnings/errors were found.*",
        ".*warnings or errors were found.*",
        ".*This document has errors that must be fixed before.*",
        ".*Document content looks like.*"
    };

    /**
     * Compiled expressions defining lines to discard.
     */
    private static final RE[] discardREs;

    static {
        // Compile regular expressions.
        try {
            warnErrorRE = new RE(warnErrorRegExp, 0, RESyntax.RE_SYNTAX_POSIX_EXTENDED);

            discardREs = new RE[discardRegExps.length];
            for (int idx = 0; idx < discardRegExps.length; idx++) {
                discardREs[idx] = new RE(discardRegExps[idx], RE.REG_MULTILINE,
                                         RESyntax.RE_SYNTAX_POSIX_EXTENDED);
            }
        } catch (REException except) {
            throw new XMLCError(except);
        }
    }

    /**
     * Constructor.
     */
    public TidyErrorHandler(ErrorReporter reporter,
                            String fileName) {
        // All print methods are overridden, so super is not actually used.
        super(System.err);
        this.reporter = reporter;
        this.fileName = fileName;
    }

    /**
     * Convert string containing a Tidy integer, which might contain a `,'
     * to a number.
     */
    private int convertTidyInt(String intStr) {
        // Remove commas
        char[] strBuf = new char[intStr.length()];
        int cIdx = 0;
        for (int sIdx = 0; sIdx < intStr.length(); sIdx++) {
            char c = intStr.charAt(sIdx);
            if (c != ',') {
                strBuf[cIdx++] = c;
            }
        }
        return Integer.valueOf(new String(strBuf, 0 ,cIdx)).intValue();
    }

    /**
     * Process an warning or error match.
     */
    private void processWarnError(REMatch match) {
        int lineNum = convertTidyInt(match.toString(1));
        String which = match.toString(2);
        String msg = match.toString(3);

        if (which.equals("Warning")) {
            reporter.warning(msg, fileName, lineNum);
        } else {
            reporter.error(msg, fileName, lineNum);
        }
    }

    /**
     * Process completed line.
     */
    private void processLine() {
        REMatch match =  warnErrorRE.getMatch(outbuf);
        if (match != null) {
            processWarnError(match);
        } else {
            // See if this should be discarded
            int idx;
            for (idx = 0; idx < discardREs.length; idx++) {
                if (discardREs[idx].isMatch(outbuf)) {
                    break;
                }
            }
            if (idx == discardREs.length) {
                // Didn't match a discard pattern, output it.
                reporter.warning("XMLC Did not detect this Tidy message: "
                                 + outbuf);
            }
        }
            
        outbuf.setLength(0);
    }

    /**
     * Flush the stream.
     */
    public void flush() {
    }

    /**
     * Close the stream.
     */
    public void close() {
    }

    /**
     * Flush the stream and check its error state.
     */
    public boolean checkError() {
        return false;
    }

    /**
     * Write a single character.
     */
    public void write(int c) {
        outbuf.append(c);
    }

    /**
     * Write a portion of an array of characters.
     */
    public void write(char buf[], int off, int len) {
        outbuf.append(buf, off, len);
    }

    /**
     * Write a portion of a string.
     */
    public void write(String s, int off, int len) {
        outbuf.append(s.substring(off, off+len));
    }

    /**
     * Write a string.
     */
    public void write(String s) {
        outbuf.append(s);
    }

    /**
     * Finish the line.
     */
    public void println() {
        processLine();
    }

    /**
     * Print a boolean, and then finish the line.
     */
    public void println(boolean x) {
        outbuf.append(x);
        processLine();
    }

    /**
     * Print a character, and then finish the line.
     */
    public void println(char x) {
        outbuf.append(x);
        processLine();
    }

    /**
     * Print an integer, and then finish the line.
     */
    public void println(int x) {
        outbuf.append(x);
        processLine();
    }

    /**
     * Print a long, and then finish the line.
     */
    public void println(long x) {
        outbuf.append(x);
        processLine();
    }

    /**
     * Print a float, and then finish the line.
     */
    public void println(float x) {
        outbuf.append(x);
        processLine();
    }

    /**
     * Print a double, and then finish the line.
     */
    public void println(double x) {
        outbuf.append(x);
        processLine();
    }

    /**
     * Print an array of characters, and then finish the line.
     */
    public void println(char x[]) {
        outbuf.append(x);
        processLine();
    }

    /**
     * Print a String, and then finish the line.
     */
    public void println(String x) {
        outbuf.append(x);
        processLine();
    }

    /**
     * Print an Object, and then finish the line.
     */
    public void println(Object x) {
        outbuf.append(x);
        processLine();
    }
}
