/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.xmlc.misc;

import org.enhydra.xml.io.InputSourceOps;
import java.io.Reader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.Iterator;
import org.xml.sax.InputSource;

/**
 * A file reader that keeps a map of byte offsets to line numbers. This is
 * used by the Swing HTML parser interface to provide reasonable error
 * messages.
 */
public class LineNumberMapReader extends Reader {
    /**
     * Input file for file name.
     */
    private InputSource fInputSource;

    /**
     * Contained reader.
     */
    private Reader fReader;

    /**
     * Structure to map byte offset to line numnber.
     */
    public class LineNumber {
        public final String fileName;  // File were encountered.
        public final int lineNum;      // one-based line number.
        public final int byteOffset;   // zero-based byte offset.

        /**
         * Constructor
         */
        LineNumber(String fName,
                  int lNum,
                  int offset) {
            fileName = fName;
            lineNum = lNum;
            byteOffset = offset;
        }
    }

    /**
     * Map of line number to offset.
     */
    private ArrayList fMap = new ArrayList();

    /**
     * Current line-number in the file.
     */
    private int fCurLineNum = 1;
    
    /**
     * Current offset in the file.
     */
    private int fCurOffset = 0;

    /**
     * Was the previous character a return?
     */
    private boolean fPrevWasReturn;

    /**
     * Construct a new reader for the specified file.
     */
    public LineNumberMapReader(InputSource inputSource)
            throws IOException {
        fInputSource = inputSource;
        fReader = new LineNumberReader(InputSourceOps.open(fInputSource));
        addLine();
    }

    /**
     * Add a line to the map.
     */
    private void addLine() {
        fMap.add(new LineNumber(fInputSource.getSystemId(), fCurLineNum,
                                fCurOffset));
    }

    /**
     * Get the filname end line number for a byte offset.  A one-based line
     * number is returned.
     */
    public LineNumber getLineNumber(int off) {
        // Start out with the first entry to remove special cases.
        LineNumber lineEntry = null;
        LineNumber prevLineEntry = (LineNumber)fMap.get(0);

        // Search for an entry.
        Iterator entries = fMap.iterator();
        while (entries.hasNext()) {
            lineEntry = (LineNumber)entries.next();
            if (off < lineEntry.byteOffset) {
                return prevLineEntry;
            }
            prevLineEntry = lineEntry;
        }
        return prevLineEntry;
    }

    /**
     * Count a charcter read (and hope this gets inlined).
     */
    private void countChar(char ch) {
        fCurOffset++;
        if (fPrevWasReturn) {
            fPrevWasReturn = false;
        } else {
            switch (ch) {
            case '\r':
                fPrevWasReturn = true;
            case '\n':
                fCurLineNum++;
                addLine();
            }
        }
    }

    /**
     * Read a character.
     *
     * @see Reader#read
     * @exception IOException If an I/O error occurs
     */
    public int read() throws IOException {
        int ch = fReader.read();
        if (ch >= 0) {
            countChar((char)ch);
        }
        return ch;
    }

    /**
     * Read characters into a portion of an array.
     *
     * @see Reader#read
     * @exception IOException If an I/O error occurs
     */
    public int read(char cbuf[], int off, int len) throws IOException {
        int num = fReader.read(cbuf, off, len);
        if (num < 0) {
            return num;
        }
        for (int cnt = 0; cnt < num; cnt++) {
            countChar(cbuf[off + cnt]);
        }
        return num;
    }

    /**
     * Close the stream.
     *
     * @see Reader#close
     * @exception  IOException  If an I/O error occurs
     */
    public void close() throws IOException {
        fReader.close();
    }
}
