/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraSelectiveClassLoader.java,v 1.3.6.1 2000/11/16 09:08:42 markd Exp $
 */

package org.enhydra.xml.xmlc.reloading;

import org.enhydra.xml.xmlc.XMLCFactory;
import java.io.PrintWriter;
import java.net.URL;
import java.security.AccessControlContext;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import org.enhydra.classloader.EnhydraClassLoader;
import com.lutris.logging.LogChannel;

/**
 * Selective based on the EnhydraClassLoader, which restricts loading to a
 * single class and its inner classes.
 *
 * @author Wes Saeger
 * @author Mark Diekhans
 * @version $Revision: 1.3.6.1 $
 * @see java.lang.ClassLoader
 * @see java.security.SecureClassLoader
 * @see org.enhydra.classloader.EnhydraClassLoader
 */
class EnhydraSelectiveClassLoader extends EnhydraClassLoader {
    /**
     * The class name this classloaded is restricted to, along with its inner
     * classes.
     */
    private String fRestrictClassName;

    /**
     * Logging interface.
     */
    private XMLCLogger fLogger;

    /**
     * Construct an instance of the class loader with the specified parent
     * class loader, class name, and class path.
     *
     * @param parent The parent class loader for delegation.
     * @param className The name of the class (and inner classes) this class loader
     *        is restricted to loading.
     * @param classPath The array of <CODE>URL</CODE>s to be used for class path.
     *        current class path.
     * @param logger XMLC logger or null for no logging.
     */
    protected EnhydraSelectiveClassLoader(ClassLoader parent,
                                          String className,
                                          URL[] classPath,
                                          XMLCLogger logger) {
        super(parent, classPath,
              ((logger instanceof EnhdyraXMLCLogger)
               ? ((EnhdyraXMLCLogger)logger).getLogChannel() : null));
        fRestrictClassName = className;
        fLogger = logger
    }

    /**
     * Check if the class should be loaded by this classloader based
     * on the classname.
     */
    private boolean loadRestrictionCheck(String className) {
        boolean accepted = className.startsWith(fRestrictClassName)
            && ((className.length() == fRestrictClassName.length())
                || (className.charAt(fRestrictClassName.length()) == '$'));
        if (fDebugLogWriter != null) {
            ResourceLoaderFactory.logRestrictionCheck(fDebugLogWriter, this,
                                                      fRestrictClassName,
                                                      className,
                                                      accepted);
        }
        return accepted;
    }

    /**
     * Attempt to load a class if the name of the class matches the the
     * restriction.  If the class does not match the restriction, delegate it
     * to the parent.  If it does match the restriction, it will never be
     * delegated, even if not found.
     *
     * @param className The name of the class to be loaded.
     * @param resolve Set to <CODE>true</CODE> for class resolution,
     *        <CODE>false</CODE> for no resolution.
     * @return <CODE>Class</CODE> object representing the loaded class.
     * @exception ClassNotFoundException if the class could not be loaded.
     */
    public Class loadClass(String className,
                           boolean resolve) throws ClassNotFoundException {
        Class cl;

        // Check for the restricted or an inner class
        if (loadRestrictionCheck(className)) {
            // Matches
            cl = findClass(className);
            if (resolve) {
                resolveClass(cl);
            }
        } else {
            // Does not match
            cl = super.loadClass(className, resolve);
        }
        return cl;
    }

    /**
     * Factory to create a new instance of the class loader.
     *
     * @param parent The parent class loader for delegation.  If null, then
     *        the system class loader is used as the parent.
     * @param classPath The array of <CODE>URL</CODE>s to be used for class path.
     * @param log The <CODE>LogChannel</CODE> for logging.  Maybe null.
     */
    public static EnhydraSelectiveClassLoader newInstance(ClassLoader parent,
                                                          final String className,
                                                          final URL[] classPath,
                                                          final LogChannel logChannel) {
        AccessControlContext callersAccessContext = AccessController.getContext();
        final ClassLoader fparent
            = (parent == null) ? getSystemClassLoader() : parent;
        
        EnhydraSelectiveClassLoader loader = (EnhydraSelectiveClassLoader)AccessController.doPrivileged(new PrivilegedAction() {
            public Object run() {
                return new EnhydraSelectiveClassLoader(fparent, className,
                                                       classPath, logChannel);
            }
        });

        loader.setAccessContext(callersAccessContext);

        return loader;
    }
}
