/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ResourceLoaderFactory.java,v 1.2.2.1.4.2 2000/11/16 09:08:42 markd Exp $
 */

package org.enhydra.xml.xmlc.reloading;

import org.enhydra.xml.xmlc.XMLCLogger;
import java.util.Hashtable;
import java.lang.reflect.Constructor;
import java.io.PrintWriter;

/**
 * Factory for creating a ResourceLoader wrapper class for a particular class
 * loader.  This takes a class loader class name and returns a instance
 * of the appropriate ResourceLoader.
 *
 * The class is pre-configured with ResourceLoader mappings for the
 * following class loaders:
 * <UL>
 * <LI> <code>com.lutris.classloader.MultiClassLoader</code>
 * <LI> <code>org.apache.tomcat.util.NetworkClassLoader</code>
 * <LI> <code>org.enhydra.classloader.EnhydraClassLoader</code>
 * <LI> <code>java.net.URLClassLoader</code>
 * </UL>
 *
 * @see ResourceLoader
 * @see MultiResourceLoader
 * @see TomcatResourceLoader
 */
public final class ResourceLoaderFactory {
    /**
     * Mapping of class loader class name to resource loader class.
     * May have exceptions in it if the load of a standard resource
     * loader fails.
     */
    private static Hashtable fLoaderTable = new Hashtable();

    /**
     * ResourceLoader constructor signature.
     */
    private static final Class[] fResLoaderConstructorSignature = {
        ClassLoader.class,
        XMLCLogger.class
    };

    /*
     * Initialize with pre-defined resource loaders.
     */
    static {
        registerPreDefinedResourceLoader("com.lutris.classloader.MultiClassLoader",
                                         "org.enhydra.xml.xmlc.reloading.MultiResourceLoader");
        registerPreDefinedResourceLoader("org.apache.tomcat.util.NetworkClassLoader",
                                         "org.enhydra.xml.xmlc.reloading.TomcatResourceLoader");
        registerPreDefinedResourceLoader("org.enhydra.classloader.EnhydraClassLoader",
                                         "org.enhydra.xml.xmlc.reloading.EnhydraResourceLoader");
        registerPreDefinedResourceLoader("java.net.URLClassLoader",
                                         "org.enhydra.xml.xmlc.reloading.URLResourceLoader");
    }


    /**
     * Disallow instantiation.
     */
    private ResourceLoaderFactory() {
    }

    /**
     * Register a resource loader class name for a class loader.
     *
     * @param classLoaderName The class name of the class loader.
     * @param resourceLoaderName The class name of the resource loader
     *  It must implement {@link ResourceLoader ResourceLoader} and 
     *  contain a public constructor taking no arguments.  This resource loader
     *  must be loaded by the system class loader.
     *
     * @see ResourceLoader
     */
    public static void registerResourceLoader(String classLoaderName,
                                              String resourceLoaderName) {
        try {
            Class resourceLoaderClass;
            // Load with our class loader, if it exists.
            ClassLoader loader = ResourceLoaderFactory.class.getClassLoader();
            if (loader != null) {
                resourceLoaderClass = loader.loadClass(resourceLoaderName);
            } else {
                resourceLoaderClass = ResourceLoaderFactory.class.forName(resourceLoaderName);
            }
            fLoaderTable.put(classLoaderName, resourceLoaderClass);
        } catch (ClassNotFoundException except) {
            throw new XMLCReloadingException("Can't load ResourceLoader class \""
                                             + resourceLoaderName + "\"", except);
        }
    }

    /**
     * Register a predefined resource loader.  If it is not found (because its
     * wasn't compiled).  Insert the exception in the loader table to throw if
     * its requested.
     */
    private static void registerPreDefinedResourceLoader(String classLoaderName,
                                                         String resourceLoaderName) {
        try {
            registerResourceLoader(classLoaderName, resourceLoaderName);
        } catch (XMLCReloadingException except) {
            fLoaderTable.put(classLoaderName, except);
        }
    }

    /**
     * Register a resource loader class for a class loader.
     *
     * @param classLoaderName The class name of the class loader.
     * @param resourceLoaderClass The class for the resource loader.
     *  It must implement {@link ResourceLoader ResourceLoader} and 
     *  contain a public constructor taking no arguments.  This resource loader
     *  can be loaded by any class loader.
     *
     * @see ResourceLoader
     */
    public static void registerResourceLoader(String classLoaderName,
                                              Class resourceLoaderClass) {
        fLoaderTable.put(classLoaderName, resourceLoaderClass);
    }

    /**
     * Find a ResourceLoader for a ClassLoader class by class name.
     * This allows classloaders that extend other known class loaders by
     * searching the inheritance chain.
     * @return The ResourceLoader class.
     */
    private static Class findResourceLoaderClass(ClassLoader classLoader) {
        // Walk the inheritance chain
        Class loaderClass = classLoader.getClass();
        Object loaderEntry = null;

        while (loaderClass != null) {
            loaderEntry = fLoaderTable.get(loaderClass.getName());
            if (loaderEntry != null) { 
                break;
            }
            loaderClass = loaderClass.getSuperclass();
        }
        if (loaderEntry == null) {
            throw new XMLCReloadingException("No XMLC ResourceLoader defined for class loader \""
                                             + classLoader.getClass().getName() + "\"");
        }
        if (loaderEntry instanceof Throwable) {
            throw new XMLCReloadingException("XMLC ResourceLoader class loader could not be loader for \""
                                             + classLoader.getClass().getName() + "\"",
                                             (Throwable)loaderEntry);
        }
        return (Class)loaderEntry;
    }

    /**
     * Create a ResourceLoader associated with a specified class loader.
     *
     * @param classLoader The class loader to create a ResourceLoader for.
     *  The ResourceLoader class must be registered with the factory.
     * @param logger The logger to pass to the resource loader.
     * @exception XMLCReloadingException If there is no registered resource 
     *  loader or the resource loader can't be created.
     */
    public static ResourceLoader create(ClassLoader classLoader,
                                        XMLCLogger logger) {

        Class resourceLoaderClass = findResourceLoaderClass(classLoader);
        try {
            Constructor constr = resourceLoaderClass.getConstructor(fResLoaderConstructorSignature);
            return (ResourceLoader)constr.newInstance(new Object[]{classLoader, logger});
        } catch (Throwable except) {
            throw new XMLCReloadingException("Can't create ResourceLoader class for "
                                             + classLoader.getClass().getName() + "\"",
                                             except);
        }
    }

    /**
     * Utility method used by selective class to log the result of a
     * restriction check.
     * @param logger XMLC logger to write to.
     * @param selectiveClassLoader The selective class loader class.
     * @param restrictClassName The class the loader is restricted to loading.
     * @param className The class being requested.
     * @param accepted Will the loader load this class?
     */
    public static void logRestrictionCheck(XMLCLogger logger,
                                           ClassLoader selectiveClassLoader,
                                           String restrictClassName,
                                           String className,
                                           boolean accepted) {
        if (logger.debugEnabled()) {
            logger.logDebug("loadRestrictionCheck: "
                            + "loader=" + selectiveClassLoader.getClass()
                            + ", restrict=" + restrictClassName
                            + ", request=" + className
                            + ", accepted=" + accepted);
        }
    }
}
