/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TomcatResourceLoader.java,v 1.1.8.2 2000/11/16 09:08:42 markd Exp $
 */

package org.enhydra.xml.xmlc.reloading;

import java.io.File;
import java.net.URL;
import java.util.Enumeration;
import java.util.ArrayList;
import org.enhydra.xml.xmlc.XMLCLogger;
import com.lutris.classloader.MultiClassLoader;
import org.apache.tomcat.util.NetworkClassLoader;

/*
 * Observed NetworkClassLoader issues:
 *   - The class loader does not search the classpath in order.
 *   - Caching of contents of zip might not be wise, there could
 *     be large files in a zip.  This also results in classes
 *     from zips being in memory twice.  Ack, it actually copies
 *     the entire zip to memory.
 *   - The decision to close the jar files should be revisited.
 *     Opening a jar file is maybe expensive due to reading the
 *     directory.  It's not clear if this is a good trade off
 *     for allowing jar files to be modified, especially since
 *     the loader will not pick up the modifications and its not
 *     clear what will happen if the loader opens a jar file
 *     at the same time its being updated.
 *   - Errors are ignored in many places, making debugging of
 *     failures difficult.
 *   - If it implemented NetworkClassLoader.findResource(), life
 *     would be much better.
 */

// FIXME: This doesn't take into account parent loader.
// FIXME: need to get useful message when file is in zip.

/**
 * ResourceLoader class for the Tomcat classloader.  This is not
 * ideal, as it doesn't take into account the parent loader or
 * chances in the classpath.  This works for classloader
 * <code>oorg.apache.tomcat.util.NetworkClassLoader</code> and
 * descendent <code>org.apache.tomcat.core/ServletClassLoader</code>.
 */
public class TomcatResourceLoader implements ResourceLoader {
    /**
     * Class loader we are associated with.
     */
    private NetworkClassLoader fClassLoader;

    /**
     * Logging interface.
     */
    private XMLCLogger fLogger;

    /**
     * Cached class path to search, in the order that the NetworkClass
     * loader will search it.
     */
    private URL[] fSearchClassPath;

    /**
     * MultiClassLoader filter.
     * Note: Duplicated from MultiResourceLoader since both will be
     * obsolete soon.
     */
    private class ClassFilter implements MultiClassLoader.ClassFilter {
        /** Class name to load */
        private String fClassName;

        /**
         * Constructor.
         */
        public ClassFilter(String className) {
            fClassName = className;
        }

        /**
         * Class filter check.
         * @see MultiClassLoader.ClassFilter
         */
        public int loadCheck(String checkClassName) {
            if (checkClassName.startsWith(fClassName)) {
                // Check for same or inner class
                if ((checkClassName.length() == fClassName.length())
                    || (checkClassName.charAt(fClassName.length()) == '$'))
                    return MUST_LOAD;
                }
            return DONT_LOAD;
        }
    }

    /**
     * Constructor.
     * @see ResourceLoader
     */
    public TomcatResourceLoader(ClassLoader classLoader,
                                XMLCLogger logger) {
        fClassLoader = (NetworkClassLoader)classLoader;
        fLogger = logger;
    }

    /**
     * @see ResourceLoader#getClassLoader
     */
    public ClassLoader getClassLoader() {
        return fClassLoader;
    }

    /**
     * Get the class path associate with this resource loader.
     */
    public URL[] getClassPath() {
        Enumeration pathEnum = fClassLoader.getURLs();
        ArrayList path = new ArrayList();
        
        URL[] urls = MultiClassLoader.getSystemClassPath();
        for  (int idx = 0; idx < urls.length; idx++) {
            path.add(urls[idx]);
        }
        while (pathEnum.hasMoreElements()) {
            path.add(pathEnum.nextElement());
        }
        return (URL[])path.toArray(new URL[path.size()]);
    }

    /**
     * Create a selective class loader for the specified class and
     * its subclasses.  All other classes are delegated to the 
     * ClassLoader assocation with this ResourceLoader. 
     */
    public ClassLoader createSelectiveClassLoader(String className) {
        //FIXME: should probably not include sys class path.
        return new MultiSelectiveClassLoader(className, getClassPath(),
                                             fClassLoader, fLogger);
    }
}
