/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: URLSelectiveClassLoader.java,v 1.2.6.1 2000/11/16 09:08:42 markd Exp $
 */

package org.enhydra.xml.xmlc.reloading;

import java.io.PrintWriter;
import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.AccessControlContext;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import org.enhydra.xml.xmlc.XMLCLogger;

/**
 * Selective classloader derived from the URLClassLoader.  An instance
 * of this class will load a specific class and it's inner classes.
 * An inner class is determined by comparing the class name using the
 * generated inner class naming convention.
 *
 * @see java.net.URLClassLoader;
 */
class URLSelectiveClassLoader extends URLClassLoader {
    /** Name of class that will be loaded */
    private String fRestrictClassName;

    /**
     * Logging interface.
     */
    private XMLCLogger fLogger;

    /**
     * Constructor, create a new instance of the classloader restricted
     * to loading a single class name.
     */
    public URLSelectiveClassLoader(String restrictClassName,
                                   URL[] path,
                                   ClassLoader parent,
                                   XMLCLogger logger) {
        super(path, parent);
        fRestrictClassName = restrictClassName;
        fLogger = logger;
    }

    /**
     * Check if the class should be loaded by this classloader based
     * on the classname.
     */
    private boolean loadRestrictionCheck(String className) {
        boolean accepted = className.startsWith(fRestrictClassName)
            && ((className.length() == fRestrictClassName.length())
                || (className.charAt(fRestrictClassName.length()) == '$'));
        ResourceLoaderFactory.logRestrictionCheck(fLogger, this,
                                                  fRestrictClassName,
                                                  className,
                                                  accepted);
        return accepted;
    }

    /**
     * Attempt to load a class if the name of the class matches the the
     * restriction.  If the class does not match the restriction, delegate it
     * to the parent.  If it does match the restriction, it will never be
     * delegated, even if not found.
     *
     * @param className The name of the class to be loaded.
     * @param resolve Set to <CODE>true</CODE> for class resolution,
     *        <CODE>false</CODE> for no resolution.
     * @return <CODE>Class</CODE> object representing the loaded class.
     * @exception ClassNotFoundException if the class could not be loaded.
     */
    public Class loadClass(String className,
                           boolean resolve) throws ClassNotFoundException {
        Class cl;

        // Check for the restricted or an inner class
        if (loadRestrictionCheck(className)) {
            // Matches
            cl = findClass(className);
            if (resolve) {
                resolveClass(cl);
            }
        } else {
            // Does not match, URL class loader will send to parent
            cl = super.loadClass(className, resolve);
        }
        return cl;
    }

    /**
     * Factory to create a new instance of the class loader.
     *
     * @param parent The parent class loader for delegation.  If null, then
     *        the system class loader is used as the parent.
     * @param classPath The array of <CODE>URL</CODE>s to be used for class path.
     * @param xmlcFactory Factory that is creating us; used to get log writer.
     */
    public static URLSelectiveClassLoader newInstance(ClassLoader parent,
                                                      final String className,
                                                      final URL[] classPath,
                                                      final XMLCLogger logger) {
        AccessControlContext callersAccessContext = AccessController.getContext();
        final ClassLoader fparent
            = (parent == null) ? getSystemClassLoader() : parent;

        URLSelectiveClassLoader loader = (URLSelectiveClassLoader)AccessController.doPrivileged(new PrivilegedAction() {
            public Object run() {
                return new URLSelectiveClassLoader(className, classPath,
                                                   fparent, logger);
            }
        });

        //FIXME: loader.setAccessContext(callersAccessContext);

        return loader;
    }
}
