/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package tests.org.enhydra.xml.xmlc;

//FIXME: rename autocomp to reloading in all test stuff

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.reloading.XMLCReloadingFactory;
import org.enhydra.xml.xmlc.reloading.StandardRecompilationHandler;
import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import com.lutris.testutil.*;
import com.lutris.logging.LogChannel;
import com.lutris.logging.Logger;
import com.lutris.util.FatalExceptionError;
import java.io.*;
import java.util.ArrayList;
import java.util.Properties;
import java.util.StringTokenizer;
import java.net.URL;
import java.net.URLClassLoader;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import org.w3c.dom.*;

/**
 * Tests for auto-compilation and class reloading.
 * 
 * The command line is: <BR>
 * <CODE>
 *  TestAutoComp [options] interfaceName [outfile]
 * </CODE>
 *
 * The following options are supported:
 * <UL>
 * <LI> -dump - Dump the DOM to stdout.
 * <LI> -sourcefile file - Path to the source file.  The root directory
 *      of the tree containing the file must be on the classpath.  This will
 *      will be overwritten.   
 * <LI> -copyfile file - File to copy over source file to generate an
 *      error or change the results.
 * <LI> -classloader name Use the specified classloader.  One of
 *      `enhydra', `multi', `url'.
 * </UL>
 * The directory containing 
 */
public class TestAutoComp {
    /**
     * Our package
     */
    private static final String TEST_PACKAGE_FILE
        = JavaLang.getPackageName(TestAutoComp.class.getName()).replace('.', '/');

    /**
     * Subpackages that that will actually contain tests XMLC doc class, but
     * not test framework classes.
     */
    private static final String[] TEST_SUBPACKAGE_FILE_PREFIXES = {
        TEST_PACKAGE_FILE + "/html/swing/",
        TEST_PACKAGE_FILE + "/html/tidy/",
        TEST_PACKAGE_FILE + "/xml/xerces/"
    };

    /**
     * Valid class loaders.
     */
    private static final String CL_ENHYDRA = "enhydra";
    private static final String CL_MULTI = "multi";
    private static final String CL_URL = "url";

    /*
     * Parsed from command line.
     */
    private boolean dump;
    private String interfaceName;
    private String outFile;
    private File sourceFile;
    private File copyFile;
    private String classLoaderName = CL_URL;

    /*
     * Objects used during the tests.
     */
    private Class origClass;
    private XMLObject origObject;

    /**
     * Usage message.
     */
    private static final String USAGE = "wrong # args: TestAutoComp [options] interfaceName [output]";

    /**
     * Test recording object.
     */
    private static Tester tester = new Tester();

    /**
     * Class loader that rejects attempts to load one of the test classes.
     * This is used as a parent to the class loader used in the test and
     * passed it back to the class loader the we want to use instead
     * of letting the system class loader get the class.  It serves as
     * a filter without us having to play classpath games.
     */
    private final class TestFilterClassLoader extends ClassLoader {
        /**
         * Test if a file path should be be returned to the delegator or passed
         * to the parent.
         */
        private boolean returnToDelegator(String fileName) {
            for (int i = 0; i < TEST_SUBPACKAGE_FILE_PREFIXES.length; i++) {
                if (fileName.startsWith(TEST_SUBPACKAGE_FILE_PREFIXES[i])) {
                    return true;
                }
            }
            return false;
        }

        /**
         * @see ClassLoader.getResource
         */
        public URL getResource(String name) {
            if (returnToDelegator(name)) {
                return null;
            } else {
                return getParent().getResource(name);
            }
        }

        /**
         * @see ClassLoader.getResourceAsStream
         */
        public InputStream getResourceAsStream(String name) {
            if (returnToDelegator(name)) {
                return null;
            } else {
                return getParent().getResourceAsStream(name);
            }
        }

        /**
         * @see ClassLoader.loadClass
         */
        public Class loadClass(String name) throws ClassNotFoundException {
            if (returnToDelegator(name.replace('.', '/') + ".class")) {
                throw new ClassNotFoundException(name);
            } else {
                return getParent().loadClass(name);
            }
        }
    }

    /**
     * Generate a failure error message and log.
     */
    private void failure(String testId,
                         String msg) {
        Thread.dumpStack();
        TestUtils.log("Failure: " + testId + ": " + msg);
        tester.failure(testId, msg);
    }

    /**
     * Add a note to the log file.
     */
    private void log(String msg) {
        TestUtils.log(msg);
    }

    /**
     * Add a note and exception to the log file.
     */
    private void log(String msg, Throwable except) {
        TestUtils.log(msg, except);
    }

    /**
     * Get the system class path, with URLs in the format
     * that is desired by URLClassLoader (ending in `/' for
     * directories).
     */
    private URL[] getSystemClassPath() throws IOException {
        Properties sysProps = System.getProperties();
        String sysCP = (String)sysProps.get("java.class.path");
        String ps = (String)sysProps.get("path.separator");
        ArrayList urls = new ArrayList();
        StringTokenizer tokens = new StringTokenizer(sysCP, ps);
        while (tokens.hasMoreTokens()) {
            String file = tokens.nextToken();
            if (!(file.endsWith(".jar") || file.endsWith(".zip"))) {
                file += "/";
            }
            urls.add(new URL("file", null, -1, file));
        }
        return (URL[])urls.toArray(new URL[urls.size()]);
    }

    /**
     * Create a MultiClassLoader for tests. Used with Enhydra 3.*.
     * Use reflection to avoid compile errors on 4.0.
     */
    private ClassLoader mkMultiClassLoader(ClassLoader filterParent) {
        try {
            Class cl = Class.forName("com.lutris.classloader.MultiClassLoader");
            // MultiClassLoader(URL[], LogChannel)
            Constructor constr = cl.getConstructor(new Class[]{URL[].class,
                                                               LogChannel.class});
            ClassLoader loader = (ClassLoader)constr.newInstance(new Object[]{getSystemClassPath(),
                                                                              TestUtils.getLogChannel()});
            
            Method setParent = cl.getMethod("setParent", new Class[]{ClassLoader.class});
            setParent.invoke(loader, new Object[]{filterParent});
            return loader;
        } catch (Exception except) {
            throw new XMLCError(except);
        }
    }

    /**
     * Create an EnhydraClassLoader for tests. Used with Enhydra 4.*.
     * Use reflection to avoid compile errors on 3.0.
     */
    private ClassLoader mkEnhydraClassLoader(ClassLoader filterParent) {
        try {
            Class cl = Class.forName("org.enhydra.classloader.EnhydraClassLoader");
            // newInstance(ClassLoader, URL[], LogChannel)
            Method newInstance = cl.getMethod("newInstance",
                                              new Class[]{ClassLoader.class,
                                                          URL[].class,
                                                          LogChannel.class});
            return (ClassLoader)newInstance.invoke(null,
                                                   new Object[]{filterParent,
                                                                getSystemClassPath(),
                                                                TestUtils.getLogChannel()});
        } catch (Exception except) {
            throw new XMLCError(except);
        }
    }

    /**
     * Create a URLClassLoader for tests.
     */
    private ClassLoader mkURLClassLoader(ClassLoader filterParent) {
        try {
            return URLClassLoader.newInstance(getSystemClassPath(),
                                              filterParent);
        } catch (Exception except) {
            throw new XMLCError(except);
        }
    }

    /**
     * Create a class loader that includes the system class path.
     * This filters to only load test classes.
     */
    private ClassLoader mkTestClassLoader() {
        ClassLoader filterParent = new TestFilterClassLoader();
        if (classLoaderName.equals(CL_MULTI)) {
            return mkMultiClassLoader(filterParent);
        } else if (classLoaderName.equals(CL_ENHYDRA)) {
            return mkEnhydraClassLoader(filterParent);
        } else if (classLoaderName.equals(CL_URL)) {
            return mkURLClassLoader(filterParent);
        } else {
            throw new XMLCError("bug: invalid classloader");
        }
    }
    
    /**
     * Find the source file for a class on the CLASSPATH.
     */
    private File findClassFile(ClassLoader loader,
                               String className)
            throws ClassNotFoundException {
        URL url = loader.getResource(className.replace('.', '/') + ".class");
        return new File(url.getFile());
    }

    /**
     * Touch a class file.
     */
    private void touchClassFile(ClassLoader loader,
                                String className)
        throws ClassNotFoundException, XMLCException {
        TestUtils.touchFile(findClassFile(loader, className));
    }

    /**
     * Verify a delegate that must have changed.
     */
    private void verifyChangedDelegate(String testId,
                                       XMLObject newObject,
                                       XMLObject prevObject) {
        XMLObject delegate = newObject.getDelegate();
        if (delegate == null) {
            failure(testId + ".1", "delegate is null");
        } else {
            if (delegate.getClass() == newObject.getClass()) {
                failure(testId + ".2", "delegate class same as loaded class");
            }
            if (delegate.getClass().getClassLoader() == newObject.getClass().getClassLoader()) {
                failure(testId + ".3", "delegate class loader same as loaded class");
            }
        }
        if ((prevObject != null) && (prevObject.getDelegate() == delegate)) {
            failure(testId + ".4", "delegate same as previous delegate");
        }
    }

    /**
     * Create initial test object.
     */
    private void createTestObject(String testId,
                                  String className,
                                  ClassLoader loader) throws Exception {
        log("Test-" + testId + ".1");
        origClass = loader.loadClass(className);

        origObject = (XMLObject)origClass.newInstance();
        if (origObject.getDelegate() != null) {
            failure(testId + ".2", "delegate not null");
        }
    }

    /*
     * Test loading of a class given a class name.
     */
    private XMLObject doTest(String testId,
                             String interfaceName) throws Exception {
        XMLObject finalObj = null;
        String className = interfaceName + "Impl";

        ClassLoader loader = mkTestClassLoader();
        LogChannel logChannel =  TestUtils.getLogChannel();
        XMLCReloadingFactory factory
            = new XMLCReloadingFactory(new StandardRecompilationHandler(),
                                       loader, new EnhydraXMLCLogger(logChannel));

        // Load class for the first time, setting object variables.
        createTestObject(testId, className, loader);

        // Create another instance through the factory.
        log("Test-" + testId + ".2.1");
        XMLObject obj1 = factory.create(className);
        if (obj1.getClass() != origClass) {
            failure(testId + ".2.2", "class not the same as original");
        }
        if (origObject.getDelegate() != null) {
            failure(testId + ".2.3", "delegate not null");
        }
        if (origObject.getDocument() == null) {
            failure(testId + ".2.4", "document is null");
        }
        if (factory.getReloadCount() != 0) {
            failure(testId + ".2.5", "reload count wrong; got "
                    + factory.getReloadCount());
        }
        if (factory.getRecompileCount() != 0) {
            failure(testId + ".2.6", "recompile count wrong; got "
                    + factory.getRecompileCount());
        }

        // Modify the class file and try again.
        log("Test-" + testId + ".3.1");
        touchClassFile(loader, className);
        XMLObject obj2 = factory.create(className);
        if (obj2.getClass() != origClass) {
            failure(testId + ".3.2", "class not the same as original");
        }
        if (obj2.getDocument() == null) {
            failure(testId + ".3.4", "document is null");
        }
        if (factory.getReloadCount() != 1) {
            failure(testId + ".3.4", "reload count wrong; got " 
                    + factory.getReloadCount());
        }
        if (factory.getRecompileCount() != 0) {
            failure(testId + ".3.5", "recompile count wrong; got "
                    + factory.getRecompileCount());
        }
        verifyChangedDelegate(testId + ".3.6", obj2, null);
        finalObj = obj2;

        if (sourceFile != null) {
            if (copyFile != null) {
                // Overwrite it.
                log("Test: overwritting " + sourceFile);
                TestUtils.copyFile(copyFile, sourceFile);
            } else {
                // Just change mod date
                log("Test: touching " + sourceFile);
                TestUtils.touchFile(sourceFile);
            }
            log("Test-" + testId + ".4.1");
            XMLObject obj3 = factory.create(className);
            if (obj3.getClass() != origClass) {
                failure(testId + ".4.2", "class not the same as original");
            }
            if (origObject.getDocument() == null) {
                failure(testId + ".4.3", "document is null");
            }
            if (factory.getReloadCount() != 2) {
                failure(testId + ".4.4", "reload count wrong; got "
                        + factory.getReloadCount());
            }
            if (factory.getRecompileCount() != 1) {
                failure(testId + ".4.5", "recompile count wrong; got "
                        + factory.getRecompileCount());
            }
            verifyChangedDelegate(testId + ".4.6", obj3, obj2);
            finalObj = obj3;
            
        }
        if (finalObj == null) {
            throw new XMLCError("final object is null");
        }
        return finalObj;
    }

    /*
     * Exit based on the error count.
     */
    private void finish() {
        if (tester.errorCnt != 0) {
            System.err.println("**** " + tester.errorCnt + " tests failed");
            System.exit(1);
        }
    }

    /**
     * Run tests.
     */
    private void runTests(String[] args) {
        int argIdx = 0;
        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
            if (args[argIdx].equals("-dump")) {
                dump = true;
            } else if (args[argIdx].equals("-sourcefile")) {
                TestUtils.checkOptArg(1, argIdx, args);
                sourceFile = new File(args[++argIdx]);
            } else if (args[argIdx].equals("-copyfile")) {
                TestUtils.checkOptArg(1, argIdx, args);
                copyFile = new File(args[++argIdx]);
            } else if (args[argIdx].equals("-classloader")) {
                TestUtils.checkOptArg(1, argIdx, args);
                classLoaderName = args[++argIdx];
                if (!(classLoaderName.equals(CL_ENHYDRA)
                      || classLoaderName.equals(CL_MULTI)
                      || classLoaderName.equals(CL_URL))) {
                    TestUtils.fatal("Error: -classloader name \"" + classLoaderName + "\"");
                }
            } else {
                TestUtils.fatal("Error: invalid option \"" + args[argIdx]
                                + "\", expected one of ");
            }
            argIdx++;
        }
        int argsLeft = args.length-argIdx;
        if ((argsLeft < 1) || (argsLeft > 2)) {
            TestUtils.fatal(USAGE);
        }
        interfaceName = args[argIdx];
        outFile = (argsLeft > 1) ? args[argIdx+1] : null;
        
        try {
            XMLObject xmlObj = doTest("1", interfaceName);
            if (dump) {
                PrintWriter out = new PrintWriter(System.out);
                DOMInfo.printTree("DOM hierarchy", xmlObj, 
                                  out);
                out.flush();
            }
            if (outFile != null) {
                TestUtils.outputDocument(xmlObj, outFile, null);
            }
            finish();
        } catch (Throwable error) {
            TestUtils.handleError(error);
        }
    }

    /*
     * Test entry point.
     */
    public static void main(String[] args) {
        new TestAutoComp().runTests(args);
    }
}
